/* controls.h -- controls for mvmf applications

	Copyright 2003,2004,2005,2006,2007
		by Mark E. Mallett, MV Communications, Inc.

	See the "LICENSE" file for terms.

One of the ideas in mvmf is to provide some administrative controls
related to email delivery.  e.g. with procmail, it's assumed that
every user has access to all tools available at the shell.  With
mvmf that may or may not be true, depending on the wishes of the
site administrator.

Equally important are user-settable controls.  These may include timeouts,
storage limits, and what-have-you.

Controls are manipulated by name and type, via the functions defined
herein.

*/


#ifndef	H_CONTROLS		/* For multiple inclusions */
#define	H_CONTROLS


    /* Control types (for CNTRL, see below) */
typedef enum {
    CTL_INT,			/* An integer value */
    CTL_STR,			/* Some string */
}  CNTRLTYPE;


    /* Control access types */
typedef enum {
    CA_USER,			/* User */
    CA_ADMIN,			/* Admin */
}  CNTRLACC;


    /* Data for an integer control value */
typedef struct {
    long	cv_int;		/* The integer value */
    long	cv_abslow;	/* Absolute (admin) low limit */
    long	cv_abshigh;	/* Absolute (admin) high limit */
    long	cv_userlow;	/* Low limit that user can not exceed */
    long	cv_userhigh;	/* High limit that user can not exceed */
}  CNTRLINT;


    /* Data for a string control value */
typedef struct {
    BOOL	cs_adminF;	/* TRUE if only admin can set this */
    UBYTE	*cs_strP;	/* The string value */
#if 0    /* maybe a future enhancement */
    BOOL	(*cs_val)();	/* Validation function for value */
#endif
}  CNTRLSTR;


    /* A control.  Name and info about control. */

typedef struct {
    char	*ctl_nameP;		/* The name of the control */
    CNTRLTYPE	ctl_type;		/* The type of control */
    void	*ctl_infoP;		/* More info about the control */
}  CNTRL;  /* ran out of vowels today */


	/* The defining structs here look like they could just contain the
	   appropriate control struct instead of being defined separately,
	   but I'd like to keep them separated so one could be changed
	   without changing the other.
	*/

    /* An entry for defining an integer control.  */
typedef struct {
    char	*cti_nameP;	/* The control name */
    int		cti_int;	/* The integer value */
    long	cti_abslow;	/* Absolute (admin) low limit */
    long	cti_abshigh;	/* Absolute (admin) high limit */
    long	cti_userlow;	/* Low limit that user can not exceed */
    long	cti_userhigh;	/* High limit that user can not exceed */
}  CNTRLDEF_INT;


    /* An entry for defining a string control */
typedef struct {
    char	*cts_nameP;	/* The control name */
    char	*cts_strP;	/* The control value */
    BOOL	cts_adminF;	/* TRUE if only admin can set it */
}  CNTRLDEF_STR;



/* Declarations for functions contained in this module: */

    /* Define an integer control */
CNTRL	*cntrl_define_int PROTO( (ESTR *nameP,
				   long value,
				   long abslow, long abshigh,
				   long userlow, long userhigh) );

CNTRL	*cntrl_define_str PROTO( (ESTR *nameP,
				   ESTR *valueP,
				   BOOL adminF) );

    /* Find a control */
CNTRL	*cntrl_find PROTO( (ESTR *nameP) );

    /* Initialize the controls module.  Optional for the application, as
       it will be called as needed by the internal functions.
    */
BOOL	cntrl_init PROTO( (void) );

    /* Get the value of an integer control, by name */
long	cntrl_int_get PROTO( (ESTR *nameP, long notfound) );

    /* Set an integer control, by name */
BOOL	cntrl_int_set PROTO( (ESTR *nameP, CNTRLACC cat, long val) );

    /* Set integer control limits, by name */
BOOL	cntrl_int_setlim PROTO( (ESTR *nameP, CNTRLACC cat,
				 long low, long high) );

    /* Set an integer control, by control pointer */
BOOL	cntrl_intc_set PROTO( (CNTRLINT *ciP, CNTRLACC cat, long val) );

    /* Set integer control limits, by control pointer */
BOOL	cntrl_intc_setlim PROTO( (CNTRLINT *ciP, CNTRLACC cat,
				 long low, long high) );

    /* Get the value of a string control, by name */
UBYTE	*cntrl_str_get PROTO( (ESTR *nameP) );

    /* Set a string control, by name */
BOOL	cntrl_str_set PROTO( (ESTR *nameP, CNTRLACC cat, ESTR *valP) );

    /* Set a string control, by control pointer */
BOOL	cntrl_strc_set PROTO( (CNTRLSTR *csP, CNTRLACC cat, ESTR *valP) );

    /* Define integer controls out of a table */
BOOL	cntrl_table_def_int PROTO( (CNTRLDEF_INT *defP) );

    /* Define string controls out of a table */
BOOL	cntrl_table_def_str PROTO( (CNTRLDEF_STR *defP) );

    /* Walk the list of controls */
void	cntrl_walk PROTO( (void (*rtc)() ) );

#endif	/* H_CONTROLS */
