/* mfl-eval.h -- definitions for the mfl-eval module

	Copyright 2003,2004,2005,2006,2007
		by Mark E. Mallett, MV Communications, Inc.

	See the "LICENSE" file for terms.

Stuff relating to mfl-eval

*/


#ifndef	H_MFL_EVAL		/* For multiple inclusions */
#define	H_MFL_EVAL



/* Declarations for functions contained in this module: */


    /* Execute "address of" node */
STSCODE	eval_addr PROTO( (VALUE *rvalP, OPINFO *opP, VALUE *valP ) );

    /* Execute "array reference" node (binary operator) */
STSCODE	eval_aref PROTO( (VALUE *rvalP, OPINFO *opP, VALUE *val1P,
			    VALUE *val2P ) );

    /* Execute binary arithmetic operator node */
STSCODE	eval_arith PROTO( (VALUE *rvalP, OPINFO *opP, VALUE *val1P,
			     VALUE *val2P ) );

    /* Execute "assign" binary operator node (assignment operator) */
STSCODE	eval_assign PROTO( (VALUE *rvalP, OPINFO *opP, VALUE *val1P,
			      VALUE *val2P ) );

    /* Execute "bitwise complement" node */
STSCODE	eval_bitcomp PROTO( (VALUE *rvalP, OPINFO *opP, VALUE *valP ) );

    /* Execute "comma" binary operator node */
STSCODE	eval_comma PROTO( (VALUE *rvalP, OPINFO *opP, VALUE *val1P,
			     VALUE *val2P ) );

    /* Execute "condex" binary operator node (conditional expression) */
STSCODE	eval_condex PROTO( (VALUE *rvalP, OPINFO *opP, VALUE *val1P,
			      PTNODE *cvnodeP ) );

    /* Execute "pointer dereference" node */
STSCODE	eval_deref PROTO( (VALUE *rvalP, OPINFO *opP, VALUE *valP ) );

    /* Execute binary "logical AND" operator node */
STSCODE	eval_land PROTO( (VALUE *rvalP, OPINFO *opP, VALUE *val1P,
			     PTNODE *term2P ) );

    /* Execute binary "logical OR" operator node */
STSCODE	eval_lor PROTO( (VALUE *rvalP, OPINFO *opP, VALUE *val1P,
			     PTNODE *term2P ) );

    /* Execute "mineq" binary operator node ( "-=" ) */
STSCODE	eval_mineq PROTO( (VALUE *rvalP, OPINFO *opP, VALUE *val1P,
			     VALUE *val2P ) );

    /* Execute "minus" binary operator node (subtraction) */
STSCODE	eval_minus PROTO( (VALUE *rvalP, OPINFO *opP, VALUE *val1P,
			     VALUE *val2P ) );

    /* Evaluate a tree node */
STSCODE	eval_node PROTO( (PTNODE *treeP, VALUE *rvalP) );

    /* Evaluate a tree node that's an operator*/
STSCODE	eval_operator PROTO( (PTNODE *treeP, VALUE *rvalP) );

    /* Execute "opeq" binary operator node (generic operator=) */
STSCODE	eval_opeq PROTO( (VALUE *rvalP, OPINFO *opP, VALUE *val1P,
			    VALUE *val2P ) );

    /* Execute "plus" binary operator node (addition) */
STSCODE	eval_plus PROTO( (VALUE *rvalP, OPINFO *opP, VALUE *val1P,
			    VALUE *val2P ) );

    /* Execute "plus" binary operator node (addition) for pointers */
STSCODE	eval_plusptr PROTO( (VALUE *rvalP, OPINFO *opP, VALUE *val1P,
			    VALUE *val2P ) );

    /* Execute "plus" binary operator node (addition) for strings */
STSCODE	eval_plusstr PROTO( (VALUE *rvalP, OPINFO *opP, VALUE *val1P,
			    VALUE *val2P ) );

    /* Execute "pluseq" binary operator node ( += operator ) */
STSCODE	eval_pluseq PROTO( (VALUE *rvalP, OPINFO *opP, VALUE *val1P,
			      VALUE *val2P ) );

    /* Execute "post-decrement" node */
STSCODE	eval_postdecr PROTO( (VALUE *rvalP, OPINFO *opP, VALUE *valP ) );

    /* Execute "post-increment" node */
STSCODE	eval_postincr PROTO( (VALUE *rvalP, OPINFO *opP, VALUE *valP ) );

    /* Execute "pre-decrement" node */
STSCODE	eval_predecr PROTO( (VALUE *rvalP, OPINFO *opP, VALUE *valP ) );

    /* Execute "pre-increment" node */
STSCODE	eval_preincr PROTO( (VALUE *rvalP, OPINFO *opP, VALUE *valP ) );

    /* Execute "prefix minus" node */
STSCODE	eval_preminus PROTO( (VALUE *rvalP, OPINFO *opP, VALUE *valP ) );

    /* Execute "prefix not" node */
STSCODE	eval_prenot PROTO( (VALUE *rvalP, OPINFO *opP, VALUE *valP ) );

    /* Execute "prefix plus" node */
STSCODE	eval_preplus PROTO( (VALUE *rvalP, OPINFO *opP, VALUE *valP ) );

    /* Execute relational binary operator node */
STSCODE	eval_rel PROTO( (VALUE *rvalP, OPINFO *opP, VALUE *val1P,
			   VALUE *val2P ) );

    /* Evaluate a tree */
STSCODE	eval_tree PROTO( (PTNODE *treeP, VALUE *rvalP) );

    /* Evaluate a tree, return no value */
STSCODE	eval_tree_nil PROTO( (PTNODE *treeP) );

    /* Evaluate a tree, return long */
STSCODE	eval_tree_long PROTO( (PTNODE *treeP, long *longP) );

    /* Evaluate a tree, return word */
STSCODE	eval_tree_word PROTO( (PTNODE *treeP, WORD *wordP) );

    /* Character support functions */
BOOL	evalc_iscontrol PROTO( (UBYTE ch) );
BOOL	evalc_isdigit PROTO( (UBYTE ch) );
BOOL	evalc_ishighbit PROTO( (UBYTE ch) );
BOOL	evalc_islower PROTO( (UBYTE ch) );
BOOL	evalc_isupper PROTO( (UBYTE ch) );

    /* These are arithmetic support functions. */
double	evals_fadd PROTO( (double d1, double d2) );
double	evals_fdiv PROTO( (double d1, double d2) );
double	evals_feq PROTO( (double d1, double d2) );
double	evals_fge PROTO( (double d1, double d2) );
double	evals_fgt PROTO( (double d1, double d2) );
double	evals_fle PROTO( (double d1, double d2) );
double	evals_flt PROTO( (double d1, double d2) );
double	evals_fmult PROTO( (double d1, double d2) );
double	evals_fne PROTO( (double d1, double d2) );
double	evals_fsub PROTO( (double d1, double d2) );
long	evals_iadd PROTO( (long l1, long l2) );
long	evals_ibitand PROTO( (long l1, long l2) );
long	evals_ibitor PROTO( (long l1, long l2) );
long	evals_idiv PROTO( (long l1, long l2) );
long	evals_ieq PROTO( (long l1, long l2) );
long	evals_ige PROTO( (long l1, long l2) );
long	evals_igt PROTO( (long l1, long l2) );
long	evals_ile PROTO( (long l1, long l2) );
long	evals_ilshift PROTO( (long l1, long l2) );
long	evals_ilt PROTO( (long l1, long l2) );
long	evals_imod PROTO( (long l1, long l2) );
long	evals_imult PROTO( (long l1, long l2) );
long	evals_ine PROTO( (long l1, long l2) );
long	evals_irshift PROTO( (long l1, long l2) );
long	evals_isub PROTO( (long l1, long l2) );

    /* String support functions */
STSCODE	scmp_eq PROTO( (VALUE *rvalP, VALUE *val1P, VALUE *val2P) );
STSCODE	scmpi_eq PROTO( (VALUE *rvalP, VALUE *val1P, VALUE *val2P) );
STSCODE	scmp_ge PROTO( (VALUE *rvalP, VALUE *val1P, VALUE *val2P) );
STSCODE	scmpi_ge PROTO( (VALUE *rvalP, VALUE *val1P, VALUE *val2P) );
STSCODE	scmp_gt PROTO( (VALUE *rvalP, VALUE *val1P, VALUE *val2P) );
STSCODE	scmpi_gt PROTO( (VALUE *rvalP, VALUE *val1P, VALUE *val2P) );
STSCODE	scmp_le PROTO( (VALUE *rvalP, VALUE *val1P, VALUE *val2P) );
STSCODE	scmpi_le PROTO( (VALUE *rvalP, VALUE *val1P, VALUE *val2P) );
STSCODE	scmp_lt PROTO( (VALUE *rvalP, VALUE *val1P, VALUE *val2P) );
STSCODE	scmpi_lt PROTO( (VALUE *rvalP, VALUE *val1P, VALUE *val2P) );
STSCODE	scmp_ne PROTO( (VALUE *rvalP, VALUE *val1P, VALUE *val2P) );
STSCODE	scmpi_ne PROTO( (VALUE *rvalP, VALUE *val1P, VALUE *val2P) );
STSCODE	sregex_eq PROTO( (VALUE *rvalP, VALUE *val1P, VALUE *val2P) );
STSCODE	sregex_ne PROTO( (VALUE *rvalP, VALUE *val1P, VALUE *val2P) );
STSCODE	swild_eq PROTO( (VALUE *rvalP, VALUE *val1P, VALUE *val2P) );
STSCODE	swildi_eq PROTO( (VALUE *rvalP, VALUE *val1P, VALUE *val2P) );
STSCODE	swild_ne PROTO( (VALUE *rvalP, VALUE *val1P, VALUE *val2P) );
STSCODE	swildi_ne PROTO( (VALUE *rvalP, VALUE *val1P, VALUE *val2P) );


#endif	/* H_MFL_EVAL */
