/* mfl-func.h -- parsing and execution of mfl functions

	Copyright 2003,2004,2005,2006,2007
		by Mark E. Mallett, MV Communications, Inc.

	See the "LICENSE" file for terms.

Parsing / execution of MFL functions, plus some built-in function
routines.

*/


#ifndef	H_MFL_FUNC		/* For multiple inclusions */
#define	H_MFL_FUNC


    /* Types of functions */
typedef enum {
    FUNC_NOTDEF,			/* Undefined */
    FUNC_BUILTIN,			/* Built-in (in C) */
    FUNC_MFL,				/* An MFL function */
}  FUNCTYPE;



    /* Information about a formal argument to a function */
typedef struct {
    char	*fa_nameP;		/* Name */
    TYPE	*fa_typeP;		/* Type */
}  FUNCFARG;


    /* Information about all formal arguments to a function */
typedef struct {
    int		ffl_argC;		/* Number of arguments */
    int		ffl_argL;		/* Length of argument array */
    FUNCFARG	*ffl_argP;		/* List of arguments */
}  FUNCFARGS;


    /* Encapsulated info about a function.  This is hooked
       into the VALUE part of a variable, in the p_genP VALPTR pointer.
    */
typedef struct {
    FUNCTYPE	fn_type;		/* Type of function */
    TYPE	*fn_rtypeP;		/* Return type */
    void	*fn_valP;		/* Ptr to function value */
    FUNCFARGS	fn_fargs;		/* Function arguments */
}  FUNC;


    /* Information about a built-in function.  Used in registering
       built-in functions.
    */
typedef struct {
    char	*bif_nameP;		/* Function name */
    STSCODE	(*bif_bif)();		/* Function */
}  BIFINF;


/* Declarations for functions contained in this module: */

    /* Evaluate a function reference found in a parse tree
       (i.e., execute the function)
    */
STSCODE	eval_funcref PROTO( (VALUE *rvalP, PTNODE *fnnodeP, PTNODE *argtreeP) );

    /* Find a name in a function arguments list */
int	fargs_find PROTO( (FUNCFARGS *fargsP, char *anameP) );

    /* Lookup a function by name */
FUNC	*func_lookup PROTO( (char *fnameP) );

    /* Register a function */
FUNC	*func_register PROTO( (char *fnameP, VSYM *vsymP,
			       FUNCTYPE ftype, TYPE *fvtypeP,
			       FUNCFARGS *fargsP, void *fvalP) );

    /* Get the nth function argument out of the parse tree */
PTNODE	*func_treearg PROTO( (PTNODE *treeP, int argN) );

    /* Evaluate the nth function argument, return temp ESTR */
ESTR	*func_treearg_estr PROTO( (char *fnameP, PTNODE *treeP,
				   int argN, VALUE *rvalP) );

    /* Evaluate the nth function argument,
	return temp ESTR with new string copy.
    */
ESTR	*func_treearg_estr_new PROTO( (char *fnameP, PTNODE *treeP,
				       int argN ) );

    /* Evaluate the nth function argument,
	return temp ESTR with tmpstr string copy.
    */
ESTR	*func_treearg_estr_tmpstr PROTO( (char *fnameP, PTNODE *treeP,
				       int argN ) );

    /* Evaluate the nth function argument, return long value */
STSCODE	func_treearg_long PROTO( (char *fnameP, PTNODE *treeP,
				   int argN, long *longP) );

    /* Evaluate the nth function argument, return nul-terminated string */
char	*func_treearg_str PROTO( (char *fnameP, PTNODE *treeP,
				   int argN, VALUE *rvalP,
				   int maxlen) );

    /* Evaluate the nth function argument, return string value */
STSCODE	func_treearg_string PROTO( (char *fnameP, PTNODE *treeP,
				   int argN, VALUE *strvalP) );

    /* Evaluate the nth function argument, return value coerced to string */
STSCODE	func_treearg_string_coerce PROTO( (char *fnameP, PTNODE *treeP,
				   int argN, VALUE *strvalP) );

    /* Evaluate the nth function argument, return word value */
STSCODE	func_treearg_word PROTO( (char *fnameP, PTNODE *treeP,
				   int argN, WORD *wordP) );

    /* Reference a function argument on the current frame */
STSCODE	mfl_func_arg PROTO( (VALUE *rvalP, int argN) );

    /* Grow the current stack frame */
STSCODE mfl_func_arg_frame_grow PROTO( (int argC) );

    /* "Mark" the current frame context on the frame */
int	mfl_func_arg_frame_mark PROTO( (void) );

    /* Pop to previous stack frame */
STSCODE	mfl_func_arg_frame_pop PROTO( (void) );

    /* Set the new stack frame */
STSCODE	mfl_func_arg_frame_set PROTO( (int frameX) );

    /* Register built-in functions from a BIFINF table */
STSCODE	mfl_func_bifreg PROTO( (BIFINF *biP) );

    /* Evaluate (execute) an MFL function */
STSCODE	mfl_func_exec PROTO( (VALUE *rvalP, FUNC *funcP, PTNODE *argtreeP) );

    /* Function hook: call a function if it's defined. */
BOOL	mfl_func_hook PROTO( (char *fnameP) );

    /* Add a directory to the hook path */
BOOL	mfl_func_hook_dir_add PROTO( (ESTR *dirnameP) );

    /* Try to find a hook function */
BOOL	mfl_func_hook_find PROTO( (char *hnameP) );

    /* Initialize the functions module */
BOOL	mfl_func_init PROTO( (void) );

    /* Parse a function definition, once the type and name have been seen */
STSCODE parse_func_def PROTO( (MML_IN *inP, PTNODE **treePP, char *vnameP) );


    /* Parse a function call, once the function name and open
        paren have been seen.
    */
STSCODE	parse_func_call PROTO( (MML_IN *inP, TOKEN *symtokP, PTNODE **treePP) );


    /* Parse a function value (i.e., a function body) if present. */
STSCODE parse_func_value PROTO( ( MML_IN *inP, PTNODE *trnodeP ) );

#endif	/* H_MFL_FUNC */
