/* mfl-parse.h -- definitions for the mfl language parsing module

	Copyright 2003,2004,2005,2006,2007
		by Mark E. Mallett, MV Communications, Inc.

	See the "LICENSE" file for terms.

Stuff for the language parser

*/


#ifndef	H_MFL_PARSE		/* For multiple inclusions */
#define	H_MFL_PARSE


/* Declarations for functions contained in this module: */

/* Many of the parse routines could be static, I suppose, but we
   leave them this way to make it easier to split up the files.
*/

    /* Parse (top level parse) */
STSCODE	mfl_parse PROTO( (MML_IN *inP, PTNODE **treePP) );

    /* Parse single or compound statement */
STSCODE	mfl_parse_soc PROTO( (MML_IN *inP, PTNODE **treePP) );

    /* Parse built in statement or function */
STSCODE	parse_bis PROTO( (MML_IN *inP, PTNODE **treePP, TOKVAL *tkvalP) );

    /* Parse built in statement: "break" */
STSCODE	parse_bis_break PROTO( (MML_IN *inP, PTNODE *nodeP) );

    /* Parse built in statement: "C" */
STSCODE	parse_bis_c PROTO( (MML_IN *inP, PTNODE *nodeP) );

    /* Parse built in statement: "continue" */
STSCODE	parse_bis_continue PROTO( (MML_IN *inP, PTNODE *nodeP) );

    /* Parse built in statement: "do" */
STSCODE	parse_bis_do PROTO( (MML_IN *inP, PTNODE *nodeP) );

    /* Parse built in statement: "for" */
STSCODE	parse_bis_for PROTO( (MML_IN *inP, PTNODE *nodeP) );

    /* Parse built in statement: "if" */
STSCODE	parse_bis_if PROTO( (MML_IN *inP, PTNODE *nodeP) );

    /* Parse built in statement: "pv$" */
STSCODE	parse_bis_pv_ PROTO( (MML_IN *inP, PTNODE *nodeP) );

    /* Parse built in statement: "return" */
STSCODE	parse_bis_return PROTO( (MML_IN *inP, PTNODE *nodeP) );

    /* Parse built in statement: "sizeof" */
STSCODE	parse_bis_sizeof PROTO( (MML_IN *inP, PTNODE *nodeP) );

    /* Parse built in statement: "while" */
STSCODE	parse_bis_while PROTO( (MML_IN *inP, PTNODE *nodeP) );

    /* Parse compound statement */
STSCODE	parse_compound PROTO( (MML_IN *inP, PTNODE **treePP) );

    /* Parse a statement that is a declaration */
STSCODE	parse_decl_stmt PROTO( (MML_IN *inP, TNSYM *tsymP, PTNODE **treePP) );

    /* Parse any initializer value for a declaration */
STSCODE parse_decl_value PROTO( (MML_IN *inP, PTNODE *trnodeP) );

    /* Parse a declaration initializer for an aggregate */
STSCODE parse_decl_value_aggr PROTO( (MML_IN *inP, PTNODE *trnodeP) );

    /* Parse declaration initializer expression list(s) */
STSCODE parse_decl_value_list PROTO( (MML_IN *inP, PTNODE **treePP) );

    /* Parse a declaration initializer for a scalar */
STSCODE parse_decl_value_scalar PROTO( (MML_IN *inP, PTNODE *trnodeP) );

    /* Parse the two branches of a conditional expression */
STSCODE	parse_expr_cond PROTO( (MML_IN *inP, PTNODE **treePP) );

    /* Parse one or more expressions separated by commas */
STSCODE	parse_expr_list PROTO( (MML_IN *inP, PTNODE **treePP) );

    /* Parse prefix operator in an expresion */
STSCODE	parse_expr_preop PROTO( (MML_IN *inP, int lprio, TOKEN *tokP,
				 PTNODE **treePP) );

    /* Parse the right side of an expression */
STSCODE	parse_expr_right PROTO( (MML_IN *inP, int lprio, PTNODE **treePP) );

    /* Parse a C statement that is an expression */
STSCODE	parse_expr_stmt PROTO( (MML_IN *inP, PTNODE **treePP) );

    /* Parse the semicolon that's expected next */
STSCODE	parse_semi PROTO( (MML_IN *inP) );

    /* Parse type: aggregate */
STSCODE	parse_type_agg PROTO( (MML_IN *inP, TYPE **typePP,
			       TYPECLASS tclass) );

    /* Parse an aggregate's name or lead-in */
STSCODE	parse_type_aggname PROTO( (MML_IN *inP, TYPE **typePP, BOOL *newfP,
				   TYPECLASS aclass) );

    /* Parse base type definition */
STSCODE	parse_type_base PROTO( (MML_IN *inP, TNSYM *tsymP, TYPE **typePP) );

    /* Parse type: enum */
STSCODE	parse_type_enum PROTO( (MML_IN *inP, TYPE **typePP) );

    /* Parse a modified int type (e.g., unsigned) */
STSCODE	parse_type_modint PROTO( (MML_IN *inP, TYPE **typePP, int tmod) );

    /* Parse a type usage, yielding reference type info and
       potentially a variable name.
    */
STSCODE	parse_type_usage PROTO( (MML_IN *inP, TYPE *btypeP,
			   TYPE **tnodePP, char **vnamePP,
			   ODINFO *infoP) );

    /* Parse a type usage, yielding a parse tree that represents the
       type usage, and potentially a variable name.
    */
STSCODE	parse_type_usage_tree PROTO( (MML_IN *inP, PTNODE **treePP,
				  char **vnamePP) );

    /* Parse a typecast */
STSCODE	parse_typecast PROTO( (MML_IN *inP, PTNODE **treePP,
				   TNSYM *tsymP) );

    /* Parse typedef statement */
STSCODE	parse_typedef PROTO( (MML_IN *inP, PTNODE **treePP) );

    /* Free storage associated with parse tree */
STSCODE	parsetree_free PROTO( (PTNODE *treeP) );

    /* Make a tree node of a given type */
STSCODE	parsetree_mknode PROTO( (PTNODE **nodePP, PNTYPE type) );

    /* Make a tree node out of a token */
STSCODE	parsetree_mktknode PROTO( (PTNODE **nodePP, TOKEN *tokP) );

    /* Print a representation of a tree */
void	parsetree_print PROTO( (FILE *fP, int level, PTNODE *treeP) );

#endif	/* H_MFL_PARSE */
