/* mfl-token.h -- definitions for mfl-token

	Copyright 2003,2004,2005,2006,2007
		by Mark E. Mallett, MV Communications, Inc.

	See the "LICENSE" file for terms.

mfl-token contains functions dealing with parsing of mfl tokens
from an input stack

*/


#ifndef	H_MFL_TOKEN		/* For multiple inclusions */
#define	H_MFL_TOKEN

#define	MFL_TOKSIZE	100	/* Default token size/limit */
#define	MFL_TOKMAX	10000	/* Maximum token size (e.g. for LITSTR) */


    /* Types of tokens, lexically speaking */
typedef enum {
	TK_UNK,				/* Unknown (temp use) */
	TK_BIS,				/* Built-in statement */
	TK_SYM,				/* Symbol */
	TK_OPR,				/* Operator */
	TK_LITSTR,			/* String literal */
	TK_SEMI,			/* Semicolon */
	TK_INTEGER,			/* Integer */
	TK_FLOAT,			/* Floating point value */
	TK_CHAR,			/* Character token */
	TK_EOL,				/* End of Line */
	TK_ELLIPSIS,			/* Ellipsis */
	TK_EOF,				/* End of file */
}  TOKTYPE;


    /* Token value */
typedef union {
    char	tv_char;
    int		tv_int;
    long	tv_long;
    float	tv_float;
    void	*tv_ptr;
}  TOKVAL;


    /* A structure used to encapsulate info about a parsed token.
       This is in support of language parsing and thus has special
       stuff.
    */
typedef struct token {
	/* To avoid allocating and freeing token structures the
	   token module maintains some lists.  The next/last links
	   are in support of that overhead.  Everyone else, don't touch..
	*/
    struct token *tk_nextP;	/* Next one */
    struct token *tk_prevP;	/* Previous one */
    TOKTYPE	tk_type;	/* Type */
    BSTR	*tk_bstrP;	/* Contains token text */
    BSTR	*tk_rawP;	/* May contain raw text if different from
				   the token text.  The raw text may be
				   required for macro storage (vs tokenized
				   input).  It can also enable pushback of
				   the raw input text for re-interpretation.
				*/
    TOKVAL	tk_val;		/* Contains any value */
    BOOL	tk_rawF;	/* True if tk_rawP has valid data. */
}  TOKEN;


/* Declarations for functions contained in this module: */

    /* Push a token back onto the input stream */
BOOL	mfl_tokback PROTO( (MML_IN *inP, TOKEN *tokP) );

    /* Restore tokback state from previously pushed one */
void	mfl_tokback_restore PROTO( (TOKEN *pbtokP) );

    /* Save current tokback state and clear */
TOKEN	*mfl_tokback_save PROTO( (void) );

    /* Release a token [back] to the free pool */
void	mfl_tokrelease PROTO( (TOKEN *tokP) );

    /* Reserve a token from the free pool */
void	mfl_tokreserve PROTO( (TOKEN *tokP) );

    /* Input up to and including the next preprocessor directive */
BOOL	mfl_ppftoken PROTO( (MML_IN *inP) );

    /* Input the next significant token */
TOKEN	*mfl_sigtoken PROTO( (MML_IN *inP) );

    /* Input the next token */
TOKEN	*mfl_token PROTO( (MML_IN *inP) );

    /* Return token type string */
char	*mfl_toktype PROTO( (TOKTYPE type) );

#endif	/* H_MFL_TOKEN */
