/* mfl-type.h -- definitions for the mfl-type module

	Copyright 2003,2004,2005,2006,2007
		by Mark E. Mallett, MV Communications, Inc.

	See the "LICENSE" file for terms.

Stuff dealing with data types

*/


#ifndef	H_MFL_TYPE		/* For multiple inclusions */
#define	H_MFL_TYPE



/* Declarations for functions contained in this module: */


    /* Access a DVS object */
DVS	*dvs_access PROTO( ( DVS *dvsP ) );

    /* Make a new DVS object */
DVS	*dvs_new PROTO( (int size, int max) );

    /* Release a DVS reference */
void	dvs_release PROTO( (DVS *dvsP) );

    /* Add an access to a free-standing VALPTR */
void	fsv_access PROTO( (FSV *fsvP) );

    /* Create a new free-standing VALPTR */
FSV	*fsv_new PROTO( (void) );

    /* Release access to a free-standing VALPTR */
void	fsv_release PROTO( (FSV *fsvP) );

    /* Delete opaque dynamic info storage */
BOOL	odinfo_delete PROTO( (ODINFO *infoP) );

    /* Get a temporary VALUE object */
VALUE	*tmpval PROTO( (void) );

    /* Find an aggregate member definition for an aggregate type */
AGGMEM	*type_agg_member PROTO( (TYPE *typeP, char *mnameP) );

    /* Find an aggregate member definition for a ptr-to-aggregate type */
AGGMEM	*type_aggp_member PROTO( (TYPE *typeP, char *mnameP) );

    /* Make or lookup an aggregate type by class and name */
TNSYM	*type_aggname PROTO( (int tclass, char *anameP) );

    /* See if a type is a string */
BOOL	type_is_string PROTO( (TYPE *typeP) );

    /* Make an aggregate type by type class */
TYPSYM	*type_mkagg PROTO( (int tclass) );

    /* Print a description of a type */
STSCODE	type_print PROTO( (FILE *fP, TYPE *typeP) );

    /* Make or find a type that's a pointer to an existing type */
TYPSYM	*type_ptr PROTO( (TYPE *stypeP) );

    /* Comparison routine for TYPSYM types */
int	type_symtab_comp PROTO( ( TYPSYM *symP, TYPE *keyP ) );

    /* Hash routine for TYPSYM types */
UWORD	type_symtab_hash PROTO( ( TYPE *keyP ) );

    /* Key association routine for string type key */
BOOL	type_symtab_keyassoc PROTO( ( TYPSYM *symP, TYPE *keyP ) );

    /* Key disassociation routine for string type key */
BOOL	type_symtab_keydis PROTO( ( TYPSYM *symP ) );

    /* Find an unsigned int type of the desired size */
TYPE	*type_uint PROTO( (int size) );

    /* Suggest a PO_xxx value for a type class */
int	typeclass_po PROTO( (TYPECLASS tc) );

    /* Access a value */
void	val_access PROTO( (VALUE *valP) );

    /* Access a pointer value */
void	val_access_ptr PROTO( (VALUE *valP) );

    /* Assign from one value to another */
STSCODE	val_assign PROTO( (VALUE *dvalP, VALUE *svalP ) );

    /* val_assign for ptr type */
STSCODE	val_assign_ptr PROTO( (VALUE *dvalP, VALUE *svalP ) );

    /* val_assign for string type */
STSCODE	val_assign_string PROTO( (VALUE *dvalP, VALUE *svalP ) );

    /* Propagate type attributes */
void	val_attrprop PROTO( (VALUE *rvalP, VALUE *val1P, VALUE *val2P) );

    /* Coerce the type of a value */
STSCODE	val_coerce PROTO( (VALUE *valP, TYPE *typeP, BOOL forceF ) );

    /* Coerce the type of a pointer value */
STSCODE	val_coerce_ptr PROTO( (VALUE *valP, TYPE *typeP ) );

    /* Find a type that's common to two types */
TYPE	*val_common_type PROTO( (VALUE *val1P, VALUE *val2P) );

    /* Dereference a pointer */
STSCODE	val_deref PROTO( (VALUE *dvalP, VALUE *svalP) );

    /* Evaluate a node or tree, return val_estr (a special ESTR) */
STSCODE	val_estr_eval PROTO( (PTNODE *treeP, ESTR **esPP, VALUE *valP) );

    /* Finish access to a val_estr */
void	val_estr_finish PROTO( (ESTR *esP) );

    /* Access a string buffer representing a value */
ESTR	*val_estr_tmp PROTO( (VALUE *valP) );

    /* Finish with a value (release and nill it) */
void	val_finish PROTO( (VALUE *valP) );

    /* Get a floating point number out of a value */
double	val_get_float PROTO( (VALUE *valP) );

    /* Get an integer out of a value */
long	val_get_int PROTO( (VALUE *valP) );

    /* Get a string reference out of a value */
REFSTR	*val_get_string PROTO( (VALUE *valP) );

    /* Initialize/instantiate a value */
STSCODE	val_init PROTO( (VALUE *valP, TYPE *typeP) );

    /* See if a value is an integer */
BOOL	val_isint PROTO( (VALUE *valP, BOOL mustF) );

    /* Get an lvalue */
STSCODE	val_lvalue PROTO( (VALUE *dvalP, VALUE *svalP ) );

    /* Reference a member of an aggregate */
STSCODE	val_member PROTO( (VALUE *dvalP, VALUE *svalP, char *mnameP ) );

    /* Make a value unused (turn it to nill) */
void	val_nill PROTO( (VALUE *valP) );

    /* Reference a member of a pointer to an aggregate */
STSCODE	val_pmember PROTO( (VALUE *dvalP, VALUE *svalP, char *mnameP ) );

    /* Print a value and its type */
STSCODE	val_print PROTO( (FILE *fP, VALUE *valP, int level, char *tagP) );

    /* Print an aggregate value (helper for val_print) */
STSCODE	val_print_agg PROTO( (FILE *fP, VALUE *valP, int level, char *tagP) );

    /* Store a floating point (double) into a value */
void	val_put_float PROTO( (VALUE *valP, double dval) );

    /* Store an integer into a value */
void	val_put_int PROTO( (VALUE *valP, long dint) );

    /* Make a new reference to an existing value */
void	val_ref PROTO( (VALUE *svalP, VALUE *rvalP ) );

    /* Release a value (clear any access that it holds) */
void	val_release PROTO( (VALUE *valP) );

    /* Release a pointer value */
void	val_release_ptr PROTO( (VALUE *valP) );

    /* Make sure we have an rvalue */
STSCODE	val_rvalue PROTO( (VALUE *valP) );

    /* Set a value to an integer constant */
void	val_set_int PROTO( (VALUE *valP, int isize, long int) );

    /* Set a value to a string */
STSCODE	val_set_string PROTO( (VALUE *valP, ESTR *esP) );

    /* Set a value to a string from a refstr */
STSCODE	val_set_string_refstr PROTO( (VALUE *valP, REFSTR *rsP) );

    /* Set a value to a pointer to an existing refstr */
STSCODE	val_set_string_refstr_ptr PROTO( (VALUE *valP, REFSTR *rsP) );

    /* Set a value to a string pointer */
STSCODE	val_set_string_ptr PROTO( (VALUE *valP, ESTR *esP) );

    /* Compare two values that are strings */
STSCODE val_string_cmp PROTO( (int *resultP,
			       VALUE *patvalP, VALUE *strvalP,
			       STRMATTYPE mtype,
			       UWORD mflags) );

    /* Create/steal a string value from another value */
STSCODE	val_string_val PROTO( (VALUE *rvalP, VALUE *svalP, BOOL stealF) );

    /* Resolve a value pointer (to a real pointer) */
void	*valptr_resolve PROTO( (VALPTR *vptrP) );


#endif	/* H_MFL_TYPE */
