/* mfl.h -- General definitions for mail filter language

	Copyright 2003,2004,2005,2006,2007
		by Mark E. Mallett, MV Communications, Inc.

	See the "LICENSE" file for terms.

Contains definitions for the basic and detailed elements of the
scripting language used in this application.

*/


#ifndef	H_MFL		/* For multiple inclusions */
#define	H_MFL

    /* Define this to enable concatenation of adjacent string
       literals (  "A" "B" becomes "AB" ).  This creates problems
       with sieve syntax, so probably best not to do it.
    */
/* #define ADJACENT_LITSTR_CONCAT	*/

    /* Define this to allow redeclaration of variables.  I don't
       recommend enabling this, and this may be removed as an option
       in the future.
    */
/* #define MFL_ALLOW_REDECLARATIONS	*/


    /* Define this to enable some strict enum checks.  Recommended
       and probably won't remain optional.
    */
#define	MFL_STRICT_ENUMS



#define	NIL		((char *)-1)	/* Special non-value */


    /* Hash table sizes */
#define	HSZCUSP		103		/* CUSPs */
#define	HSZBIS		23		/* Builtin statements */
#define	HSZGLOBAL	103		/* Global variables */
#define	HSZPP		103		/* Preprocessor names */
#define	HSZSTRUCT	103		/* Structs */
#define	HSZTYPE		103		/* Types */
#define	HSZTPOOL	103		/* Type pool */


    /* Sizes of various storage types. */
#define	SZINT		4		/* Size for default int */
#define	SZSHORT		2		/* Short int */
#define	SZLONG		4		/* Long int */
#define	SZPTR		4		/* To contain an address */


    /* Operator rankings (priorities); low values are low ranks */
#define	ORCOMMA		1		/* Expression sequencer */
#define	ORASSIGN	2		/* Assignment and assign-ops */
#define	ORCOND		3		/* Conditional expression */
#define	ORLOR		4		/* Logical OR */
#define	ORLAND		5		/* Logical AND */
#define	ORBITOR		6		/* Bitwise inclusive OR */
#define	ORBITXOR	7		/* Bitwise exclusive OR */
#define	ORBITAND	8		/* Bitwise AND */
#define	ORRELEQ		9		/* Relational equality */
#define	ORRELREL	10		/* Relational relative */
#define	ORSHIFT		11		/* Shifting */
#define	ORADD		12		/* Addition */
#define	ORMULT		13		/* Multiplications */
#define	ORPREFIX	14		/* All prefix operators */
#define	ORPOSTFIX	15		/* All postfix operators */
#define	ORMEMBER	16		/* Structure membership */


    /* Type flags */
#define	TF_MOD		0x000f		/* Mask for modifier code */
#define	 TMSTRUCT	 0x1		/*  struct */
#define	 TMUNION	 0x2		/*  union */
#define	 TMENUM	 	 0x3		/*  enum */
#define	 TMSHORT	 0x8		/*  short */
#define	 TMLONG	 	 0x9		/*  long */
#define	 TMUNSIGNED	 0xa		/*  unsigned */
#define	 TMSIGNED	 0xb		/*  signed */
#define	 TMTYPEDEF	 0xf		/*  typedef */

#define	TF_DEFINED	0x0010		/* If defined */
#define	TF_BUILTIN	0x0020		/* Builtin... */
#define	TF_AGGR		0x0040		/* Aggregate */


    /* Value flags */
#define	VF_CONST	0x0001		/* Constant value */

    /* Valid lvalue.  This is something that can be assigned into.  */
#define	VF_LVALUE	0x0002

    /* Rvalue is present- meaning that the value can be taken. */
#define	VF_RVALUE	0x0004



    /* Special indicator values for VALPTR type offset value */
#define	PO_UNDEF	-1		/* Pointer not defined */
#define	PO_SIMPLE	-2		/* Points to the v_rvalue element */
#define	PO_PRIVATE	-3		/* Has some private meaning for
					   this type.
					*/
#define	PO_REFSTR	-4		/* Pointer is a REFSTR */
#define	PO_FSV		-5		/* Pointer is to FSV
					   (free-standing value) */


    /* Type Classes */
typedef enum {
	TC_VOID,			/* The NULL type */
	TC_UINT,			/* Unsigned integer */
	TC_INT,				/* Signed integer */
	TC_FLOAT,			/* Floating point */
	TC_ENUM,			/* Enumeration */
	TC_UNION,			/* Union */
	TC_STRUCT,			/* Structure */
	TC_DIM,				/* Dimensioned (array) */
	TC_PTR,				/* Pointer to something */
	TC_FUNCT,			/* Function returning something */
	TC_STRING,			/* A string (REFSTR) */
}  TYPECLASS;


    /* Parse node types */
typedef enum {
	PN_DECL,			/* Declaration of a variable (etc) */
	PN_EXPR,			/* Expression */
	PN_PTR,				/* Pointer type declaration */
	PN_DIM,				/* Dimension type declaration */
	PN_FUNCDEF,			/* Function definition of some sort */
	PN_FUNCREF,			/* Function reference */
	PN_OPR,				/* Operator */
	PN_CAST,			/* Typecast */
	PN_INTEGER,			/* Integer term */
	PN_FLOAT,			/* Floating point term */
	PN_SYM,				/* Symbol term */
	PN_LITSTR,			/* String literal */
	PN_TYPEREF,			/* Type reference */
	PN_TYPEDEF,			/* Type definition */
	PN_BIS,				/* Builtin statement */
	PN_LSEQ,			/* Left-sequential placeholder */
	PN_ARG,				/* An actual argument */

#ifdef	SIEVE
	PN_SVBIS,			/* Sieve built-in statement */
#endif	/* SIEVE */

}  PNTYPE;



    /* Opaque dynamic info.  Contains a pointer to something and
       optionally how to delete it.  NOTE: if there is no deletion
       routine, data is assumed to be non-deletable.
    */
typedef struct {
    void	*od_P;			/* The info */
    BOOL	(*od_del)();		/* How to delete it. */
}  ODINFO;

    /* Built-in statement info */
typedef struct {
    char	*bis_symP;		/* Symbol */
    STSCODE	(*bis_parse)();		/* Parse routine */
    STSCODE	(*bis_exec)();		/* Execution routine */
    BOOL	bis_stmtF;		/* If it can be used as a statement */
    BOOL	bis_termF;		/* If it can be used as a term */
}  BISINFO;


    /* Type definition.  A type is either a fundamental type (e.g.
       int, char) or it modifies some other thing.  ty_nextP is a
       pointer to the thing it is built from-- this is usually another
       TYPE but it could be something else (e.g. an aggregate).
       Types are the same if they have the same attributes and
       the same ty_nextP reference.  Array types may be retyped if
       their dimension changes (e.g. in the case of dynamic-bound
       arrays).
    */

typedef struct {
    int		ty_refC;		/* Reference count */
    UBYTE	ty_class;		/* Type class */
    UBYTE	ty_flags;		/* Type flags */
    int		ty_size;		/* Size */
    int		ty_dim;			/* Dimension, if any */
    void	*ty_nextP;		/* Pointer to next thing,
					   (another type or a definition
					   of the aggregate) */
}  TYPE;



    /* Member of an aggregate */
typedef struct {
    UWORD	am_offs;		/* Offset into aggregate */
    char	*am_nameP;		/* Member name */
    TYPE	*am_typeP;		/* Member type */
}  AGGMEM;


    /* Aggregate type */
typedef struct {
    UWORD	ag_refC;		/* Reference count */
    UWORD	ag_memC;		/* Member count */
    UWORD	ag_size;		/* Aggregate size */
    struct tnsym *ag_tnsymP;		/* Ptr to name entry */
    AGGMEM	*ag_memP;		/* Members */
}   TYPEAG;



    /* Dynamically variable/value storage.  Some simple values
       are kept inside a VALUE struct, but more complex (or larger)
       data is dynamically allocated and pointed to.  Since the
       storage can change (e.g. if an array grows), there's a two
       level storage structure.  One is the fixed part, which never
       changes (and it's therefore safe to point to).  The other is
       the part that is subject to reallocation.  Code that uses
       this storage must do an indirection through this handle.
    */
typedef struct {
    WORD	dv_refC;		/* Reference count */
    size_t	dv_size;		/* The size of the data */
    size_t	dv_max;			/* Maximum data size if non-zero */
    UBYTE	*dv_dataP;		/* Pointer to the data */
}  DVS;


    /* A pointer.  A pointer is one of:

	DVS pointer: a pointer to the data object and a byte
	offset into the data.  This is indicated by a non-negative
	byte offset value.

	Otherwise, a negative byte offset indicates a special storage
	type.  This is one of the PO_XXX values.
    */
typedef struct {
    int		p_byteX;		/* Byte offset or PO_XXX value */
    union {
      DVS	*p_dvsP;		/* Ptr to DVS struct */
      void	*p_genP;		/* Ptr to generic thing */
    }	p_ptr;			/* the pointer */
}  VALPTR;


    /* Value data.  A simple data item can be one of several
       different kinds of thing: this type is the union of those
       things.
    */
typedef union {
    char	d_byte;			/* 1-byte int */
    WORD	d_word;			/* 2-byte int */
    long	d_long;			/* 4-byte int */
    float	d_float;		/* standard float */
    double	d_double;		/* double float */
    VALPTR	d_ptr;			/* A pointer */
}  VALDATA;



    /* A free-standing value (simple value as opposed to VALUE).
       Used when we need to create a pointed-to object that isn't
       tied to anything ("anything" such a symbol).
    */

typedef struct {
    BOOL	fv_ptrF;	/* Set if data is a pointer */
    int		fv_refC;	/* Reference count */
    VALDATA	fv_data;	/* The data.  note that this is not
				   really a VALDATA: it just must be
				   large enough to contain any simple
				   data item, such as might be found
				   in a VALDATA.
				*/
}  FSV;


    /* Value yielded by evaluator.

       A value is either an lvalue or rvalue (colloquially, left
       hand side or right hand side of an assignment).  An lvalue
       is a reference to where the value is stored (e.g. into a
       variable's storage or into an array or structure kept in
       DVS storage).  An rvalue is the actual value.  The lvalue
       is kept as long as possible so that a reference or
       assignment may be done, but is turned into an rvalue when
       the actual data is needed.

       Data about the value is stored here.  This may be one of
       several things:

       The actual value, if VF_RVALUE is set and the value is a
       simple one that can be contained in the VALUE struct;

       A pointer to the actual value, if VF_RVALUE is set and the
       value is not a simple one;

       The location of the value storage, if VF_LVALUE is set
       (this is stored in the ptr member).
    */

typedef struct {
    UBYTE	v_flags;		/* Flags */
    TYPE	*v_typeP;		/* Type of value */
    VALDATA	v_data;
  } VALUE;



    /* Operator info */
typedef struct {
    char	*op_str;		/* String for operator */
    char	op_prio;		/* Evaluation priority */
    STSCODE	(*op_eval)();		/* Evaluation routine */
    STSCODE	(*op_preeval)(); 	/* Prefix operation eval routine */
    STSCODE	(*op_posteval)();	/* Postfix operation eval routine */
    double	(*op_feval)();		/* Floating point evaluation */
    long	(*op_ieval)();		/* Integer evaluation */
    STSCODE	(*op_seval)();		/* String evaluation */
}  OPINFO;



    /* SYMTAB entry: built-in statement symbol table */
typedef struct {
    SYMSTR	bss_sym;		/* Symbol prolog */
    WORD	bss_num;		/* Statement index */
}  BISSYM;


   /* SYMTAB ENTRY: type description symbol table */
typedef struct {
    SYMENT	ts_sym;			/* Symbol prolog */
    TYPE	ts_type;		/* Type description */
}  TYPSYM;


/* SYMTAB ENTRY: type-name and struct name symbol tables */
typedef struct tnsym {
    SYMSTR	tn_sym;			/* Symbol prolog */
    UBYTE	tn_flags;		/* Flag bits */
    TYPE	*tn_typeP;		/* Type definition */
}  TNSYM;
			      

    /* SYMTAB ENTRY: the variables symbol table */
typedef struct {
    SYMSTR	vs_sym;			/* Symbol prolog */
    VALUE	vs_value;		/* Value for variable */
}  VSYM;



    /* Parse tree node */
typedef struct ptnode {
    UBYTE	   pt_type;		/* Node type */
    struct ptnode *pt_leftP;		/* Ptr to left node.  Left node is
    					   typically the continuation;
					   or the left side of a binary
					   operation. */
    struct ptnode *pt_rightP;		/* Ptr to right node. */
    union {
      int	 pv_int;	/* Integer data */
      long	 pv_long;	/* Long data */
      float	 pv_float;	/* Floating data */
          }	   pt_data;
    ODINFO	pt_info;		/* Any attached special info */
    ODINFO	pt_xinfo;		/* Any extended info */
  } PTNODE;



    /* About a cusp -- commonly used service program */
typedef struct cusp {
    UBYTE	*cp_pathP;		/* Path to the service program */
}  CUSP;


/* Declarations for functions contained in this module: */


#endif	/* H_MFL */
