/* mvmf-locking.h -- definitions for mvmf-locking module.

	Copyright 2003,2004,2005,2006,2007
		by Mark E. Mallett, MV Communications, Inc.

	See the "LICENSE" file for terms.

Functions dealing with file locking for mvmf.

*/


#ifndef	H_MVMF_LOCKING		/* For multiple inclusions */
#define	H_MVMF_LOCKING

    /* Parameters for dotlocking... */
#define	DOTLOCK_TRYEVERY	1	/* Sleep this long between tries */
#define	DOTLOCK_MAXTRIES	30	/* Try this many times */
#define	DOTLOCK_MAXAGE		(2*3600)  /* Delete troublesome dotlock
					     older than this many seconds
					  */

    /* Maximum number of lock types for any given lock */
#define	LOCKMAX		5

    /* Various locking strategies (not all may be available) */
typedef enum {
    LCK_DOTLOCK,		/* dotlocking */
    LCK_FCNTL,			/* fcntl locks */
    LCK_FLOCK,			/* flock */
    LCK_LOCKF			/* lockf library call */
}  LOCKTYPE;


    /* Control for a single lock */
typedef struct {
    BOOL	lk_lockedF;	/* If lock was taken */
    LOCKTYPE	lk_type;	/* Lock type */
}  LOCKINFO;


    /* Control for a set of locks */
typedef struct {
    int		ls_lockC;	/* Number of lock types wanted */
    LOCKINFO	ls_locks[ LOCKMAX ];  /* The locks */
}  LOCKSET;
    


/* Declarations for functions contained in this module: */

    /* Establish a dotlock lock */
BOOL	lock_dotlock PROTO( (UBYTE *lfnameP) );

#ifdef	HAVE_FCNTL
    /* Establish an fcntl lock */
BOOL	lock_fcntl PROTO( (int lfd) );
#endif	/* HAVE_FCNTL */

#ifdef	HAVE_FLOCK
    /* Establish an flock lock */
BOOL	lock_flock PROTO( (int lfd) );
#endif	/* HAVE_FLOCK */

#ifdef	HAVE_LOCKF
    /* Establish a lockf lock */
BOOL	lock_lockf PROTO( (int lfd) );
#endif	/* HAVE_LOCKF */

    /* Establish lock(s) via a lockset */
BOOL	lock_lockset PROTO( (LOCKSET *lsP, int lfd, UBYTE *lfnameP) );

    /* Parse a lock spec into a lockset */
void	lock_parse PROTO( (char *lockspecP, LOCKSET *lsP) );

    /* Release a dotlock lock */
void	unlock_dotlock PROTO( (UBYTE *lfnameP) );


#ifdef	HAVE_FCNTL
    /* Release an fcntl lock */
void	unlock_fcntl PROTO( (int lfd) );
#endif	/* HAVE_FCNTL */

#ifdef	HAVE_FLOCK
    /* Release an flock lock */
void	unlock_flock PROTO( (int lfd) );
#endif	/* HAVE_FLOCK */

#ifdef	HAVE_LOCKF
    /* Release a lockf lock */
void	unlock_lockf PROTO( (int lfd) );
#endif	/* HAVE_LOCKF */

    /* Release lock(s) via a lockset */
void	unlock_lockset PROTO( (LOCKSET *lsP, int lfd, UBYTE *lfnameP) );


#endif	/* H_MVMF_LOCKING */
