/* mvmf.h -- generic mvmf definitions

	Copyright 2003,2004,2005,2006,2007
		by Mark E. Mallett, MV Communications, Inc.

	See the "LICENSE" file for terms.

Provides definitions universal to mvmf modules

*/


#ifndef	H_MVMF			/* For multiple inclusions */
#define	H_MVMF

#include <mml/mml.h>		/* Everyone needs this, just do it here */

/* mml is autoconf'd too, and all mml headers include autoconf defs.
   Some of mvmf's autoconf defs will conflict.  I don't really know how
   to get around this other than by this hack.  Maybe I'll learn more
   one day.
*/
#undef PACKAGE
#undef PACKAGE_BUGREPORT
#undef PACKAGE_NAME
#undef PACKAGE_STRING
#undef PACKAGE_TARNAME
#undef PACKAGE_VERSION
#undef VERSION

#include "mvmf-autoconf.h"	/* stuff from configure. */


/***********************************************************************
 *                                                                     *
 *   Configuration items that you may want to look at and/or change    *
 *                                                                     *
 ***********************************************************************/


    /* Stuff that's in MVMF_LIBDIR and MVMF_SYSDIR, the user-level
       and system-level shared/config data directories, respectively.
       Those symbols and paths are generated by the configure process
       and are defined in the configure-generated header file.
    */


/* ***** Files within MVMF_LIBDIR ***** */

    /* Information about dnsbl stuff. */
#define	LIBDNSBL	MVMF_LIBDIR "/dnsbl.conf"



/* ***** Files within MVMF_SYSDIR ***** */

    /* Where general binaries (executable) are */
#define	MVMF_SYSBIN	MVMF_SYSDIR "/bin"


    /* port number for mvmcas */
#define	MVMCAS_PORT	2830

       

/***********************************************************************
 *                                                                     *
 *   Configuration items that you are less likely to need to change    *
 *                                                                     *
 ***********************************************************************/

    /* Maximum email header line length we'll deal with */
#define	MAX_HEADER_LENGTH	70000

    /* Default directory for temporary files, and environment variable
       name that will override this default
    */
#define	TMPDIR		"/var/tmp"
#define	TMPDIRENV	"TMPDIR"



/***********************************************************************
 *                                                                     *
 *    Beyond here are things that you probably don't want to touch.    *
 *                                                                     *
 ***********************************************************************/

    /* Handy multiplers */
#define	K	1024		/* You know, K */
#define	Ki	1000		/* and Ki for 1000 */
#define	M	(1024L * 1024)	/* A mega */
#define	Mi	(1000L * 1000)	/* A million */

    /* Set this to something unparsable to catch things we want to do. */
#define	TODO	/* blah blah */

    /* Some exit codes */
#define	MEX_OK		0	/* Success */
#define	MEX_HARD	1	/* Hard error */
#define	MEX_SOFT	2	/* Soft error */
#define	MEX_OKCONT	3	/* Success, but continue (pass-through) */
#define	MEX_FORKFAIL	-101	/* Some failure in child when making a fork */

    /* Some QMAIL exit codes. */
#define	QMEX_OK		0	/* Success */
#define	QMEX_STOP	99	/* Success, proceeed no farther */
#define	QMEX_HARD	100	/* Hard error */
#define	QMEX_SOFT	111	/* Soft error */



    /* Status codes passed around by some functions.  These are
       used in different ways by different functions; not all
       codes have meaning to all applications, etc.  And some
       functions/applications overload the use of some of these
       codes with meanings that aren't necessarily suggested by
       the names.  That's just a tradeoff to avoid the explosion
       of SCxxx names.
    */
typedef enum {
    SCNOTHING,			/* No specific status code set */
    SCOK,			/* OK */
    SCERR,			/* Some error */
    SCNOTDONE,			/* Not fatal, but aborted */
    SCDONE,			/* Things are done. */
    SCBREAK,			/* Break from loop iteration */
    SCCONTINUE,			/* Continue to next loop iteration */
    SCRETURN,			/* Return from a function */
    SCACCEPT,			/* Some action or thing is accepted */
    SCREJECT,			/* Some action or thing is rejected */
}  STSCODE;


    /* Mail directory layout types */
typedef enum {
	/* As yet unknown */
    ML_UNK,


	/* Standard filesystem usage.  A mailbox can appear anywhere
	   in the filesystem as indicated by the path portion of
	   the "folder" name.
	*/
    ML_FS,

	/* Maildir++ .  All folders exist within the Maildir.  Encoding
	   of folder names is probably hardcoded.
	*/
    ML_MDPP,
}  MLTYPE;


/* Declarations for functions contained in this module: */

#endif	/* H_MVMF */
