/* mvmfmisc.h -- some general stuff.

	Copyright 2005,2006,2007
		by Mark E. Mallett, MV Communications, Inc.

	See the "LICENSE" file for terms.

Some general things that might be useful in mvmf.

*/


#ifndef	H_MML_MVMFMISC		/* For multiple inclusions */
#define	H_MML_MVMFMISC


    /* Some things have simple attributes that get handled in
       common ways at a deep level (perhaps guided by logic at a
       higher level).  These may be of the following types.
    */

typedef struct {			/* Integer attribute */
    BOOL	ai_setF;		/* Whether it's been set */
    long	ai_val;			/* The value */
}  ATTR_INT;

typedef struct {			/* String attribute */
    BOOL	as_allocF;		/* If value was alloc'd */
    UBYTE	*as_strP;		/* String value */
}  ATTR_STR;



/* Declarations for functions contained in this module: */

    /* Clear an integer attribute */
void	attr_int_clear PROTO( (ATTR_INT *aiP) );

    /* Set an integer attribute */
void	attr_int_set PROTO( (ATTR_INT *aiP, long val) );

    /* Clear a string attribute */
void	attr_str_clear PROTO( (ATTR_STR *asP) );

    /* Set a string attribute */
BOOL	attr_str_set PROTO( (ATTR_STR *asP, UBYTE *strP, BOOL copyF) );

    /* Mark a string attribute as not set */
void	attr_str_unset PROTO( (ATTR_STR *asP) );

    /* Decode base64 from counted string into BSTR */
BOOL	b64_decode_bstr_str_n PROTO( (BSTR *dstP, UBYTE *srcP, int srcL) );

    /* Decode base64 from counted string into tmpstr */
UBYTE	*b64_decode_tmpstr_str_n PROTO( (UBYTE *srcP, int srcL, int *dstLP) );

    /* Access the DNS handle */
MML_DNS	*dns_handle PROTO( (void) );

    /* Rotate a filename through n suffixes, optionally install new file */
BOOL	file_rotate PROTO( (char *fnameP, char *newfnameP, int verC) );

    /* Decode IP address from string */
BOOL	ip_decode_str PROTO( (UBYTE *srcP, in_addr_t *ipP) );

    /* Add a directory name to a path */
BOOL	path_dir_add PROTO( (LL_HEAD *pathP, ESTR *dirnameP) );

    /* Exec a program, opening up standard file descriptors */
pid_t	pexec_std PROTO( (char *pathP, char **argPP,
			  int *coutfdP, int *cinfdP, int *cerrfdP) );

    /* Finish a program previously exec'd via pexec_std() */
int	pexec_std_end PROTO( (pid_t pid, int coutfd, int cinfd, int cerrfd) );

    /* Sanitize the process (e.g. after forking) */
BOOL	sanitize_process PROTO( (void) );

    /* Get whitespace-separated token from string, handling errors */
BOOL	ws_token PROTO( (UBYTE **bufPP, UBYTE *tokP, int tokL) );

    /* Get token, interpret as int */
BOOL	ws_token_int PROTO( (UBYTE **bufPP, long *lvP) );

#endif	/* H_MML_MVMFMISC */
