/* mvmtr.h -- definitions for the mvmtr application

	Copyright 2003,2004,2005,2006,2007
		by Mark E. Mallett, MV Communications, Inc.

	See the "LICENSE" file for terms.

General definitions for mvmtr

*/


#ifndef	H_MVMTR		/* For multiple inclusions */
#define	H_MVMTR


/***********************************************************************
 *                                                                     *
 *   Configuration items that you may want to look at and/or change    *
 *                                                                     *
 ***********************************************************************/

	/* Where the mvmta control files are. */
#ifdef	QMAIL
    /* You might want the mvmta stuff here: */
/* # define MVMTRDIR	QMAILDIR "/control"	*/
    /* But I would suggest letting the definition below take effect. */
#endif	/* QMAIL */

    /* Default location for mvmtr files if not defined above */
#ifndef	MVMTR_SYSDIR
# define MVMTR_SYSDIR	MVMF_SYSDIR "/mvmtr"
#endif	/* MVMTR_SYSDIR */


    /* System control file: the mfl-based config file (rc file).  Note
       that this will be relative to the installed system-wide
       include path.  Since MVMTR_SYSDIR will be in the include
       path, we just use the .mfl filename here.
    */
#define	MVMTR_SYSRC	  "mvmtr.mfl"

    /* Environment variable that you can set to override MVMTR_SYSRC. */
#define	MVMTR_RCENV	"MVMTR_RC"

    /* System control file: the address map file */
#define	MVMTRSYS_ADDRMAP  MVMTR_SYSDIR "/addrmap.cdb"




/***********************************************************************
 *                                                                     *
 *   Configuration items that you are less likely to need to change    *
 *                                                                     *
 ***********************************************************************/




/***********************************************************************
 *                                                                     *
 *    Beyond here are things that you probably don't want to touch.    *
 *                                                                     *
 ***********************************************************************/


    /* Decision values for mail.
       Positive values are final.
       Negate the value if it is tentative and can be overriden.
       0 is don't know (this must be 0!)
       Would like to use an enum here, but it's handy to have a
         signed value (per above), which you can't do reliably with enum.
    */
#define	MDV_DUNNO	0		/* Don't know yet */
#define	MDV_ACCEPT	1		/* Accept */
#define	MDV_PASS	2		/* Accept + pass some sender checks */
#define	MDV_REJECT	3		/* Reject */
#define	MDV_KNOWN	4		/* Known, but not accepted/rejected */
#define	MDV_UNKNOWN	5		/* Unknown, e.g. an outside address */
#define	MDV_ERROR	6		/* Some error happened. */
#define	MDV_MAXVAL	MDV_ERROR	/* Max legit value */


    /* Session delivery policies.  These are persistent policies that
       can be set during a session to affect all later attempts
       during the session.
    */
typedef enum {
	SDP_ACCEPT,			/* SOP, acceptance */
	SDP_DEFER,			/* Defer (4xx) */
	SDP_REJECT			/* Reject (5xx) */
}  SDPOLICY;


    /* Operating modes */
typedef enum {
	OM_NONE,			/* None specified */
	OM_NORMAL,			/* Normal */
	OM_VERSION,			/* Print version info */
	OM_CONTROLS			/* Print controls */
} OPMODE;


    /* SMTP mail client block types */
typedef	enum {
	MCBLOCK_NONE,			/* Not blocked */
	MCBLOCK_CONN,			/* Block the connection */
	MCBLOCK_GREET,			/* Block by greeting & exiting */
	MCBLOCK_SMTP			/* Block by stalling in SMTP */
}  MCBLOCK;

    /* Encapsulates info for accessing a cdb file */
typedef struct {
    int		cdb_fd;		/* The fd */
    struct cdb	cdb_cdb;	/* The cdb struct */
}  CDBWRAP;


/* Declarations for functions contained in this module: */

#endif	/* H_MVMTR */
