/* omm.h -- "open mail message" management stuff

	Copyright 2003,2004,2005,2006,2007
		by Mark E. Mallett, MV Communications, Inc.

	See the "LICENSE" file for terms.

"omm" provides a semi-opaque interface to manipulate parts of the
"open mail message" -- which is the primary message that we are
dealing with in the delivery environment.

It probably would have been better, especially now that some of
these functions have been forked into more generic message
handling stuff, to call it something other than the "open"
message, e.g. the "active" message or the "incoming" message or
something.  Maybe that's another change in the future..

*/


#ifndef	H_OMM			/* For multiple inclusions */
#define	H_OMM


    /* Limits on how much of the body we'll attempt to cache.  No real
       reason to use powers of two, it's just an old geezer thing.
    */
#define	OMM_BODYCACHE_MIN	(2L * 1024)
#define	OMM_BODYCACHE_MAX	(250L * 1024)

    /* Always attempt to cache this much of any body part, even if the
       total limit has been reached.
    */
#define	OMM_BODYCACHE_GTY	(1L * 1024)


    /* Structure to track an open mail message. */
typedef struct omm {
    MMSG	*om_mmsgP;	/* Ptr to the message object */
    MSGPART	*om_curmpP;	/* Ptr to currently selected message part */
    LL_HEAD	om_actions;	/* Actions that have been done */
    char	**om_ripPP;	/* Ptr to list of responsible IPs (of
				   systems that we know to have sent
				   the message)
				*/
    int		om_ripC;	/* # of entries in om_ripp */
    BSTR	*om_errsP;	/* Any error messages that should be
				   inserted into the body
				*/
    char	*om_aesenderP;	/* Ascertained/approximated envelope sender */
    BOOL	om_storingF;	/* TRUE if a storing operation is happening */
    BOOL	om_fwdingF;	/* TRUE if a forwarding operation is happening */
    BOOL	om_pipingF;	/* TRUE if a pipe operation is happening */

    BOOL	om_ikeepF;	/* Implicit "keep" */
    BOOL	om_wantbounceF;	/* TRUE to request a bounce */
    BOOL	om_wantrejectF;	/* TRUE to request a SMTP refusal. */

    BOOL	om_rejectF;	/* TRUE if a "reject" has been done */
    BOOL	om_bouncedF;	/* TRUE if a bounce has been done. */

    int		om_deliveryC;	/* # of successful deliveries */
    int		om_softfailC;	/* # of "soft" delivery failures  */
    int		om_hardfailC;	/* # of hard delivery failures */
}  OMM;


/* Declarations for functions contained in this module: */

    /* Add an action to the actions list */
BOOL	omm_action_add PROTO( (OMM *ommP, ESTR *acnameP, ESTR *acvalP) );

    /* Check for an action on the actions list */
BOOL	omm_action_check PROTO( (OMM *ommP, ESTR *acnameP, ESTR *acvalP) );

    /* Handle a delivery error */
void	omm_delivery_error PROTO( (OMM *ommP, int err, char *dtypeP) );

    /* Free/finish/release an omm */
BOOL	omm_free PROTO( (OMM *ommP) );

    /* Setup an omm from an already-open file */
void	*omm_openfp PROTO( (FILE *fP, long bodycache) );

    /* Setup an omm from an open MML_IN input stack */
void	*omm_openip PROTO( (MML_IN *inP, long bodycache) );

    /* Add an entry to the list of responsible sender IPs */
BOOL	omm_rip_add PROTO( (OMM *ommP, ESTR *ripP) );

    /* Write any cached error text to a file stream */
BOOL	omm_write_errs PROTO( (OMM *ommP, FILE *ofP, BOOL bracketF ) );

#endif	/* H_OMM */
