/* qmail-queue-if.h -- definitions for mvmf's "qmail-queue-if" module

	Copyright 2005,2006,2007
		by Mark E. Mallett, MV Communications, Inc.

	See the "LICENSE" file for terms.

Definitions and functions for interfacing to qmail's qmail-queue
program.

*/


#ifdef	QMAIL		/* Only enabled if qmail stuff is. */

#ifndef	H_QMAIL_QUEUE_IF		/* For multiple inclusions */
#define	H_QMAIL_QUEUE_IF

    /* Status codes returned by the qmail-queue program. */
#define	QQS_OK		0	/* OK, message accepted */
#define	QQS_ELONGSTRING	11	/* Envelope string too long */
#define	QQS_REJECTED	31	/* Recipients rejected */
#define	QQS_MEMORY	51	/* Out of memory */
#define	QQS_TIMEOUT	52	/* Timeout */
#define	QQS_WRITE	53	/* Write error */
#define	QQS_ACCEPT	54	/* Error reading data we're sending it */
#define	QQS_READ	55	/* Error reading config file */
#define	QQS_NET_OUT	56	/* Can't initiate some net connection */
#define	QQS_QMAILDIR	61	/* Some problem with the qmail directory */
#define	QQS_QUEUE_DIR	62	/* Problem with the queue directory */
#define	QQS_QUEUE_PID	63	/* Problem with queue/pid */
#define	QQS_QUEUE_MESS	64	/* Problem with queue/mess */
#define	QQS_QUEUE_INTD	65	/* Problem with queue/intd */
#define	QQS_QUEUE_TODO	66	/* Problem with queue/todo */
#define	QQS_TMP_REJECT	71	/* Recipients temporarily rejected */
#define	QQS_MSCON_TIMEOUT 72	/* Connection to mailserver timeout */
#define	QQS_MSCON_REJECT 73	/* Connection to mailserver rejected */
#define	QQS_MSCON_ABORT	74	/* Connection to mailserver aborted */
#define	QQS_ERR		81	/* Internal error */
#define	QQS_ENVERR	91	/* Envelope format error */

#define	QQS_EXEC	120	/* Internal code: can't exec qmail-queue.
				   Value chosen by imitating djb code but
				   it could be anything not used by
				   qmail-queue.
				*/


    /* A handle for interfacing to qmail-queue */
typedef struct {
    FILE	*qq_msgfP;		/* FILE to write message to */
    FILE	*qq_envfP;		/* FILE to write envelope to */
    pid_t	qq_pid;			/* qmail-queue process ID */
    BOOL	qq_errF;		/* If any error happened */
}  QMQIF;


/* Declarations for functions contained in this module: */

    /* Close a pipe to qmail-queue */
STSCODE	qmail_queue_close PROTO( (QMQIF *qqP, BOOL keepF) );

    /* Open a pipe to qmail-queue */
QMQIF	*qmail_queue_open PROTO( (void) );

    /* Write a message to qmail-queue */
BOOL	qmail_queue_write_message PROTO( (QMQIF *qqP, MSGPART *mpP) );

    /* Write a recipient address to qmail-queue */
BOOL	qmail_queue_write_recipient PROTO( (QMQIF *qqP, UBYTE *addrP) );

    /* Write the sender address to qmail-queue */
BOOL	qmail_queue_write_sender PROTO( (QMQIF *qqP, UBYTE *addrP) );

    /* Write some message text to qmail-queue */
BOOL	qmail_queue_write_text PROTO( (QMQIF *qqP, UBYTE *textP ) );

#endif	/* H_QMAIL_QUEUE_IF */

#endif	/* QMAIL */
