/* xmtp.h -- definitions for mvmf's "xmtp" module

	Copyright 2003,2004,2005,2006,2007
		by Mark E. Mallett, MV Communications, Inc.

	See the "LICENSE" file for terms.

This contains some primitive functions to help engage in an xMTP
dialog (SMTP, LMTP, SUBMISSION, and what-have-you).

*/


#ifndef	H_XMTP		/* For multiple inclusions */
#define	H_XMTP

    /* Some limits */
#define	XMTP_CMD_MAX	512	/* RFC2821: command line max */
#define	XMTP_DOMAIN_MAX	255	/* RFC2821: domain name max */
#define	XMTP_LOCAL_MAX	64	/* RFC2821: address local-part max */
#define	XMTP_PATH_MAX	255	/* RFC2821: length of a path */

#define	XMTP_REPMAX	10000	/* Maximum reply buffer size */
#define	XMTP_REPCHUNK	1000	/* Expansion chunk for reply buffer */

    /* How command lines are terminated (in the context of getbtrec().
       Lines can end with CRLF or LF, and we flush the rest of
       the line when the buffer is full.  (Application must detect this.)
    */
#define	XMTP_CMD_TXR	((TXR_END_CRLF)|(TXR_END_LF))

    /* How non-command (i.e., text) lines are terminated (in the context
       of getbtrec().  Lines can end with CRLF or LF, and we stop
       reading when the buffer is full.
    */
#define	XMTP_DATA_TXR	((TXR_END_CRLF)|(TXR_END_LF)|TXR_END_FULL)


    /* Things returned by xmtp_parse_xxx functions */
typedef enum {
    XPR_OK,			/* OK result */
    XPR_OFL,			/* Buffer overflow */
    XPR_SYNTAX,			/* Syntax error */
    XPR_ERR,			/* Some internal error */
}  XPRSLT;



    /* Control information for an xMTP dialog */
typedef struct {
	/* Note that input and output may be the same fd;
	   neither is necessarily buffered.
	*/
    BIO		mt_in;		/* Buffered I/O for input */
    BIO		mt_out;		/* Buffered I/O for output */
    MML_IN	*mt_inP;	/* Input stream */
    BSTR	*mt_cmbsP;	/* BSTR for received commands */
    UBYTE	*mt_cmdP;	/* Ptr to command keyword (in mt_cmdsP) */
    int		mt_cmdL;	/* Length of command keyword */
    UBYTE	*mt_cmdtailP;	/* Ptr to rest of command */

    BOOL	mt_ecodesF;	/* Use extended result codes */

}  XMTP;



    /* Structure for a RFC2821/2 style "mailbox" */
typedef struct {
    UBYTE	*mb_pathP;		/* Ptr to an allocated path buffer */
    BOOL	mb_adlF;		/* If there was an a-d-l component */
    int		mb_atX;			/* Index of '@' in path */
    int		mb_len;			/* Length of the path. */
}  XMTPMBX;



/* Declarations for functions contained in this module: */

    /* Input stream handler routine(s): XMTP dotstuffed stream */
int	in_h_xmtp_dss_getc PROTO( ( MML_INS *instrP ) );

    /* Push a dot-stuffed input stream onto an input stack */
BOOL	in_stream_push_xmtp_dss PROTO( (MML_IN *inP, BIO *bioP,
					char *nameP) );


    /* Get and prepare a command from xMTP peer */
STSCODE	xmtp_getcmd PROTO( (XMTP *xmtP) );

    /* Input a single line of text from xMTP peer */
STSCODE	xmtp_getline PROTO( (XMTP *xmtP, BSTR *bsP) );

    /* Get a message from an xMTP peer (a la DATA) */
STSCODE	xmtp_getmsg PROTO( (XMTP *xmtP, OMM **ommPP, ULONG maxsize) );

    /* Init control structure for client dialog */
XMTP	*xmtp_init PROTO( (int fdin, int fdout) );


    /* Allocate an XMTPMBX struct */
XMTPMBX	*xmtp_mbox_alloc PROTO( (void) );

    /* Free a previously allocated XMTPMBX struct */
void	xmtp_mbox_dealloc PROTO( (XMTPMBX *mbxP) );

    /* Parse a source route (a-d-l) from e.g. an xMTP command buffer */
XPRSLT	xmtp_parse_adl PROTO( (UBYTE **cmdPP, UBYTE *pathP, int pathL) );

    /* Parse an "atom" from e.g. an xMTP command buffer */
XPRSLT	xmtp_parse_atom PROTO( (UBYTE **cmdPP, UBYTE *pathP, int pathL) );

    /* Parse a domain part from e.g. an xMTP command buffer */
XPRSLT	xmtp_parse_domain PROTO( (UBYTE **cmdPP, UBYTE *pathP, int pathL) );

    /* Parse a localpart from e.g. an xMTP command buffer */
XPRSLT	xmtp_parse_localpart PROTO( (UBYTE **cmdPP, UBYTE *pathP, int pathL) );

    /* Parse a mailbox string from e.g. an xMTP command */
XPRSLT	xmtp_parse_mailbox PROTO( (UBYTE **cmdPP, XMTPMBX **mbxPP) );

    /* Parse a path string from e.g. an xMTP command */
XPRSLT	xmtp_parse_path PROTO( (UBYTE **cmdPP, XMTPMBX **mbxPP) );

    /* Parse a quoted string from e.g. an xMTP command buffer */
XPRSLT	xmtp_parse_quoted_string PROTO( (UBYTE **cmdPP, UBYTE *pathP, int pathL) );

    /* Send an xMTP reply */
BOOL	xmtp_reply PROTO( (XMTP *xmtP, BOOL nowF, int rcode, char *ecodeP,
			   char *fmtP, ...) );


#endif	/* H_XMTP */
