/* mfl-bif-misc.c -- MFL functions: misc

	Copyright 2003,2004,2005,2006,2007
		by Mark E. Mallett, MV Communications, Inc.

	See the "LICENSE" file for terms.

MFL built-in-functions of random nature (either in a class that doesn't
really warrant a special module, or that just have no better place to
be at the moment.)

*/

#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include <mml/mml.h>

#include <mml/mml-alloc.h>
#include <mml/mml-ll.h>
#include <mml/mml-str.h>
#include <mml/mml-in.h>
#include <mml/mml-iopf.h>
#include <mml/mml-dns.h>
#include <mml/mml-symtab.h>

#include "mvmf.h"
#include "controls.h"
#include "mvmfmisc.h"
#include "mfl.h"
#include "mfl-pp.h"
#include "mfl-token.h"
#include "mfl-type.h"
#include "mfl-parse.h"
#include "mfl-eval.h"
#include "mfl-func.h"
#include "mfl-misc.h"

#include "data.h"
#include "msgpart.h"
#include "omm.h"
#include "msgstore.h"
#include "errwarn.h"

/* Local definitions */


/* External data referenced */


extern	BOOL	Mfl_traceF;		/* Whether tracing */


/* External routines used */


/* Local routines and forward references */


    /* Built-in functions contained herein */
static	STSCODE	bif_env_cwd();
static	STSCODE	bif_env_cwd_set();
static	STSCODE	bif_env_variable();
static	STSCODE	bif_env_variable_assign();
static	STSCODE	bif_env_homedir();
static	STSCODE	bif_ip_rev();
static	STSCODE	bif_mfl_hookdir_add();
static	STSCODE	bif_mfl_incdir_add();
static	STSCODE	bif_mfl_parse();
static	STSCODE	bif_mfl_protect_admin();
static	STSCODE	bif_mfl_run_fname();
static	STSCODE	bif_mfl_run_string();
static	STSCODE	bif_mfl_trace();
static	STSCODE	bif_msns_attr_int_set();
static	STSCODE	bif_msns_attr_string_set();
static	STSCODE	bif_msns_define();
static	STSCODE	bif_msns_tailhook();
static	STSCODE	bif_msst_attr_int_set();
static	STSCODE	bif_msst_attr_string_set();

/* Local data publicly available */


    /* Table of built-in functions to define */
	BIFINF	Bifinf__[] = {
/*  function name               handler function		*/
{  "$env_cwd",			bif_env_cwd		},
{  "$env_cwd_set",		bif_env_cwd_set		},
{  "$env_homedir",		bif_env_homedir		},
{  "$env_variable",		bif_env_variable	},
{  "$env_variable_assign",	bif_env_variable_assign	},
{  "$ip_rev",			bif_ip_rev		},
{  "$mfl_hookdir_add",		bif_mfl_hookdir_add	},
{  "$mfl_incdir_add",		bif_mfl_incdir_add	},
{  "$mfl_parse",		bif_mfl_parse		},
{  "$mfl_protect_admin",	bif_mfl_protect_admin	},
{  "$mfl_run_fname",		bif_mfl_run_fname	},
{  "$mfl_run_string",		bif_mfl_run_string	},
{  "$mfl_trace",		bif_mfl_trace		},
{  "$msns_attr_int_set",	bif_msns_attr_int_set	},
{  "$msns_attr_string_set",	bif_msns_attr_string_set },
{  "$msns_define",		bif_msns_define		},
{  "$msns_tailhook",		bif_msns_tailhook	},
{  "$msst_attr_int_set",	bif_msst_attr_int_set	},
{  "$msst_attr_string_set",	bif_msst_attr_string_set },

{  NULL,		NULL			}
		};


/* Private data */


/*

*//* bif__init(args)

	Initialize the misc built-in-functions module

Accepts :

	<nothing>

Returns :

	<value>		TRUE if OK
			FALSE if not.

Notes :

	Called as part of bif module(s) initialization to do any
	extra setup required after functions have been registered.

*/

BOOL
bif__init NOARGLIST
{
    return ( TRUE );
}


/***********************************************************************
 *                                                                     *
 *                     Internal built-in functions                     *
 *                                                                     *
 ***********************************************************************/

/* Built-in functions are called with standard arguments, like this:

     built-in-function( fnameP, rvalP, treeP )

	fnameP is the function name that was referenced
	rvalP is a VALUE: where the return value is placed.
	treeP is the address of the argument list.

  The function returns a STSCODE status code.

*/


/*

*//* bif_env_cwd -- Return the current working directory

    string $env_cwd()

    Return the path to the current working directory.
    Returns empty string if can't get cwd.

*/

static STSCODE
bif_env_cwd ARGLIST( (fnameP, rvalP, treeP) )
   NFARG( char		*fnameP )	/* Function name */
   NFARG( VALUE		*rvalP )	/* Return value ptr */
    FARG( PTNODE	*treeP )	/* Argument tree */
{
	char		*pathP;
	char		cwd[PATH_MAX +1];

    if ( ( pathP = getcwd( &cwd[0], sizeof(cwd) ) ) == NULL )
	pathP = "";

    val_set_string( rvalP, estr_tmp_str( pathP ) );

    return ( SCOK );
}
/*

*//* bif_env_cwd_set -- Set the current working directory

    BOOL $env_cwd_set( string path )

    Sets the current working directory.

    Returns 1 if successful, 0 if not.

*/

static STSCODE
bif_env_cwd_set ARGLIST( (fnameP, rvalP, treeP) )
   NFARG( char		*fnameP )	/* Function name */
   NFARG( VALUE		*rvalP )	/* Return value ptr */
    FARG( PTNODE	*treeP )	/* Argument tree */
{
	STSCODE		status;
	char		*pathP;		/* The desired path string */
	VALUE		pathval;	/* Value for path */

    pathP = func_treearg_str( fnameP, treeP, 0, &pathval, PATH_MAX );
    if ( pathP == NULL ) {
	warning( "%s requires a valid string path argument.", fnameP );
	status = SCNOTDONE;
    }
    else {
	if ( chdir( pathP ) )
	    status = SCOK;
	else
	    status = SCDONE;		/* Done, but not changed */

	val_release( &pathval );
    }

    val_set_int( rvalP, 4, ( status == SCOK ) ? 1 : 0 );

    /* Change SCDONE (function call OK, but couldn't chdir) to SCOK */
    if ( status == SCDONE )
	status = SCOK;

    return ( status );
}
/*

*//* bif_env_homedir -- Return the home directory

    string $env_homedir()

    Return the path to the home directory.
    Returns empty string if can't get it.

    Note: assumes global variable MyhomedirP has been set up.

*/

static STSCODE
bif_env_homedir ARGLIST( (fnameP, rvalP, treeP) )
   NFARG( char		*fnameP )	/* Function name */
   NFARG( VALUE		*rvalP )	/* Return value ptr */
    FARG( PTNODE	*treeP )	/* Argument tree */
{
	char		*pathP;

    if ( ( pathP = MyhomedirP ) == NULL )
	pathP = "";

    val_set_string( rvalP, estr_tmp_str( pathP ) );

    return ( SCOK );
}
/*

*//* bif_env_variable -- get environment variable.

    string  *$env_variable( name )

    get environment variable 'name' and return string.

    Returns empty string if name is not found.  See also $env_variable_assign()

*/

static STSCODE
bif_env_variable ARGLIST( (fnameP, rvalP, treeP) )
   NFARG( char		*fnameP )	/* Function name */
   NFARG( VALUE		*rvalP )	/* Return value ptr */
    FARG( PTNODE	*treeP )	/* Argument tree */
{
	STSCODE		status;
	UBYTE		*strP;		/* result of getenv() */
	char		*nameP;		/* Name of environment variable */
	VALUE		nameval;	/* Value for name */

    nameP = func_treearg_str( fnameP, treeP, 0, &nameval, 100 );
    if ( nameP == NULL ) {
	warning( "%s requires a valid string argument.", fnameP );
	strP = "";
	status = SCNOTDONE;
    }
    else {
	strP = (UBYTE *)getenv( nameP );
	if ( strP == NULL )
	    strP = "";			/* Env var doesn't exist, return blank */
	val_release( &nameval );

	status = SCOK;
    }

    val_set_string( rvalP, estr_tmp_str( strP ) );

    return ( status );
}
/*

*//* bif_env_variable_assign -- get environment variable.

    BOOL  $env_variable_assign( destvar, name )

    get environment variable 'name' and assign to variable 'destvar'.

    also see $env_variable();

*/

static STSCODE
bif_env_variable_assign ARGLIST( (fnameP, rvalP, treeP) )
   NFARG( char		*fnameP )	/* Function name */
   NFARG( VALUE		*rvalP )	/* Return value ptr */
    FARG( PTNODE	*treeP )	/* Argument tree */
{
	UBYTE		ch;
	STSCODE		status;
	BYTE		*nameP;
	BYTE		*strP;
	PTNODE		*destargP;	/* Arg: destination variable */
	VALUE		destval, nameval;
	VALUE		tempval;

    /* Get destination reference */
    if ( ( destargP = func_treearg( treeP, 0 ) ) == NULL ) {
	warning( "%s: first arg must be destination variable.", fnameP );
	return ( SCNOTDONE );
    }
    status = eval_tree( destargP, &destval );
    if ( status == SCOK ) {
	/* Get environment variable name */
	nameP = func_treearg_str( fnameP, treeP, 1, &nameval, 100 );
	if ( nameP == NULL ) {
	    warning( "%s: environment variable name required.", fnameP );
	    status = SCNOTDONE;
	}
	else {
	    strP = getenv( nameP );

	    if ( strP != NULL ) {
		/* Set up a temporary VALUE with the string, so we can
		   to the assign to the lvalue from it.
		*/
		status = val_set_string( &tempval, estr_tmp_str( strP ) );
		if ( status == SCOK ) {
		    status = val_assign( &destval, &tempval );
		    if ( status == SCOK )
			val_set_int( rvalP, 4, 1 );
		    val_release( &tempval );
		}
	    }
	    else {
		/* No such environment variable. */
		status = SCNOTDONE;
	    }
	    val_release( &nameval );
	}
	val_release( &destval );
    }

    val_set_int( rvalP, 4, ( status == SCOK ) ? 1 : 0 );

    if ( status == SCNOTDONE )
	status = SCOK;

    return ( status );
}
/*

*//* bif_ip_rev -- reverse the octets in an IP string

    string $ip_rev( string ipP )

    Returns the string that has the octets reversed, e.g.
    for DNSBL or in-addr lookups.

*/

static STSCODE
bif_ip_rev ARGLIST( (fnameP, rvalP, treeP) )
   NFARG( char		*fnameP )	/* Function name */
   NFARG( VALUE		*rvalP )	/* Return value ptr */
    FARG( PTNODE	*treeP )	/* Argument tree */
{
	STSCODE		status;
	char		*ipP;		/* Input IP string */
	char		*revP;		/* String with reversed octets */
	VALUE		ipval;		/* Value for input IP string */
	UBYTE		ip_octets[4];

    ipP = func_treearg_str( fnameP, treeP, 0, &ipval, 20 );
    if ( ipP == NULL ) {
	warning( "%s requires a valid string argument.", fnameP );
	status = SCNOTDONE;
    }
    else {
	/* Get octets of IP address */
	mdns_ip4_iptoba( ntohl( inet_addr( ipP ) ), &ip_octets[0] );

	/* Need a string for the reversed octets.. */
	revP = tmpstr( 21 );
	if ( ( revP == NULL ) ||
	     ( snprintf( revP, 20, "%d.%d.%d.%d",
			     ip_octets[3],
			     ip_octets[2],
			     ip_octets[1],
			     ip_octets[0] ) >= 20 ) )
	    status = SCNOTDONE;
	else {
	    status = SCOK;
	}

	val_release( &ipval );
    }

    /* Give empty string if problem */
    if ( status != SCOK ) {
	revP = "";
	if ( status != SCERR )
	    status = SCOK;
    }

    val_set_string( rvalP, estr_tmp_str( revP ) );

    return ( status );
}
/*

*//* bif_mfl_hookdir_add -- Add a directory to the hook function search list

    BOOL $mfl_hookdir_add( string dir )

    dir is the directory path to add to the hook function search list

    Requires admin capability

*/

static STSCODE
bif_mfl_hookdir_add ARGLIST( (fnameP, rvalP, treeP) )
   NFARG( char		*fnameP )	/* Function name */
   NFARG( VALUE		*rvalP )	/* Return value ptr */
    FARG( PTNODE	*treeP )	/* Argument tree */
{
	STSCODE		status;
	ESTR		*dnameP;	/* The directory name to add */
	VALUE		dnameval;

    /* Get directory name */
    dnameP = func_treearg_estr( fnameP, treeP, 0, &dnameval );
    if ( dnameP == NULL ) {
	warning( "%s requires a directory/path string.", fnameP );
	status = SCNOTDONE;
    }
    else {
	if ( cntrl_int_get( estr_tmp_str( "admin" ), 0 ) == 0 )
	    status = SCNOTDONE;
	else if ( !mfl_func_hook_dir_add( dnameP ) )
	    status = SCNOTDONE;
	else
	    status = SCOK;

	val_release( &dnameval );
    }
    val_set_int( rvalP, 4, ( status == SCOK ) ? 1 : 0 );

    return ( status );
}
/*

*//* bif_mfl_incdir_add -- Add one to the list of include directories (paths)

    BOOL $mfl_incdir_add( string type, string dir )

    type is "system" or "user";

    dir is the directory path to add for the specified type.

*/

static STSCODE
bif_mfl_incdir_add ARGLIST( (fnameP, rvalP, treeP) )
   NFARG( char		*fnameP )	/* Function name */
   NFARG( VALUE		*rvalP )	/* Return value ptr */
    FARG( PTNODE	*treeP )	/* Argument tree */
{
	STSCODE		status;
	MFLINCTYPE	type;		/* Include type */
	ESTR		*typeP;		/* include path type */
	ESTR		*pathP;		/* The path to add */
	VALUE		typeval, pathval;

    typeP = pathP = NULL;
    status = SCOK;

    /* Get type and directory */
    typeP = func_treearg_estr( fnameP, treeP, 0, &typeval );
    if ( typeP == NULL ) {
	    warning( "%s requires an include type (\"system\" or \"user\").",
			fnameP );
	    status = SCNOTDONE;
    }
    else if ( ( pathP = func_treearg_estr( fnameP, treeP, 1, &pathval ) )
			== NULL ) {
	warning( "%s requires a directory/path string.", fnameP );
	status = SCNOTDONE;
    }
    else {
	if ( estr_icmp_str( typeP, (UBYTE *)"system" ) == 0 )
	    type = INC_SYSTEM;
	else if ( estr_icmp_str( typeP, (UBYTE *)"user" ) == 0 )
	    type = INC_USER;
	else {
	    warning( "%s: unknown include type \"%s\"",
			fnameP, typeP->es_strP );
	    status = SCNOTDONE;
	}
    }

    if ( status == SCOK ) {
	if ( !pp_incdir( type, pathP ) )
	    status = SCNOTDONE;
    }

    /* Release values as required */
    if ( pathP != NULL )
	val_release( &pathval );
    if ( typeP != NULL )
	val_release( &typeval );

    val_set_int( rvalP, 4, ( status == SCOK ) ? 1 : 0 );

    return ( status );
}
/*

*//* bif_mfl_parse -- show parse tree

    BOOL $mfl_parse( any_mfl_code )

    Note: mfl code is simply passed as an argument, NOT as a string.

*/

static STSCODE
bif_mfl_parse ARGLIST( (fnameP, rvalP, treeP) )
   NFARG( char		*fnameP )	/* Function name */
   NFARG( VALUE		*rvalP )	/* Return value ptr */
    FARG( PTNODE	*treeP )	/* Argument tree */
{
    /* Print the argument tree if any */
    parsetree_print( stdout, 0, treeP );

    /* Return good status */
    val_set_int( rvalP, 4, 1 );
    return ( SCOK );
}
/*

*//* bif_mfl_protect_admin -- run user MFL code, protecting admin status

    BOOL $mfl_protect_admin( any_mfl_code )

    Note: mfl code is simply passed as an argument, NOT as a string.

*/

static STSCODE
bif_mfl_protect_admin ARGLIST( (fnameP, rvalP, treeP) )
   NFARG( char		*fnameP )	/* Function name */
   NFARG( VALUE		*rvalP )	/* Return value ptr */
    FARG( PTNODE	*treeP )	/* Argument tree */
{
	STSCODE		status;
	int		admin;		/* Old admin value */
	CNTRL		*cntrlP;
	CNTRLINT	*ciP;

    /* Get current admin setting and revoke it */
    if ( ( ( cntrlP = cntrl_find( estr_tmp_str( "admin" ) ) ) == NULL ) ||
         ( cntrlP->ctl_type != CTL_INT ) ) {
	warning( "%s: can't get admin control info!", fnameP );
	val_set_int( rvalP, 4, 0 );
	status = SCNOTDONE;
    }
    else {
	ciP = (CNTRLINT *)cntrlP->ctl_infoP;
	admin = ciP->cv_int;
	cntrl_intc_set( ciP, CA_ADMIN, 0 );

	/* exec the argument tree */
	status = exec_treeval( treeP, rvalP, FALSE );

	/* Return admin setting to its old state */
	cntrl_intc_set( ciP, CA_ADMIN, admin );
    }

    /* Done */
    return ( status );
}
/*

*//* bif_mfl_run_fname -- Run mfl code from a named file

    int $mfl_run_fname( string fname, string modes )

    The named file is opened and mfl inside it is run

    modes is a modes string per the mfl_run_xxx functions.

    The return value is -1 if the file couldn't be run,
      otherwise 0 or 1 as returned by mfl_run_fname().

*/

static STSCODE
bif_mfl_run_fname ARGLIST( (fnameP, rvalP, treeP) )
   NFARG( char		*fnameP )	/* Function name */
   NFARG( VALUE		*rvalP )	/* Return value ptr */
    FARG( PTNODE	*treeP )	/* Argument tree */
{
	STSCODE		status;
	BOOL		execedF;	/* If we did the exec */
	ESTR		*fnesP;		/* Filename string */
	ESTR		*msesP;		/* Modes string */
	VALUE		fnval, msval;	/* fname and modes string values */

    execedF = FALSE;

    /* Get the arguments */
    if ( ( fnesP = func_treearg_estr( fnameP, treeP, 0, &fnval ) ) == NULL ) {
	warning( "%s: first argument must be valid filename string.", fnameP );
	status = SCNOTDONE;
    }
    else {
	msesP = func_treearg_estr( fnameP, treeP, 1, &msval );
	if ( msesP == NULL ) {
	    warning( "%s: second argument must be valid modes string.",
			fnameP );
	    status = SCNOTDONE;
	}
	else {
	    /* Run the file */
	    execedF = TRUE;
	    status = mfl_run_fname( rvalP, fnesP, msesP );

	    val_release( &msval );
	}
	val_release( &fnval );
    }

    if ( !execedF )
	val_set_int( rvalP, 4, -1 );

    return ( status );
}
/*

*//* bif_mfl_run_string -- Run mfl code from a string

    BOOL $mfl_run_string( string mflstr, string modes )

    mfl code is passed in the string argument.

    modes is a modes string per the mfl_run_xxx functions.

*/

static STSCODE
bif_mfl_run_string ARGLIST( (fnameP, rvalP, treeP) )
   NFARG( char		*fnameP )	/* Function name */
   NFARG( VALUE		*rvalP )	/* Return value ptr */
    FARG( PTNODE	*treeP )	/* Argument tree */
{
	STSCODE		status;
	ESTR		*strP;		/* String to run */
	ESTR		*msesP;		/* Modes string */
	VALUE		strval;		/* Value for string */
	VALUE		msval;		/* Value for modes string */

    if ( ( strP = func_treearg_estr( fnameP, treeP, 0, &strval ) ) == NULL ) {
	warning( "%s requires a string argument.", fnameP );
	status = SCNOTDONE;
    }
    else {
	if ( ( msesP = func_treearg_estr( fnameP, treeP, 1, &msval ) )
			== NULL ) {
	    warning( "%s requires a modes argument.", fnameP );
	    status = SCNOTDONE;
	}
	else {
	    status = mfl_run_str( NULL, strP, msesP );
	    val_release( &msval );
	}
	val_release( &strval );
    }

    val_set_int( rvalP, 4, ( status == SCOK ) ? 1 : 0 );

    if ( status == SCNOTDONE )
	status = SCOK;

    return ( status );
}
/*

*//* bif_mfl_trace -- Set the trace flag

    int  $mfl_trace( traceF )

    Sets the MFL trace flag to the passed value, and returns the previous
    value of the trace flag.  The trace flag controls whether or not
    MFL should do any tracing:  0 is disabled, non-zero is enabled.

    Specific tracing types are further enabled or disabled via control
    values.

*/

static STSCODE
bif_mfl_trace ARGLIST( (fnameP, rvalP, treeP) )
   NFARG( char		*fnameP )	/* Function name */
   NFARG( VALUE		*rvalP )	/* Return value ptr */
    FARG( PTNODE	*treeP )	/* Argument tree */
{
	WORD		traceF;

    /* Get the new flag value */
    if ( func_treearg_word( fnameP, treeP, 0, &traceF ) == SCOK ) {
	/* Return the old value, and set the new one */
	val_set_int( rvalP, 2, Mfl_traceF );
	Mfl_traceF = traceF;
    }
    else {
	/* Just return the old value. */
	val_set_int( rvalP, 2, Mfl_traceF );
    }

    return ( SCOK );
}
/*

*//* bif_msns_attr_int_set -- Set a message store namespace integer attribute

    BOOL $msns_attr_int_set( string prefix, string attr, int val )

    prefix is the namespace prefix;

    attr is the name of the attribute to set

    val is the value to set

*/

static STSCODE
bif_msns_attr_int_set ARGLIST( (fnameP, rvalP, treeP) )
   NFARG( char		*fnameP )	/* Function name */
   NFARG( VALUE		*rvalP )	/* Return value ptr */
    FARG( PTNODE	*treeP )	/* Argument tree */
{
	STSCODE		status;
	long		val;		/* The value */
	UBYTE		*pfxstrP;	/* tmp string for prefix */
	UBYTE		*attrstrP;	/* tmp string for attr name */
	ESTR		*prefixP;	/* Namespace prefix */
	ESTR		*attrP;		/* Attribute name */
	MSNS		*msnsP;		/* Namespace handle */
	VALUE		pfxval, attrval;

    prefixP = attrP = NULL;
    status = SCOK;

    /* Get the prefix, attribute name, and value arguments */
    prefixP = func_treearg_estr( fnameP, treeP, 0, &pfxval );
    if ( prefixP == NULL ) {
	warning( "%s requires a prefix string.", fnameP );
	status = SCNOTDONE;
    }
    else if ( ( attrP = func_treearg_estr( fnameP, treeP, 1, &attrval ) )
			== NULL ) {
	warning( "%s requires an attribute name.", fnameP );
	status = SCNOTDONE;
    }
    else if ( ( status = func_treearg_long( fnameP, treeP, 2, &val ) ) != SCOK ) {
	warning( "%s requires an integer value", fnameP );
    }
    else {
	/* Get convenient nul-terminated strings */
	pfxstrP = tmpstr_estr( prefixP );
	attrstrP = tmpstr_estr( attrP );
	
	if ( ( msnsP = msns_find( pfxstrP ) ) == NULL ) {
	    warning( "%s: can't find prefix \"%s\".", fnameP, pfxstrP );
	    status = SCNOTDONE;
	}
	else {
	    if ( !(*msnsP->ns_msstP->st_attr_int_set)( 
				msnsP->ns_gtdataP, msnsP->ns_stdataP,
				attrstrP, val ) )
		status = SCNOTDONE;
	    else
		status = SCOK;
	}
    }

    /* Release values as required */
    if ( attrP != NULL )
	val_release( &attrval );
    if ( prefixP != NULL )
	val_release( &pfxval );

    val_set_int( rvalP, 4, ( status == SCOK ) ? 1 : 0 );

    return ( status );
}
/*

*//* bif_msns_attr_string_set -- Set a message store namespace string attribute

    BOOL $msns_attr_string_set( string prefix, string attr, string val )

    prefix is the namespace prefix;

    attr is the name of the attribute to set

    val is the value to set

*/

static STSCODE
bif_msns_attr_string_set ARGLIST( (fnameP, rvalP, treeP) )
   NFARG( char		*fnameP )	/* Function name */
   NFARG( VALUE		*rvalP )	/* Return value ptr */
    FARG( PTNODE	*treeP )	/* Argument tree */
{
	STSCODE		status;
	UBYTE		*pfxstrP;	/* tmp string for prefix */
	UBYTE		*attrstrP;	/* tmp string for attr name */
	UBYTE		*valstrP;	/* tmp string for value name */
	ESTR		*prefixP;	/* Namespace prefix */
	ESTR		*attrP;		/* Attribute name */
	ESTR		*valP;		/* Value */
	MSNS		*msnsP;		/* Namespace handle */
	VALUE		pfxval, attrval, valval;

    prefixP = attrP = valP = NULL;
    status = SCOK;

    /* Get the prefix, attribute name, and value arguments */
    prefixP = func_treearg_estr( fnameP, treeP, 0, &pfxval );
    if ( prefixP == NULL ) {
	warning( "%s requires a prefix string.", fnameP );
	status = SCNOTDONE;
    }
    else if ( ( attrP = func_treearg_estr( fnameP, treeP, 1, &attrval ) )
			== NULL ) {
	warning( "%s requires an attribute name.", fnameP );
	status = SCNOTDONE;
    }
    else if ( ( valP = func_treearg_estr( fnameP, treeP, 2, &valval ) )
			== NULL ) {
	warning( "%s requires a value.", fnameP );
	status = SCNOTDONE;
    }
    else {
	/* Get convenient nul-terminated strings */
	pfxstrP = tmpstr_estr( prefixP );
	attrstrP = tmpstr_estr( attrP );
	valstrP = tmpstr_estr( valP );
	
	if ( ( msnsP = msns_find( pfxstrP ) ) == NULL ) {
	    warning( "%s: can't find prefix \"%s\".", fnameP, pfxstrP );
	    status = SCNOTDONE;
	}
	else {
	    if ( !(*msnsP->ns_msstP->st_attr_str_set)( 
				msnsP->ns_gtdataP, msnsP->ns_stdataP,
				attrstrP, valstrP ) )
		status = SCNOTDONE;
	    else
		status = SCOK;
	}
    }

    /* Release values as required */
    if ( attrP != NULL )
	val_release( &attrval );
    if ( prefixP != NULL )
	val_release( &pfxval );
    if ( valP != NULL )
	val_release( &valval );

    val_set_int( rvalP, 4, ( status == SCOK ) ? 1 : 0 );

    return ( status );
}
/*

*//* bif_msns_define -- Define a message store namespace

    BOOL $msns_define( string prefix, string sstype )

    prefix is the namespace prefix;

    sstype is the specific storage type (e.g. "maildir").

    Note: Most if not all namespace definitions will need completing
      via $msns_attr_xxx_set functions

*/

static STSCODE
bif_msns_define ARGLIST( (fnameP, rvalP, treeP) )
   NFARG( char		*fnameP )	/* Function name */
   NFARG( VALUE		*rvalP )	/* Return value ptr */
    FARG( PTNODE	*treeP )	/* Argument tree */
{
	STSCODE		status;
	char		*pfxstrP;	/* tmp string for prefix */
	char		*sststrP;	/* tmp string for storage type */
	ESTR		*prefixP;	/* Namespace prefix */
	ESTR		*sstypeP;	/* Specific storage type */
	MSNS		*msnsP;		/* Namespace handle */
	MSST		*msstP;		/* msgstore specific type handle */
	VALUE		pfxval, sstval;

    prefixP = sstypeP = NULL;
    status = SCOK;

    /* Get the prefix and storage type args */
    prefixP = func_treearg_estr( fnameP, treeP, 0, &pfxval );
    if ( prefixP == NULL ) {
	warning( "%s requires a prefix string.", fnameP );
	status = SCNOTDONE;
    }
    else if ( ( sstypeP = func_treearg_estr( fnameP, treeP, 1, &sstval ) )
			== NULL ) {
	warning( "%s requires a specific storage type.", fnameP );
	status = SCNOTDONE;
    }
    else {
	/* Get convenient nul-terminated strings */
	pfxstrP = tmpstr_estr( prefixP );
	sststrP = tmpstr_estr( sstypeP );
	
	if ( ( msstP = msst_find( sststrP ) ) == NULL ) {
	    warning( "%s: no msgstore type \"%s\".", fnameP, sststrP );
	    status = SCNOTDONE;
	}
	else if ( ( msnsP = msns_register( pfxstrP ) ) == NULL ) {
	    warning( "%s: can't register with prefix \"%s\".",
			fnameP, pfxstrP );
	    status = SCNOTDONE;
	}
	else {
	    /* Make the association. */
	    msnsP->ns_msstP = msstP;

	    /* Instantiate the general type data */
	    msnsP->ns_gtdataP = (*msstP->st_gtdata_get)( NULL );
	    if ( msnsP->ns_gtdataP == NULL )
		status = SCERR;
	    else {
		/* Instantiate the specific type data */
		msnsP->ns_stdataP = (*msstP->st_stdata_get)( NULL );
		if ( msnsP->ns_stdataP == NULL )
		    status = SCERR;
		else
		    status = SCOK;
	    }
	}
    }

    /* Release values as required */
    if ( sstypeP != NULL )
	val_release( &sstval );
    if ( prefixP != NULL )
	val_release( &pfxval );

    val_set_int( rvalP, 4, ( status == SCOK ) ? 1 : 0 );

    return ( status );
}
/*

*//* bif_msns_tailhook -- Set a message store namespace tail translation hook

    BOOL $msns_tailhook( string prefix, string hookname );

    prefix is the namespace prefix;

    hookname is the name of the MFL function to call for tail translation.

    Note: intended to be temporary until we have mfl lambdas.

*/

static STSCODE
bif_msns_tailhook ARGLIST( (fnameP, rvalP, treeP) )
   NFARG( char		*fnameP )	/* Function name */
   NFARG( VALUE		*rvalP )	/* Return value ptr */
    FARG( PTNODE	*treeP )	/* Argument tree */
{
	STSCODE		status;
	UBYTE		*pfxstrP;	/* tmp string for prefix */
	UBYTE		*hookstrP;	/* tmp string for hook name */
	ESTR		*prefixP;	/* Namespace prefix */
	ESTR		*hookP;		/* Attribute name */
	MSNS		*msnsP;		/* Namespace handle */
	VALUE		pfxval, hookval;

    prefixP = hookP = NULL;
    status = SCOK;

    /* Get the prefix and hook name arguments */
    prefixP = func_treearg_estr( fnameP, treeP, 0, &pfxval );
    if ( prefixP == NULL ) {
	warning( "%s requires a prefix string.", fnameP );
	status = SCNOTDONE;
    }
    else if ( ( hookP = func_treearg_estr( fnameP, treeP, 1, &hookval ) )
			== NULL ) {
	warning( "%s requires a function (hook) name.", fnameP );
	status = SCNOTDONE;
    }
    else {
	/* Get convenient nul-terminated strings */
	pfxstrP = tmpstr_estr( prefixP );
	hookstrP = tmpstr_estr( hookP );
	
	if ( ( msnsP = msns_find( pfxstrP ) ) == NULL ) {
	    warning( "%s: can't find prefix \"%s\".", fnameP, pfxstrP );
	    status = SCNOTDONE;
	}
	else if ( !attr_str_set( &msnsP->ns_transhook, hookstrP, TRUE ) )
	    status = SCNOTDONE;
	else
	    status = SCOK;
    }

    /* Release values as required */
    if ( prefixP != NULL )
	val_release( &pfxval );
    if ( hookP != NULL )
	val_release( &hookval );

    val_set_int( rvalP, 4, ( status == SCOK ) ? 1 : 0 );

    return ( status );
}
/*

*//* bif_msst_attr_int_set -- Set a message store type's integer attribute

    BOOL $msst_attr_int_set( string name, string attr, int val )

    name is the name of the specific msgstore type

    attr is the name of the attribute to set

    val is the value to set

*/

static STSCODE
bif_msst_attr_int_set ARGLIST( (fnameP, rvalP, treeP) )
   NFARG( char		*fnameP )	/* Function name */
   NFARG( VALUE		*rvalP )	/* Return value ptr */
    FARG( PTNODE	*treeP )	/* Argument tree */
{
	STSCODE		status;
	long		val;		/* The value */
	UBYTE		*namestrP;	/* tmp string for type name */
	UBYTE		*attrstrP;	/* tmp string for attr name */
	ESTR		*nameP;		/* Type name */
	ESTR		*attrP;		/* Attribute name */
	MSST		*msstP;		/* msgstore specific type handle */
	VALUE		nameval, attrval;

    nameP = attrP = NULL;
    status = SCOK;

    /* Get the type name, attribute name, and value args */
    nameP = func_treearg_estr( fnameP, treeP, 0, &nameval );
    if ( nameP == NULL ) {
	warning( "%s requires a msgstore type name string.", fnameP );
	status = SCNOTDONE;
    }
    else if ( ( attrP = func_treearg_estr( fnameP, treeP, 1, &attrval ) )
			== NULL ) {
	warning( "%s requires an attribute name.", fnameP );
	status = SCNOTDONE;
    }
    else if ( ( status = func_treearg_long( fnameP, treeP, 2, &val ) ) != SCOK ) {
	warning( "%s requires an integer value", fnameP );
    }
    else {
	/* Get convenient nul-terminated strings */
	namestrP = tmpstr_estr( nameP );
	attrstrP = tmpstr_estr( attrP );
	
	if ( ( msstP = msst_find( namestrP ) ) == NULL ) {
	    warning( "%s: no msgstore type \"%s\".", fnameP, namestrP );
	    status = SCNOTDONE;
	}
	else {
	    if ( !(*msstP->st_attr_int_set)( 
				msstP->st_gtdataP, msstP->st_stdataP,
				attrstrP, val ) )
		status = SCNOTDONE;
	    else
		status = SCOK;
	}
    }

    /* Release values as required */
    if ( attrP != NULL )
	val_release( &attrval );
    if ( nameP != NULL )
	val_release( &nameval );

    val_set_int( rvalP, 4, ( status == SCOK ) ? 1 : 0 );

    return ( status );
}
/*

*//* bif_msst_attr_string_set -- Set a message store type's string attribute

    BOOL $msst_attr_string_set( string name, string attr, string val )

    name is the name of the specific message store type

    attr is the name of the attribute to set

    val is the value to set

*/

static STSCODE
bif_msst_attr_string_set ARGLIST( (fnameP, rvalP, treeP) )
   NFARG( char		*fnameP )	/* Function name */
   NFARG( VALUE		*rvalP )	/* Return value ptr */
    FARG( PTNODE	*treeP )	/* Argument tree */
{
	STSCODE		status;
	UBYTE		*namestrP;	/* tmp string for type name */
	UBYTE		*attrstrP;	/* tmp string for attr name */
	UBYTE		*valstrP;	/* tmp string for value name */
	ESTR		*nameP;		/* msgstore specific type name */
	ESTR		*attrP;		/* Attribute name */
	ESTR		*valP;		/* Value */
	MSST		*msstP;		/* msgstore specific type handle */
	VALUE		nameval, attrval, valval;

    nameP = attrP = valP = NULL;
    status = SCOK;

    /* Get/evaluate the arguments */
    nameP = func_treearg_estr( fnameP, treeP, 0, &nameval );
    if ( nameP == NULL ) {
	warning( "%s requires a msgstore type name string.", fnameP );
	status = SCNOTDONE;
    }
    else if ( ( attrP = func_treearg_estr( fnameP, treeP, 1, &attrval ) )
			== NULL ) {
	warning( "%s requires an attribute name.", fnameP );
	status = SCNOTDONE;
    }
    else if ( ( valP = func_treearg_estr( fnameP, treeP, 2, &valval ) )
			== NULL ) {
	warning( "%s requires a value.", fnameP );
	status = SCNOTDONE;
    }
    else {
	/* Get convenient nul-terminated strings */
	namestrP = tmpstr_estr( nameP );
	attrstrP = tmpstr_estr( attrP );
	valstrP = tmpstr_estr( valP );
	
	if ( ( msstP = msst_find( namestrP ) ) == NULL ) {
	    warning( "%s: no msgstore type \"%s\".", fnameP, namestrP );
	    status = SCNOTDONE;
	}
	else {
	    if ( !(*msstP->st_attr_str_set)( 
				msstP->st_gtdataP, msstP->st_stdataP,
				attrstrP, valstrP ) )
		status = SCNOTDONE;
	    else
		status = SCOK;
	}
    }

    /* Release values as required */
    if ( attrP != NULL )
	val_release( &attrval );
    if ( nameP != NULL )
	val_release( &nameval );
    if ( valP != NULL )
	val_release( &valval );

    val_set_int( rvalP, 4, ( status == SCOK ) ? 1 : 0 );

    return ( status );
}
