/* mfl-bif-msgpart.c -- MFL functions: messagepart stuff

	Copyright 2003,2004,2005,2006,2007
		by Mark E. Mallett, MV Communications, Inc.

	See the "LICENSE" file for terms.

MFL built-in-functions relating to message parts and messages

*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include <mml/mml.h>

#include <mml/mml-alloc.h>
#include <mml/mml-ll.h>
#include <mml/mml-str.h>
#include <mml/mml-in.h>
#include <mml/mml-iopf.h>
#include <mml/mml-dns.h>
#include <mml/mml-symtab.h>

#include "mvmf.h"
#include "mfl.h"
#include "mfl-token.h"
#include "mfl-type.h"
#include "mfl-eval.h"
#include "mfl-func.h"
#include "mfl-misc.h"

#include "msgpart.h"
#include "omm.h"
#include "data.h"
#include "errwarn.h"

/* Local definitions */


/* External data referenced */

extern	OMM	*OmmP;			/* Open mail message */

/* External routines used */


/* Local routines and forward references */

static	STSCODE	val_set_hdr PROTO( (VALUE *rvalP, void *hdrP) );
static	void	*get_hdrarg PROTO( (char *fnameP, PTNODE *treeP, int argN) );

    /* Built-in functions contained herein */
static	STSCODE	bif_msg_address();
static	STSCODE	bif_msg_envelope();
static	STSCODE	bif_msg_open_fname();
static	STSCODE	bif_msg_reject();
static	STSCODE	bif_msg_rip_add();
static	STSCODE	bif_msg_rip_n();
static	STSCODE	bif_msg_rip_nth();
static	STSCODE	bif_msgpart_cur_body();
static	STSCODE	bif_msgpart_cur_delete();
static	STSCODE	bif_msgpart_cur_epilog();
static	STSCODE	bif_msgpart_cur_size_body();
static	STSCODE	bif_msgpart_cur_size_hdr();
static	STSCODE	bif_msgpart_cur_undelete();
static	STSCODE	bif_msgpart_go_child();
static	STSCODE	bif_msgpart_go_next();
static	STSCODE	bif_msgpart_go_parent();
static	STSCODE	bif_msgpart_go_top();
static	STSCODE	bif_msgpart_hdr_first();
static	STSCODE	bif_msgpart_hdr_last();
static	STSCODE	bif_msgpart_hdr_next();
static	STSCODE	bif_msgpart_hdr_prev();

/* Local data publicly available */

    /* Table of built-in functions to define */
	BIFINF	Bifinf_msgpart[] = {
/*  function name               handler function		*/
{  "$msg_address",		bif_msg_address		},
{  "$msg_envelope",		bif_msg_envelope	},
{  "$msg_open_fname",		bif_msg_open_fname	},
{  "$msg_reject",		bif_msg_reject		},
{  "$msg_rip_add",		bif_msg_rip_add		},
{  "$msg_rip_n",		bif_msg_rip_n		},
{  "$msg_rip_nth",		bif_msg_rip_nth		},
{  "$msgpart_cur_body",		bif_msgpart_cur_body	},
{  "$msgpart_cur_delete",	bif_msgpart_cur_delete	},
{  "$msgpart_cur_epilog",	bif_msgpart_cur_epilog	},
{  "$msgpart_cur_size_body",	bif_msgpart_cur_size_body },
{  "$msgpart_cur_size_hdr",	bif_msgpart_cur_size_hdr },
{  "$msgpart_cur_undelete",	bif_msgpart_cur_undelete },
{  "$msgpart_go_child",		bif_msgpart_go_child	},
{  "$msgpart_go_next",		bif_msgpart_go_next	},
{  "$msgpart_go_parent",	bif_msgpart_go_parent	},
{  "$msgpart_go_top",		bif_msgpart_go_top	},
{  "$msgpart_hdr_first",	bif_msgpart_hdr_first	},
{  "$msgpart_hdr_last",		bif_msgpart_hdr_last	},
{  "$msgpart_hdr_next",		bif_msgpart_hdr_next	},
{  "$msgpart_hdr_prev",		bif_msgpart_hdr_prev	},

{  NULL,		NULL			}
		};


/* Private data */

static	TYPE	*Ty_hdrP;	/*   ($HDR$) */
static	TYPE	*Ty_hdrpP;	/*   ($HDR$ *) */
static	TYPE	*Ty_hdriP;	/*   ($HDRi$) */
static	TYPE	*Ty_hdripP;	/*   ($HDRi$ *) */

static	int	Ty_hdr_x_h_tailp;  /* offset of h_tailP in $HDR$ */
static	int	Ty_hdr_x_h_namep;  /* offset of h_nameP in $HDR$ */
static	int	Ty_hdri_x_hi_hdr;  /* offset of hi_hdr in $HDRi$ */

/*

*//* bif_msgpart_init(args)

	Initialize the msgpart built-in-functions module

Accepts :

	<nothing>

Returns :

	<value>		TRUE if OK
			FALSE if not.

Notes :

	Called as part of bif module(s) initialization to do any
	extra setup required after functions have been registered.

*/

BOOL
bif_msgpart_init NOARGLIST
{
	int		i;
	TNSYM		*tnsymP;
	TYPSYM		*tpsymP;
	TYPE		*typeP;
	AGGMEM		*agmemP;
	char		mfl[1000];	/* Buffer for mfl code */

    /* Initialize the $HDR$ type that's available to scripts. */
    strcpy( &mfl[0], "\n\
typedef struct $hdr$ {\n\
    string *h_nameP;\n\
    string *h_tailP;\n\
} $HDR$;\n\
" );
    mfl_run_str( NULL, estr_tmp_str( &mfl[0] ), estr_tmp_str( "Mc" ) );
    tnsymP = type_aggname( TC_STRUCT, "$hdr$" );
    if ( ( tnsymP == NULL ) ||
         ( ( typeP = tnsymP->tn_typeP ) == NULL ) ||
	 ( ( typeP->ty_flags & TF_DEFINED ) == 0 ) ) {
	warning( "bif_msgpart_init: unable to find \"$hdr$\" struct." );
	return ( FALSE );
    }
    Ty_hdrP = typeP;		/* ($HDR$) type */
    tpsymP = type_ptr( Ty_hdrP );  /* ($HDR$ *) type */
    if ( tpsymP == NULL ) {
	warning( "bif_msgpart_init: can't find/make '$HDR$ *' type." );
	return ( FALSE );
    }
    Ty_hdrpP = &(tpsymP->ts_type);

    /* Remember offset(s) for future reference */
    if ( ( agmemP = type_agg_member( Ty_hdrP, "h_tailP" ) ) == NULL ) {
	warning( "bif_msgpart_init: can't find $HDR$ member 'h_tailP'" );
	return ( FALSE );
    }
    Ty_hdr_x_h_tailp = agmemP->am_offs;
    if ( ( agmemP = type_agg_member( Ty_hdrP, "h_nameP" ) ) == NULL ) {
	warning( "bif_msgpart_init: can't find $HDR$ member 'h_nameP'" );
	return ( FALSE );
    }
    Ty_hdr_x_h_namep = agmemP->am_offs;



    /* The $HDRi$ mfl type is used internally to encapsulate both
      an internal header pointer and the script-visible $HDR$
      structure.
    */
    i = snprintf( &mfl[0], sizeof(mfl), "\n\
typedef struct $hdri$ {\n\
    unsigned char hi_hdrP[%d];\n\
    $HDR$ hi_hdr;\n\
} $HDRi$;\n\
",
	sizeof(void *) );
    if ( i >= sizeof(mfl) ) {
	warning( "bif_msgpart_init: buffer overflow making $HDRi$" );
	return ( FALSE );
    }
    mfl_run_str( NULL, estr_tmp_str( &mfl[0] ), estr_tmp_str( "Mc" ) );

    tnsymP = type_aggname( TC_STRUCT, "$hdri$" );
    if ( ( tnsymP == NULL ) ||
         ( ( typeP = tnsymP->tn_typeP ) == NULL ) ||
	 ( ( typeP->ty_flags & TF_DEFINED ) == 0 ) ) {
	warning( "bif_msgpart_init: unable to find \"$hdri$\" struct. " );
	return ( FALSE );
    }
    Ty_hdriP = typeP;		/* ($HDRi$) type */
    tpsymP = type_ptr( Ty_hdriP );  /* ($HDRi$ *) type */
    if ( tpsymP == NULL ) {
	warning( "bif_msgpart_init: can't find/make '$HDRi$ *' type." );
	return ( FALSE );
    }
    Ty_hdripP = &(tpsymP->ts_type);

    /* Remember offset(s) for future reference */
    if ( ( agmemP = type_agg_member( Ty_hdriP, "hi_hdr" ) ) == NULL ) {
	warning( "bif_msgpart_init: can't find $HDRi$ member 'hi_hdr'" );
	return ( FALSE );
    }
    Ty_hdri_x_hi_hdr = agmemP->am_offs;
    
    return ( TRUE );
}


/***********************************************************************
 *                                                                     *
 *                     Internal built-in functions                     *
 *                                                                     *
 ***********************************************************************/

/* Built-in functions are called with standard arguments, like this:

     built-in-function( fnameP, rvalP, treeP )

	fnameP is the function name that was referenced
	rvalP is a VALUE: where the return value is placed.
	treeP is the address of the argument list.

  The function returns a STSCODE status code.

*/


/*

*//* bif_msg_address -- extract an email address from a string

    string  $msg_address( string s )

    Returns the next address from the given string, starting
    at the current fetch point and adjusting the fetch point
    for any next time.

*/

static STSCODE
bif_msg_address ARGLIST( (fnameP, rvalP, treeP) )
   NFARG( char		*fnameP )	/* Function name */
   NFARG( VALUE		*rvalP )	/* Return value ptr */
    FARG( PTNODE	*treeP )	/* Argument tree */
{
	STSCODE		status;
	REFSTR		*rsP;		/* REFSTR for the passed string */
	UBYTE		*addrP;		/* Returned address string */
	VALUE		strval;

    /* Access the given string */
    status = func_treearg_string( fnameP, treeP, 0, &strval );
    if ( status == SCOK ) {
	val_rvalue( &strval );
	if ( ( rsP = val_get_string( &strval ) ) == NULL ) {
	    warning( "%s: bad/uninitialized string", fnameP );
	    status = SCNOTDONE;
	}
	else {
	    /* Get next address */
	    addrP = msgpart_refstr_addr( (MSGPART *)NULL, rsP,
					  AP_LOCAL|AP_DOMAIN );

	    if ( addrP != NULL )
		val_set_string( rvalP, estr_tmp_str( addrP ) );
	    else
		status = SCNOTDONE;
	}

	val_release( &strval );
    }

    if ( status != SCOK ) {
	val_set_string( rvalP, estr_tmp_str( "" ) );
	if ( status != SCERR )
	    status = SCOK;
    }

    return ( status );
}
/*

*//* bif_msg_envelope -- get envelope information

    string  $msg_envelope( thing )

    thing is:
	"sender"	for the envelope sender  (may also use "from" )
	"recipient"	for the envelope recipient (or "to" )

    Note: unrecognized name is a benign event, just yielding "".

*/

static STSCODE
bif_msg_envelope ARGLIST( (fnameP, rvalP, treeP) )
   NFARG( char		*fnameP )	/* Function name */
   NFARG( VALUE		*rvalP )	/* Return value ptr */
    FARG( PTNODE	*treeP )	/* Argument tree */
{
	STSCODE		status;
	UBYTE		*strP;
	ESTR		*nameP;		/* Evaluated variable name */
	VALUE		nameval;

    /* Find the envelope item name */
    nameP = func_treearg_estr( fnameP, treeP, 0, &nameval );
    if ( nameP == NULL ) {
	warning( "%s: first arg must be envelope item name", fnameP );
	status = SCNOTDONE;
    }
    else {
	/* Get the source of the info. */
	if ( ( estr_icmp_str( nameP, (UBYTE *)"recipient" ) == 0 ) ||
	     ( estr_icmp_str( nameP, (UBYTE *)"to" ) == 0 ) ) {
	    strP = ( E_RecipP == NULL ) ?
			 NULL : refstr_bstr(E_RecipP)->bs_bP;
	}
	else if ( ( estr_icmp_str( nameP, (UBYTE *)"from" ) == 0 ) ||
	          ( estr_icmp_str( nameP, (UBYTE *)"sender" ) == 0 ) ) {
	    strP = ( E_SenderP == NULL ) ?
			 NULL : refstr_bstr(E_SenderP)->bs_bP;
	}
	else
	    strP = NULL;

	if ( strP == NULL )
	    strP = (UBYTE *)"";
	status = val_set_string( rvalP, estr_tmp_str( strP ) );

	val_release( &nameval );
    }

    if ( status != SCOK ) {
	val_set_int( rvalP, 4, 0 );
	if ( status == SCNOTDONE )
	    status = SCOK;
    }

    return ( status );
}
/*

*//* bif_msg_open_fname -- open incoming message from named file

    int  $msg_open_fname( fname, bodycache )

    Opens primary/incoming message from the named file.  If the filename
      argument is blank (empty string), opens message from stdin.

    bodycache is the number of bytes to cache in memory (approximately;
      according to msgpart stuff), or zero for a default.

    Returns zero on failure, non-zero on success.

*/

static STSCODE
bif_msg_open_fname ARGLIST( (fnameP, rvalP, treeP) )
   NFARG( char		*fnameP )	/* Function name */
   NFARG( VALUE		*rvalP )	/* Return value ptr */
    FARG( PTNODE	*treeP )	/* Argument tree */
{
	STSCODE		status;
	long		bodycache;
	FILE		*fP;
	char		*nameP;		/* The filename */
	VALUE		nameval;

    /* Get bodycache arg first since there's no context to keep for it */
    status = func_treearg_long( fnameP, treeP, 1, &bodycache );

    if ( status == SCOK ) {
	nameP = func_treearg_str( fnameP, treeP, 0, &nameval, PATH_MAX );
	if ( nameP == NULL ) {
	    warning( "%s requires a valid filename string.", fnameP );
	    status = SCNOTDONE;
	}
	else {
	    if ( OmmP != NULL ) {
		warning( "%s: a message is already open.", fnameP );
		fP = NULL;
	    }
	    /* Empty string means stdin */
	    else if ( *nameP == NUL )
		fP = stdin;
	    else if ( ( fP = fopen( (char *)nameP, "r" ) ) == NULL ) {
		warning( "%s: can't open \"%s\"", fnameP, nameP );
	    }

	    if ( fP == NULL )
		status = SCNOTDONE;
	    else if ( ( OmmP = omm_openfp( fP, bodycache ) ) == NULL ) {
		if ( fP != stdin )
		    fclose( fP );
		status = SCNOTDONE;
	    }

	    /* Done with name value */
	    val_release( &nameval );
	}
    }
	
    val_set_int( rvalP, 4, ( status == SCOK ) ? 1 : 0 );

    if ( status == SCNOTDONE )
	status = SCOK;

    return ( status );
}
/*

*//* bif_msg_reject -- request a bounce and/or rejection

    string  $msg_reject( rejectonly )

    rejectonly is 0 (FALSE) to request whatever kind of rejection we
      can (bounce and/or SMTP refusal).  Set to 1 to indicate that
      only an SMTP time refusal is requested (no bounces).

    Note: requesting a "bounce" is still subject to mvmf's overall
      decisionmaking.  bounces may be inhibited for a variety of
      reasons, up to and including the fact that they are likely to
      be hostile.

*/

static STSCODE
bif_msg_reject ARGLIST( (fnameP, rvalP, treeP) )
   NFARG( char		*fnameP )	/* Function name */
   NFARG( VALUE		*rvalP )	/* Return value ptr */
    FARG( PTNODE	*treeP )	/* Argument tree */
{
	STSCODE		status;
	WORD		rejectonlyF;

    /* rejectonlyF is first arg */
    status = func_treearg_word( fnameP, treeP, 0, &rejectonlyF );
    if ( status == SCOK ) {
	if ( OmmP != NULL ) {
	    /* Always want reject, and conditionally want bounce. */
	    OmmP->om_wantrejectF = TRUE;
	    OmmP->om_wantbounceF = ( rejectonlyF == 0 );
	}
	else {
	    warning( "%s: there is no message open.", fnameP );
	}
    }

    if ( status != SCOK )
	val_set_int( rvalP, 4, 0 );

    return ( status );
}
/*

*//* bif_msg_rip_add -- add responsible/sender IP

    string  $msg_rip_add( ip )

    ip is a string; an IP address of a system that has sent this email
       message somewhere along the way;

    Adds the responsible IP address to the list of known IP addresses.
    These are used by the default dnsbl operation.

*/

static STSCODE
bif_msg_rip_add ARGLIST( (fnameP, rvalP, treeP) )
   NFARG( char		*fnameP )	/* Function name */
   NFARG( VALUE		*rvalP )	/* Return value ptr */
    FARG( PTNODE	*treeP )	/* Argument tree */
{
	STSCODE		status;
	ESTR		*ipP;		/* Evaluated IP string */
	VALUE		ipval;		/* Result VALUE for IP string */

    if ( OmmP == NULL ) {
	warning( "%s: no message is open.", fnameP );
	return ( SCNOTDONE );
    }

    /* Evaluate and access the IP string */
    ipP = func_treearg_estr( fnameP, treeP, 0, &ipval );
    if ( ipP != NULL ) {
	if ( omm_rip_add( OmmP, ipP ) )
	    status = SCOK;
	else
	    status = SCNOTDONE;
	val_release( &ipval );
    }
    else
	status = SCNOTDONE;

    val_set_int( rvalP, 4, status );

    return ( status );
}
/*

*//* bif_msg_rip_n -- Get the number of RIP entries

    int  $msg_rip_n()

    Returns the number of entries in the responsible IP list.

*/

static STSCODE
bif_msg_rip_n ARGLIST( (fnameP, rvalP, treeP) )
   NFARG( char		*fnameP )	/* Function name */
   NFARG( VALUE		*rvalP )	/* Return value ptr */
    FARG( PTNODE	*treeP )	/* Argument tree */
{
    if ( OmmP == NULL ) {
	warning( "%s: no message is open.", fnameP );
	return ( SCNOTDONE );
    }

    /* Just return the count */

    val_set_int( rvalP, 4, OmmP->om_ripC );

    return ( SCOK );
}
/*

*//* bif_msg_rip_nth -- Get the nth RIP entry

    string  $msg_rip_nth( ripN )

    Returns the nth entry in the responsible IP list, 0 thru n-1.
    Any out-of-range value of n returns an empty string.

*/

static STSCODE
bif_msg_rip_nth ARGLIST( (fnameP, rvalP, treeP) )
   NFARG( char		*fnameP )	/* Function name */
   NFARG( VALUE		*rvalP )	/* Return value ptr */
    FARG( PTNODE	*treeP )	/* Argument tree */
{
	STSCODE		status;
	WORD		ripN;
	char		*ipP;

    if ( OmmP == NULL ) {
	warning( "%s: no message is open.", fnameP );
	status = SCNOTDONE;
    }
    else {
	/* Get the ripN arg */
	status = func_treearg_word( fnameP, treeP, 0, &ripN );
    }

    if ( ( status == SCOK ) &&
	 ( ripN >= 0 )
	 && ( ripN < OmmP->om_ripC ) ) {
	/* Valid index, access the string */
	ipP = OmmP->om_ripPP[ ripN ];
    }
    else {
	/* Out of bounds or other problem; return empty string. */
	ipP = "";
    }

    val_set_string( rvalP, estr_tmp_str( ipP ) );

    return ( SCOK );
}
/*

*//* bif_msgpart_cur_body -- Return pointer to cached body

    string  *$msgpart_cur_body()

    Returns pointer to the cached part of the body (*not* the child),
      or NULL if no cached body.

*/

static STSCODE
bif_msgpart_cur_body ARGLIST( (fnameP, rvalP, treeP) )
   NFARG( char		*fnameP )	/* Function name */
   NFARG( VALUE		*rvalP )	/* Return value ptr */
    FARG( PTNODE	*treeP )	/* Argument tree */
{
	STSCODE		status;
	MSGPART		*mpP;

    if ( OmmP == NULL ) {
	warning( "%s: there is no mail message.", fnameP );
        status = SCNOTDONE;
    }
    else if ( ( mpP = OmmP->om_curmpP ) == NULL ) {
	/* No current message part?  how can that be.. */
	status = SCDONE;
    }
    else if ( mpP->mp_body.ps_partbsP == NULL ) {
	/* No cached body */
	status = SCDONE;
    }
    else {
	status = val_set_string_ptr( rvalP,
			    estr_tmp_nstr( mpP->mp_body.ps_partbsP->bs_bP,
					   mpP->mp_body.ps_partbsP->bs_bC ) );
    }

    /* Return NULL if no success */
    if ( status != SCOK ) {
	val_set_string_ptr( rvalP, NULL );

	/* change SCDONE soft error to SCOK so MFL will proceed. */
	if ( status == SCDONE )
	    status = SCOK;
    }

    return ( status );
}
/*

*//* bif_msgpart_cur_delete -- Mark the current messagepart for deletion

    BOOL  $msgpart_cur_delete()

    Returns TRUE if successful
            FALSE if not (e.g. no message open or no child parts)

*/

static STSCODE
bif_msgpart_cur_delete ARGLIST( (fnameP, rvalP, treeP) )
   NFARG( char		*fnameP )	/* Function name */
   NFARG( VALUE		*rvalP )	/* Return value ptr */
    FARG( PTNODE	*treeP )	/* Argument tree */
{
	BOOL		retval;

    if ( ( OmmP != NULL ) &&
         ( OmmP->om_curmpP != NULL ) ) {
	OmmP->om_curmpP->mp_deleteF = TRUE;
	retval = TRUE;
    }
    else
	retval = FALSE;

    /* Supply the return code */
    val_set_int( rvalP, 4, retval );

    return ( SCOK );
}
/*

*//* bif_msgpart_cur_epilog -- Return pointer to cached epilog

    string  *$msgpart_cur_epilog()

    Returns pointer to the cached part of the epilog mime part
      or NULL if no cached epilog.

*/

static STSCODE
bif_msgpart_cur_epilog ARGLIST( (fnameP, rvalP, treeP) )
   NFARG( char		*fnameP )	/* Function name */
   NFARG( VALUE		*rvalP )	/* Return value ptr */
    FARG( PTNODE	*treeP )	/* Argument tree */
{
	STSCODE		status;
	MSGPART		*mpP;

    if ( OmmP == NULL ) {
	warning( "%s: there is no mail message.", fnameP );
        status = SCNOTDONE;
    }
    else if ( ( mpP = OmmP->om_curmpP ) == NULL ) {
	/* No current message part?  how can that be.. */
	status = SCDONE;
    }
    else if ( mpP->mp_epilog.ps_partbsP == NULL ) {
	/* No cached body */
	status = SCDONE;
    }
    else {
	status = val_set_string_ptr( rvalP,
			    estr_tmp_nstr( mpP->mp_epilog.ps_partbsP->bs_bP,
					   mpP->mp_epilog.ps_partbsP->bs_bC ) );
    }

    /* Return NULL if no success */
    if ( status != SCOK ) {
	val_set_string_ptr( rvalP, NULL );

	/* change SCDONE soft error to SCOK so MFL will proceed. */
	if ( status == SCDONE )
	    status = SCOK;
    }

    return ( status );
}
/*

*//* bif_msgpart_cur_size_body -- Get the body size of current messagepart

    long $msgpart_cur_size_body( BOOL childrenF )

    Returns the size of the current msgpart's body, or -1 if not open.

    If childrenF is FALSE, only the size of the body part strings will
      be returned (prolog/body plus epilog).  When childrenF is TRUE,
      the total size of all child parts plus their headers is added in.

    The size is approximate, as described in various places.

*/

static STSCODE
bif_msgpart_cur_size_body ARGLIST( (fnameP, rvalP, treeP) )
   NFARG( char		*fnameP )	/* Function name */
   NFARG( VALUE		*rvalP )	/* Return value ptr */
    FARG( PTNODE	*treeP )	/* Argument tree */
{
	STSCODE		status;
	BOOL		childrenF;
	WORD		wval;
	LONG		count;

    /* Get the children flag */
    status = func_treearg_word( fnameP, treeP, 0, &wval );
    if ( status != SCOK )
	status = SCNOTDONE;
    else if ( ( OmmP == NULL ) || ( OmmP->om_curmpP == NULL ) ) {
	warning( "%s: no message is open", fnameP );
	status = SCNOTDONE;
    }
    else {
	childrenF = ( wval != 0 );	/* Yes I'm anal */
	count = msgpart_size_body( OmmP->om_curmpP, childrenF );
    }

    if ( status != SCOK )
	count = -1;			/* Error value */
    if ( status == SCNOTDONE )
	status = SCOK;

    /* Supply the return code */
    val_set_int( rvalP, 4, count );

    return ( status );
}
/*

*//* bif_msgpart_cur_size_hdr -- Get the header size of current messagepart

    long $msgpart_cur_size_hdr()

    Returns the size of the current msgpart's headers, or -1 if not open.

    The count represents the size of the headers in memory, which may be
      different from the size that would be stored on disk or on the
      wire.

*/

static STSCODE
bif_msgpart_cur_size_hdr ARGLIST( (fnameP, rvalP, treeP) )
   NFARG( char		*fnameP )	/* Function name */
   NFARG( VALUE		*rvalP )	/* Return value ptr */
    FARG( PTNODE	*treeP )	/* Argument tree */
{
	STSCODE		status;
	LONG		count;

    if ( ( OmmP == NULL ) || ( OmmP->om_curmpP == NULL ) ) {
	warning( "%s: no message is open", fnameP );
	status = SCNOTDONE;
    }
    else {
	count = msgpart_size_hdr( OmmP->om_curmpP );
	status = SCOK;
    }

    if ( status != SCOK )
	count = -1;			/* Error value */
    if ( status == SCNOTDONE )
	status = SCOK;

    /* Supply the return code */
    val_set_int( rvalP, 4, count );

    return ( status );
}
/*

*//* bif_msgpart_cur_undelete -- Unmark the current messagepart for deletion

    BOOL  $msgpart_cur_undelete()

    Returns TRUE if successful
            FALSE if not (e.g. no message open or no child parts)

*/

static STSCODE
bif_msgpart_cur_undelete ARGLIST( (fnameP, rvalP, treeP) )
   NFARG( char		*fnameP )	/* Function name */
   NFARG( VALUE		*rvalP )	/* Return value ptr */
    FARG( PTNODE	*treeP )	/* Argument tree */
{
	BOOL		retval;

    if ( ( OmmP != NULL ) &&
         ( OmmP->om_curmpP != NULL ) ) {
	OmmP->om_curmpP->mp_deleteF = FALSE;
	retval = TRUE;
    }
    else
	retval = FALSE;

    /* Supply the return code */
    val_set_int( rvalP, 4, retval );

    return ( SCOK );
}
/*

*//* bif_msgpart_go_child -- go to first child message part

    BOOL  $msgpart_go_child()

    Returns TRUE if successful
            FALSE if not (e.g. no message open or no child parts)

    "go to" means set the current messagepart in the open message

*/

static STSCODE
bif_msgpart_go_child ARGLIST( (fnameP, rvalP, treeP) )
   NFARG( char		*fnameP )	/* Function name */
   NFARG( VALUE		*rvalP )	/* Return value ptr */
    FARG( PTNODE	*treeP )	/* Argument tree */
{
	BOOL		retval;
	LL_NODE		*nodeP;

    if ( ( OmmP != NULL ) &&
         ( ( nodeP = OmmP->om_curmpP->mp_parts.llh_firstP ) != NULL ) ) {
	/* Set to the child */
	OmmP->om_curmpP = (MSGPART *)nodeP->lln_valP;
	retval = TRUE;
    }
    else
	retval = FALSE;

    /* Supply the return code */
    val_set_int( rvalP, 4, retval );

    return ( SCOK );
}
/*

*//* bif_msgpart_go_next -- go to next message part

    BOOL  $msgpart_go_next()

    Returns TRUE if successful
            FALSE if not (e.g. no message open or no next)

    "go to" means set the current messagepart in the open message

*/

static STSCODE
bif_msgpart_go_next ARGLIST( (fnameP, rvalP, treeP) )
   NFARG( char		*fnameP )	/* Function name */
   NFARG( VALUE		*rvalP )	/* Return value ptr */
    FARG( PTNODE	*treeP )	/* Argument tree */
{
	BOOL		retval;
	LL_NODE		*nodeP;		/* Linked list node */

    if ( ( OmmP != NULL ) &&
         ( ( nodeP = OmmP->om_curmpP->mp_slnodeP ) != NULL ) &&
	 ( ( nodeP = nodeP->lln_nextP ) != NULL ) ) {
	/* Set to the next message part */
	OmmP->om_curmpP = (MSGPART *)nodeP->lln_valP;
	retval = TRUE;
    }
    else
	retval = FALSE;

    /* Supply the return code */
    val_set_int( rvalP, 4, retval );

    return ( SCOK );
}
/*

*//* bif_msgpart_go_parent -- go to parent message part

    BOOL  $msgpart_go_parent()

    Returns TRUE if successful
            FALSE if not (e.g. no message open or no parent)

    "go to" means set the current messagepart in the open message

*/

static STSCODE
bif_msgpart_go_parent ARGLIST( (fnameP, rvalP, treeP) )
   NFARG( char		*fnameP )	/* Function name */
   NFARG( VALUE		*rvalP )	/* Return value ptr */
    FARG( PTNODE	*treeP )	/* Argument tree */
{
	BOOL		retval;

    if ( ( OmmP != NULL ) && ( OmmP->om_curmpP->mp_parentP != NULL ) ) {
	/* Set to the parent message part */
	OmmP->om_curmpP = OmmP->om_curmpP->mp_parentP;
	retval = TRUE;
    }
    else
	retval = FALSE;

    /* Supply the return code */
    val_set_int( rvalP, 4, retval );

    return ( SCOK );
}
/*

*//* bif_msgpart_go_top -- go to top message part

    BOOL  $msgpart_go_top()

    Returns TRUE if successful
            FALSE if not (e.g. no message open)

    "go to" means set the current messagepart in the open message
*/

static STSCODE
bif_msgpart_go_top ARGLIST( (fnameP, rvalP, treeP) )
   NFARG( char		*fnameP )	/* Function name */
   NFARG( VALUE		*rvalP )	/* Return value ptr */
    FARG( PTNODE	*treeP )	/* Argument tree */
{
	BOOL		retval;

    if ( OmmP != NULL ) {
	/* Set to the top message part */
	OmmP->om_curmpP = OmmP->om_mmsgP->mm_topmpP;
	retval = TRUE;
    }
    else
	retval = FALSE;

    
    /* Supply the return code */
    val_set_int( rvalP, 4, retval );

    return ( SCOK );
}
/*

*//* bif_msgpart_hdr_first -- get first header

    $HDR$  *$msgpart_hdr_first()

    Returns ptr to a $HDR$ struct for the first header

*/

static STSCODE
bif_msgpart_hdr_first ARGLIST( (fnameP, rvalP, treeP) )
   NFARG( char		*fnameP )	/* Function name */
   NFARG( VALUE		*rvalP )	/* Return value ptr */
    FARG( PTNODE	*treeP )	/* Argument tree */
{
	STSCODE		status;
	void		*hdrP;

    if ( OmmP == NULL )
	status = SCNOTDONE;
    else {
	hdrP = msgpart_hdr_next( OmmP->om_curmpP, NULL, NULL );
	if ( hdrP == NULL )
	    status = SCNOTDONE;
	else {
	    status = val_set_hdr( rvalP, hdrP );
	}
    }

    if ( status != SCOK )
	val_set_int( rvalP, 4, 0 );

    return ( SCOK );
}
/*

*//* bif_msgpart_hdr_last -- get last header

    $HDR$  *$msgpart_hdr_last()

    Returns ptr to a $HDR$ struct for the last header

*/

static STSCODE
bif_msgpart_hdr_last ARGLIST( (fnameP, rvalP, treeP) )
   NFARG( char		*fnameP )	/* Function name */
   NFARG( VALUE		*rvalP )	/* Return value ptr */
    FARG( PTNODE	*treeP )	/* Argument tree */
{
	STSCODE		status;
	void		*hdrP;

    if ( OmmP == NULL )
	status = SCNOTDONE;
    else {
	hdrP = msgpart_hdr_prev( OmmP->om_curmpP, NULL, NULL );
	if ( hdrP == NULL )
	    status = SCNOTDONE;
	else {
	    status = val_set_hdr( rvalP, hdrP );
	}
    }

    if ( status != SCOK )
	val_set_int( rvalP, 4, 0 );

    return ( SCOK );
}
/*

*//* bif_msgpart_hdr_next -- get next header

    $HDR$  *$msgpart_hdr_next( hdrP )

    hdrP is ($HDR$ *), the result of a prior $msgpart_hdr_xxx().

    Returns ptr to a $HDR$ struct for the next header

*/

static STSCODE
bif_msgpart_hdr_next ARGLIST( (fnameP, rvalP, treeP) )
   NFARG( char		*fnameP )	/* Function name */
   NFARG( VALUE		*rvalP )	/* Return value ptr */
    FARG( PTNODE	*treeP )	/* Argument tree */
{
	STSCODE		status;
	void		*hdrP;

    if ( OmmP == NULL )
	status = SCNOTDONE;
    else if ( ( hdrP = get_hdrarg( fnameP, treeP, 0 ) ) == NULL )
	status = SCNOTDONE;
    else {
	hdrP = msgpart_hdr_next( OmmP->om_curmpP, hdrP, NULL );
	if ( hdrP == NULL )
	    status = SCNOTDONE;
	else {
	    status = val_set_hdr( rvalP, hdrP );
	}
    }

    if ( status != SCOK )
	val_set_int( rvalP, 4, 0 );

    return ( SCOK );
}
/*

*//* bif_msgpart_hdr_prev -- get previous header

    $HDR$  *$msgpart_hdr_prev( hdrP )

    hdrP is ($HDR$ *), the result of a prior $msgpart_hdr_xxx().

    Returns ptr to a $HDR$ struct for the previous header

*/

static STSCODE
bif_msgpart_hdr_prev ARGLIST( (fnameP, rvalP, treeP) )
   NFARG( char		*fnameP )	/* Function name */
   NFARG( VALUE		*rvalP )	/* Return value ptr */
    FARG( PTNODE	*treeP )	/* Argument tree */
{
	STSCODE		status;
	void		*hdrP;

    if ( OmmP == NULL )
	status = SCNOTDONE;
    else if ( ( hdrP = get_hdrarg( fnameP, treeP, 0 ) ) == NULL )
	status = SCNOTDONE;
    else {
	hdrP = msgpart_hdr_prev( OmmP->om_curmpP, hdrP, NULL );
	if ( hdrP == NULL )
	    status = SCNOTDONE;
	else {
	    status = val_set_hdr( rvalP, hdrP );
	}
    }

    if ( status != SCOK )
	val_set_int( rvalP, 4, 0 );

    return ( SCOK );
}


/***********************************************************************
 *                                                                     *
 *                     Internal support routines                       *
 *                                                                     *
 ***********************************************************************/



/*

*//* get_hdrarg( fnameP, treeP, argN )

	Get the header pointer passed to a built-in-function

Accepts :

	fnameP		Function name (for messages)
	treeP		Ptr to argument tree passed to bif
	argN		Argument number that the hdr is passed in

Returns :

	<value>		Ptr to header object if found
			NULL if not

Notes :

	Error/warning messages are given here
*/

static void *
get_hdrarg ARGLIST( (fnameP, treeP, argN) )
   NFARG( char		*fnameP )	/* Function name */
   NFARG( PTNODE	*treeP )	/* Passed tree */
    FARG( int		argN )		/* Arg number */
{
	UBYTE		*dataP;
	PTNODE		*hdrargP;	/* Subtree for header arg */
	VALUE		hdrval;		/* To get the value.. */
	void		*hdrP;

    if ( ( hdrargP = func_treearg( treeP, argN ) ) == NULL ) {
	warning( "%s: arg #%d (header pointer) not supplied.",
			fnameP, argN );
	return ( NULL );
    }

    /* Evaluate */
    if ( eval_tree( hdrargP, &hdrval ) != SCOK )
	return ( NULL );
    if ( hdrval.v_typeP != Ty_hdrpP ) {
	warning( "%s: arg #%d not $HDR$ pointer.", fnameP, argN );
	return ( NULL );
    }

    val_rvalue( &hdrval );
    dataP = (UBYTE *)valptr_resolve( &hdrval.v_data.d_ptr );
    if ( dataP == NULL ) {
	warning( "%s: arg #%d bad $HDR$ pointer.", fnameP, argN );
	return ( NULL );
    }
    dataP -= Ty_hdri_x_hi_hdr;
    hdrP = *(void **)dataP;

    val_finish( &hdrval );

    return ( hdrP );
}
/*

*//* val_set_hdr(rvalP, hdrP)

	Set a value to a $HDR$ type

Accepts :

	rvalP		Where to built the result value
	hdrP		Pointer to header in a messagepart

Returns :

	<value>		status code

Notes :

	Constructs a VALUE in rvalP this value will contain a reference to
	the mfl-visible $HDR$ struct within an internal-purpose $HDRi$

*/

static STSCODE
val_set_hdr ARGLIST( (rvalP, hdrP) )
   NFARG( VALUE		*rvalP )	/* Value to return/build */
    FARG( void		*hdrP )		/* Ptr to header handle */
{
	FSV		*nfsvP;		/* FSV for the name */
	FSV		*tfsvP;		/* FSV for the tail */
	DVS		*dvsP;		/* dynamic storage */
	REFSTR		*nameP, *tailP;	/* Ptrs to header parts */
	VALPTR		*vptrP;

    /* Allocate dynamic storage for the encapsulating $HDRi$ data */
    dvsP = dvs_new( Ty_hdriP->ty_size, Ty_hdriP->ty_size );
    if ( dvsP == NULL )
	return ( SCERR );

    /* Also need freestanding values for the string pointers. */
    nfsvP = fsv_new();
    tfsvP = fsv_new();
    if ( ( nfsvP == NULL ) || ( tfsvP == NULL ) )
	return ( SCERR );

    /* The VALUE will reference the $HDR$ within the DVS $HDRi$ */
    rvalP->v_flags = VF_RVALUE;
    rvalP->v_typeP = Ty_hdrpP;
    rvalP->v_data.d_ptr.p_byteX = Ty_hdri_x_hi_hdr;
    rvalP->v_data.d_ptr.p_ptr.p_genP = dvsP;

    /* The DVS will contain the actual data.
       First store the hdr pointer.  We've allowed space for
       it..
    */
    *(void **)dvsP->dv_dataP = hdrP;

    /* Get refstrs for header parts */
    msgpart_hdr_parts( OmmP->om_curmpP, hdrP, &nameP, &tailP );

    /* Setup ptr to name using intermediate FSV */
    nfsvP->fv_ptrF = TRUE;
    nfsvP->fv_data.d_ptr.p_byteX = PO_REFSTR;
    nfsvP->fv_data.d_ptr.p_ptr.p_genP = nameP;
    vptrP = (VALPTR *)(dvsP->dv_dataP + Ty_hdri_x_hi_hdr + Ty_hdr_x_h_namep);
    vptrP->p_byteX = PO_FSV;
    vptrP->p_ptr.p_genP = nfsvP;
    

    /* Setup ptr to tail, also using intermediate FSV */
    tfsvP->fv_ptrF = TRUE;
    tfsvP->fv_data.d_ptr.p_byteX = PO_REFSTR;
    tfsvP->fv_data.d_ptr.p_ptr.p_genP = tailP;
    vptrP = (VALPTR *)(dvsP->dv_dataP + Ty_hdri_x_hi_hdr + Ty_hdr_x_h_tailp);
    vptrP->p_byteX = PO_FSV;
    vptrP->p_ptr.p_genP = tfsvP;

    return ( SCOK );
}
