/* mfl-bif-str.c -- MFL functions: strings

	Copyright 2003,2004,2005,2006,2007
		by Mark E. Mallett, MV Communications, Inc.

	See the "LICENSE" file for terms.

MFL built-in-functions relating to strings

*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include <mml/mml.h>

#include <mml/mml-alloc.h>
#include <mml/mml-ll.h>
#include <mml/mml-str.h>
#include <mml/mml-in.h>
#include <mml/mml-iopf.h>
#include <mml/mml-dns.h>
#include <mml/mml-symtab.h>

#include "mvmf.h"
#include "mfl.h"
#include "mfl-token.h"
#include "mfl-type.h"
#include "mfl-eval.h"
#include "mfl-func.h"
#include "mfl-misc.h"

#include "data.h"
#include "errwarn.h"

/* Local definitions */


/* External data referenced */


/* External routines used */

extern	TYPSYM	*TystrpP;		/* (string *) */

/* Local routines and forward references */

static	STSCODE	str_count_xx PROTO( (char *fnameP, VALUE *rvalP,
				   PTNODE *treeP, BOOL (*rtc)( UBYTE ch ) ) );

static	STSCODE	str_findx PROTO( (char *fnameP, VALUE *rvalP,
				   PTNODE *treeP, BOOL igncaseF) );

static	STSCODE	str_findx_byte PROTO( (char *fnameP, VALUE *rvalP,
				   PTNODE *treeP, BOOL igncaseF) );

static	STSCODE	str_has_xx PROTO( (char *fnameP, VALUE *rvalP,
				   PTNODE *treeP, BOOL (*rtc)( UBYTE ch ) ) );

static	STSCODE	str_is_xx PROTO( (char *fnameP, VALUE *rvalP,
				   PTNODE *treeP, BOOL (*rtc)( UBYTE ch ) ) );

    /* Built-in functions contained herein */
static	STSCODE	bif_str_bx();
static	STSCODE	bif_str_bx_set();
static	STSCODE	bif_str_byte();
static	STSCODE	bif_str_debug();

static	STSCODE	bif_str_count_control();
static	STSCODE	bif_str_count_digit();
static	STSCODE	bif_str_count_highbit();
static	STSCODE	bif_str_count_lower();
static	STSCODE	bif_str_count_upper();

static	STSCODE	bif_str_find();
static	STSCODE	bif_str_find_byte();
static	STSCODE	bif_str_findi();
static	STSCODE	bif_str_findi_byte();
static	STSCODE	bif_str_find_token_delimited();

static	STSCODE	bif_str_has_control();
static	STSCODE	bif_str_has_digit();
static	STSCODE	bif_str_has_highbit();
static	STSCODE	bif_str_has_lower();
static	STSCODE	bif_str_has_upper();

static	STSCODE	bif_str_is_control();
static	STSCODE	bif_str_is_digit();
static	STSCODE	bif_str_is_highbit();
static	STSCODE	bif_str_is_lower();
static	STSCODE	bif_str_is_upper();

static	STSCODE	bif_str_length();
static	STSCODE	bif_str_lower();
static	STSCODE	bif_str_match();
static	STSCODE	bif_str_quote_regex();
static	STSCODE	bif_str_quote_wild();
static	STSCODE	bif_str_sub();
static	STSCODE	bif_str_supe();
static	STSCODE	bif_str_union();
static	STSCODE	bif_str_upper();


/* Local data publicly available */

    /* Table of built-in functions to define */
	BIFINF	Bifinf_str[] = {
/*  function name               handler function		*/
{  "$str_bx",			bif_str_bx	},
{  "$str_bx_set",		bif_str_bx_set	},
{  "$str_byte",			bif_str_byte	},
{  "$str_debug",		bif_str_debug	},
{  "$str_count_control",	bif_str_count_control	},
{  "$str_count_digit",		bif_str_count_digit	},
{  "$str_count_highbit",	bif_str_count_highbit	},
{  "$str_count_lower",		bif_str_count_lower	},
{  "$str_count_upper",		bif_str_count_upper	},
{  "$str_find",			bif_str_find	},
{  "$str_find_byte",		bif_str_find_byte	},
{  "$str_findi",		bif_str_findi	},
{  "$str_findi_byte",		bif_str_findi_byte	},
{  "$str_find_token_delimited",	bif_str_find_token_delimited	},
{  "$str_has_control",		bif_str_has_control	},
{  "$str_has_digit",		bif_str_has_digit	},
{  "$str_has_highbit",		bif_str_has_highbit	},
{  "$str_has_lower",		bif_str_has_lower	},
{  "$str_has_upper",		bif_str_has_upper	},
{  "$str_is_control",		bif_str_is_control	},
{  "$str_is_digit",		bif_str_is_digit	},
{  "$str_is_highbit",		bif_str_is_highbit	},
{  "$str_is_lower",		bif_str_is_lower	},
{  "$str_is_upper",		bif_str_is_upper	},
{  "$str_length",		bif_str_length	},
{  "$str_lower",		bif_str_lower	},
{  "$str_match",		bif_str_match	},
{  "$_match",			bif_str_match	},  /* temporary */
{  "$str_quote_regex",		bif_str_quote_regex },
{  "$str_quote_wild",		bif_str_quote_wild },
{  "$str_sub",			bif_str_sub	},
{  "$str_supe",			bif_str_supe	},
{  "$str_union",		bif_str_union	},
{  "$str_upper",		bif_str_upper	},

{  NULL,		NULL			}
		};


/* Private data */


/*

*//* bif_str_init(args)

	Initialize the string built-in-functions module

Accepts :

	<nothing>

Returns :

	<value>		TRUE if OK
			FALSE if not.

Notes :

	Called as part of bif module(s) initialization to do any
	extra setup required after functions have been registered.

*/

BOOL
bif_str_init NOARGLIST
{
    return ( TRUE );
}


/***********************************************************************
 *                                                                     *
 *                     Internal built-in functions                     *
 *                                                                     *
 ***********************************************************************/

/* Built-in functions are called with standard arguments, like this:

     built-in-function( fnameP, rvalP, treeP )

	fnameP is the function name that was referenced
	rvalP is a VALUE: where the return value is placed.
	treeP is the address of the argument list.

  The function returns a STSCODE status code.

*/


/*

*//* bif_str_bx -- Return the current byte index

    int  $str_bx( string )

    Returns the current value of the string's byte index.

    Each string has a byte index that provides a little context of
      where one is looking in a string.  This is mostly for internal
      use but it's proven useful for MFL access in some cases.

*/

static STSCODE
bif_str_bx ARGLIST( (fnameP, rvalP, treeP) )
   NFARG( char		*fnameP )	/* Function name */
   NFARG( VALUE		*rvalP )	/* Return value ptr */
    FARG( PTNODE	*treeP )	/* Argument tree */
{
	STSCODE		status;
	REFSTR		*rsP;
	VALUE		strval;

    /* Access the passed string */
    status = func_treearg_string( fnameP, treeP, 0, &strval );
    if ( status == SCOK ) {
	val_rvalue( &strval );
	if ( ( rsP = val_get_string( &strval ) ) == NULL ) {
	    warning( "%s: bad/uninitialized string", fnameP );
	    status = SCNOTDONE;
	}
	else
	    val_set_int( rvalP, 4, rsP->rs_bX );

	val_release( &strval );
    }

    if ( status != SCOK )
	val_set_int( rvalP, 4, -1 );

    return ( status );
}
/*

*//* bif_str_bx_set -- Set the current byte index

    int  $str_bx_set( string, int n )

    Sets the passed string's byte index

    n is silently constrained to be within the string's length

    Returns the previous value, or -1 if some error.

*/

static STSCODE
bif_str_bx_set ARGLIST( (fnameP, rvalP, treeP) )
   NFARG( char		*fnameP )	/* Function name */
   NFARG( VALUE		*rvalP )	/* Return value ptr */
    FARG( PTNODE	*treeP )	/* Argument tree */
{
	STSCODE		status;
	WORD		newX;
	REFSTR		*rsP;
	VALUE		strval;

    /* Get the new index value */
    status = func_treearg_word( fnameP, treeP, 1, &newX );
    if ( status == SCOK ) {
	/* Get the string reference */
	status = func_treearg_string( fnameP, treeP, 0, &strval );

	if ( status == SCOK ) {
	    val_rvalue( &strval );
	    if ( ( rsP = val_get_string( &strval ) ) == NULL ) {
		warning( "%s: bad/uninitialized string", fnameP );
		status = SCNOTDONE;
	    }
	    else {
		if ( newX < -1 )
		    newX = -1;
		else if ( newX > refstr_len( rsP ) )
		    newX = refstr_len(rsP);

		/* Return the old value */
		val_set_int( rvalP, 4, rsP->rs_bX );

		/* Set the new */
		rsP->rs_bX = newX;
	    }
	    val_release( &strval );
	}
    }

    if ( status != SCOK )
	val_set_int( rvalP, 4, -1 );

    return ( status );
}
/*

*//* bif_str_byte -- Get a byte value from a string

    BOOL  $str_byte( string, byteN )

    Returns the value of the nth byte of the passed string.  e.g.:

        str_byte( (string)"hello", 0 )

    returns the integer value of the character 'h'.

*/

static STSCODE
bif_str_byte ARGLIST( (fnameP, rvalP, treeP) )
   NFARG( char		*fnameP )	/* Function name */
   NFARG( VALUE		*rvalP )	/* Return value ptr */
    FARG( PTNODE	*treeP )	/* Argument tree */
{
	STSCODE		status;
	WORD		byteX;
	int		ch;
	REFSTR		*rsP;
	BSTR		*bsP;
	VALUE		strval;

    /* Get the byte # wanted */
    status = func_treearg_word( fnameP, treeP, 1, &byteX );
    if ( status == SCOK )
	/* Get the string reference */
	status = func_treearg_string( fnameP, treeP, 0, &strval );

    if ( status == SCOK ) {
	val_rvalue( &strval );
	if ( ( rsP = val_get_string( &strval ) ) == NULL ) {
	    warning( "%s: bad/uninitialized string", fnameP );
	    status = SCNOTDONE;
	}
	else {
	    if ( ( byteX < 0 ) || ( byteX >= refstr_len( rsP ) ) ) {
		warning( "%s: byte index (%d) out of range.", fnameP, byteX );
		status = SCNOTDONE;
	    }
	    else {
		bsP = refstr_bstr( rsP );
		ch = bsP->bs_bP[ refstr_startx( rsP ) + byteX ];
	    }
	}
	val_release( &strval );
    }

    if ( status == SCOK )
	val_set_int( rvalP, 4, ch );
    else
	val_set_int( rvalP, 4, 0 );

    return ( status );
}
/*

*//* bif_str_debug -- string debug

    BOOL  $str_debug( string )

    Returns TRUE if successful
            FALSE if not (e.g. no message open or no child parts)

*/

static STSCODE
bif_str_debug ARGLIST( (fnameP, rvalP, treeP) )
   NFARG( char		*fnameP )	/* Function name */
   NFARG( VALUE		*rvalP )	/* Return value ptr */
    FARG( PTNODE	*treeP )	/* Argument tree */
{
	STSCODE		status;
	BOOL		strvalF;	/* If srcstr value is present */
	REFSTR		*rsP, *nrsP;
	REFSTR_BASE	*rsbP;
	LL_NODE		*llP;
	VALUE		srcstr;

    status = func_treearg_string( fnameP, treeP, 0, &srcstr );

    if ( status == SCOK ) {
	strvalF = TRUE;
	val_rvalue( &srcstr );
	if ( ( rsP = val_get_string( &srcstr) ) == NULL ) {
	    warning( "%s: bad/uninitialized string", fnameP );
	    status = SCNOTDONE;
	}
    }
    else
	strvalF = FALSE;

    if ( status == SCOK ) {
	printf( "rsP=%lx  rs_refC=%d start=%d(=%d) end=%d(=%d) len=%d bx=%d\n",
	         rsP,
		 rsP->rs_refC,
		 rsP->rs_startX, refstr_startx(rsP),
		 rsP->rs_endX, refstr_endx(rsP),
		 refstr_len(rsP),
		 rsP->rs_bX );
	printf( "   <" );
	fwrite( refstr_bstr( rsP )->bs_bP + refstr_startx( rsP ),
		          1, refstr_len( rsP ), stdout );
	printf( ">\n" );
	printf( "\n" );

	rsbP = rsP->rs_baseP;
	printf( "base=%lx\n", rsbP );
	printf( "\n" );

	printf( "Attached REFSTR(s):\n" );
	for ( llP = rsbP->bbs_ll.llh_firstP;
			llP != NULL;
				llP = llP->lln_nextP ) {
	    nrsP = (REFSTR *)llP->lln_valP;
	    printf( "rsP=%lx  rs_refC=%d start=%d(=%d) end=%d(=%d) len=%d bX=%d\n",
			nrsP,
			nrsP->rs_refC,
			nrsP->rs_startX, refstr_startx(nrsP),
			nrsP->rs_endX, refstr_endx(nrsP),
			refstr_len(nrsP),
			nrsP->rs_bX );
	    printf( "   <" );
	    fwrite( refstr_bstr( nrsP )->bs_bP + refstr_startx( nrsP ),
		          1, refstr_len( nrsP ), stdout );
	    printf( ">\n" );
	}
	printf( "\n" );
    }

    /* Release value(s) and return. */
    if ( strvalF )
	val_release( &srcstr );

    val_set_int( rvalP, 4, status == SCOK );

    return ( status );
}
/*

*//* bif_str_count_xxx -- see if a string count a type of character.

    A set of related functions that simply hand off to str_count_xx()
      These count the number of characters of a particular class that
      occur in a string, up to an optional maximum count.

    int  $str_count_control( string s [, int max] );
    int  $str_count_digit( string s [, int max] );
    int  $str_count_highbit( string s [, int max] );
    int  $str_count_lower( string s [, int max] );
    int  $str_count_upper( string s [, int max] );

    max, if given and positive, specifies a maximum value for the count;
    if the count reaches this amount, that value is returned.

    Returns the character count.

*/

static STSCODE
bif_str_count_control ARGLIST( (fnameP, rvalP, treeP) )
   NFARG( char		*fnameP )	/* Function name */
   NFARG( VALUE		*rvalP )	/* Return value ptr */
    FARG( PTNODE	*treeP )	/* Argument tree */
{
    return ( str_count_xx( fnameP, rvalP, treeP, evalc_iscontrol ) );
}
static STSCODE
bif_str_count_digit ARGLIST( (fnameP, rvalP, treeP) )
   NFARG( char		*fnameP )	/* Function name */
   NFARG( VALUE		*rvalP )	/* Return value ptr */
    FARG( PTNODE	*treeP )	/* Argument tree */
{
    return ( str_count_xx( fnameP, rvalP, treeP, evalc_isdigit ) );
}
static STSCODE
bif_str_count_highbit ARGLIST( (fnameP, rvalP, treeP) )
   NFARG( char		*fnameP )	/* Function name */
   NFARG( VALUE		*rvalP )	/* Return value ptr */
    FARG( PTNODE	*treeP )	/* Argument tree */
{
    return ( str_count_xx( fnameP, rvalP, treeP, evalc_ishighbit ) );
}
static STSCODE
bif_str_count_lower ARGLIST( (fnameP, rvalP, treeP) )
   NFARG( char		*fnameP )	/* Function name */
   NFARG( VALUE		*rvalP )	/* Return value ptr */
    FARG( PTNODE	*treeP )	/* Argument tree */
{
    return ( str_count_xx( fnameP, rvalP, treeP, evalc_islower ) );
}
static STSCODE
bif_str_count_upper ARGLIST( (fnameP, rvalP, treeP) )
   NFARG( char		*fnameP )	/* Function name */
   NFARG( VALUE		*rvalP )	/* Return value ptr */
    FARG( PTNODE	*treeP )	/* Argument tree */
{
    return ( str_count_xx( fnameP, rvalP, treeP, evalc_isupper ) );
}
/*

*//* bif_str_find -- find a string within a string

    string *$str_find( string, pstring [, nth] );

	Find the nth occurance of pstring within string.

*/

static STSCODE
bif_str_find ARGLIST( (fnameP, rvalP, treeP) )
   NFARG( char		*fnameP )	/* Function name */
   NFARG( VALUE		*rvalP )	/* Return value ptr */
    FARG( PTNODE	*treeP )	/* Argument tree */
{
    /* Hand off to str_findx() */
    return( str_findx( fnameP, rvalP, treeP, FALSE ) );
}
/*

*//* bif_str_find_byte -- find a character within a string

    string *$str_find_byte( string, char [, nth] );

	Find the nth occurance of pstring within string.

*/

static STSCODE
bif_str_find_byte ARGLIST( (fnameP, rvalP, treeP) )
   NFARG( char		*fnameP )	/* Function name */
   NFARG( VALUE		*rvalP )	/* Return value ptr */
    FARG( PTNODE	*treeP )	/* Argument tree */
{
    /* Hand off to str_findx_byte() */
    return( str_findx_byte( fnameP, rvalP, treeP, FALSE ) );
}
/*

*//* bif_str_findi -- find a string within a string, ignoring case

    string *$str_findi( string, pstring [, nth] );

	Find the nth occurance of pstring within string, ignoring case

*/

static STSCODE
bif_str_findi ARGLIST( (fnameP, rvalP, treeP) )
   NFARG( char		*fnameP )	/* Function name */
   NFARG( VALUE		*rvalP )	/* Return value ptr */
    FARG( PTNODE	*treeP )	/* Argument tree */
{
    /* Hand off to str_findx() */
    return( str_findx( fnameP, rvalP, treeP, TRUE ) );
}
/*

*//* bif_str_findi_byte -- find a character within a string, ignoring case

    string *$str_findi_byte( string, char [, nth] );

	Find the nth occurance of pstring within string, ignoring case

*/

static STSCODE
bif_str_findi_byte ARGLIST( (fnameP, rvalP, treeP) )
   NFARG( char		*fnameP )	/* Function name */
   NFARG( VALUE		*rvalP )	/* Return value ptr */
    FARG( PTNODE	*treeP )	/* Argument tree */
{
    /* Hand off to str_findx_byte() */
    return( str_findx_byte( fnameP, rvalP, treeP, TRUE ) );
}
/*

*//* bif_str_find_token_delimited -- find a token via delimiters

    string *$str_find_token_delimited( string, vsep, isep [, flags] )

*/

static STSCODE
bif_str_find_token_delimited ARGLIST( (fnameP, rvalP, treeP) )
   NFARG( char		*fnameP )	/* Function name */
   NFARG( VALUE		*rvalP )	/* Return value ptr */
    FARG( PTNODE	*treeP )	/* Argument tree */
{
	STSCODE		status;
	UBYTE		ch;
	int		rsX;		/* source refstr start index */
	int		bX;		/* Byte index */
	int		sX, len;	/* Start index and length */
	char		*fstrP;		/* String for flags */
	WORD		tflags;		/* Tokenizing flags */
	ESTR		*fesP;		/* For flags */
	ESTR		*vsepP, *isepP;	/* The passed vsep/isep strings */
	REFSTR		*srsP, *rrsP;	/* Source/result strings */
	VALUE		srcstr;
	VALUE		tempval;
	TKNB		tknb;		/* Token boundaries */

    /* First look for/at the optional flags argument */
    tflags = 0;
    if ( func_treearg( treeP, 3 ) != NULL ) {
	fesP = func_treearg_estr( fnameP, treeP, 3, &tempval );
	if ( fesP == NULL ) {
	    val_set_int( rvalP, 4, 0 );
	    return ( SCNOTDONE );
	}
	fstrP = (char *)fesP->es_strP;
	len = estr_len( fesP );
	while ( len-- > 0 ) {
	    ch = *fstrP++;
	    if ( ( ch == 'Q' ) || ( ch == 'q' ) )
		tflags |= TKF_NOQUOTSTR;
	    else if ( ( ch == 'S' ) || ( ch == 's' ) )
		tflags |= TKF_NOSKIPLEAD;

	    /* Just ignore chars we don't recognize */
	}
	val_release( &tempval );
    }

    /* Access the parms */
    status = func_treearg_string( fnameP, treeP, 0, &srcstr );
    if ( status != SCOK ) {
	val_set_int( rvalP, 4, 0 );
	return ( status );
    }

    if ( ( vsepP = func_treearg_estr_tmpstr( fnameP, treeP, 1 ) ) == NULL ) {
	isepP = NULL;
	status = SCNOTDONE;
    }
    else if ( ( isepP = func_treearg_estr_tmpstr( fnameP, treeP, 2 ) ) == NULL )
	status = SCNOTDONE;
    else {
	val_rvalue( &srcstr );
	if ( ( srsP = val_get_string( &srcstr ) ) == NULL ) {
	    warning( "%s: bad/uninitialized string", fnameP );
	    status = SCNOTDONE;
	}
	else {
	    rsX = refstr_startx( srsP );
	    bX = rsX + ( srsP->rs_bX < 0 ? 0 : srsP->rs_bX );
	    len = refstr_endx( srsP ) - rsX;

	    tknb.tb_startX = bX;

	    bX = find_token_delimited( refstr_bstr( srsP )->bs_bP,
				        len,
					&tknb,
					vsepP->es_strP, isepP->es_strP,
					tflags );
	    if ( bX <= 0 ) {
		/* No more tokens, set to end and return null */
		bX = len;
		val_set_int( rvalP, 4, 0 );
	    }
	    else {
		/* Found a token, make new refstr value */
		rrsP = refstr_ref( srsP, rsX + tknb.tb_startX,
					 rsX + tknb.tb_endX );
		if ( rrsP == NULL )
		    status = SCERR;
		else {
		    status = val_set_string_refstr_ptr( rvalP, rrsP );
		    if ( status != SCOK )
			refstr_free( rrsP );
		}
	    }
	}
    }

    if ( status == SCOK ) {
	/* Set up new byte index */
	srsP->rs_bX = bX - rsX;
    }
    else {
	val_set_int( rvalP, 4, 0 );
    }

    val_release( &srcstr );

    return ( status );
}
/*

*//* bif_str_has_xxx -- see if a string has a type of character.

    A set of related functions that simply hand off to str_has_xx()
      These test to see if a character class occurs at least once in a
      string.

    int  $str_has_control( string s );
    int  $str_has_digit( string s );
    int  $str_has_highbit( string s );
    int  $str_has_lower( string s );
    int  $str_has_upper( string s );

    Returns 1 if so, 0 if not.

*/

static STSCODE
bif_str_has_control ARGLIST( (fnameP, rvalP, treeP) )
   NFARG( char		*fnameP )	/* Function name */
   NFARG( VALUE		*rvalP )	/* Return value ptr */
    FARG( PTNODE	*treeP )	/* Argument tree */
{
    return ( str_has_xx( fnameP, rvalP, treeP, evalc_iscontrol ) );
}
static STSCODE
bif_str_has_digit ARGLIST( (fnameP, rvalP, treeP) )
   NFARG( char		*fnameP )	/* Function name */
   NFARG( VALUE		*rvalP )	/* Return value ptr */
    FARG( PTNODE	*treeP )	/* Argument tree */
{
    return ( str_has_xx( fnameP, rvalP, treeP, evalc_isdigit ) );
}
static STSCODE
bif_str_has_highbit ARGLIST( (fnameP, rvalP, treeP) )
   NFARG( char		*fnameP )	/* Function name */
   NFARG( VALUE		*rvalP )	/* Return value ptr */
    FARG( PTNODE	*treeP )	/* Argument tree */
{
    return ( str_has_xx( fnameP, rvalP, treeP, evalc_ishighbit ) );
}
static STSCODE
bif_str_has_lower ARGLIST( (fnameP, rvalP, treeP) )
   NFARG( char		*fnameP )	/* Function name */
   NFARG( VALUE		*rvalP )	/* Return value ptr */
    FARG( PTNODE	*treeP )	/* Argument tree */
{
    return ( str_has_xx( fnameP, rvalP, treeP, evalc_islower ) );
}
static STSCODE
bif_str_has_upper ARGLIST( (fnameP, rvalP, treeP) )
   NFARG( char		*fnameP )	/* Function name */
   NFARG( VALUE		*rvalP )	/* Return value ptr */
    FARG( PTNODE	*treeP )	/* Argument tree */
{
    return ( str_has_xx( fnameP, rvalP, treeP, evalc_isupper ) );
}
/*

*//* bif_str_is_xxx -- see if a string has only a type of character

    A set of related functions that simply hand off to str_is_xx()
      These test to see if a string is composed entirely of a class
      of character.

    int  $str_is_control( string s );
    int  $str_is_digit( string s );
    int  $str_is_highbit( string s );
    int  $str_is_lower( string s );
    int  $str_is_upper( string s );

    Returns 1 if so, 0 if not.

*/

static STSCODE
bif_str_is_control ARGLIST( (fnameP, rvalP, treeP) )
   NFARG( char		*fnameP )	/* Function name */
   NFARG( VALUE		*rvalP )	/* Return value ptr */
    FARG( PTNODE	*treeP )	/* Argument tree */
{
    return ( str_is_xx( fnameP, rvalP, treeP, evalc_iscontrol ) );
}
static STSCODE
bif_str_is_digit ARGLIST( (fnameP, rvalP, treeP) )
   NFARG( char		*fnameP )	/* Function name */
   NFARG( VALUE		*rvalP )	/* Return value ptr */
    FARG( PTNODE	*treeP )	/* Argument tree */
{
    return ( str_is_xx( fnameP, rvalP, treeP, evalc_isdigit ) );
}
static STSCODE
bif_str_is_highbit ARGLIST( (fnameP, rvalP, treeP) )
   NFARG( char		*fnameP )	/* Function name */
   NFARG( VALUE		*rvalP )	/* Return value ptr */
    FARG( PTNODE	*treeP )	/* Argument tree */
{
    return ( str_is_xx( fnameP, rvalP, treeP, evalc_ishighbit ) );
}
static STSCODE
bif_str_is_lower ARGLIST( (fnameP, rvalP, treeP) )
   NFARG( char		*fnameP )	/* Function name */
   NFARG( VALUE		*rvalP )	/* Return value ptr */
    FARG( PTNODE	*treeP )	/* Argument tree */
{
    return ( str_is_xx( fnameP, rvalP, treeP, evalc_islower ) );
}
static STSCODE
bif_str_is_upper ARGLIST( (fnameP, rvalP, treeP) )
   NFARG( char		*fnameP )	/* Function name */
   NFARG( VALUE		*rvalP )	/* Return value ptr */
    FARG( PTNODE	*treeP )	/* Argument tree */
{
    return ( str_is_xx( fnameP, rvalP, treeP, evalc_isupper ) );
}
/*

*//* bif_str_length -- length of string

    int  $str_length( string s )

    Returns the length of the passed string.

*/

static STSCODE
bif_str_length ARGLIST( (fnameP, rvalP, treeP) )
   NFARG( char		*fnameP )	/* Function name */
   NFARG( VALUE		*rvalP )	/* Return value ptr */
    FARG( PTNODE	*treeP )	/* Argument tree */
{
	STSCODE		status;
	REFSTR		*rsP;		/* String ref */
	VALUE		strval;		/* value for string */

    /* Access the passed string */
    status = func_treearg_string( fnameP, treeP, 0, &strval );
    if ( status == SCOK ) {
	val_rvalue( &strval );
	if ( ( rsP = val_get_string( &strval ) ) == NULL ) {
	    warning( "%s: bad/uninitialized string", fnameP );
	    status = SCNOTDONE;
	}
	else
	    val_set_int( rvalP, 4, refstr_len( rsP ) );

	val_release( &strval );
    }

    if ( status != SCOK )
	val_set_int( rvalP, 4, -1 );

    return ( status );
}
/*

*//* bif_str_lower -- string lowercase

    string  $str_lower( string s )

    Returns copy of input string s with any uppercase ASCII chars mapped
        to lowercase.

*/

static STSCODE
bif_str_lower ARGLIST( (fnameP, rvalP, treeP) )
   NFARG( char		*fnameP )	/* Function name */
   NFARG( VALUE		*rvalP )	/* Return value ptr */
    FARG( PTNODE	*treeP )	/* Argument tree */
{
	int		len;
	UBYTE		ch;
	UBYTE		*strP;
	ESTR		*esP;		/* ESTR for new string */

    /* Evaluate and make a new string copy of the argument */
    if ( ( esP = func_treearg_estr_tmpstr( fnameP, treeP, 0 ) ) == NULL ) {
	warning( "%s requires a string argument.", fnameP );
    }
    else {
	/* Map this copy to lowercase */
	strP = esP->es_strP;
	for ( len = estr_len( esP ); len-- > 0; ++strP ) {
	    ch = *strP;
	    if ( isupper( ch ) )
		*strP = tolower( ch );
	}
    }

    if ( esP != NULL )
	val_set_string( rvalP, esP );
    else
	val_set_int( rvalP, 4, 0 );

    return ( SCOK );
}
/*

*//* bif_str_match -- Get matched [sub]string

    BOOL  $str_match( partN )

    Returns the nth remembered match part of the last string match,
    or the entire matched if part is 0.

*/

static STSCODE
bif_str_match ARGLIST( (fnameP, rvalP, treeP) )
   NFARG( char		*fnameP )	/* Function name */
   NFARG( VALUE		*rvalP )	/* Return value ptr */
    FARG( PTNODE	*treeP )	/* Argument tree */
{
	STSCODE		status;
	int		len;
	WORD		partnum;
	UBYTE		*strP;

    status = func_treearg_word( fnameP, treeP, 0, &partnum );
    if ( status == SCOK ) {
	/* Get the matched string */
	strP = strmatched_part( partnum, &len );
	if ( strP == NULL )
	    strP = (UBYTE *)"";

	/* Return the string */
	val_set_string( rvalP, estr_tmp_nstr( strP, len ) );
    }

    return ( status );
}
/*

*//* bif_str_quote_regex -- quote for regex use

    string  $str_quote_regex( string s )

    Returns copy of input string s, quoting any characters that are
       special to regular expressions.

*/

static STSCODE
bif_str_quote_regex ARGLIST( (fnameP, rvalP, treeP) )
   NFARG( char		*fnameP )	/* Function name */
   NFARG( VALUE		*rvalP )	/* Return value ptr */
    FARG( PTNODE	*treeP )	/* Argument tree */
{
	STSCODE		status;
	int		len;
	REFSTR		*srsP;		/* The input (source) refstr */
	REFSTR		*rrsP;		/* New refstr */
	UBYTE		*sP;		/* Source pointer */
	BSTR		*dbsP;		/* Destination bstr */
	PTNODE		*argP;
	VALUE		strval;		/* Input value for string */

    /* Evaluate the argument, forced to a string */
    status = func_treearg_string_coerce( fnameP, treeP, 0, &strval );
    if ( status == SCOK ) {
	val_rvalue( &strval );
	if ( ( srsP = val_get_string( &strval ) ) == NULL ) {
	    warning( "%s: bad/uninitialized string", fnameP );
	    status = SCNOTDONE;
	}
	else {
	    len = refstr_len( srsP );
	    rrsP = refstr_new( "bif_str_quote_regex", len, 0 );
	    if ( rrsP == NULL ) {
		warning( "%s: can't make new string.", fnameP );
		status = SCERR;
	    }
	}

	if ( status == SCOK ) {
	    /* Put new string into new destination buffer. */
	    sP = refstr_bstr( srsP )->bs_bP + refstr_startx( srsP );
	    dbsP = refstr_bstr( rrsP );

	    if ( !bstr_quote_str_n( dbsP, sP, len, QF_REGEX ) ) {
		refstr_free( rrsP );
		status = SCNOTDONE;
	    }
	    else
		bstr_nulterm( dbsP );

	    /* The new refstr is defined with {-1,-1} start/end, so
	       that should already be set.
	    */
	}

	val_release( &strval );		/* Done with this now */
    }

    if ( status == SCOK )
	status = val_set_string_refstr( rvalP, rrsP );

    if ( status != SCOK ) {
	val_set_int( rvalP, 4, 0 );
	if ( status == SCNOTDONE )
	    status = SCOK;
    }

    return ( status );
}
/*

*//* bif_str_quote_wild -- quote for wild use

    string  $str_quote_wild( string s )

    Returns copy of input string s, quoting any characters that are
       special to regular expressions.

*/

static STSCODE
bif_str_quote_wild ARGLIST( (fnameP, rvalP, treeP) )
   NFARG( char		*fnameP )	/* Function name */
   NFARG( VALUE		*rvalP )	/* Return value ptr */
    FARG( PTNODE	*treeP )	/* Argument tree */
{
	STSCODE		status;
	int		len;
	REFSTR		*srsP;		/* The input (source) refstr */
	REFSTR		*rrsP;		/* New refstr */
	UBYTE		*sP;		/* Source pointer */
	BSTR		*dbsP;		/* Destination bstr */
	VALUE		strval;		/* Input value for string */

    /* Evaluate the argument, forced to a string */
    status = func_treearg_string_coerce( fnameP, treeP, 0, &strval );
    if ( status == SCOK ) {
	val_rvalue( &strval );
	if ( ( srsP = val_get_string( &strval ) ) == NULL ) {
	    warning( "%s: bad/uninitialized string", fnameP );
	    status = SCNOTDONE;
	}
	else {
	    len = refstr_len( srsP );
	    rrsP = refstr_new( "bif_str_quote_wild", len, 0 );
	    if ( rrsP == NULL ) {
		warning( "%s: can't make new string.", fnameP );
		status = SCERR;
	    }
	}

	if ( status == SCOK ) {
	    /* Put new string into new destination buffer. */
	    sP = refstr_bstr( srsP )->bs_bP + refstr_startx( srsP );
	    dbsP = refstr_bstr( rrsP );

	    if ( !bstr_quote_str_n( dbsP, sP, len, QF_WILD ) ) {
		refstr_free( rrsP );
		status = SCNOTDONE;
	    }
	    else
		bstr_nulterm( dbsP );

	    /* The new refstr is defined with {-1,-1} start/end, so
	       that should already be set.
	    */
	}

	val_release( &strval );		/* Done with this now */
    }

    if ( status == SCOK )
	status = val_set_string_refstr( rvalP, rrsP );

    if ( status != SCOK ) {
	val_set_int( rvalP, 4, 0 );
	if ( status == SCNOTDONE )
	    status = SCOK;
    }

    return ( status );
}
/*

*//* bif_str_sub -- generate a substring reference

    string *$str_sub( string, start, len )

*/

static STSCODE
bif_str_sub ARGLIST( (fnameP, rvalP, treeP) )
   NFARG( char		*fnameP )	/* Function name */
   NFARG( VALUE		*rvalP )	/* Return value ptr */
    FARG( PTNODE	*treeP )	/* Argument tree */
{
	STSCODE		status;
	WORD		sX, len;	/* Start index and length */
	PTNODE		*xargP;		/* Arg: for index values */
	REFSTR		*srsP, *rrsP;	/* Source/result strings */
	VALUE		srcstr;

    /* Get the start point and length.  Since these are optional
       arguments we have to do it the long way around.
    */
    if ( ( xargP = func_treearg( treeP, 1 ) ) == NULL ) {
	sX = len = -1;			/* Default to entire string */
    }
    else {
	status = eval_tree_word( xargP, &sX );
	if ( status != SCOK )
	    return ( status );

	/* Length */
	if ( ( xargP = func_treearg( treeP, 2 ) ) == NULL ) {
	    len = -1;			/* Default to all of it */
	}
	else {
	    status = eval_tree_word( xargP, &len );
	    if ( status != SCOK )
		return ( status );
	}
    }

    /* Access the source string */
    status = func_treearg_string( fnameP, treeP, 0, &srcstr );
    if ( status == SCOK ) {
	val_rvalue( &srcstr );
	if ( ( srsP = val_get_string( &srcstr ) ) == NULL ) {
	    warning( "%s: bad/uninitialized string", fnameP );
	    status = SCNOTDONE;
	}
	else {
	    /* Make a new reference string */
	    if ( len >= 0 )
		len = ( sX >= 0 ? sX : 0 ) + len;
	    rrsP = refstr_ref( srsP, sX, len );
	    if ( rrsP == NULL )
		status = SCERR;
	}

	if ( status == SCOK ) {
	    /* Make a new string pointer value using this REFSTR */
	    status = val_set_string_refstr_ptr( rvalP, rrsP );
	    if ( status != SCOK )
		refstr_free( rrsP );
	}

	/* Release the string */
	val_release( &srcstr );
    }

    if ( status != SCOK ) {
	val_set_int( rvalP, 4, 0 );

	if ( status == SCNOTDONE )
	    status = SCOK;
    }

    return ( status );
}
/*

*//* bif_str_supe -- generate a sub-or-super-string reference

    string *$str_supe( string, startX, endX )

	Generate a string reference to the input string, with start
	and ends moved by the specified number of bytes.

*/

static STSCODE
bif_str_supe ARGLIST( (fnameP, rvalP, treeP) )
   NFARG( char		*fnameP )	/* Function name */
   NFARG( VALUE		*rvalP )	/* Return value ptr */
    FARG( PTNODE	*treeP )	/* Argument tree */
{
	STSCODE		status;
	WORD		sX, eX;		/* Start and end offsets */
	PTNODE		*xargP;		/* Arg: for index values */
	REFSTR		*srsP, *rrsP;	/* Source/result strings */
	VALUE		srcstr;

    /* Get start and end offsets. */
    status = func_treearg_word( fnameP, treeP, 1, &sX );
    if ( status == SCOK )
	status = func_treearg_word( fnameP, treeP, 2, &eX );

    if ( status == SCOK ) {
	/* Access the source string */
	status = func_treearg_string( fnameP, treeP, 0, &srcstr );
	if ( status == SCOK ) {
	    val_rvalue( &srcstr );
	    if ( ( srsP = val_get_string( &srcstr ) ) == NULL ) {
		warning( "%s: bad/uninitialized string", fnameP );
		status = SCNOTDONE;
	    }
	    else {
		/* Make a new reference string mapping all of source string */
		rrsP = refstr_ref( srsP, 0, refstr_len( srsP ) );
		if ( rrsP == NULL )
		    status = SCERR;
		else {
		    /* Adjust the new string by the desired offsets */
		    rrsP->rs_startX += sX;
		    if ( rrsP->rs_startX < 0 )
			rrsP->rs_startX = 0;
		    else if ( rrsP->rs_startX > refstr_bstr( rrsP )->bs_bC )
			rrsP->rs_startX = refstr_bstr( rrsP )->bs_bC;

		    rrsP->rs_endX += eX;
		    if ( rrsP->rs_endX < 0 )
			rrsP->rs_endX = 0;
		    if ( rrsP->rs_endX > refstr_bstr( rrsP )->bs_bC )
			rrsP->rs_endX = refstr_bstr( rrsP )->bs_bC;

		    if ( rrsP->rs_endX < rrsP->rs_startX )
			rrsP->rs_endX = rrsP->rs_startX;
		}
	    }

	    if ( status == SCOK ) {
		/* Make a new string pointer value using this REFSTR */
		status = val_set_string_refstr_ptr( rvalP, rrsP );
		if ( status != SCOK )
		    refstr_free( rrsP );
	    }
	}

	/* Release string value */
	val_release( &srcstr );
    }

    if ( status != SCOK ) {
	val_set_int( rvalP, 4, 0 );

	if ( status == SCNOTDONE )
	    status = SCOK;
    }

    return ( status );
}
/*

*//* bif_str_union -- A union of two related strings

    string *$str_union( string1, string2 )

	Where string1 and string2 refer to the same base string, return
	a new substring that maps the maximum area included by the two
	strings.

*/

static STSCODE
bif_str_union ARGLIST( (fnameP, rvalP, treeP) )
   NFARG( char		*fnameP )	/* Function name */
   NFARG( VALUE		*rvalP )	/* Return value ptr */
    FARG( PTNODE	*treeP )	/* Argument tree */
{
	STSCODE		status;
	int		sX, eX;		/* Start and end offsets */
	int		tX;		/* Temp offset value */
	REFSTR		*rs1P, *rs2P;	/* The two source strings */
	REFSTR		*rrsP;		/* The result string */
	BSTR		*bsP;		/* Underlying BSTR */
	VALUE		s1val, s2val;	/* The two source string values */

    /* Access the two strings */
    status = func_treearg_string( fnameP, treeP, 0, &s1val );
    if ( status == SCOK ) {
	status = func_treearg_string( fnameP, treeP, 1, &s2val );
	if ( status == SCOK ) {
	    val_rvalue( &s1val );
	    val_rvalue( &s2val );
	    if ( ( rs1P = val_get_string( &s1val ) ) == NULL ) {
		warning( "%s: bad/uninitialized first string", fnameP );
		status = SCNOTDONE;
	    }
	    else if ( ( rs2P = val_get_string( &s2val ) ) == NULL ) {
		warning( "%s: bad/uninitialized second string", fnameP );
		status = SCNOTDONE;
	    }
	    else if ( ( bsP = refstr_bstr( rs1P ) ) !=
			      refstr_bstr( rs2P ) ) {
		warning( "%s: strings do not refer to the same base", fnameP );
		status = SCNOTDONE;
	    }

	    /* Make a new reference string; we'll set start/end manually */
	    else if ( ( rrsP = refstr_ref( rs1P, 0, 0 ) ) == NULL ) {
		status = SCERR;
	    }

	    else {
		/* All ok.  Set up start and end points. */
		sX = refstr_startx( rs1P );
		tX = refstr_startx( rs2P );
		if ( tX < sX )
		    sX = tX;

		eX = refstr_endx( rs1P );
		tX = refstr_endx( rs2P );
		if ( tX > eX )
		    eX = tX;

		/* Sanitize */
		if ( sX < 0 )
		    sX = 0;
		if ( eX > bsP->bs_bC )
		    eX = bsP->bs_bC;
		if ( sX >= eX )
		    sX = eX;

		/* Store */
		rrsP->rs_startX = sX;
		rrsP->rs_endX = eX;

		/* Set up the string pointer return value. */
		status = val_set_string_refstr_ptr( rvalP, rrsP );
		if ( status != SCOK )
		    refstr_free( rrsP );
	    }
	    val_release( &s2val );
	}
	val_release( &s1val );
    }

    if ( status != SCOK ) {
	val_set_int( rvalP, 4, 0 );

	if ( status == SCNOTDONE )
	    status = SCOK;
    }

    return ( status );
}
/*

*//* bif_str_upper -- string uppercase

    string  $str_upper( string s )

    Returns copy of input string s with any lowercase ASCII chars mapped
        to uppercase.

*/

static STSCODE
bif_str_upper ARGLIST( (fnameP, rvalP, treeP) )
   NFARG( char		*fnameP )	/* Function name */
   NFARG( VALUE		*rvalP )	/* Return value ptr */
    FARG( PTNODE	*treeP )	/* Argument tree */
{
	int		len;
	UBYTE		ch;
	UBYTE		*strP;
	ESTR		*esP;		/* ESTR for new string */

    /* Evaluate and make a new string copy of the argument */
    if ( ( esP = func_treearg_estr_tmpstr( fnameP, treeP, 0 ) ) == NULL ) {
	warning( "%s requires a string argument.", fnameP );
    }
    else {
	/* Map this copy to uppercase */
	strP = esP->es_strP;
	for ( len = estr_len( esP ); len-- > 0; ++strP ) {
	    ch = *strP;
	    if ( islower( ch ) )
		*strP = toupper( ch );
	}
    }

    if ( esP != NULL )
	val_set_string( rvalP, esP );
    else
	val_set_int( rvalP, 4, 0 );

    return ( SCOK );
}



/***********************************************************************
 *                                                                     *
 *                     Internal support routines                       *
 *                                                                     *
 ***********************************************************************/



/*

*//* str_count_xx( fnameP, rvalP, treeP, rtc )

	Count the occurances of a type of character in a string

Accepts :

	fnameP		Name of built-in-function
	rvalP		Where to put the return value
	treeP		Argument tree
	rtc		The character test function to call

Returns :

	<value>		Status code

Notes :

	Called on behalf of bif_str_count_xxx() functions.

*/

static STSCODE
str_count_xx ARGLIST( (fnameP, rvalP, treeP, rtc) )
   NFARG( char		*fnameP )	/* Function name */
   NFARG( VALUE		*rvalP )	/* Return value ptr */
   NFARG( PTNODE	*treeP )	/* Argument tree */
    FARG( BOOL		(*rtc)( UBYTE ch ) )	/* Function to call */
{
	STSCODE		status;
	UBYTE		ch;
	int		len;
	WORD		maxC;		/* Maximum if given */
	int		cC;		/* The count we return */
	PTNODE		*argP;		/* Argument subtree */
	UBYTE		*strP;		/* String buffer */
	REFSTR		*rsP;		/* String reference ptr */
	VALUE		strval;		/* Value to access the string with */

    /* See if a max was given */
    if ( ( argP = func_treearg( treeP, 1 ) ) != NULL ) {
	status = eval_tree_word( argP, &maxC );
	if ( maxC < 0 )			/* Regardless of status */
	    maxC = 0;
    }
    else {
	maxC = 0;			/* None given, count all */
	status = SCOK;
    }

    /* Access the passed string */
    if ( status == SCOK )
	status = func_treearg_string( fnameP, treeP, 0, &strval );

    if ( status == SCOK ) {
	val_rvalue( &strval );
	if ( ( rsP = val_get_string( &strval ) ) == NULL ) {
	    warning( "%s: bad/uninitialized string", fnameP );
	    status = SCNOTDONE;
	}
	else {
	    /* Examine the string and count chars */
	    strP = refstr_bstr( rsP )->bs_bP + refstr_startx( rsP );
	    len = refstr_len( rsP );
	    for ( cC = 0; len-- > 0; ) {
		ch = *strP++;
		if ( (*rtc)(ch) ) {
		    ++cC;
		    if ( ( maxC > 0 ) && ( cC == maxC ) )
			break;
		}
	    }
	}

	val_release( &strval );
    }

    if ( status != SCOK )
	cC = 0;

    val_set_int( rvalP, 4, cC );

    return ( status );
}
/*

*//* str_findx( fnameP, rvalP, treeP, igncaseF )

	Find a pattern within a string

Accepts :

	fnameP		Name of built-in-function
	rvalP		Where to put the return value
	treeP		Argument tree
	igncaseF	TRUE to ignore case

Returns :

	<value>		Status code

Notes :

	Called on behalf of bif_str_find() and bif_str_findi() functions

*/

static STSCODE
str_findx ARGLIST( (fnameP, rvalP, treeP, igncaseF) )
   NFARG( char		*fnameP )	/* Function name */
   NFARG( VALUE		*rvalP )	/* Return value ptr */
   NFARG( PTNODE	*treeP )	/* Argument tree */
    FARG( BOOL		igncaseF )	/* TRUE to ignore case */
{
	STSCODE		status;
	WORD		nth;		/* nth occurance wanted */
	int		sX, eX;		/* Start and end of searchable
					   range in base string.
					*/
	int		bX;		/* Offset where comparing */
	int		plen;		/* Pattern length */
	int		cmpval;
	PTNODE		*argP;		/* Argument subtree */
	ESTR		*pesP;		/* Pattern string */
	REFSTR		*brsP;		/* Base string */
	REFSTR		*rrsP;		/* Result string */
	BSTR		*bsP;		/* Base BSTR */
	VALUE		strval, patval;	/* String and pattern values */

    /* Get the optional 'nth' argument if given */
    if ( ( argP = func_treearg( treeP, 2 ) ) != NULL ) {
	status = eval_tree_word( argP, &nth );
	if ( nth == 0 )			/* Regardless of status */
	    nth = 0;
    }
    else {
	nth = 1;			/* None given, find first */
	status = SCOK;
    }

    /* Access the strings */
    if ( status == SCOK ) {
	pesP = func_treearg_estr_tmpstr( fnameP, treeP, 1 );
	if ( pesP == NULL )
	    status = SCNOTDONE;
    }
    if ( status == SCOK ) {
	status = func_treearg_string( fnameP, treeP, 0, &strval );
	if ( status == SCOK ) {
	    val_rvalue( &strval );
	    if ( ( brsP = val_get_string( &strval ) ) == NULL ) {
		warning( "%s: bad/uninitialized string", fnameP );
		status = SCNOTDONE;
	    }

	    if ( status == SCOK ) {
		plen = estr_len( pesP );
		bsP = refstr_bstr( brsP );

TODO		/* Very brute force, should be improved. */

		cmpval = -1;
		if ( nth > 0 ) {
		    /* Forward search.  Search boundaries are from the
		       current access point to the end of the string.
		    */
		    sX = refstr_startx( brsP );
		    if ( brsP->rs_bX > 0 )
			sX += brsP->rs_bX;
		    eX = refstr_endx( brsP ) - plen;

		    for ( bX = sX; bX <= eX; ++bX ) {
			if ( igncaseF )
			    cmpval = strnicmp( (char *)&bsP->bs_bP[bX],
					       (char *)pesP->es_strP,
					       plen );
			else
			    cmpval = strncmp( (char *)&bsP->bs_bP[bX],
					      (char *)pesP->es_strP,
					      plen );

			if ( ( cmpval == 0 ) && ( --nth == 0 ) ) {
			    /* Update the bX to be after the match */
			    brsP->rs_bX = bX + plen - refstr_startx( brsP );
			    break;
			}
		    }
		}
		else {
		    /* Backwards search.  Search boundaries are from the
		       beginning of the string to the current access point.
		    */
		    sX = refstr_startx( brsP );
		    if ( brsP->rs_bX >= 0 )
			eX = sX + brsP->rs_bX - plen;
		    else
			eX = refstr_endx( brsP ) - plen;

		    for ( bX = eX; bX >= sX; --bX ) {
			if ( igncaseF )
			    cmpval = strnicmp( (char *)&bsP->bs_bP[bX],
					       (char *)pesP->es_strP,
					       plen );
			else
			    cmpval = strncmp( (char *)&bsP->bs_bP[bX],
					      (char *)pesP->es_strP,
					      plen );

			if ( ( cmpval == 0 ) && ( ++nth == 0 ) ) {
			    /* Update the bX to be before the match */
			    brsP->rs_bX = bX - refstr_startx( brsP );
			    break;
			}
		    }
		}

		/* Here we have cmpval=0 if matched, with bX set to the
		   place where the match occured.
		*/
		if ( cmpval == 0 ) {
		    /* Make new reference string and set start/end by hand */
		    if ( ( rrsP = refstr_ref( brsP, 0, 0 ) ) == NULL )
			status = SCERR;
		    else {
			rrsP->rs_startX = bX;
			rrsP->rs_endX = bX + plen;

			/* Set up the return value */
			status = val_set_string_refstr_ptr( rvalP, rrsP );
			if ( status != SCOK )
			    refstr_free( rrsP );
		    }
		}
		else {
		    /* There was no match; this will return a NULL string. */
		    status = SCNOTDONE;
		}
	    }

	    val_release( &strval );
	}
    }

    if ( status != SCOK ) {
	val_set_int( rvalP, 4, 0 );

	if ( status == SCNOTDONE )
	    status = SCOK;
    }

    return ( status );
}
/*

*//* str_findx_byte( fnameP, rvalP, treeP, igncaseF )

	Find a byte within a string

Accepts :

	fnameP		Name of built-in-function
	rvalP		Where to put the return value
	treeP		Argument tree
	igncaseF	TRUE to ignore case

Returns :

	<value>		Status code

Notes :

	Called on behalf of bif_str_find_byte() and
	   bif_str_findi_byte() functions

	Seems like way too much code for this simple function.

*/

static STSCODE
str_findx_byte ARGLIST( (fnameP, rvalP, treeP, igncaseF) )
   NFARG( char		*fnameP )	/* Function name */
   NFARG( VALUE		*rvalP )	/* Return value ptr */
   NFARG( PTNODE	*treeP )	/* Argument tree */
    FARG( BOOL		igncaseF )	/* TRUE to ignore case */
{
	STSCODE		status;
	BYTE		fb, tb;		/* Byte to find and to test */
	WORD		bval;		/* Byte value to find */
	WORD		nth;		/* nth occurance wanted */
	int		sX, eX;		/* Start and end of searchable
					   range in base string.
					*/
	int		bX;		/* Offset where comparing */
	int		cmpval;
	BYTE		*bP;
	PTNODE		*argP;		/* Argument subtree */
	REFSTR		*brsP;		/* Base string */
	BSTR		*bsP;		/* Base BSTR */
	VALUE		strval;		/* value for string */

    /* Get the byte value */
    status = func_treearg_word( fnameP, treeP, 1, &bval );
    if ( status == SCOK ) {
	/* validate */
	if ( ( bval < 0 ) || ( bval > 255 ) ) {
	    warning( "%s: bad byte value %d", fnameP, bval );
	    status = SCNOTDONE;
	}
	else {
	    /* Make it lowercase if we are ignoring case */
	    fb = igncaseF ? tolower( bval ) : bval;
	}
    }

    /* Get the optional 'nth' argument if given */
    if ( status == SCOK ) {
	if ( ( argP = func_treearg( treeP, 2 ) ) != NULL ) {
	    status = eval_tree_word( argP, &nth );
	    if ( nth == 0 )		/* Regardless of status */
		nth = 0;
	}
	else {
	    nth = 1;			/* None given, find first */
	    status = SCOK;
	}
    }

    /* Access the string */
    if ( status == SCOK ) {
	status = func_treearg_string( fnameP, treeP, 0, &strval );
	if ( status == SCOK ) {
	    val_rvalue( &strval );
	    if ( ( brsP = val_get_string( &strval ) ) == NULL ) {
		warning( "%s: bad/uninitialized string", fnameP );
		status = SCNOTDONE;
	    }
	    else {
		bsP = refstr_bstr( brsP );

		cmpval = -1;
		if ( nth > 0 ) {
		    /* Forward search.  Search boundaries are from the
		       current access point to the end of the string.
		    */
		    sX = refstr_startx( brsP );
		    if ( brsP->rs_bX > 0 )
			sX += brsP->rs_bX;
		    eX = refstr_endx( brsP ) -1;

		    for ( bX = sX, bP = &bsP->bs_bP[bX];
		    		bX <= eX; ++bX ) {
			tb = *bP++;
			if ( igncaseF )
			    tb = tolower( tb );
			cmpval = tb - fb;
			if ( ( cmpval == 0 ) && ( --nth == 0 ) ) {
			    /* Update the bX to be after the match */
			    brsP->rs_bX = bX + 1 - refstr_startx( brsP );
			    break;
			}
		    }
		}
		else {
		    /* Backwards search.  Search boundaries are from the
		       beginning of the string to the current access point.
		    */
		    sX = refstr_startx( brsP );
		    if ( brsP->rs_bX >= 0 )
			eX = sX + brsP->rs_bX -1;
		    else
			eX = refstr_endx( brsP ) -1;

		    for ( bX = eX, bP = &bsP->bs_bP[eX];
				bX >= sX; --bX ) {
			tb = *bP--;
			if ( igncaseF )
			    tb = tolower( tb );
			cmpval = tb - fb;
			if ( ( cmpval == 0 ) && ( ++nth == 0 ) ) {
			    /* Update the bX to be before (at) the match */
			    brsP->rs_bX = bX - refstr_startx( brsP );
			    break;
			}
		    }
		}

		/* Here we have cmpval=0 if matched, with bX set to the
		   place where the match occured.
		*/
		if ( cmpval == 0 ) {
		    /* Set the return value */
		    val_set_int( rvalP, 4, bX - refstr_startx( brsP ) );
		}
		else {
		    /* There was no match; this will do the error return */
		    status = SCNOTDONE;
		}
	    }
	    val_release( &strval );
	}
    }

    if ( status != SCOK ) {
	val_set_int( rvalP, 4, -1 );

	if ( status == SCNOTDONE )
	    status = SCOK;
    }

    return ( status );
}
/*

*//* str_has_xx( fnameP, rvalP, treeP, rtc )

	See if a string has a type of character

Accepts :

	fnameP		Name of built-in-function
	rvalP		Where to put the return value
	treeP		Argument tree
	rtc		The character test function to call

Returns :

	<value>		Status code

Notes :

	Called on behalf of bif_str_has_xxx() functions.

*/

static STSCODE
str_has_xx ARGLIST( (fnameP, rvalP, treeP, rtc) )
   NFARG( char		*fnameP )	/* Function name */
   NFARG( VALUE		*rvalP )	/* Return value ptr */
   NFARG( PTNODE	*treeP )	/* Argument tree */
    FARG( BOOL		(*rtc)( UBYTE ch ) )	/* Function to call */
{
	STSCODE		status;
	BOOL		resultF;	/* The result we return */
	UBYTE		ch;
	int		len;
	UBYTE		*strP;		/* String buffer */
	REFSTR		*rsP;		/* String reference ptr */
	VALUE		strval;		/* Value to access the string with */

    resultF = FALSE;

    /* Access the passed string */
    status = func_treearg_string( fnameP, treeP, 0, &strval );
    if ( status == SCOK ) {
	val_rvalue( &strval );
	if ( ( rsP = val_get_string( &strval ) ) == NULL ) {
	    warning( "%s: bad/uninitialized string", fnameP );
	    status = SCNOTDONE;
	}
	else {
	    /* Examine the string */
	    strP = refstr_bstr( rsP )->bs_bP + refstr_startx( rsP );
	    len = refstr_len( rsP );
	    while ( len-- > 0 ) {
		ch = *strP++;
		if ( (*rtc)(ch) ) {
		    resultF = TRUE;
		    break;
		}
	    }
	}

	val_release( &strval );
    }

    val_set_int( rvalP, 4, resultF );

    return ( status );
}
/*

*//* str_is_xx( fnameP, rvalP, treeP, rtc )

	See if a string contains only a specific class of character

Accepts :

	fnameP		Name of built-in-function
	rvalP		Where to put the return value
	treeP		Argument tree
	rtc		The character test function to call

Returns :

	<value>		Status code

Notes :

	Called on behalf of bif_str_is_xxx() functions.

*/

static STSCODE
str_is_xx ARGLIST( (fnameP, rvalP, treeP, rtc) )
   NFARG( char		*fnameP )	/* Function name */
   NFARG( VALUE		*rvalP )	/* Return value ptr */
   NFARG( PTNODE	*treeP )	/* Argument tree */
    FARG( BOOL		(*rtc)( UBYTE ch ) )	/* Function to call */
{
	STSCODE		status;
	BOOL		resultF;	/* The result we return */
	UBYTE		ch;
	int		len;
	UBYTE		*strP;		/* String buffer */
	REFSTR		*rsP;		/* String reference ptr */
	VALUE		strval;		/* Value to access the string with */

    /* Access the passed string */
    status = func_treearg_string( fnameP, treeP, 0, &strval );
    if ( status == SCOK ) {
	val_rvalue( &strval );
	if ( ( rsP = val_get_string( &strval ) ) == NULL ) {
	    warning( "%s: bad/uninitialized string", fnameP );
	    status = SCNOTDONE;
	    resultF = FALSE;
	}
	else {
	    /* Examine the string */
	    strP = refstr_bstr( rsP )->bs_bP + refstr_startx( rsP );
	    len = refstr_len( rsP );
	    resultF = TRUE;
	    while ( len-- > 0 ) {
		ch = *strP++;
		if ( !(*rtc)(ch) ) {
		    resultF = FALSE;
		    break;
		}
	    }
	}

	val_release( &strval );
    }
    else
	resultF = FALSE;

    val_set_int( rvalP, 4, resultF );

    return ( status );
}
