/* data.c -- generic data 

	Copyright 2003,2004,2005,2006,2007
		by Mark E. Mallett, MV Communications, Inc.

	See the "LICENSE" file for terms.

This is a collection of generic data elements that might be of use to
various programs in this project area.  Not all data elements are
necessarily expected to be used.  However, data elements that are not
extremely generic should not be put in here.

There are also a couple of misc functions in here because it's
easier than making a new module, and they do sort of go here.

*/

#include <stdlib.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <pwd.h>
#include <netinet/in.h>
#include <arpa/nameser.h>
#include <resolv.h>

#include "mvmf.h"

#include <mml/mml-alloc.h>
#include <mml/mml-ll.h>
#include <mml/mml-str.h>
#include <mml/mml-in.h>
#include <mml/mml-iopf.h>
#include <mml/mml-dns.h>

#include "controls.h"

#define	D_DATA				/* Define data here */
#include "data.h"			/* This will do it.. */


/* Local definitions */



/* External data referenced */



/* External routines used */

extern	char	*getenv();

/* Local data publicly available */

	/* Names of the days of week, 0 = Sunday */
	char	*Daynames[] = {
		"Sunday",	"Monday",	"Tuesday",
		"Wednesday",	"Thursday",	"Friday",
		"Saturday"
	};

	/* Names of the months, 0 = January */
	char	*Monthnames[] = {
		"January",	"February",	"March",
		"April",	"May",		"June",
		"July",		"August",	"September",
		"October",	"November",	"December"
	};


/* Local routines and forward references */



/* Private data */



/*

*//* data_directory( createF )

	Return path to program's data directory

Accepts :

	createF		Whether to create the directory if it doesn't exist.

Returns :

	<value>		Path to validated data directory,
			NULL if no such directory or directory doesn't exist.

Notes :

	This will look for an application control (see control.h) named
	"data_directory" which specifies the path to a directory containing
	the program's private data.  If there is no such control,
	this function will return NULL.

	The primary use of this function is to make sure the data directory
	exists.  "createF" can be passed as TRUE to attempt to create it.

*/

char *
data_directory ARGLIST( (createF) )
    FARG( BOOL		createF )	/* Whether to create the directory */
{
	char		*dirnameP;
	struct stat	statblk;

    dirnameP = (char *)cntrl_str_get( estr_tmp_str( "data_directory" ) );
    if ( dirnameP == NULL )
	return ( NULL );

    /* See if the directory exists */
    if ( stat( dirnameP, &statblk ) < 0 ) {
	if ( !createF )
	    return ( NULL );

	/* Attempt to create it, and then recurse to verify. */
	if ( mkdir( dirnameP, 0750 ) < 0 )
	    return ( NULL );
	return ( data_directory( FALSE ) );
    }

    /* The file exists, make sure it's a directory */
    if ( ( statblk.st_mode & S_IFMT ) != S_IFDIR )
	return ( NULL );

   /* Seems to be there.  Return it. */
   return ( dirnameP );
}
/*

*//* data_init()

	Initialize generic data

Accepts :

	<nothing>

Returns :

	<value>		TRUE if OK
			FALSE if error

Notes :


*/


BOOL
data_init NOARGLIST
{
	char		*strP;
	char		*userP;
	struct passwd	*pwdP;

    /* Setup temporary directory */
    strP = getenv( TMPDIRENV );
    if ( ( strP == NULL ) || ( *strP == NUL ) )
	strP = TMPDIR;			/* Use default */
    TmpdirP = newstr( "data_init", "TmpdirP", strP );


    /* Get my uid and user information.  Note that at this point, the
       USER environment variable *must* resolve to our user id.
    */
    userP = getenv( "USER" );
    if ( userP == NULL ) {
	warning( "data_init: USER environment variable missing." );
	return ( FALSE );
    }
    pwdP = getpwnam( userP );
    if ( pwdP == NULL ) {
	warning( "data_init: USER environment variable \"%s\" matches no user.",
				userP );
	return ( FALSE );
    }
    if ( pwdP->pw_uid != getuid() ) {
	warning( "data_init: USER environment variable \"%s\" is incorrect.",
				userP );
	return ( FALSE );
    }
    
    Myuid = pwdP->pw_uid;
    MyusernameP = newstr( "data_init", "My username", userP );
    MyhomedirP = newstr( "data init", "My homedir", pwdP->pw_dir );

    From_F = FALSE;			/* No "From " seen yet */

    /* Initialize the E_xxx envelope strings */
    if ( ( E_RecipP = refstr_new( "E_RecipP", 100, 0 ) ) == NULL )
	return ( FALSE );
    if ( ( E_SenderP = refstr_new( "E_SenderP", 100, 0 ) ) == NULL )
	return ( FALSE );


#ifdef	QMAIL
    /* Get envelope information from environment */

    /* Envelope sender */
    if ( ( ( strP = getenv( "SENDER" ) ) != NULL ) &&
	 ( *strP != NUL ) ) {
	bstr_cat( refstr_bstr( E_SenderP ), strP );
    }

    /* Envelope recipient */
    if ( ( ( strP = getenv( "RECIPIENT" ) ) != NULL ) &&
	 ( *strP != NUL ) ) {
	bstr_cat( refstr_bstr( E_RecipP ), strP );
    }

#endif	/* QMAIL */

    bstr_nulterm( refstr_bstr( E_SenderP ) );
    bstr_nulterm( refstr_bstr( E_RecipP ) );


    if ( ( Gbs1P = bstr_new( "Gbsp", 15000, 0 ) ) == NULL )
	return ( FALSE );
    if ( ( Gbs2P = bstr_new( "Gbs1p", 15000, 0 ) ) == NULL )
	return ( FALSE );
    if ( ( Gbs3P = bstr_new( "Gbs2p", 15000, 0 ) ) == NULL )
	return ( FALSE );

    if ( ( GtkbsP = bstr_new( "Gtkbsp", 15000, 0 ) ) == NULL )
	return ( FALSE );

    if ( ( GtknsP = tkns_new( GtkbsP, 100,
			     (UBYTE *)"", (UBYTE *)" \t,;:",
			     FALSE ) ) == NULL )
	return ( FALSE );

    return ( TRUE );
}
