/* msgt_file.h -- definitions for the msgstore generic type "file"

	Copyright 2006,2007
		by Mark E. Mallett, MV Communications, Inc.

	See the "LICENSE" file for terms.

Definitions for the msgstore generic type "file" handler

*/


#ifndef	H_MSGT_FILE		/* For multiple inclusions */
#define	H_MSGT_FILE


    /* General type data that goes with this type. */
typedef struct {
    FILE	*gf_fP;			 /* File handle if needed */
    MLTYPE	gf_mltype;		 /* Folder layout type */
    BOOL	gf_oqfromF;		 /* Quote "From " lines? */
    BOOL	gf_blendF;		 /* Blank line at end required? */
    UBYTE	gf_afproot[ PATH_MAX +1 ]; /* Any folder path root */
    UBYTE	gf_afp[ PATH_MAX +1 ];   /* Absolute folder path */
    UBYTE	gf_aflkp[ PATH_MAX +1 ]; /* Absolute path to folder lock */
    LOCKSET	gf_locks;		 /* Locks */
    ATTR_STR	gf_locking;		 /* How to do locking */
}  MSGT_FILE_DATA;


    /* This general type defines an extended interface to each
       specific type, so that common functions may do specific
       type functions in a general way.  This structure defines
       that interface.
    */
typedef struct {
    BOOL	(*fi_folder_check) PROTO( (MSMSG *msmsgP) );
					/* Check an existing folder.  The
					   "check" result is dependant
					   on the type.
					*/

    BOOL	(*fi_msg_new) PROTO( (OMM *ommP, MSMSG *msmsgP) );
					/* Begin new message */

    BOOL	(*fi_msg_end) PROTO( (MSMSG *msmsgP) );
					/* End new message */

    void	(*fi_msg_abort) PROTO( (MSMSG *msmsgP) );
					/* Abort new message */
} MSSTIF_FILE;

    

/* Declarations for functions contained in this module: */

    /* Initialize msgt_file */
BOOL	msgt_file_init PROTO( (void) );


/* Declarations for functions that are provided for use by specific type
   handlers that are of this general type.
*/

    /* Set an integer attribute value */
BOOL	msgt_file_attr_int_set PROTO( (MSGT_FILE_DATA *mfdataP,
				       char *nameP, long val) );

    /* Set a string attribute value */
BOOL	msgt_file_attr_str_set PROTO( (MSGT_FILE_DATA *mfdataP,
				       char *nameP, char *valP) );

    /* Instantiate general type data */
MSGT_FILE_DATA *msgt_file_gtdata_get PROTO( (MSGT_FILE_DATA *srcP) );

    /* Release general type data */
void	msgt_file_gtdata_release PROTO( (MSGT_FILE_DATA *mfdataP) );

    /* Write a message */
BOOL	msgt_file_msg_write PROTO( (OMM *ommP, MSMSG *msmsgP) );


#endif	/* H_MSGT_FILE */
