/* msst-file-mbox.c -- msgstore specific file type "mbox"

	Copyright 2006,2007
		by Mark E. Mallett, MV Communications, Inc.

	See the "LICENSE" file for terms.

Stuff to handle the msgstore specific file type "mbox"

*/

#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <errno.h>
#include <limits.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <string.h>
#include <time.h>
#include <sys/time.h>

#include <mml/mml.h>
#include <mml/mml-alloc.h>
#include <mml/mml-ll.h>
#include <mml/mml-str.h>
#include <mml/mml-in.h>
#include <mml/mml-iopf.h>
#include <mml/mml-dns.h>

#include "mvmf.h"
#include "controls.h"
#include "mvmfmisc.h"
#include "msgpart.h"
#include "omm.h"
#include "data.h"
#include "mvmf-locking.h"
#include "msgstore.h"
#include "msgt-file.h"
#include "msst-file-mbox.h"


/* Local definitions */



/* External data referenced */

extern	MSGT	Msgt_file;		/* Control for "file" general type */


/* External routines used */



/* Local data publicly available */



/* Local routines and forward references */

    /* MSST handler functions: */
static	BOOL	msst_mbox_attr_int_set PROTO( (void *gtdataP, void *stdataP,
					       char *nameP, long val) );
static	BOOL	msst_mbox_attr_str_set PROTO( (void *gtdataP, void *stdataP,
					       char *nameP, char *valP) );
static	void	*msst_mbox_gtdata_get PROTO( (void *srcP) );
static	void	msst_mbox_gtdata_release PROTO( (void *gtdataP) );
static	void	*msst_mbox_stdata_get PROTO( (void *srcP) );
static	void	msst_mbox_stdata_release PROTO( (void *stdataP) );
static	BOOL	msst_mbox_msg_write PROTO( (OMM *ommP, MSMSG *msmsgP) );

    /* MSSTIF_FILE handler functions (generic/specific interface): */
static	BOOL	msstif_mbox_folder_check PROTO( (MSMSG *msmsgP) );
static	void	msstif_mbox_msg_abort PROTO( (MSMSG *msmsgP) );
static	BOOL	msstif_mbox_msg_end PROTO( (MSMSG *msmsgP) );
static	BOOL	msstif_mbox_msg_new PROTO( (OMM *ommP, MSMSG *msmsgP) );

    /* Other support functions */
static	BOOL	mbox_check PROTO( (char *fnameP) );


/* Private data */

    /* The MSSTIF_FILE definition, providing specific type info
       used by general type functions of this type.
    */
static	MSSTIF_FILE Msstif_file_mbox = {
	msstif_mbox_folder_check,	/* Check folder for this type */
	msstif_mbox_msg_new,		/* Begin new mbox message */
	msstif_mbox_msg_end,		/* End new mbox message */
	msstif_mbox_msg_abort		/* Abort new mbox message */
};


    /* The MSST definition for the msst_file_mbox type. */
static	MSST	Msst_file_mbox = {
	"mbox",				/* The name */

	&Msgt_file,			/* Ptr to general type */
	&Msstif_file_mbox,		/* Specific type interface */

	NULL,				/* General data */
	NULL,				/* Specific data */

		/* Handlers */
	msst_mbox_gtdata_get,		/* Get general data */
	msst_mbox_gtdata_release,	/* Release general data */
	msst_mbox_gtdata_get,		/* Get specific data */
	msst_mbox_gtdata_release,	/* Release specific data */
	msst_mbox_attr_int_set,		/* Set integer attribute */
	msst_mbox_attr_str_set,		/* Set string attribute */
	msst_mbox_msg_write		/* Write a message */

};
	

/*

*//* msst_file_mbox_init()

	Initialize the specific file type handler

Accepts :


Returns :

	<value>		STSCODE status code:
			  SCOK if successful
			  something else if not.

Notes :


*/

BOOL
msst_file_mbox_init NOARGLIST
{
	MSGT_FILE_DATA	*mfdataP;	/* general file data */

    /* Get initial (template) general type data */
    Msst_file_mbox.st_gtdataP = msst_mbox_gtdata_get( NULL );
    if ( ( mfdataP = Msst_file_mbox.st_gtdataP ) == NULL ) {
	warning( "msst_file_mbox_init: failed to initialize gtdata." );
	return ( FALSE );
    }
    /* Tweak for this specific type */
    mfdataP->gf_oqfromF = TRUE;		/* Quote "From " lines */
    mfdataP->gf_blendF = TRUE;		/* Need blank line at end */


    /* Get initial (template) specific type data */
    Msst_file_mbox.st_stdataP = msst_mbox_stdata_get( NULL );
    if ( Msst_file_mbox.st_stdataP == NULL ) {
	warning( "msst_file_mbox_init: failed to initialize stdata." );
	return ( FALSE );
    }

    /* Register this type */
    if ( !msst_register( &Msst_file_mbox ) )
	return ( FALSE );

    return ( TRUE );
}

/***********************************************************************
 *                                                                     *
 *                          MSST HANDLERS                              *
 *                                                                     *
 *   These are handler functions which support msst functions for      *
 *   this type.                                                        *
 *                                                                     *
 ***********************************************************************/

/*

*//* msst_mbox_attr_int_set( gtdataP, stdataP, nameP, val )

	Set an integer attribute value for this specific type

Accepts :

	gtdataP		Ptr to general data
	stdataP		Ptr to specific data
	nameP		The attribute name
	val		The value to set for that attribute

Returns :

	<value>		TRUE if OK
			FALSE if not.

Notes :


*/

static BOOL
msst_mbox_attr_int_set ARGLIST( (gtdataP, stdataP, nameP, val) )
   NFARG( void		*gtdataP )	/* General data */
   NFARG( void		*stdataP )	/* Specific data */
   NFARG( char		*nameP )	/* Attribute */
    FARG( long		val )		/* The value to set */
{
    /* Hand off to the general type's value setting */
    return ( msgt_file_attr_int_set( (MSGT_FILE_DATA *)gtdataP,
				      nameP, val ) );
}
/*

*//* msst_mbox_attr_int_set( gtdataP, stdataP, nameP, valP )

	Set a string attribute value for this specific type

Accepts :

	gtdataP		Ptr to general data
	stdataP		Ptr to specific data
	nameP		The attribute name
	valP		The value to set for that attribute

Returns :

	<value>		TRUE if OK
			FALSE if not.

Notes :


*/

static BOOL
msst_mbox_attr_str_set ARGLIST( (gtdataP, stdataP, nameP, valP) )
   NFARG( void		*gtdataP )	/* General data */
   NFARG( void		*stdataP )	/* Specific data */
   NFARG( char		*nameP )	/* Attribute name */
    FARG( char		*valP )		/* The value to set */
{
    /* Hand off to the general type's value setting */
    return ( msgt_file_attr_str_set( (MSGT_FILE_DATA *)gtdataP,
				      nameP, valP ) );
}
/*

*//* msst_mbox_gtdata_get( srcP )

	Instantiate a new copy of general data for this specific type

Accepts :

	srcP		Ptr to source/template data to make a copy of

Returns :

	<value>		Pointer to the general data
			NULL if failed

Notes :

	If the source is NULL, the template data associated with this
	specific type is used as a source.  (That may, in turn, be
	NULL, in which case the general type getter will fill it in.)

*/

void *
msst_mbox_gtdata_get ARGLIST( (srcP) )
    FARG( void		*srcP)		/* Ptr to source to copy from */
{
	MSGT_FILE_DATA	*mfdataP;	/* New data being made */

    /* Use this specific type's template data if no source was given */
    if ( srcP == NULL )
	srcP = Msst_file_mbox.st_gtdataP;

    /* Use general type function to get a copy */
    mfdataP = msgt_file_gtdata_get( (MSGT_FILE_DATA *)srcP );

    /* No further msst refinements needed. */

    return ( (void *)mfdataP );
}
/*

*//* msst_mbox_gtdata_release( gtdataP )

	Get rid of a copy of general data for msgt file type

Accepts :

	gtdataP		Ptr to general data to release

Returns :

	<nothing>

Notes :


*/

void
msst_mbox_gtdata_release ARGLIST( (gtdataP) )
    FARG( void		*gtdataP )	/* Data to release */
{
    /* We can safely use the general type release for all the work */
    msgt_file_gtdata_release( (MSGT_FILE_DATA *)gtdataP );
}
/*

*//* msst_mbox_stdata_get( srcP )

	Instantiate a new copy of general data for this specific type

Accepts :

	srcP		Ptr to source/template data to make a copy of

Returns :

	<value>		Pointer to the general data
			NULL if failed

Notes :

	This specific type has, as yet, no specific data, so this
	function just returns a dummy address, to pretend to succeed.

*/

void *
msst_mbox_stdata_get ARGLIST( (srcP) )
    FARG( void		*srcP)		/* Ptr to source to copy from */
{
    /* Return dummy address */
    return ( (void *)&Msst_file_mbox );
}
/*

*//* msst_mbox_stdata_release( stdataP )

	Get rid of a copy of general data for msgt file type

Accepts :

	stdataP		Ptr to general data to release

Returns :

	<nothing>

Notes :

	This specific type has no specific type data, so this function
	is simply a no-op.

*/

void
msst_mbox_stdata_release ARGLIST( (stdataP) )
    FARG( void		*stdataP )	/* Data to release */
{
}
/*

*//* msst_mbox_msg_write( ommP, msmsgP )

	Write a message for this specific type

Accepts :

	ommP		Ptr to the open message to be written
	msmsgP		Ptr to msgstore message handle for writing

Returns :

	<value>		TRUE if OK
			FALSE if not.

Notes :

	The general type interface provides most (or all!) of the
	logic for writing a message to a file, as guided by the
	parameters associated with this message.

*/

static BOOL
msst_mbox_msg_write ARGLIST( (ommP, msmsgP) )
   NFARG( OMM		*ommP )		/* Open message handle */
    FARG( MSMSG		*msmsgP )	/* Mailstore message for writing */
{
	BOOL		okF;
	MSGT_FILE_DATA	*mfdataP;	/* General type's data */

    mfdataP = (MSGT_FILE_DATA *)msmsgP->mm_gtdataP;

    /* Set up the path to the base lockfile name, which is the same
       as the folder name.  
    */
    if ( snprintf( (char *)&mfdataP->gf_aflkp[0], sizeof(mfdataP->gf_aflkp),
		   "%s",
		     (char *)&mfdataP->gf_afp[0]
		 ) >= sizeof( mfdataP->gf_aflkp ) ) {
	warning( "Lock path too long for \"%s.lock\"",
			(char *)&mfdataP->gf_afp[0] );
	return ( FALSE );
    }

    /* Use the general type's function */
    okF = msgt_file_msg_write( ommP, msmsgP );

    return ( okF );
}

/***********************************************************************
 *                                                                     *
 *                       MSSTIF_FILE HANDLERS                          *
 *                                                                     *
 *   These are handler functions which support specific type interface *
 *   functions between the general type and this specific type         *
 *                                                                     *
 ***********************************************************************/

/*

*//* msstif_mbox_folder_check( msmsgP )

	Specific type check for a folder

Accepts :

	msmsgP		Ptr to the mailstore message handle

Returns :

	<value>		TRUE if OK
			FALSE if not.

Notes :

	A pointer to some general type data is known to be present at this
	point.

	For this general type (file), the check is to see if an
	existing folder is of this specific type.

*/

static BOOL
msstif_mbox_folder_check ARGLIST( (msmsgP) )
    FARG( MSMSG		*msmsgP )	/* Mailstore message handle */
{
	MSGT_FILE_DATA	*mfdataP;	/* general type (file) data */

    mfdataP = msmsgP->mm_gtdataP;	/* Get general data pointer */

    /* Hand off to a check on the absolute path. */
    return ( mbox_check( (char *)&mfdataP->gf_afp[0] ) );
}
/*

*//* msstif_mbox_msg_abort( msmsgP )

	Abort a mbox message already started

Accepts :

	msmsgP		Ptr to mailstore message handle

Returns :

	<none>

Notes :


*/

static void
msstif_mbox_msg_abort ARGLIST( (msmsgP) )
    FARG( MSMSG		*msmsgP )	/* mailbox being written */
{
	MSGT_FILE_DATA	*mfdataP;	/* general type (file) data */
	MSST_MBOX_DATA	*mbxdP;		/* type-specific (mbox) data */

    mfdataP = msmsgP->mm_gtdataP;	/* Get general data */
    mbxdP = msmsgP->mm_stdataP;		/* Get type-specific data ptr */

    /* Truncate the file to where it was when we started */
    ftruncate( fileno(mfdataP->gf_fP), mbxdP->mbx_orgsize );

    /* Take care of locks */
    unlock_lockset( &mfdataP->gf_locks, -1, &mfdataP->gf_aflkp[0] );

    /* Close (& unlock more) */
    fclose( mfdataP->gf_fP );
    mfdataP->gf_fP = NULL;

    /* some higher level will deal with msmsg further. */
}
/*

*//* msstif_mbox_msg_end( msmsgP )

	End a mbox message being written

Accepts :

	msmsgP		ptr to handle for mailbox being written

Returns :

	<value>		TRUE if OK
			FALSE if not.

Notes :

	On failure, the mailbox will still be in some opened state; caller
	 is expected to do further cleanup (such as calling msg_abort() )

*/

static BOOL
msstif_mbox_msg_end ARGLIST( (msmsgP) )
    FARG( MSMSG		*msmsgP )	/* mailbox being written handle */
{
	int		status;
	MSGT_FILE_DATA	*mfdataP;	/* general type (file) data */
	MSST_MBOX_DATA	*mbxdP;		/* type-specific (mbox) data */

    mfdataP = msmsgP->mm_gtdataP;	/* Get general data */
    mbxdP = msmsgP->mm_stdataP;		/* Get type-specific data ptr */

    /* Try to flush the buffers */
    if ( fflush( mfdataP->gf_fP ) < 0 ) {
	/* Something ended badly. */
	return ( FALSE );
    }

    /* Unlock any locks */
    unlock_lockset( &mfdataP->gf_locks, fileno(mfdataP->gf_fP ),
			&mfdataP->gf_aflkp[0] );

    /* Close (&unlock more) */
    if ( fclose( mfdataP->gf_fP ) != 0 )
	return ( FALSE );		/* Error, return badly */

    /* Higher level will deal with the MSMSG */

    return ( TRUE );
}
/*

*//* msstif_mbox_msg_new( ommP, msmsgP )

	Create a new message in a mbox folder

Accepts :

	ommP		Ptr to open message handle
	msmsgP		Ptr to working message handle

Returns :

	<value>		TRUE if OK
			FALSE if not

Notes :


*/

static BOOL
msstif_mbox_msg_new ARGLIST( (ommP, msmsgP) )
   NFARG( OMM		*ommP )		/* Ptr to open message handle */
    FARG( MSMSG		*msmsgP )	/* Ptr to MSMSG in progress */
{
	int		len;
	time_t		timenow;
	char		*strP;		/* Some string */
	MSFOLDER	*msfP;		/* Folder */
	MSGT_FILE_DATA	*mfdataP;	/* general type (file) data */
	MSST_MBOX_DATA	*mbxdP;		/* type-specific (mbox) data */
	struct tm	*tmP;
	char		buf[200];

    msfP = msmsgP->mm_folderP;		/* Get folder ptr */
    mfdataP = msmsgP->mm_gtdataP;	/* Get general data */
    mbxdP = msmsgP->mm_stdataP;		/* Get type-specific data ptr */

    /* Open file if it exists */
    mfdataP->gf_fP = fopen( (char *)&mfdataP->gf_afp[0], "a" );
    if ( mfdataP->gf_fP == NULL )
	return ( FALSE );

    /* Make any lock(s) required */
    lock_parse( (char *)mfdataP->gf_locking.as_strP, &mfdataP->gf_locks );
    if ( !lock_lockset( &mfdataP->gf_locks, fileno(mfdataP->gf_fP),
		        &mfdataP->gf_aflkp[0] ) ) {

	/* Locking failed.  Can't do generic abort here as that presumes
	   that we got the lock.  Instead, close up and abandon.
	*/

	/* First unlock any locks that we successfully acquired. */
	unlock_lockset( &mfdataP->gf_locks, fileno(mfdataP->gf_fP ),
			&mfdataP->gf_aflkp[0] );

	/* Close the write handle */
	fclose( mfdataP->gf_fP );
	mfdataP->gf_fP = NULL;

	return ( FALSE );
    }

    /* Remember the file position */
    mbxdP->mbx_orgsize = ftell( mfdataP->gf_fP );

    /* Write a "From " header if one isn't present. */
    if ( ( mfdataP->gf_fP != NULL ) && ( !From_F ) ) {
	time( &timenow );
	tmP = gmtime( &timenow );

	/* Figure out an address to put on the "From " line.
	   We'll use:
	     - the envelope sender if present (e.g. from the MTA);
	     - the ascertained envelope sender if one was computed;
	     - "mailer-daemon" as a last resort.
	*/
	if ( refstr_len( E_SenderP ) > 0 )
	    strP = (char *)refstr_bstr( E_SenderP )->bs_bP;
	else if ( ( ommP->om_aesenderP != NULL ) &&
		  ( *ommP->om_aesenderP != NUL ) )
	    strP = ommP->om_aesenderP;
	else
	    strP = "Mailer-Daemon";
	snprintf( &buf[0], sizeof(buf), "From %s %s",
			strP, asctime( tmP ) );
	len = strlen( buf );
	if ( buf[len -1] == '\n' )
	   buf[ --len ] = NUL;
	if ( fprintf( mfdataP->gf_fP, "%s +0000\n", buf ) < 0 ) {
	    msstif_mbox_msg_abort( msmsgP );
	    mfdataP->gf_fP = NULL;
	}
    }

    return ( mfdataP->gf_fP != NULL );
}

/***********************************************************************
 *                                                                     *
 *                     Internal support routines                       *
 *                                                                     *
 ***********************************************************************/

/*

*//* mbox_check( fnameP )

	Check a file to see if it's a mbox

Accepts :

	fnameP		Name of file that might be a mbox

Returns :

	<value>		TRUE if it's a mbox
			FALSE if not.

Notes :

	Mostly we assume that a file is a mailbox if the file exists
	and is a regular file.  There are probably some checks that
	could be done to see if the file can't be an mbox.

*/

static BOOL
mbox_check ARGLIST( (fnameP) )
    FARG( char		*fnameP )	/* Filename */
{
	int		i;
	struct stat	statblk;	/* For stat of a file */

    /* Make sure it's a regular file or a symbolic link to one. */
    if ( ( lstat( fnameP, &statblk ) < 0 ) ||
         ( ( statblk.st_mode & S_IFMT ) != S_IFREG ) )
	return ( FALSE );

TODO /* Should probably do more checks here against the file contents. */

    /* Seems to check out, so we'll say it can be an mbox. */
    return ( TRUE );
}
