#!/bin/sh
#
#	Copyright 2003,2004,2005,2006,2007
#		by Mark E. Mallett, MV Communications, Inc.
#
#	See the "LICENSE" file for terms.
# 
#  Create prog-version.h for each new build based on version data file.
#
#  Arguments:
#     prog function
#
#  'prog' is the name of the program to control the version of.
#       The version data is stored in 'prog'.version
#  'function' is one of:
#      "fix" to increment fix number
#      "minor" to increment minor number
#      "major" to increment major number
#      "gen" to generate the prog-version.h file
    
if [ $# -ne 2 ] ; then
    echo "$0 [prog] major    -- increment major number"
    echo "$0 [prog] minor    -- increment minor number"
    echo "$0 [prog] fix      -- increment fix number"
    echo "$0 [prog] gen      -- generate prog-version.h"
    echo ""
    echo "'prog' is the program name;"
    echo "    data file will be 'prog'.version."
    echo "    .h file will be 'prog'-version.h"
    exit 1
fi

PROG="$1"
FUNC="$2"

UCPROG=`echo $PROG | tr '[a-z]' '[A-Z]'`

# Make sure version file exists
if [ ! -s "${PROG}.version" ] ; then
    echo Initializing "${PROG}.version"..
    echo "0" "1" "0" > "${PROG}.version"
fi

major=`grep -v ^# "${PROG}.version" | awk '{print $1}'`
minor=`grep -v ^# "${PROG}.version" | awk '{print $2}'`
fixnum=`grep -v ^# "${PROG}.version" | awk '{print $3}'`

if [ "$FUNC" = "major" ] ; then
    major=`expr $major + 1`
    minor="0"
    fixnum="0"
    echo $major $minor $fixnum > "${PROG}.version"
elif [ "$FUNC" = "minor" ] ; then
    minor=`expr $minor + 1`
    fixnum="0"
    echo $major $minor $fixnum > "${PROG}.version"
elif [ "$FUNC" = "fix" ] ; then
    fixnum=`expr $fixnum + 1`
    echo $major $minor $fixnum > "${PROG}.version"
elif [ "$FUNC" = "gen" ] ; then
    verstr=`echo $major $minor $fixnum |\
            awk '{printf "%d.%d.%d", PROG, $2, $3;}'`
    rm -f "${PROG}-version.h"
    ( cat << !END_OF_HEADER!
/* ${PROG}-version.h -- version information.

	Copyright 2003,2004,2005,2006,2007
		by Mark E. Mallett, MV Communications, Inc.

	See the "LICENSE" file for terms.

   Automatically generated by version.sh on `date`

*/

#ifndef	H_${UCPROG}_VERSION	/* Facilitate multiple inclusions */
#define	H_${UCPROG}_VERSION

#define	${UCPROG}_VER_MAJOR	$major
#define	${UCPROG}_VER_MINOR	$minor
#define	${UCPROG}_VER_FIXNUM	$fixnum

#define	${UCPROG}_VERSION	"$verstr"

#endif	/* H_${UCPROG}_VERSION */
!END_OF_HEADER!
) > ${PROG}-version.h

else
    echo "unknown argument \"$FUNC\" -- use no arguments for usage text."
    exit 1
fi
