/* mcdb.h -- definitions for the mvmcas "mcdb" module

	Copyright 2006,2007
		by Mark E. Mallett, MV Communications, Inc.

	See the "LICENSE" file for terms.

Low-level database functions for mvmcas

Note that there is no corresponding "mcdb" source module.  There are
one or more modules that implement the database functions in a different
way.  The mcdb interface should be generic, though, and followed by
the storage-specific module(s).

*/


#ifndef	H_MCDB		/* For multiple inclusions */
#define	H_MCDB



/* Declarations for functions contained in this module: */

    /* Initialize mcdb module */
BOOL	mcdb_init PROTO( (void) );

    /* Initialize mcdb module again (runtime checks for new stuff) */
BOOL	mcdb_init_again PROTO( (void) );

    /* The mail client "access" functions will access data via the db
       implementation.  The corresponding mail client data is
       exclusively locked for access by the caller, and is guaranteed
       to remain locked (and the pointed-to data available) until the
       data is released.  The MCLIENT pointer returned may be a direct
       pointer into the implementation's in-memory cache, or it may be
       a copy.  If caller modifies the returned value in any way, it
       must be aware that it may be modifying the db cache, regardless
       of whether an update-on-release is later signalled.
    */

    /* Get client info for any mail client type */
MCLIENT *mcdb_mclient_access PROTO( (MCNAME *mcnameP) );

    /* Get client info for specific mail client types */
MCLIENT	*mcdb_mclient_access_auth PROTO( (UBYTE *authnameP) );
MCLIENT	*mcdb_mclient_access_ip PROTO( (MCNAME_IP4 *ipnameP) );


    /* The mail client "delete" function will delete the previously
       accessed mail client from the database.  It will also perform any
       needed release (e.g. mcdb_mclient_release).
    */
BOOL	mcdb_mclient_delete PROTO( (MCLIENT *mcP) );

    /* The mail client "release" functions will release access to a
       mail client data previously gained via "access".  The caller may
       set a flag that indicates that an update should be done; this is
       merely a signal that the cached data has been changed and should
       (at some point) be updated in more permanent storage.  Caller
       should be aware that with or without the update flag, any
       changes to the MCLIENT data that was returned by "access"
       function may change the cached in-memory data.
    */
    /* Release client info */
void	mcdb_mclient_release PROTO( (MCLIENT *mcP, BOOL updateF) );

    /* Print some brief stats into a buffer */
void	mcdb_mclient_stats PROTO( (BSTR *bsP) );

    /* Flag that a mail client db entry has changed an should be saved */
void	mcdb_mclient_update PROTO( (MCLIENT *mcP) );

    /* Synchronize any cached memory and disk */
void	mcdb_sync PROTO( (void) );

#endif	/* H_MCDB */
