/* mcoccur.h -- definitions for the mvmcas mail client occurance module

	Copyright 2006,2007
		by Mark E. Mallett, MV Communications, Inc.

	See the "LICENSE" file for terms.

Stuff relating to mail client occurances.
Also has some code for agent manipulation.

Note:  When a caller must gain access to both a mail client occurance
and an agent occurance simultaneously, all callers must access them in
the same order, so as to avoid deadlocks.  Access the mail client
occurence first, the agent second.

*/


#ifndef	H_MCOCCUR		/* For multiple inclusions */
#define	H_MCOCCUR


    /* Flags in interested-agent flagwords */
#define	IA_NEXT		0x0001	/* Always notify next pass */
#define	IA_SIGCHG	0x0002	/* Notify of significant changes */
#define	IA_ALL		0x0004	/* Notify of everything */



/* Declarations for functions contained in this module: */

    /* Get mail client occurance for any mail client type */
MCOCCUR	*mcoccur_access PROTO( (MCNAME *mcnameP) );

    /* Get mail client occurance for specific mail client types */
MCOCCUR	*mcoccur_access_auth PROTO( (UBYTE *authnameP) );
MCOCCUR	*mcoccur_access_ip PROTO( (MCNAME_IP4 *ipnameP) );

    /* Establish agent interest in mail client occurance */
void	mcoccur_agint_add PROTO( (MCOCCUR *mcoccP, AGENT *agentP, UWORD flags) );

    /* Remove agent interest in mail client occurance */
void	mcoccur_agint_remove PROTO( (MCOCCUR *mcoccP, AGENT *agentP) );

    /* Remove all interest connections for an agent */
void	mcoccur_agint_remove_all PROTO( (AGENT *agentP) );

    /* Process the beginning of a mail client occurance */
void	mcoccur_begin PROTO( (MCOCCUR *mcoccP, MCLIENT *heldmcP) );

    /* Process the end of a mail client occurance */
void	mcoccur_end PROTO( (MCOCCUR *mcoccP, MCLIENT *heldmcP) );

    /* Initialize mcoccur module */
BOOL	mcoccur_init PROTO( (void) );

    /* Send a notification about a mail client occurance */
void	mcoccur_notify PROTO( (UWORD flags,
			       BOOL (*fmtrtc)(), BSTR *nbsP,
			       MCOCCUR *mcoccP, AGENT *agentP,
			       MCLIENT *mcP) );

    /* Release mail client occurance */
void	mcoccur_release PROTO( (MCOCCUR *mcoccP) );

    /* Remove all interest connections for a mail client */
void	mcoccur_remove_all PROTO( (MCOCCUR *mcoccP) );

    /* Print some brief stats into a buffer */
void	mcoccur_stats PROTO( (BSTR *bsP) );

    /* Find or establish a non-persistent agent */
AGENT	*npagent_access PROTO( (AGENT_CTL *actlP, void *infoP, int infoL) );

    /* Acquire a lock on a non-persistent agent */
BOOL	npagent_lock PROTO( (AGENT *agentP) );

    /* Release access to non-persistent agent */
void	npagent_release PROTO( (AGENT *agentP) );

    /* Release a lock on a non-persistent agent */
BOOL	npagent_unlock PROTO( (AGENT *agentP) );

    /* Create a new persistent agent */
AGENT	*pagent_create PROTO( (AGENT_CTL *actlP, void *infoP) );

    /* Delete a persistent agent */
void	pagent_delete PROTO( (AGENT *agentP) );

#endif	/* H_MCOCCUR */
