/* mvmcas-data.c -- Global data for mvmcas

	Copyright 2006,2007
		by Mark E. Mallett, MV Communications, Inc.

	See the "LICENSE" file for terms.

Some global data for mvmcas is here, plus any relevant functions.

*/

#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <errno.h>
#include <limits.h>
#include <sys/types.h>
#include <netinet/in.h>
#include <fcntl.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <pwd.h>
#include <signal.h>

#include <mml/mml.h>
#include <mml/mml-alloc.h>
#include <mml/mml-ll.h>
#include <mml/mml-str.h>
#include <mml/mml-in.h>
#include <mml/mml-iopf.h>
#include <mml/mml-bio.h>
#include <mml/mml-dns.h>

#include "cdb.h"

#include "mvmf.h"
#include "mvmfmisc.h"
#include "mvmcas.h"

#include "msgpart.h"
#include "omm.h"
#include "msgstore.h"
#include "xmtp.h"
#include "smtpd.h"


/* Local definitions */



/* External data referenced */



/* External routines used */



/* Local data publicly available */

    /* These things needed to resolve references in library
       code.  Need to find a way not to make that be true.
       ---------------------- */
	char	*MailboxP = NULL;	/* Default mailbox */
    /* ---------------------- */


    /* General variables */
	BOOL	Debug = 0;		/* Debugging level */
	char	*MyhostnameP;		/* Our hostname */
	struct passwd	*MypwdP;	/* Passwd info for user */
	OMM	*OmmP;			/* The open mail message handle */
	int	Rgov_bad_incr;		/* Bad rgov increment */
	int	Rgov_bad_limit;		/* Bad rgov limit */
	int	Rgov_good_incr;		/* Good rgov increment */
	int	Rgov_good_limit;	/* Good rgov limit */
	long	Rgov_decay_int;		/* How often (seconds) to decay rgov */
	long	Rgov_decay_max;		/* Max interval to apply decay over */
	int	Rgov_decay_mult;	/* Decay rgov by multiplying by this */
	int	Rgov_decay_div;		/*   and then dividing by this */
	struct timeval Starttime;	/* Wall clock when we start */
	long	Tcp_connC;		/* TCP connections */
	long	Udp_connC;		/* UDP connections */
	char	*UsehostnameP;		/* Hostname to use (e.g. for greeting) */

	OPMODE	Opmode = OM_NONE;	/* Operating mode */


    /* Table of IP4 bitmasks, indexed by mask length */
	in_addr_t IP4mask[] = {
		0x00000000,		/* 0 */
		0x80000000,		/* 1 */
		0xc0000000,		/* 2 */
		0xe0000000,		/* 3 */
		0xf0000000,		/* 4 */
		0xf8000000,		/* 5 */
		0xfc000000,		/* 6 */
		0xfe000000,		/* 7 */
		0xff000000,		/* 8 */
		0xff800000,		/* 9 */
		0xffc00000,		/* 10 */
		0xffe00000,		/* 11 */
		0xfff00000,		/* 12 */
		0xfff80000,		/* 13 */
		0xfffc0000,		/* 14 */
		0xfffe0000,		/* 15 */
		0xffff0000,		/* 16 */
		0xffff8000,		/* 17 */
		0xffffc000,		/* 18 */
		0xffffe000,		/* 19 */
		0xfffff000,		/* 20 */
		0xfffff800,		/* 21 */
		0xfffffc00,		/* 22 */
		0xfffffe00,		/* 23 */
		0xffffff00,		/* 24 */
		0xffffff80,		/* 25 */
		0xffffffc0,		/* 26 */
		0xffffffe0,		/* 27 */
		0xfffffff0,		/* 28 */
		0xfffffff8,		/* 29 */
		0xfffffffc,		/* 30 */
		0xfffffffe,		/* 31 */
		0xffffffff		/* 32 */
	};



/* Local routines and forward references */



/* Private data */


/*

*//* mvmcas_data_init()

	Do initialization of data in this module

Accepts :


Returns :

	<value>		TRUE if OK
			FALSE if not.

Notes :


*/

BOOL
mvmcas_data_init NOARGLIST
{
	uid_t		myuid;
	char		*userP;
	char		*strP;
	struct passwd	*pwdP;
	char		path[ PATH_MAX +1 ];

    /* Get hostname */
    if ( gethostname( &path[0], sizeof(path) ) < 0 ) {
	perror( "gethostname" );
	return ( FALSE );
    }
    MyhostnameP = newstr( "mvmcas_data_init", "hostname", &path[0] );
    if ( MyhostnameP == NULL )
	return ( FALSE );

    /* Get home directory. */
    myuid = getuid();			/* My uid */

    /* If there's a USER environment variable and it matches my uid,
       we'll use that info.
    */
    userP = getenv( "USER" );
    if ( userP != NULL ) {
	pwdP = getpwnam( userP );
	if ( ( pwdP == NULL ) || ( pwdP->pw_uid != myuid ) )
	    /* Not a match, use first user matching my uid */
	    pwdP = getpwuid( myuid );
    }
    else
	pwdP = getpwuid( myuid );
    if ( pwdP == NULL ) {
	fprintf( stderr, "Can't find user info!\n" );
	return ( FALSE );
    }
    MypwdP = pwdP;

    /* Init the rgov decay values */
    Rgov_bad_incr = RG_BAD_INCR;
    Rgov_bad_limit = RG_BAD_LIMIT;
    Rgov_good_incr = RG_GOOD_INCR;
    Rgov_good_limit = RG_GOOD_LIMIT;
    Rgov_decay_int = RG_DECAY_INT;
    Rgov_decay_max = RG_DECAY_MAX;
    Rgov_decay_mult = RG_DECAY_MULT;
    Rgov_decay_div = RG_DECAY_DIV;

    return ( TRUE );
}
