/* mvmcas-misc -- misc functions for mvmcas

	Copyright 2006,2007
		by Mark E. Mallett, MV Communications, Inc.

	See the "LICENSE" file for terms.

Some general functions for mvmcas

*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <sys/param.h>
#include <sys/types.h>
#include <netinet/in.h>
#include <ctype.h>
#include <pthread.h>
#include <time.h>

#include <mml/mml.h>
#include <mml/mml-ll.h>

#include "mvmf.h"
#include "mvmcas.h"
#include "mvmcas-misc.h"

/* Local definitions */



/* External data referenced */

extern	int	Debug;		/* Debugging level */
extern	in_addr_t IP4mask[];


/* External routines used */



/* Local data publicly available */



/* Local routines and forward references */



/* Private data */


/*

*//* mclient_init(mcP)

	Initialize (most of) an MCLIENT struct

Accepts :

	mcP		MCLIENT struct

Returns :

	<nothing>

Notes :

	Sets an mclient structure to an initialized (cleared) state.

*/

void
mclient_init ARGLIST( (mcP) )
    FARG( MCLIENT	*mcP )		/* MCLIENT ptr */
{
	time_t		timenow;

    time( &timenow );			/* We'll want this */

    /* The MCNAME part */
    mcP->mc_mcname.mn_type = MCT_NONE;
    mcP->mc_mcname.mn_nameP = NULL;

    mcP->mc_flags = 0;		/* Flags */
    mcP->mc_dbP = NULL;		/* Ptr to db access info */
    mcP->mc_pgd_sec = 0;	/* pre-greet delay in seconds */
    mcP->mc_pgd_wday = 0;	/* pre-greet delay window in days */
    mcP->mc_pgd_wcon = 0;	/* pre-greet delay window in connections */
    mcP->mc_pgd_wconC = 0;	/* # connections this window */
    mcP->mc_pgd_wtime = 0L;	/* Window start */
    mcP->mc_sstate = MCS_NONE;	/* Simple state */
    mcP->mc_sstime = timenow;	/* Time with simple state */
    mcP->mc_ostate = MCO_NONE;	/* Override state */
    mcP->mc_ostime = timenow;	/* Time with override state */
    mcP->mc_fctime = 0;		/* First connection time */
    mcP->mc_lctime = 0;		/* Last connection time */
    mcP->mc_connC = 0;		/* Connection count */
    mcP->mc_grcptC = 0;		/* Good rcpt-to count */
    mcP->mc_brcptC = 0;		/* Bad rcpt-to count */
    mcP->mc_rcptgov = 0;	/* rcpt-to governor */
    mcP->mc_rgtime = 0;		/* rcpt-to governor decay time */

    /* done */
}
/*

*//* mclient_rcptgov_set(mcP, rcptgov, actF)

	Set a mail client's rcpt governor value, with potential side effects

Accepts :

	mcP		MCLIENT struct
	rcptgov		New rcptgov value
	actF		TRUE to act on changes

Returns :

	<value>		SCxxx status code

Notes :

	Sets a new rcptgov value for a mail client.

	If actF is TRUE, the client's state may be changed by the update
	  in governor value.  Otherwise, the value is merely remembered
	  but not acted on.

*/

STSCODE
mclient_rcptgov_set ARGLIST( (mcP, rcptgov, actF) )
   NFARG( MCLIENT	*mcP )		/* MCLIENT ptr */
   NFARG( int		rcptgov )	/* New rcptgov value */
    FARG( BOOL		actF )		/* Whether to act on the value */
{
	STSCODE		status;
	time_t		sst_delta;	/* new delta for sstime */

    if ( rcptgov == mcP->mc_rcptgov )
	return ( SCOK );		/* No change */

    /* If no decay time, initialize it now */
    if ( mcP->mc_rgtime == 0 )
	time( &mcP->mc_rgtime );

    sst_delta = 0;

    if ( actF && ( rcptgov < mcP->mc_rcptgov ) ) {
	/* Value decrease, may have to upgrade the state. */
	switch ( mcP->mc_sstate ) {
	    case MCS_NONE:		/* States that don't change */
	    case MCS_GREY:
	    case MCS_BLOCK_C_ADM:
	    case MCS_BLOCK_G_ADM:
	    case MCS_BLOCK_S_ADM:
		break;

	    case MCS_BLOCK_C_AUTO:	/* Block connection, automatic */
		break;

	    case MCS_BLOCK_G_AUTO:	/* Block at greeting, automatic */
		break;

	    case MCS_BLOCK_S_AUTO:	/* Block with dialog, automatic */
		break;

	    default:
		warning( "mclient_rcptgov_set: bad client state %d",
			   mcP->mc_sstate );
		status = SCERR;
		break;
	}
    }
    else if ( actF ) {
	/* Value increase, may have to downgrade */
	switch ( mcP->mc_sstate ) {
	    case MCS_BLOCK_C_ADM:	/* States that don't change */
	    case MCS_BLOCK_G_ADM:
	    case MCS_BLOCK_S_ADM:
	    case MCS_BLOCK_C_AUTO:
		break;

	    case MCS_NONE:		/* Can go to S, G, or C */
	    case MCS_GREY:
		if ( rcptgov >= RG_INCT_SBLOCK ) {
		    mcP->mc_sstate = MCS_BLOCK_S_AUTO;
		    sst_delta = RG_TIME_SBLOCK;
		}
		/* Fall through */

	    case MCS_BLOCK_S_AUTO:	/* Can go to G or C */
		if ( rcptgov >= RG_INCT_GBLOCK ) {
		    mcP->mc_sstate = MCS_BLOCK_G_AUTO;
		    sst_delta = RG_TIME_GBLOCK;
		}
		/* Fall through */

	    case MCS_BLOCK_G_AUTO:	/* Can go to C */
		if ( rcptgov >= RG_INCT_CBLOCK ) {
		    mcP->mc_sstate = MCS_BLOCK_C_AUTO;
		    sst_delta = RG_TIME_CBLOCK;
		}
		break;


	    default:
		warning( "mclient_rcptgov_set: bad client state %d",
			   mcP->mc_sstate );
		status = SCERR;
		break;
	}
    }

    /* Update the simple-state time if needed */
    if ( sst_delta != 0 ) {
	time( &mcP->mc_sstime );
	mcP->mc_sstime += sst_delta;

	if ( Debug > 1 )
	    printf( "New state %d time %ld\n", mcP->mc_sstate, mcP->mc_sstime );
	    
    }

    /* In any event, set the new governor value */
    mcP->mc_rcptgov = rcptgov;

    return ( status );
}
/*

*//* mcname_ip4_cmp( n1P, n2P )

	Compare two MCNAME_IP4 structs

Accepts :

	n1P		First to compare
	n2P		Second to compare

Returns :

	<value>		<0 if n1P < n2P
			 0 if n1P = n2P
			>0 if n1P > n2P

Notes :


*/

int
mcname_ip4_cmp ARGLIST( (n1P, n2P) )
   NFARG( MCNAME_IP4	*n1P )		/* First to compare */
    FARG( MCNAME_IP4	*n2P )		/* Second to compare */
{
	in_addr_t	ip1, ip2;	/* The two IP addrs */

    /* Masklens must match; we'll sort by masklen first */
    if ( n1P->nip_masklen < n2P->nip_masklen )
	return ( -1 );
    if ( n1P->nip_masklen > n2P->nip_masklen )
	return ( 1 );

    /* Compare the IP values as masked */
    ip1 = ntohl( n1P->nip_addr ) & IP4mask[ n1P->nip_masklen ];
    ip2 = ntohl( n2P->nip_addr ) & IP4mask[ n2P->nip_masklen ];

    if ( ip1 < ip2 )
	return ( -1 );
    if ( ip1 == ip2 )
	return ( 0 );

    return ( 1 );
}
/*

*//* mcname_str( mcnameP, bufP, bufL )

	Generate a name string based on mail client name

Accepts :

	mcnameP		Ptr to the mail client name
	bufP		Ptr to buffer to hold the name string
	bufL		Length of the buffer

Returns :

	<value>		-1 if error (bad input)
			>= 0 for length of string.

Notes :

	If returned value is larger than bufL, the buffer is too small.

*/

int
mcname_str ARGLIST( (mcnameP, bufP, bufL) )
   NFARG( MCNAME	*mcnameP )	/* Mail client name thing */
   NFARG( UBYTE		*bufP )		/* Buffer to hold the string */
    FARG( int		bufL )		/* Length of buffer */
{
	int		len;
	MCNAME_IP4	*ipnameP;	/* for IPv4 type */
	UBYTE		ip_octets[4];	/* IP octets */

    switch ( mcnameP->mn_type ) {
	case MCT_AUTH:			/* Auth name: a string */
	    len = snprintf( (char *)bufP, bufL, "AUTH %s",
			    (char *)mcnameP->mn_nameP );
	    break;

	case MCT_IP:			/* IPv4 name */
	    ipnameP = (MCNAME_IP4 *)mcnameP->mn_nameP;
	    mdns_ip4_iptoba( ntohl( ipnameP->nip_addr ), &ip_octets[0] );
	    if ( ipnameP->nip_masklen == 32 ) {
		/* Omit the mask if /32 */
		len = snprintf( (char *)bufP, bufL, "IP %d.%d.%d.%d",
			        ip_octets[0],
			        ip_octets[1],
			        ip_octets[2],
			        ip_octets[3] );
	    }
	    else {
		len = snprintf( (char *)bufP, bufL, "IP %d.%d.%d.%d/%d",
			        ip_octets[0],
			        ip_octets[1],
			        ip_octets[2],
			        ip_octets[3],
				ipnameP->nip_masklen );
	    }
	    break;


	case MCT_IPv6:			/* IPv6 name */
	    len = -1;			/* Not handled. */
	    break;


	default:
	    warning( "mcname_str: bad type %d", mcnameP->mn_type );
	    len = -1;			/* Error */
	    break;
    }

    return ( len );
}
