/* mvmcas.h -- definitions for the mvmcas application

	Copyright 2006,2007
		by Mark E. Mallett, MV Communications, Inc.

	See the "LICENSE" file for terms.

General definitions for mvmcas

*/


#ifndef	H_MVMCAS		/* For multiple inclusions */
#define	H_MVMCAS

    /* Maximum length of a client name that is a byte string */
#define	MCNAME_BSMAX	50


    /* Pregreeting-delay window parameters.  Once a mail client has passed
       its pre-greeting delay check, it gets a window where it won't be
       checked again.  The window is expressed as either a number of
       connections or a number of days, whichever comes first.  A zero
       will represent a default; these are the defaults.
    */
#define	MCLIENT_PGD_WCON	300
#define	MCLIENT_PGD_WDAY	30


    /* If we haven't heard from a mail client in this long (in seconds)
       it may be removed from memory (subject to other conditions)
    */
#define	MCLIENT_PRUNE_AGE   (30L * 24 * 3600)


    /* If a greylisted mail client has passed its greylisting end time
       without having been heard from this long, it may be removed from
       memory.
    */
#define	MCLIENT_PRUNE_GREY   (6L * 3600)


    /* How long to time out non-persistent agents, in seconds.  This only
       specifies the comparison time -- if the prune frequency is greater
       than this, any given np-agent could stay around for the prune
       frequency time instead.
    */
#define	NPA_TIMEOUT	(5L * 60)

    /* The frequency of pruning non-persistent agents, in seconds. */
#define	NPA_PRUNE_INTERVAL	(2L * 60)

    /* Flag bits in MCLIENT flags word */
#define	MCF_AGG		0x0001	/* An aggregate: used to represent all
				   in its class (e.g. longer prefixes)
				*/
#define	MCF_PROTO	0x0002	/* A prototype: used to instantiate others
				   in its class
				*/
#define	MCF_TRANSIENT	0x0004	/* This client is short-term only; when
				   combined with MCF_PROTO, the instantiated
				   mail clients are short-term.
				*/
#define	MCF_RELAYOK	0x0008	/* OK to relay */
#define	MCF_AUTHREQ	0x0010	/* AUTH is required for any access */
#define	MCF_DNSBL	0x0020	/* DNSBL checks should be applied */
#define	MCF_NO_RGOV	0x0040	/* Do not apply rcpt-to governing */


    /* Types of server connections */
typedef enum {
    SERVE_NONE,			/* None/undef */
    SERVE_UDP,			/* UDP service */
    SERVE_TCP			/* TCP service */
}  SERVE_TYPE;



    /* Types of clients */
typedef enum {
    MCT_NONE,			/* None or unrecognized */
    MCT_IP,			/* IP address */
    MCT_IPv6,			/* IPv6 address (someday) */
    MCT_AUTH			/* AUTH id */
}  MCTYPE;


    /* Simple client states */
typedef enum {
    MCS_NONE,			/* No specific state */
    MCS_GREY,			/* In greylisting period */
    MCS_BLOCK_C_ADM,		/* Block connection, administrative */
    MCS_BLOCK_C_AUTO,		/* Block connection, automatic */
    MCS_BLOCK_G_ADM,		/* Block at greeting, administrative */
    MCS_BLOCK_G_AUTO,		/* Block at greeting, automatic */
    MCS_BLOCK_S_ADM,		/* Block with dialog, administrative */
    MCS_BLOCK_S_AUTO		/* Block with dialog, automatic */
}  MCSSTATE;


    /* Override states */
typedef enum {
   MCO_NONE,			/* None */
   MCO_PASS			/* Client gets a "Pass" */
}  MCOSTATE;


    /* agent operations structure, per communication method */
typedef struct {
    void	(*ac_send)();	/* Function to send a message to the agent.
				   Called with:
				    (AGENT *agentP, UBYTE *bufP, int bufL)
				   returning nothing.
				*/

    BOOL	(*ac_origstr)(); /* Function to provide a string that
				    represents the agent's origin (e.g.
				    its IP/port).  Called with:
				      (AGENT *agentP, UBYTE *bufP, int bufL)
				    returns TRUE or FALSE.
				 */

    void	(*ac_delinfo)(); /* Function to delete or clean up any
				    agent-handler-specific info that is
				    attached to the agent.  Called with:
				      (AGENT *agentP)
				    returns nothing.
				 */
}  AGENT_CTL;


    /* Agent info object */
typedef struct {
    AGENT_CTL	*ag_ctlP;	/* control for this kind of agent */
    time_t	ag_heard;	/* Time we last heard from this agent */
    void	*ag_infoP;	/* Agent info ptr */
    int		ag_infoL;	/* Agent info length (for np-agents) */
    void	*ag_dbP;	/* Any data needed by np-agent database */
    LL_HEAD	ag_mclients;	/* List of interested mail clients. */
}  AGENT;


    /* Mail client name, IP address */
typedef struct {
    in_addr_t	nip_addr;	/* IPv4 address */
    UBYTE	nip_masklen;	/* Mask length */
}  MCNAME_IP4;


    /* Generic mail client name, with pointers to name data */
typedef struct {
    MCTYPE	mn_type;	/* The type */
    void	*mn_nameP;	/* Ptr to the name */
}  MCNAME;


    /* Generic mail client name, with attached storage for the data.
       Useful for collecting data prior to making a MCNAME_P struct;
       can be converted to MCNAME later.
    */
typedef struct {
    MCNAME	mns_mcname;	/* MCNAME pointing to the mns_name data */
    union {
      MCNAME_IP4  nm_ip4;		  /* IP4 */
      UBYTE   nm_bs[ MCNAME_BSMAX +1 ];   /* Any generic byte string */
    }		mns_data;	/* The attached name data */
}  MCNAME_S;


    /* Information about a mail client */
typedef struct {
    MCNAME	mc_mcname;	/* Mail client name (with ptrs) */
    void	*mc_dbP;	/* Any data needed by database access */
    UWORD	mc_flags;	/* Flag bits (MCF_xxx above) */

	/* Pre-greeting delay parms/trackers */
    UBYTE	mc_pgd_sec;	/* # of seconds for pre-greeting delay */
    UBYTE	mc_pgd_wday;	/* # days per window (0 for undef) */
    UWORD	mc_pgd_wcon;	/* Connections per window (0 for undef) */
    UWORD	mc_pgd_wconC;	/* # connections this window */
    time_t	mc_pgd_wtime;	/* Time of window start */
    
	/* General parms */
    MCSSTATE	mc_sstate;	/* Simple state */
    MCOSTATE	mc_ostate;	/* Override state */
    time_t	mc_sstime;	/* Any time to go with simple state */
    time_t	mc_ostime;	/* Any time to go with override state */

	/* Other counters and such */
    time_t	mc_fctime;	/* First connection time */
    time_t	mc_lctime;	/* Last connection time */
    long	mc_connC;	/* Connection count */
    long	mc_grcptC;	/* Count of good rcpt-to */
    long	mc_brcptC;	/* Count of bad rcpt-to */
    int		mc_rcptgov;	/* rcpt-to governor. */
    time_t	mc_rgtime;	/* Last time rcptgov was decayed */
    
}  MCLIENT;




    /* Information about mail client occurance */
typedef struct {
    MCNAME	mco_mcname;	/* Mail client name */
    void	*mco_odbP;	/* Hook for the occurance database */
    LL_HEAD	mco_agents;	/* List of agents tracking this client */
}  MCOCCUR;


    

/***********************************************************************
 *                                                                     *
 *   Configuration items that you may want to look at and/or change    *
 *                                                                     *
 ***********************************************************************/


    /* Mail clients can be controlled by a "rcpt governor" .  The governor
       is a running value that indicates recent rcpt-to behavior.  There
       may be a fancier implementation at some point; the current
       experimental implementation is rather dumb.  It increments the
       governor by some amount for each invalid RCPT-TO, decrements it by
       some amount for each valid RCPT-TO, and changes mail client state
       when various thresholds are passed.  The rcpt governor value may
       also be decreased with the passage of time, as may the client
       state.  This allows a bad reputation to accumulate at one rate and
       fade away at another.  These values control the behaviour.
    */

    /* Each invalid RCPT-TO causes the governor to increment this much. */
#define	RG_BAD_INCR	5

    /* High limit for the governor */
#define	RG_BAD_LIMIT	500

    /* Each valid RCPT-TO causes the governor to increment this much. */
#define	RG_GOOD_INCR	(-1)

    /* Low limit for the governor */
#define	RG_GOOD_LIMIT	(-50)

    /* The rcpt-to governor is decayed after some amount of time has passed.
       This is the default decay interval, in seconds.
    */
#define	RG_DECAY_INT	(5 * 3600L)

    /* If a long period has passed, multiple decays are processed.
       However, this is the maximum time to apply any given decay for.
       i.e. if we haven't heard from a mail client in 3 days, but
       this maximum specifies 1 day, we'll only decay 1 day worth.
    */
#define	RG_DECAY_MAX	(24 * 3600L)

    /* Decay is simply a fractional multiplication, accomplished via
       integer arithmetic.  The fraction is specified as a multiplier
       and a divisor.
    */
#define	RG_DECAY_MULT	9
#define	RG_DECAY_DIV	10
       
       

    /* Thresholds to go to each blocking state when governor is increasing,
       and the corresponding time to stay blocked.
       (It should only be possible to get to CBLOCK by skipping GBLOCK).
    */
#define	RG_INCT_SBLOCK	25		/* Go into "session" block */
#define	RG_TIME_SBLOCK	(3 * 3600L)	/* Time to stay in session block */
#define	RG_INCT_GBLOCK	50		/* Go into "greeting" block */
#define	RG_TIME_GBLOCK	(12 * 3600L)	/* Time to stay in greeting block */
#define	RG_INCT_CBLOCK	100		/* Go into "connection" block */
#define	RG_TIME_CBLOCK	(24 * 3600L)	/* Time to stay in connection block */

    /* Thresholds to go to each unblocking state when governor is decreasing.
       Note the unlikelihood of getting out of some blocks via RCPT monitoring,
       other than via a mix of good & bad in the same session.
    */
#define	RG_DECT_GREY	0		/* Go into greylisting */
#define	RG_DECT_OPEN	(-10)		/* Unblock */



    /* Default location for mvmcas files if not defined above */
#ifndef	MVMCAS_SYSDIR
# define MVMCAS_SYSDIR	MVMF_SYSDIR "/mvmcas"
#endif	/* MVMCAS_SYSDIR */

    /* Default startfile for mvmcas */
#define	MVMCAS_RCFILE	MVMCAS_SYSDIR "/mvmcas-rc"

    /* Directory for user-supplied database info */
#define	MVMCAS_UDBDIR	MVMCAS_SYSDIR "/dbu"

    /* Directory for runtime database files */
#define	MVMCAS_RDBDIR	MVMCAS_SYSDIR "/db"


/***********************************************************************
 *                                                                     *
 *   Configuration items that you are less likely to need to change    *
 *                                                                     *
 ***********************************************************************/




/***********************************************************************
 *                                                                     *
 *    Beyond here are things that you probably don't want to touch.    *
 *                                                                     *
 ***********************************************************************/

    /* Operating modes */
typedef enum {
	OM_NONE,			/* None specified */
	OM_NORMAL,			/* Normal */
	OM_VERSION,			/* Print version info */
} OPMODE;



/* Declarations for functions contained in this module: */

#endif	/* H_MVMCAS */
