/* mvmtr-data.c -- Global data for mvmtr

	Copyright 2003,2004,2005,2006,2007
		by Mark E. Mallett, MV Communications, Inc.

	See the "LICENSE" file for terms.

Some global data for mvmtr is here, plus any relevant functions.

*/

#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <errno.h>
#include <limits.h>
#include <sys/types.h>
#include <fcntl.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <sys/ioctl.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <net/if.h>
#include <pwd.h>
#include <signal.h>
#ifndef SIOCGIFCONF	/* Should probably test @ configure time */
# include <sys/sockio.h>
#endif

#include <mml/mml.h>
#include <mml/mml-alloc.h>
#include <mml/mml-ll.h>
#include <mml/mml-str.h>
#include <mml/mml-avl.h>
#include <mml/mml-in.h>
#include <mml/mml-iopf.h>
#include <mml/mml-bio.h>
#include <mml/mml-dns.h>

#include "cdb.h"

#include "mvmf.h"
#include "mvmfmisc.h"
#include "mvmtr.h"

#include "msgpart.h"
#include "omm.h"
#include "msgstore.h"
#include "xmtp.h"
#include "smtpd.h"


/* Local definitions */



/* External data referenced */



/* External routines used */



/* Local data publicly available */

    /* These things needed to resolve references in library
       code.  Need to find a way not to make that be true.
       ---------------------- */
	char	*MailboxP = NULL;	/* Default mailbox */
    /* ---------------------- */


    /* General variables */
	long	Bodycache = OMM_BODYCACHE_MAX;  /* Amount of body to cache */
	XMTP	*ClientP;		/* Control for XMTP client */
	int	Debug;			/* Debugging level */
	char	*MyhostnameP;		/* Our hostname */
	int	MyipC;			/* Count of addresses in MyipP */
	int	MyipL;			/* Length of MyipP array */
	in_addr_t *MyipP;		/* IPv4 addresses on this host */
	pid_t	Mypid;			/* Process ID */
	struct passwd	*MypwdP;	/* Passwd info for user */
	OMM	*OmmP;			/* The open mail message handle */
	struct timeval Starttime;	/* Wall clock when we start */
	char	*UsehostnameP;		/* Hostname to use (e.g. for greeting) */

	OPMODE	Opmode = OM_NONE;	/* Operating mode */




    /* SMTP client info
       ---------------------- */
	BOOL	SCI_authreqF = FALSE;	/* If auth is required */
	MCBLOCK	SCI_block = MCBLOCK_NONE; /* Client block type */
	char	*SCI_blockcodeP;	/* Block code */
	char	*SCI_blockhowP;		/* How blocked (static string) */
	BOOL	SCI_dnsblF;		/* DNSBL check should be applied */
	in_addr_t SCI_hostip;		/* Client IP address */
	char	*SCI_hostipP;		/* Client IP address string */
	char	*SCI_hostnameP;		/* Client hostname */
	UBYTE	SCI_hostoct[4];		/* Host IP4 octets */
	char	*SCI_myhostipP;		/* Our IP that client connected to */
	char	*SCI_myhostnameP;	/* Our name that was connected to */
	BOOL	SCI_relayokF;		/* If OK to relay */
	BOOL	SCI_suppress_authF;	/* To suppress AUTH advertisement */
	time_t	SCI_t_gl = 0;		/* End of greylisting,
					   doubles as flag.
					*/
	time_t	SCI_t_pass = 0;		/* End of "free pass" -- also flag */
	int	SCI_t_greet;		/* Client greet delay */
    /* ---------------------- */



    /* SMTP Daemon state and parameters
       ---------------------- */
	BOOL	SDS_addrmapF = TRUE;	/* Whether to use addrmap */
	BOOL	SDS_authF = FALSE;	/* Have done AUTH */
	UBYTE	*SDS_authnameP = NULL;	/* Name auth'd with */
	int	SDS_badcmdC = 0;	/* Count of bad commands */
	BOOL	SDS_conn_pgdF = FALSE;	/* If pre-greeting disconnect */
	BOOL	SDS_conn_pgtF = FALSE;	/* If pre-greeting talker */
	BOOL	SDS_conn_validF = FALSE; /* Probably valid (not a probe) */
	BOOL	SDS_dataF = FALSE;	/* If DATA command seen */
	long	SDS_databytes = 0;	/* Max data bytes */
	BOOL	SDS_greetF = FALSE;	/* If we issued a greeting */
	BOOL	SDS_greylistF = FALSE;	/* If greylisting is enabled */
	BOOL	SDS_heloF = FALSE;	/* Got HELO */
	BOOL	SDS_esmtpF = FALSE;	/* If ESMTP (ehlo used) */
	UBYTE	SDS_heloname[ XMTP_DOMAIN_MAX ];
	char	*SDS_localiphostP;	/* Hostname to use when our ip
					   is in the domain part of a
					   user@[ip-address] format addres
					*/

#ifdef	MVMCAS_IN_MVMTR
	BOOL	SDS_mvmcasF = FALSE;	/* If mvmcas is enabled */
#endif	/* MVMCAS_IN_MVMTR */

#ifdef	MVMTR_QMAIL_COMPATIBILITY
	BOOL	SDS_qmcompatF = FALSE;	/* qmail-smtpd compatibility enable */
#endif	/* MVMTR_QMAIL_COMPATIBILITY */

	SDPOLICY SDS_sdp_data;		/* Session delivery policy: DATA */
	SDPOLICY SDS_sdp_rcpt;		/* Session policy for RCPT TO */
	int	SDS_to_general = SMTPD_TOSEC_GENERAL;
					/* General timeout */
	BOOL	SDS_transF = FALSE;	/* Transaction started (got MAIL FROM) */

	    /* List of recipients (rcpt-to list) */
	LL_HEAD	SDS_rcpt;		/* The list head */
	int	SDS_rcptC;		/* The list counter */

	    /* All destination addresses in a transaction */
	AVLTREE	*SDS_dest_treeP;

	    /* rcpt counters/limits */
	int	SDS_rcpt_acceptC;	/* Accepted */
	int	SDS_rcpt_badmax;	/* # bad before we close the channel */
	int	SDS_rcpt_deferC;	/* Defered (4xx) */
	int	SDS_rcpt_rejectC;	/* Rejected (5xx) */

	XMTPMBX	*SDS_rpathP;		/* Reverse path  (from MAIL FROM) */
    /* ---------------------- */



    /* CDB control file stuff
       ---------------------- */
	CDBWRAP	Addrmapdb;		/* Address map database */
    /* ---------------------- */



/* Local routines and forward references */

    /* Add an IP(v4) address to the local IP addresses list */
static	BOOL	add_local_ip_addr PROTO( (in_addr_t addr) );

    /* Get local IP addresses on this host */
static	BOOL	get_local_ip_addrs PROTO( (void) );

/* Private data */


/*

*//* mvmtr_data_init()

	Do initialization of data in this module

Accepts :


Returns :

	<value>		TRUE if OK
			FALSE if not.

Notes :


*/

BOOL
mvmtr_data_init NOARGLIST
{
	uid_t		myuid;
	char		*userP;
	char		*strP;
	struct passwd	*pwdP;
	char		path[ PATH_MAX +1 ];

    /* My process ID */
    Mypid = getpid();

    /* Get hostname */
    if ( gethostname( &path[0], sizeof(path) ) < 0 ) {
	perror( "gethostname" );
	return ( FALSE );
    }
    MyhostnameP = newstr( "mvmtr_data_init", "hostname", &path[0] );
    if ( MyhostnameP == NULL )
	return ( FALSE );

    /* Get local IP addresses on this system */
    if ( !get_local_ip_addrs() )
	return ( FALSE );

    /* Get home directory. */
    myuid = getuid();			/* My uid */

    /* If there's a USER environment variable and it matches my uid,
       we'll use that info.
    */
    userP = getenv( "USER" );
    if ( userP != NULL ) {
	pwdP = getpwnam( userP );
	if ( ( pwdP == NULL ) || ( pwdP->pw_uid != myuid ) )
	    /* Not a match, use first user matching my uid */
	    pwdP = getpwuid( myuid );
    }
    else
	pwdP = getpwuid( myuid );
    if ( pwdP == NULL ) {
	fprintf( stderr, "Can't find user info!\n" );
	return ( FALSE );
    }
    MypwdP = pwdP;

    /* Force the USER environment variable to match who we think we
       are at this point.
    */
    if ( ( userP == NULL ) ||
         ( strcmp( userP, MypwdP->pw_name ) != 0 ) ) {
	if ( setenv( "USER", MypwdP->pw_name, 1 ) != 0 ) {
	    fprintf( stderr,
		   "mvmda_init_user: can't set USER environment variable." );
	    return ( FALSE );
	}
    }

    /* Get info about the SMTP client connection. */
    strP = getenv( "TCPREMOTEIP" );	/* IP of remote host */
    if ( ( strP == NULL ) || ( *strP == NUL ) )
	strP = "unknown";
    SCI_hostipP = newstr( "mvmtr_data_init", "SCI_hostipP", strP );
    SCI_hostip = inet_addr( strP );
    mdns_ip4_iptoba( ntohl( SCI_hostip ), &SCI_hostoct[0] );
					/* Get IP4 octets */


    strP = getenv( "TCPREMOTEHOST" );	/* Name of remote host */
    if ( ( strP == NULL ) || ( *strP == NUL ) )
	strP = "unknown";
    SCI_hostnameP = newstr( "mvmtr_data_init", "SCI_hostnameP", strP );

    strP = getenv( "TCPLOCALIP" );	/* Local IP connected to */
    if ( ( strP == NULL ) || ( *strP == NUL ) )
	strP = "127.0.0.1";
    SCI_myhostipP = newstr( "mvmtr_data_init", "SCI_myhostipP", strP );

    strP = getenv( "TCPLOCALHOST" );	/* Local hostname connected to */
    if ( ( strP == NULL ) || ( *strP == NUL ) )
	strP = SCI_myhostipP;		/* .. use my IP */
    SCI_myhostnameP = newstr( "mvmtr_data_init", "SCI_myhostnameP", strP );

    strP = getenv( "RELAYCLIENT" );	/* Set if relaying is OK */
    SCI_relayokF = ( strP != NULL );

    strP = getenv( "SUPPRESS_AUTH" );	/* Set to suppress AUTH advertisement */
    SCI_suppress_authF = ( strP != NULL );

    return ( TRUE );
}

/***********************************************************************
 *                                                                     *
 *                     Internal support routines                       *
 *                                                                     *
 ***********************************************************************/

/*

*//* add_local_ip_addr( addr )

	Adds an IPv4 address to the list of local IP addresses

Accepts :

	addr		IP address to add

Returns :

	<value>		TRUE if OK
			FALSE if not.

Notes :

	Does not add duplicate addresses (returns OK, tho)

*/

static BOOL
add_local_ip_addr ARGLIST( (addr) )
    FARG( in_addr_t	addr )		/* The address to add */
{
	int		i;

    /* Don't add duplicate */
    for ( i = 0; i < MyipC; ++i )
	if ( MyipP[i] == addr )
	    return ( TRUE );

    /* Expand the list if necessary */
    if ( i == MyipL ) {
	/* Need to expand the array */
	i = MyipL + 10;			/* New size to aim for */
	if ( MyipL == 0 )
	    MyipP = (in_addr_t *)emalloc( "add_local_ip_addr", "new array",
					  i * sizeof( in_addr_t ) );
	else
	    MyipP = (in_addr_t *)erealloc( "add_local_ip_addr", "array",
					    MyipP,
					    i * sizeof( in_addr_t ) );
	if ( MyipP == NULL ) {
	    MyipL = MyipC = 0;
	    return ( FALSE );
	}

	MyipL = i;
    }

    /* Add new entry */
    MyipP[ MyipC++ ] = addr;

    return ( TRUE );
}
/*

*//* get_local_ip_addrs()

	Get the IP addresses associated with this system

Accepts :


Returns :

	<value>		TRUE if OK
			FALSE if not.

Notes :

	Does a best effort attempt.  Failure to get addresses will still
	  give a TRUE return.  FALSE is reserved for real errors like
	  problems allocating memory.

	Stores results in global values.

*/

static BOOL
get_local_ip_addrs NOARGLIST
{
	int		ifX;		/* Interface index in some form */
	int		len;
	int		sockfd;
	char		*ifnameP;
	in_addr_t	addr;
	struct ifreq	ifreq;
	long		ifdataL;
	UBYTE		*ifdataP;	/* Where ifreq data is stored */
	struct ifconf	ifconf;
	struct ifreq	*ifreqP;	/* working ifreq ptr */

    /* Need a handy fd to use the ioctls on */
    sockfd = socket( AF_INET, SOCK_DGRAM, IPPROTO_UDP );
    if ( sockfd < 0 )
        return ( TRUE );

#ifdef	SIOCGSIZIFCONF
    /* Some systems support SIOCGSIZIFCONF ioctl that will tell the
       size needed for the ifreq array.  We'll use it as a starting
       point at least.
    */
    if ( ioctl( sockfd, SIOCGSIZIFCONF, &ifdataL ) < 0 )
#else	/* SIOCGSIZIFCONF -- funky else */
	ifdataL = 10 * sizeof(struct ifreq);
#endif	/* SIOCGSIZIFCONF */

    /* Get initial buffer for ifreqs */
    ifdataP = (UBYTE *)emalloc( "get_local_ip_addrs", "ifdataP", ifdataL );
    if ( ifdataP == NULL ) {
	close( sockfd );
	return ( FALSE );
    }

    /* Loop to try to get interface list and expand buffer needed */
    for ( ; ; ) {
	ifconf.ifc_len = ifdataL;
	ifconf.ifc_req = (struct ifreq *)ifdataP;
	if ( ioctl( sockfd, SIOCGIFCONF, &ifconf ) < 0 ) {
	    warning( "get_local_ip_addrs: ioctl SIOCGIFCONF failed." );
	    close( sockfd );
	    return ( FALSE );
	}

	/* This ioctl will return the length of the data actually returned.
	   If the returned length is fairly close to the buffer length,
	   such that there wasn't room to fit one more, we probably didn't
	   get them all.  Dumb.  At any rate, keep reallocating if we
	   think this could conceivably have happened.
	*/
	if ( ( ifdataL - ifconf.ifc_len ) > ( sizeof( struct ifreq ) + 100 ) )
	    break;

	/* Try to expand the buffer, maybe */
	if ( ifdataL >= ( 1000 * sizeof( struct ifreq ) ) ) {
	    /* Gotta stop sometime */
	    warning( "get_local_ip_addrs: truncating interface list." );
	    break;
	}
	ifdataL += ( 50 * sizeof( struct ifreq ) );
	ifdataP = (UBYTE *)erealloc( "get_local_ip_addrs", "ifdataP",
				     ifdataP, ifdataL );
	if ( ifdataP == NULL ) {
	    close( sockfd );
	    return ( FALSE );
	}
    }

    /* Look through the interface list */
    ifX = 0;				/* Offset into ifdataP */

    for ( ; ;  ) {
	/* How do we know if we are at the end?  If the returned length
	   from the SOICGIFCONF ioctl won't cover this ifreq index.
	*/
	if ( ( ifX + sizeof( struct ifreq ) ) > ifconf.ifc_len )
	    break;
	ifreqP = (struct ifreq *)&ifdataP[ifX];
	ifnameP = &ifreqP->ifr_name[0];

	/* Skip ifX to the next entry */
#ifdef	HAVE_SOCKADDR_SA_LEN
	len = sizeof( ifreqP->ifr_name ) + ifreqP->ifr_addr.sa_len;
	if ( len < sizeof( struct ifreq ) )
	    len = sizeof( struct ifreq );
#else	/* HAVE_SOCKADDR_SA_LEN */
	len = sizeof( struct ifreq );
#endif	/* HAVE_SOCKADDR_SA_LEN */
	ifX += len;

	/* Get flags to see if we should pay attention */
	strcpy( &ifreq.ifr_name[0], ifnameP );
	if ( ioctl( sockfd, SIOCGIFFLAGS, &ifreq ) < 0 )
	    continue;			/* No data, just ignore */
	if ( ( ifreq.ifr_flags & IFF_UP ) == 0 )
	    continue;			/* Skip down interfaces */

	/* Handle this entry. */
	switch( ifreqP->ifr_addr.sa_family ) {
	    case AF_INET:		/* IPv4 */
		addr = ((struct sockaddr_in *)&ifreqP->ifr_addr)->sin_addr.s_addr;
		add_local_ip_addr( addr );
		break;

	    default:			/* Something we don't handle */
		break;
	}
    }

    close( sockfd );

    return ( TRUE );
}
