/* mvmtr.c -- mvmf mail transport agent receiver

	Copyright 2003,2004,2005,2006,2007
		by Mark E. Mallett, MV Communications, Inc.

	See the "LICENSE" file for terms.

Basic SMTPD daemon for mvmf

*/

#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <errno.h>
#include <limits.h>
#include <sys/types.h>
#include <string.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <pwd.h>
#include <signal.h>

#include <mml/mml.h>
#include <mml/mml-alloc.h>
#include <mml/mml-ll.h>
#include <mml/mml-str.h>
#include <mml/mml-symtab.h>
#include <mml/mml-in.h>
#include <mml/mml-iopf.h>
#include <mml/mml-bio.h>
#include <mml/mml-dns.h>

#include "cdb.h"

#include "mvmf.h"
#include "mvmfmisc.h"
#include "mvmtr.h"
#include "mvmtr-version.h"
#include "controls.h"

#include "mfl.h"
#include "mfl-misc.h"
#include "mfl-token.h"
#include "mfl-pp.h"

#include "msgpart.h"
#include "omm.h"
#include "dnsbl.h"
#include "msgstore.h"
#include "data.h"
#include "errwarn.h"
#include "xmtp.h"
#include "smtpd.h"
#include "qmail-smtpd-compat.h"

/* Local definitions */



/* External data referenced */

extern	long	Bodycache;		/* Amount of message body to cache */
extern	XMTP	*ClientP;		/* Control for XMTP client */
extern	int	Debug;			/* Debugging level */
extern	char	*MyhostnameP;		/* Our hostname */
extern	struct passwd	*MypwdP;	/* Our pwd info */
extern	OMM	*OmmP;			/* The open mail message handle */
extern	OPMODE	Opmode;			/* Operating mode */
extern	struct timeval Starttime;	/* Wall clock when we start */
extern	char	*UsehostnameP;		/* Hostname to use (e.g. for greeting) */

    /* SMTP client info
       ---------------------- */
extern	char	*SCI_hostipP;		/* Client IP address */
extern	char	*SCI_hostnameP;		/* Client hostname */
extern	char	*SCI_myhostipP;		/* Our IP that client connected to */
extern	char	*SCI_myhostnameP;	/* Our name that was connected to */
extern	BOOL	SCI_relayokF;		/* If OK to relay */
extern	BOOL	SCI_suppress_authF;	/* To suppress AUTH advertisement */
    /* ---------------------- */

extern	char	*optarg;		/* getopt() argument */
extern	int	optind;			/* getopt() index */


/* External routines used */

extern	BOOL	mvmtr_data_init PROTO( (void) );

/* Local data publicly available */



/* Local routines and forward references */

    /* A function to print a control, via cntrl_walk */
static	void	mvmtr_control_print();

    /* Initialize stuff */
static	BOOL	mvmtr_init PROTO( (void) );
static	BOOL	mvmtr_init_mfl_hooks PROTO( (void) );

    /* Normal mode control */
static	STSCODE	mvmtr_normal PROTO( (MML_IN *inP) );

static	void	usage();



/* Private data */

static	char	*StartfileP = NULL;	/* Name of startup file */
static	BOOL	StartfileF = FALSE;	/* If explicitly specified */

    /* Table of integer controls to define */
static	CNTRLDEF_INT Ctdef_int[] = {
/*
 ------- name -------		val	admin	admin	user	user
					low	 high	 low	high
*/
{ "add_x_peer_info",		  0,	    0,	    1,	   0,	   1	},
{ "addrmap",			  0,	    0,	    1,	   0,	   1	},
{ "admin",			  0,	    0,	    1,	   0,	   0	},
{ "pipe_allow",			  1,	    0,	    1,	   0,	   1	},
{ "smtp_server_autoblock_enable", 1,	    0,	    1,	   0,	   1	},
{ "smtp_server_badcmd_max",	  3,	    0,	   25,     0,	  25	},
{ "smtp_server_badrcpt_delay",	  5,	    0,	  255,     0,	 255	},
{ "smtp_server_badrcpt_max",	  2,	    0,	  100,     0,	 100	},
{ "smtp_server_client_check_defer", 0,	    0,	    1,	   0,	   1	},
{ "smtp_server_greet_delay",	 12,	    0,	  255,	   0,	 255	},
{ "smtp_server_greet_delay_max", 60,	    0,	  600,	   0,	 600	},
{ "smtp_server_greylisting",	  0,	    0,	    1,	   0,	   1	},
{ "smtp_server_helo_check_defer", 0,	    0,	    1,	   0,	   1	},
{ "smtp_server_log_commands",	  0,	    0,	    1,	   0,	   1	},
{ "smtp_server_mail_check_defer", 0,	    0,	    1,	   0,	   1	},
{ "smtp_server_msgsize_max",	  0,	    0,  100*M,	   0,  100*M	},
{ "smtp_server_ss_helo",	 -1,	   -1,	    1,	  -1,	   1	},
{ "smtp_server_strict_sessions",  1,	    0,	    1,	   0,	   1	},
{ "smtp_server_timeout",	300,	    0,	 1200,	   0,	1200	},
{ "smtp_server_timeout_helo",	 30,	    0,	  600,	   0,	 600	},

#ifdef	MVMCAS_IN_MVMTR
{ "mvmcas",			  0,	    0,	    1,	   0,	   1	},
#endif	/* MVMCAS_IN_MVMTR */

#ifdef	MVMTR_QMAIL_COMPATIBILITY
{ "qmail_smtpd_compatibility",	  1,	    0,	    1,	   0,	   1	},
#endif	/* MVMTR_QMAIL_COMPATIBILITY */


{ NULL,				  0,	    0,	    0,	   0,	   0	}
		};


    /* Table of string controls to define */
static	CNTRLDEF_STR Ctdef_str[] = {
/*
 ------- name -------		--------- val --------		adminF
*/
{ "data_directory",		NULL,				 FALSE	},
{ "email_address_primary",	NULL,				 FALSE	},
{ "log_private_name",		NULL,				 FALSE	},
{ "log_program_name",		"mvmtr",			 FALSE	},
{ "prog_checkpassword",		CHECKPASSWORD,			 FALSE  },
{ "smtp_server_client_check",	"pass auth",			 FALSE	},
{ "smtp_server_greeting",	NULL,				 FALSE  },
{ "smtp_server_helo_check",	"pass auth syntax hook me",	 FALSE	},
{ "smtp_server_localip_host",	NULL,				 FALSE  },
{ "smtp_server_mail_check",	"pass auth hook",		 FALSE	},

#ifdef	MVMCAS_IN_MVMTR
{ "mvmcas_host",		"localhost",			 FALSE  },
#endif	/* MVMCAS_IN_MVMTR */

{ NULL,			  NULL,					 FALSE	}
		};


int
main( argc, argv )
	int		argc;
	char		**argv;
{
	char		usageOK;
	int		optch;
	STSCODE		status;
	struct timezone	tz;		/* Dummy timezone */
	MML_IN		*inP;		/* Input stack object */

    /* First thing: get resource info at program start. */
    gettimeofday( &Starttime, &tz );

    /* Initial (if not permanent) client/command fd setup */
    ClientP = xmtp_init( fileno(stdin), fileno(stdout) );

    usageOK = TRUE;

    while ( usageOK &&
	   ( optch = getopt( argc, argv, "b:C:h:vVx" ) ) != EOF ) {

	switch ( optch ) {
	    case	'b':		/* Approximate amount of body to cache */
		Bodycache = atod( optarg );
		break;

	    case	'C':		/* Specify a startup file */
		StartfileP = optarg;
		break;

	    case	'h':		/* My hostname to use */
		if ( UsehostnameP == NULL )
		    UsehostnameP = optarg;
		else {
		    fprintf( stderr, "Duplicate switch \"-h\".\n" );
		    usageOK = FALSE;
		}
		break;

	    case	'v':		/* Print version info */
		if ( Opmode != OM_NONE ) {
		    fprintf( stderr, "Conflicting switch \"-v\".\n" );
		    usageOK = FALSE;
		}
		else
		    Opmode = OM_VERSION;
		break;

	    case	'V':		/* Print version and controls */
		if ( Opmode != OM_NONE ) {
		    fprintf( stderr, "Conflicting switch \"-V\".\n" );
		    usageOK = FALSE;
		}
		else
		    Opmode = OM_CONTROLS;
		break;

	    case	'x':		/* Enable debugging */
		++Debug;
		break;

	    default:
		usageOK = FALSE;
		break;
	}
    }

    /* If no operating mode selected, set to normal */
    if ( usageOK && ( Opmode == OM_NONE ) )
	Opmode = OM_NORMAL;

    /* Validate body cache size */
    if ( usageOK && ( Bodycache < OMM_BODYCACHE_MIN ) ) {
	Bodycache = OMM_BODYCACHE_MIN;
	fprintf( stderr, "Specified bodycache less than minimum, using %ld\n",
			  Bodycache );
    }
    if ( usageOK && ( Bodycache > OMM_BODYCACHE_MAX ) ) {
	Bodycache = OMM_BODYCACHE_MAX;
	fprintf( stderr, "Specified bodycache exceeds maximum, using %ld\n",
			  Bodycache );
    }

#ifdef	MVMTR_QMAIL_COMPATIBILITY
    /* qmail-smtpd might have other arguments depending the patches that
       have been applied.  e.g. hostname is common, and if SMTP AUTH patch
       is applied, a checkpasswd program may appear.  Rather than try to
       guess, we'll just skip anything else here-- user will have to
       configure those other things separately.
    */
    if ( usageOK && ( optind < argc ) )
	optind = argc;
#endif	/* MVMTR_QMAIL_COMPATIBILITY */

    /* No more command line arguments allowed */
    if ( usageOK && ( optind < argc ) )
	usageOK = FALSE;

    if ( !usageOK ) {
	usage();
	exit( 1 );
    }

    if ( !mvmtr_init() ) {
	xmtp_reply( ClientP, TRUE, 421, 0,
		       "%s startup problem: mvmtr unavailable.",
		       MyhostnameP != NULL ? MyhostnameP : "unknown" );
	exit( 1 );
    }

    if ( UsehostnameP == NULL ) {
	UsehostnameP = newstr( "main", "UsehostnameP", MyhostnameP );
	if ( UsehostnameP == NULL ) {
	    xmtp_reply( ClientP, TRUE, 421, 0,
		          "%s hostname alloc: mvmtr unavailable.",
			  MyhostnameP );
	    exit( 1 );
	}
    }

    /* Create an input stack */
    inP = in_new();
    if ( inP == NULL ) {
	xmtp_reply( ClientP, TRUE, 421, 0,
		      "%s no input stack: mvmtr unavailable.",
		      UsehostnameP );
	exit( 1 );
    }

    /* Proceed now depending on the mode */
    switch ( Opmode ) {
	case	OM_NORMAL:
	    status = mvmtr_normal( inP );
	    break;

	case	OM_VERSION:
	    printf( "mvmtr version %s\n", MVMTR_VERSION );
	    status = SCOK;
	    break;

	case	OM_CONTROLS:
	    printf( "// mvmtr version %s\n\n", MVMTR_VERSION );
	    cntrl_walk( mvmtr_control_print );
	    status = SCOK;
	    break;

	default:
	    warning( "Error: unknown Opmode %d.  This is a bug.", Opmode );
	    status = SCERR;
	    break;
    }

    exit( 0 );
}
/*

*//* usage()

	Print usage message

Accepts :

Returns :

*/


static void
usage()
{
    printf( "usage: mvmtr [flags]\n" );
    printf( "    Flags are:\n" );
    printf( "       -b  nnnn         Maximum number of bytes of body to cache.\n" );
    printf( "                         (default is %ld)\n", (long)OMM_BODYCACHE_MAX );
    printf( "       -C file          Specify a startup file to use.  File must\n" );
    printf( "                          be a name in the MFL's system include path.\n" );
    printf( "       -h hostname      Specify my hostname to use in smtpd dialog.\n" );
    printf( "       -v               Give version information.\n" );
    printf( "       -x               Increment debugging level.\n" );
    printf( "\n" );
}
/*

*//* mvmtr_control_print( ctlP )

	Print a control for control-printing mode

Accepts :

	ctlP		Ptr to control struct

Returns :

	<nothing>

Notes :

	Called via cntrl_walk()

*/

static void
mvmtr_control_print ARGLIST( (ctlP) )
    FARG( CNTRL		*ctlP )		/* The control */
{
	CNTRLINT	*ciP;		/* For integer control */
	CNTRLSTR	*csP;		/* For string control */

    switch ( ctlP->ctl_type ) {
	case CTL_INT:
	    ciP = (CNTRLINT *)ctlP->ctl_infoP;
	    printf( "$control_int_set( \"%s\", %d );\n\n",
		   ctlP->ctl_nameP,
		   ciP->cv_int );
	    break;

	case CTL_STR:
	    csP = (CNTRLSTR *)ctlP->ctl_infoP;
	    if ( csP->cs_strP == NULL )
		printf( "$control_string_set( \"%s\", 0 );\n\n",
			  ctlP->ctl_nameP );
	    else
TODO		/* Really need to quote special chars here */
		printf( "$control_string_set( \"%s\", \"%s\" );\n\n",
		          ctlP->ctl_nameP,
		          csP->cs_strP );
	    break;

	default:
	    fprintf( stderr, "mvmtr_control_print: unknown control type %d",
			ctlP->ctl_type );
	    break;
    }

    /* done */
}
/*

*//* mvmtr_init()

	Do general initialization

Accepts :


Returns :

	<value>		TRUE if OK
			FALSE if not.

Notes :


*/

static BOOL
mvmtr_init NOARGLIST
{
	int		i;
	char		path[ PATH_MAX +1 ];

    /* This brings in pthreads stubs for mml use since we don't
       use pthreads in this application.
    */
    mml_fthread();

    if ( !mvmtr_data_init() )		/* Init mvmtr data */
	return ( FALSE );

    /* Define controls used by this application */
    if ( !cntrl_table_def_int( &Ctdef_int[0] ) ) {
	fprintf( stderr, "Problem defining integer controls.\n" );
	return ( FALSE );
    }
    if ( !cntrl_table_def_str( &Ctdef_str[0] ) ) {
	fprintf( stderr, "Problem defining string controls.\n" );
	return ( FALSE );
    }

    /* Initialize lib stuff */
    if ( !data_init() )
	return ( FALSE );

    /* Initialize DNSBL */
    dnsbl_init();

    /* Initialize language stuff */
    if ( mfl_init() != SCOK )
	return ( FALSE );

    /* Populate the standard include directory search list (i.e.
       for #include <x>
    */
    if ( !pp_incdir( INC_SYSTEM, estr_tmp_str( MVMTR_SYSDIR ) ) )
	return ( FALSE );


#ifdef	MVMTR_QMAIL_COMPATIBILITY
    /* Read qmail controls.  Do this before running any MFL start file, so
       that the mvmtr start file has a chance to override anything done
       here.
    */
    if ( cntrl_int_get( estr_tmp_str( "qmail_smtpd_compatibility" ), 1 )
			!= 0 ) {
	if ( qmc_smtpd_init() != SCOK )
	    return ( FALSE );
    }
#endif	/* MVMTR_QMAIL_COMPATIBILITY */


    /* If not specified on command line, consult environment var for
       any alternative startfile.
    */
    if ( StartfileP == NULL ) {
	StartfileP = getenv( MVMTR_RCENV );
	if ( ( StartfileP != NULL ) && ( *StartfileP == NUL ) )
	    StartfileP = NULL;
    }

    /* If there's a start file at this point, it was explicitly specified. */
    if ( StartfileP != NULL )
	StartfileF = TRUE;

#ifdef	MVMTR_SYSRC		/* system-wide MFL RC file */
    if ( StartfileP == NULL )
	StartfileP = MVMTR_SYSRC;
#endif	/* MVMTR_SYSRC */

    if ( ( StartfileP != NULL ) && ( *StartfileP != NUL ) ) {

	/* Use the 'path' buffer to construct a command to exec.
	   Use @include_noerr to suppress warnings if file not found.
	*/
	snprintf( &path[0], PATH_MAX, "@include%s <%s>\n",
		   StartfileF ? "" : "_noerr",
		   StartfileP );

	/* The system-level rc file executes in admin mode.  Care must be
	   taken to relinquish that mode if user-level code is invoked.
	   This probably needs more thought and work to relieve the
	   systemrc script writer of that burden.
	*/
	cntrl_int_set( estr_tmp_str( "admin" ), CA_ADMIN, 1 );
	mfl_run_str( NULL, estr_tmp_str( &path[0] ), estr_tmp_str( "Mc" ) );
	cntrl_int_set( estr_tmp_str( "admin" ), CA_ADMIN, 0 );
    }

    if ( !mvmtr_init_mfl_hooks() )
	return ( FALSE );

    return ( TRUE );
}
/*

*//* mvmtr_init_mfl_hooks()

	Initialize standard mfl hooks

Accepts :

	<nothing>

Returns :

	<value>		TRUE if OK
			FALSE if not.

Notes :

	Subservient to mfl_init, specifically sets up any
	mfl hook functions that are needed internally and that
	haven't been set up by system or user rc files.

	Embedded mfl code is pretty ugly.

	Note that in multi-line embedded code sequences we end each
	line explicitly with a "\n\" sequence per the notes in
	mfl_run_str() function.
*/

static BOOL
mvmtr_init_mfl_hooks NOARGLIST
{
    return ( TRUE );
}
/*

*//* mvmtr_normal(inP)

	Run in normal mode

Accepts :

	inP		Ptr to input stack

Returns :

	<value>		SCxx status code

Notes :


*/

static STSCODE
mvmtr_normal ARGLIST( (inP) )
    FARG( MML_IN	*inP )		/* Input stack */
{
	STSCODE		status;

    /* It's OK to start logging now */
    log_begin();

    /* Hand off to smtpd function. */
    status = smtpd( ClientP );

    /* Finish up client */
TODO	;
    
    return ( status );
}
