/* smtpd -- mvmtr's smtpd driver

	Copyright 2003,2004,2005,2006,2007
		by Mark E. Mallett, MV Communications, Inc.

	See the "LICENSE" file for terms.

Oversees the smtpd dialog for mvmtr

*/

#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <sys/param.h>
#include <sys/types.h>
#include <sys/time.h>
#include <errno.h>
#include <sys/fcntl.h>
#include <sys/select.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#include <ctype.h>
#include <stdarg.h>

#include <mml/mml.h>
#include <mml/mml-alloc.h>
#include <mml/mml-ll.h>
#include <mml/mml-str.h>
#include <mml/mml-avl.h>
#include <mml/mml-symtab.h>
#include <mml/mml-in.h>
#include <mml/mml-iopf.h>
#include <mml/mml-bio.h>
#include <mml/mml-dns.h>

#include "cdb.h"

#include "mvmf.h"
#include "mvmtr.h"
#include "mvmtr-version.h"
#include "controls.h"
#include "msgpart.h"
#include "data.h"
#include "omm.h"
#include "xmtp.h"
#include "smtpd.h"
#include "qmail-queue-if.h"
#include "mvmfmisc.h"
#include "authmisc.h"

#include "mfl.h"
#include "mfl-token.h"
#include "mfl-type.h"
#include "mfl-func.h"
#include "mfl-misc.h"

#ifdef	MVMTR_QMAIL_COMPATIBILITY
#include "qmail-smtpd-compat.h"
#endif	/* MVMTR_QMAIL_COMPATIBILITY */

/* Local definitions */

    /* Structure of a node in the destination address AVL tree */
typedef struct {
    AVLNODE	dst_avl;	/* AVL node header */
    SMTP_DEST	dst_dest;	/* The destination stuff */
}  DESTNODE;


#ifdef	MVMCAS_IN_MVMTR

    /* mvmcas message verb table */
typedef struct {
    char	*cav_nameP;		/* Verb */
    STSCODE	(*cav_rtc)();		/* Routine to call */
}  CAVTE;

#endif	/* MVMCAS_IN_MVMTR */


/* External data referenced */

extern	int	Debug;			/* Debugging level */
extern	int	MyipC;			/* Count of addresses in Myip4P */
extern	in_addr_t *MyipP;		/* IPv4 addresses on this host */
extern	pid_t	Mypid;			/* My Process ID */
extern	OMM	*OmmP;			/* The open mail message handle */
extern	char	*UsehostnameP;		/* Our hostname to use */

    /* SMTP client info
       ---------------------- */
extern	BOOL	SCI_authreqF;		/* If auth is required */
extern	MCBLOCK	SCI_block;		/* Client block type */
extern	char	*SCI_blockcodeP;	/* Block code */
extern	char	*SCI_blockhowP;		/* How blocked */
extern	BOOL	SCI_dnsblF;		/* DNSBL check should be applied */
extern	in_addr_t SCI_hostip;		/* Client IP address */
extern	char	*SCI_hostipP;		/* Client IP address string */
extern	char	*SCI_hostnameP;		/* Client hostname */
extern	UBYTE	SCI_hostoct[4];		/* Host IP4 octets */
extern	char	*SCI_myhostipP;		/* Our IP that client connected to */
extern	char	*SCI_myhostnameP;	/* Our name that was connected to */
extern	BOOL	SCI_relayokF;		/* If OK to relay */
extern	BOOL	SCI_suppress_authF;	/* To suppress AUTH advertisement */
extern	time_t	SCI_t_gl;		/* Greylisting end (and flag) */
extern	int	SCI_t_greet;		/* Greet delay */
extern	time_t	SCI_t_pass;		/* "Free Pass" end (on some checks) */

    /* ---------------------- */


    /* SMTP Daemon state/parameter stuff
       ---------------------- */
extern	BOOL	SDS_addrmapF;		/* Whether to use addrmap */
extern	BOOL	SDS_authF;		/* Did AUTH */
extern	UBYTE	*SDS_authnameP;		/* The authenticated name */
extern	int	SDS_badcmdC;		/* Count of bad commands */
extern	BOOL	SDS_conn_pgdF;		/* If pre-greeting disconnect */
extern	BOOL	SDS_conn_pgtF;		/* If pre-greeting talker */
extern	BOOL	SDS_conn_validF;	/* Probably valid (not a probe) */
extern	BOOL	SDS_dataF;		/* If DATA seen */
extern	BOOL	SDS_greetF;		/* If we issued a greeting */
extern	BOOL	SDS_greylistF;		/* Greylisting enabled/disabled */
extern	BOOL	SDS_heloF;		/* HELO/EHLO seen */
extern	BOOL	SDS_esmtpF;		/* The HELO/EHLO was EHLO */
extern	UBYTE	SDS_heloname[];
extern	char	*SDS_localiphostP;	/* Hostname to use when our ip
					   is in the domain part of a
					   user@[ip-address] format addres
					*/

#ifdef	MVMCAS_IN_MVMTR
extern	BOOL	SDS_mvmcasF;		/* If mvmcas is enabled */
#endif	/* MVMCAS_IN_MVMTR */

#ifdef	MVMTR_QMAIL_COMPATIBILITY
extern	BOOL	SDS_qmcompatF;		/* qmail-smtpd compatibility enable */
#endif	/* MVMTR_QMAIL_COMPATIBILITY */

extern	SDPOLICY SDS_sdp_data;		/* Session delivery policy: DATA */
extern	SDPOLICY SDS_sdp_rcpt;		/* Session policy for RCPT TO */
extern	int	SDS_to_general;		/* General timeout */
extern	BOOL	SDS_transF;		/* In a transaction */

	    /* List of recipients (rcpt-to list) */
extern	LL_HEAD	SDS_rcpt;		/* The list head */
extern	int	SDS_rcptC;		/* The list counter */

	    /* All destination addresses in a transaction */
extern	AVLTREE	*SDS_dest_treeP;

	    /* rcpt counters/ */
extern	int	SDS_rcpt_acceptC;	/* Accepted */
extern	int	SDS_rcpt_badmax;	/* # bad before we close the channel */
extern	int	SDS_rcpt_deferC;	/* Defered (4xx) */
extern	int	SDS_rcpt_rejectC;	/* Rejected (5xx) */

extern	XMTPMBX	*SDS_rpathP;
    /* ---------------------- */


    /* CDB control file stuff
       ---------------------- */
extern	CDBWRAP	Addrmapdb;		/* Address map database */
    /* ---------------------- */



/* External routines used */



/* Local data publicly available */


static	UBYTE	Rcvdhdr[ 1000 ];	/* Big buffer for received line */
static	UBYTE	Xphdr[1000];		/* Buffer for X-Peer-Info line */

/* Local routines and forward references */

    /* Check out the client host */
static	int	client_check PROTO( (XMTP *xmtP) );

    /* These are the client_check functions, subservient to
       client_check, via disatch table.
    */
static	int	client_check_auth PROTO( (XMTP *xmtP) );
static	int	client_check_generic PROTO( (XMTP *xmtP) );
static	int	client_check_pass PROTO( (XMTP *xmtP) );

    /* Check out the connection */
static	int	connect_check PROTO( (XMTP *xmtP) );

    /* Support functions for the destination address AVL tree */
static	int	dest_avl_cmpkey();
static	AVLNODE	*dest_avl_mknode();
static	int	dest_avl_rmnode();

    /* Check for a generic-looking hostname */
static	int	generic_ip_check PROTO( (XMTP *xmtP,
					 in_addr_t haddr, UBYTE *hnameP) );

    /* Check the HELO/EHLO name */
static	int	helo_check PROTO( (XMTP *xmtP) );

    /* These are helo_check functions, subservient to helo_check, via
       dispatch table.  Form and function as per helo_check() interface.
    */
static	int	helo_check_auth PROTO( (XMTP *xmtP, char *heloP) );
static	int	helo_check_dns PROTO( (XMTP *xmtP, char *heloP) );
static	int	helo_check_dnsclient PROTO( (XMTP *xmtP, char *heloP) );
static	int	helo_check_generic PROTO( (XMTP *xmtP, char *heloP) );
static	int	helo_check_hook PROTO( (XMTP *xmtP, char *heloP) );
static	int	helo_check_me PROTO( (XMTP *xmtP, char *heloP) );
static	int	helo_check_pass PROTO( (XMTP *xmtP, char *heloP) );
static	int	helo_check_syntax PROTO( (XMTP *xmtP, char *heloP) );


    /* Check the MAIL FROM name */
static	int	mail_check PROTO( (XMTP *xmtP, XMTPMBX *mbxP) );

    /* These are mail_check functions, subservient to mail_check(),
       via dispatch table, using its interface regime.
    */
static	int	mail_check_auth PROTO( (XMTP *xmtP, XMTPMBX *mbxP) );
static	int	mail_check_dns PROTO( (XMTP *xmtP, XMTPMBX *mbxP) );
static	int	mail_check_hook PROTO( (XMTP *xmtP, XMTPMBX *mbxP) );
static	int	mail_check_pass PROTO( (XMTP *xmtP, XMTPMBX *mbxP) );

#ifdef	MVMTR_QMAIL_COMPATIBILITY
static	int	mail_check_qmail PROTO( (XMTP *xmtP, XMTPMBX *mbxP) );
#endif	/* MVMTR_QMAIL_COMPATIBILITY */


    /* Decode a string into an MDV (mail disposition value, aka acceptance) */
static	int	mdv_from_str PROTO( (char *mvdP) );

    /* Return a string corresponding to an MDV */
static	char	*mdv_to_str PROTO( (int mdv) );

#ifdef	MVMCAS_IN_MVMTR			/* If support for mvmcas */

    /* Check in with mvmcas */
static	STSCODE	mvmcas_check PROTO( (int tosec) );

    /* Process a message from mvmcas */
static	STSCODE	mvmcas_process PROTO( (UBYTE *msgP) );

    /* Functions for processing specific message types */
static	STSCODE	mvmcas_process_hi PROTO( (UBYTE *msgP, UBYTE *tailP) );
static	STSCODE	mvmcas_process_info PROTO( (UBYTE *msgP, UBYTE *tailP) );

    /* Send something to mvmcas */
static	STSCODE	mvmcas_send PROTO( (char *fmtP, ...) );

    /* Send an update message to mvmcas */
static	STSCODE	mvmcas_send_update PROTO( (BOOL stopF, BOOL doneF) );

    /* Initialize for talking to mvmcas */
static	STSCODE	mvmcas_start PROTO( (void) );

    /* Get a token from mvmcas message, with errors & warnings */
static	BOOL	mvmcas_token PROTO( (UBYTE **msgPP, UBYTE *tokP, int tokL) );

    /* Get a token from mvmcas message, interpret as integer */
static	BOOL	mvmcas_token_int PROTO( (UBYTE **msgPP, long *longP) );

    /* Get a token from mvmcas message, interpret as unsigned int */
static	BOOL	mvmcas_token_uint PROTO( (UBYTE **msgPP, ULONG *ulongP) );

#endif	/* MVMCAS_IN_MVMTR */



    /* Various smtp commands */
static	STSCODE	smtpd_cmd_auth PROTO( (XMTP *xmtP, UBYTE *tailP) );
static	STSCODE	smtpd_cmd_data PROTO( (XMTP *xmtP, UBYTE *tailP) );
static	STSCODE	smtpd_cmd_ehlo PROTO( (XMTP *xmtP, UBYTE *tailP) );
static	STSCODE	smtpd_cmd_helo PROTO( (XMTP *xmtP, UBYTE *tailP) );
static	STSCODE	smtpd_cmd_help PROTO( (XMTP *xmtP, UBYTE *tailP) );
static	STSCODE	smtpd_cmd_mail PROTO( (XMTP *xmtP, UBYTE *tailP) );
static	STSCODE	smtpd_cmd_noop PROTO( (XMTP *xmtP, UBYTE *tailP) );
static	STSCODE	smtpd_cmd_quit PROTO( (XMTP *xmtP, UBYTE *tailP) );
static	STSCODE	smtpd_cmd_rcpt PROTO( (XMTP *xmtP, UBYTE *tailP) );
static	STSCODE	smtpd_cmd_rset PROTO( (XMTP *xmtP, UBYTE *tailP) );

    /* Auth subcommands */
static	STSCODE	smtpd_cmd_auth_login PROTO( (XMTP *xmtP, UBYTE *tailP) );
static	STSCODE	smtpd_cmd_auth_plain PROTO( (XMTP *xmtP, UBYTE *tailP) );

    /* Common code for HELO */
static	STSCODE	smtpd_cmd_helo_first PROTO( (XMTP *xmtP, char *ecmdP,
						UBYTE **tailPP) );

    /* Check an address to see if we will accept mail for it */
static	int	smtpd_address_check PROTO( (UBYTE *addrP, int addrL,
					    int mdv) );

    /* Check the addrmap database for a recipient address */
static	int	smtpd_addrmap_check PROTO( (UBYTE *addrP, int addrL) );

    /* Open the addrmap database for later use */
static	BOOL	smtpd_addrmap_open PROTO( (void) );

    /* Set up authorized state */
static	STSCODE	smtpd_authsetup PROTO( (XMTP *xmtP, UBYTE *unameP) );

    /* Accept and process an smtp command */
static	STSCODE	smtpd_command PROTO( (XMTP *xmtP) );

    /* Access/add a destination block */
static	SMTP_DEST *smtpd_dest_access PROTO( (XMTPMBX *mbxP) );

    /* Apply an MFL hook that's used in a dialog */
static	STSCODE	smtpd_hook_dialog PROTO( (char *cmdP, int *mdvP ) );

#ifdef	QMAIL
    /* Hand off a message to qmail-queue */
static	STSCODE	smtpd_qmail_queue PROTO( (OMM *ommP) );

    /* Called for each destination by avlwalk for smtpd_qmail_queue */
static	int	smtpd_qmail_queue_dest();
#endif	/* QMAIL */

    /* Add one to the rcpt-to table */
static	STSCODE	smtpd_rcpt_add PROTO( (XMTP *xmtP, SMTP_DEST *destP) );

    /* Create "Received" header in the Rcvdhdr buffer */
static	void	smtpd_rcvdhdr PROTO( (void) );

    /* Count bad commands, possibly reply if max is reached */
static	STSCODE	smtpd_reply_bad_command PROTO( (XMTP *xmtP) );

    /* Issue message about too many rejected recipients */
static	STSCODE	smtpd_reply_badmax PROTO( (XMTP *xmtP) );

    /* Issue syntax error message */
static	STSCODE	smtpd_reply_syntax PROTO( (XMTP *xmtP) );

    /* Make sure a command has no [more] arguments */
static	STSCODE	smtpd_require_eol PROTO( (XMTP *xmtP, UBYTE *tailP) );

    /* Make sure a HELO has been seen */
static	STSCODE	smtpd_require_helo PROTO( (XMTP *xmtP) );

    /* Make sure a transaction is in progress */
static	STSCODE	smtpd_require_transaction PROTO( (XMTP *xmtP) );

    /* Check for and rewrite local-ip in user@[ip-address] address form */
static	STSCODE smtpd_rewrite_local_ip PROTO( (XMTPMBX *mbxP) );

    /* Handle an smtpd session */
static	STSCODE	smtpd_session PROTO( (XMTP *xmtP) );

    /* Set new session policy */
static	STSCODE	smtpd_session_policy PROTO( (SDPOLICY rcpt, SDPOLICY data,
					     BOOL wantendF) );

    /* Get a token, delimited by a char, into a tmpstr() */
static	UBYTE	*tmptoken_char PROTO ( (UBYTE **strPP, UBYTE delch) );


/* Private data */

#ifdef	MVMCAS_IN_MVMTR			/* If mvmcas support enabled */

static	BSTR	*Mvmcas_bstrP;		/* BSTR for sending to mvmcas */
static	BOOL	Mvmcas_expectF;		/* TRUE if we think we expect a
					   mesage from mvmcas.
					*/
static	int	Mvmcas_fd = -1;		/* fd for talking to mvmcas */
static	struct sockaddr_in Mvmcas_to;	/* Socket info for mvmcas server */

    /* Table of message types that we handle, per verb */
static	CAVTE	Mvmcas_vtbl[] = {
/*	verb		rtc		*/
{	"hi",		mvmcas_process_hi		},
{	"info",		mvmcas_process_info		},

{	NULL,		NULL				}
};

#endif	/* MVMCAS_IN_MVMTR */

	/* Some transitory SMTP dialog elements */
static	char	*SDD_ecodeP;		/* Any extended reply code */
static	char	*SDD_logP;		/* Any extra logging element */
static	char	*SDD_replyP;		/* SMTP reply overriding the default */


    /* Table of SMTP commands and their handler functions */
static	struct {
    char	*cm_nameP;	/* Command name */

    STSCODE	(*cm_rtc) PROTO( (XMTP *xmtP, UBYTE *tailP) );
				/* Handler */
}  Smtpcmds[] = {

{	"AUTH",		smtpd_cmd_auth		},
{	"DATA",		smtpd_cmd_data		},
{	"EHLO",		smtpd_cmd_ehlo		},
{	"HELO",		smtpd_cmd_helo		},
{	"HELP",		smtpd_cmd_help		},
{	"MAIL",		smtpd_cmd_mail		},
{	"NOOP",		smtpd_cmd_noop		},
{	"QUIT",		smtpd_cmd_quit		},
{	"RCPT",		smtpd_cmd_rcpt		},
{	"RSET",		smtpd_cmd_rset		},

{	NULL,		NULL			},

};



    /* Table of AUTH methods we deal with.  Table ends with NULL method
       name pointer.

       NB: Coordinate this with the capability announcement in EHLO.
       One would think that could be driven off of this table, but
       we may need finer control of the advertisement, so we don't
       do that.
    */
static struct {
    char	*au_mnameP;		/* Method name */
    STSCODE	(*au_rtc) PROTO( (XMTP *xmtP, UBYTE *argP) );
					/* Handler */
}  Smtpauths[] = {

{	"LOGIN",	smtpd_cmd_auth_login,		},
{	"PLAIN",	smtpd_cmd_auth_plain,		},

{	NULL,		NULL				}

};



    /* Table of client checks that we do.  Each check has a name
       and a function that processes it.  The function will
       return an mdv mail disposition value, as described in
       the client_check notes.
    */
static struct {
    char	*cc_nameP;		/* client check name */
    int		(*cc_rtc) PROTO( (XMTP *xmtP) );
					/* Handler */
}  Clientchecks[] = {

{	"auth",		client_check_auth,		},
{	"generic",	client_check_generic,		},
{	"pass",		client_check_pass,		},

{	NULL,		NULL				}



};


    /* Table of HELO checks that we do.  Each check has a name
       and a function that processes it.  The function will
       return an mdv mail disposition value, as described
       in the helo_check notes.
    */
static struct {
    char	*hc_nameP;		/* HELO check name */
    int		(*hc_rtc) PROTO( (XMTP *xmtP, char *heloP) );
					/* Handler */
}  Helochecks[] = {

{	"auth",		helo_check_auth,		},
{	"dns",		helo_check_dns,			},
{	"dnsclient",	helo_check_dnsclient,		},
{	"generic",	helo_check_generic,		},
{	"hook",		helo_check_hook,		},
{	"me",		helo_check_me,			},
{	"pass",		helo_check_pass,		},
{	"syntax",	helo_check_syntax,		},

{	NULL,		NULL				}

};



    /* Table of MAIL checks that we do.  Each check has a name
       and a function that processes it.  The function will
       return an mdv mail disposition value, as described in
       the mail_check notes.
    */
static struct {
    char	*mc_nameP;		/* MAIL check name */
    int		(*mc_rtc) PROTO( (XMTP *xmtP, XMTPMBX *mbxP) );
					/* Handler */
}  Mailchecks[] = {

{	"auth",		mail_check_auth,		},
{	"dns",		mail_check_dns,			},
{	"hook",		mail_check_hook,		},
{	"pass",		mail_check_pass,		},

#ifdef	MVMTR_QMAIL_COMPATIBILITY
{	"qmail",	mail_check_qmail,		},
#endif	/* MVMTR_QMAIL_COMPATIBILITY */

{	NULL,		NULL				}

};


/*

*//* smtpd(xmtP)

	Do the smtpd service.  

Accepts :

	xmtP		Ptr to XMTP control block

Returns :

	<value>		status code

Notes :

	A lot of this startup stuff should probably be forked.

*/


STSCODE
smtpd ARGLIST( (xmtP) )
    FARG( XMTP		*xmtP )		/* XMTP control block */
{
	STSCODE		status;
	int		i;
	int		mdv;
	time_t		time1;
	time_t		timenow;
	UBYTE		*bP;		/* Generic pointer */

    status = SCOK;			/* Presume success */
    SDD_logP = NULL;
    mdv = MDV_DUNNO;

    /* Initialize the recipient list */
    ll_init( &SDS_rcpt );

    /* Initialize the destination tree */
    SDS_dest_treeP = avlnew( NULL,
			     dest_avl_cmpkey,
			     dest_avl_mknode,
			     dest_avl_rmnode );
    if ( SDS_dest_treeP == NULL )
	return ( SCERR );

    /* Get start time so we can calculate delay time later */
    time( &time1 );

    /* Set session policies */
    SDS_sdp_data = SDP_ACCEPT;
    SDS_sdp_rcpt = SDP_ACCEPT;

    /* Fetch some of the control values that aren't going to change. */
    SCI_t_greet = cntrl_int_get(
			estr_tmp_str( "smtp_server_greet_delay" ), 12 );

    SDS_addrmapF = cntrl_int_get( estr_tmp_str( "addrmap" ), 0 );
    SDS_greylistF = cntrl_int_get(
			  estr_tmp_str( "smtp_server_greylisting" ), 0 );
    SDS_localiphostP = cntrl_str_get(
			  estr_tmp_str( "smtp_server_localip_host" ) );
    SDS_rcpt_badmax = cntrl_int_get(
			  estr_tmp_str( "smtp_server_badrcpt_max" ), 2 );
    SDS_to_general = cntrl_int_get( estr_tmp_str( "smtp_server_timeout" ),
				    SDS_to_general );

#ifdef	MVMTR_QMAIL_COMPATIBILITY
    SDS_qmcompatF = cntrl_int_get(
			estr_tmp_str( "qmail_smtpd_compatibility" ), 1 );
#endif	/* MVMTR_QMAIL_COMPATIBILITY */

    if ( SDS_addrmapF && !smtpd_addrmap_open() ) {
	xmtp_reply( xmtP, TRUE, 421, NULL,
		    "server startup -- can't open addrmap." );
	return ( SCNOTDONE );
    }

#ifdef	MVMCAS_IN_MVMTR			/* If mvmcas support */
    SDS_mvmcasF = cntrl_int_get( estr_tmp_str( "mvmcas" ), 0 );

    if ( SDS_mvmcasF ) {
	/* Start up conversation with mvmcas server */
	mvmcas_start();

	/* Send a message about the IP that connected to us */
	mvmcas_send( "START IP %s", SCI_hostipP );

	/* Handle any responses */
	mvmcas_check( 3 );
    }
#endif	/* MVMCAS_IN_MVMTR */

    /* If the client gets a free pass, let's log that. */
    if ( SCI_t_pass != 0 ) {
	smtpd_log( xmtP, "PASS+", NULL, "Free pass" );
	mdv = MDV_ACCEPT;
    }

    /* Do initial connect check */
    if ( ( status == SCOK ) &&
         ( mdv <= 0 ) ) {
	mdv = connect_check( xmtP );

	/* Trap and interpret any special immediate disconnect codes.
	   Others will be handled as-is.
	*/
	if ( mdv == MDV_UNKNOWN ) {
	    /* Disconnect without SMTP reply. */
	    smtpd_log( xmtP, "CONN-", SDD_logP, "Connect-check" );
	    status = SCNOTDONE;		/* Will disconnect. */
	}
    }

    /* Do a client validation unless we're deferring that. */
    if ( ( status == SCOK ) && ( mdv <= 0 ) &&
         ( cntrl_int_get( estr_tmp_str( "smtp_server_client_check_defer" ),
		        0 ) == 0 ) )
	mdv = client_check( xmtP );

    /* Honor any connection block from the assessment server.  This
       can get an override from the connect/client checks or from a free
       pass.
    */
    if ( ( status == SCOK ) &&
         ( mdv != MDV_ACCEPT ) &&
         ( SCI_block == MCBLOCK_CONN ) ) {
	smtpd_log( xmtP, "CONN-", SDD_logP, NULL );
	status = SCNOTDONE;		/* This will disconnect */
    }

    /* Now if the connect-check failed, we will be more formal about
       disconnecting.  Note that we don't check the session policy, since
       the connect-check deferral is a more relevant test.
    */
    if ( ( status == SCOK ) && ( mdv == MDV_REJECT ) ) {
	smtpd_log( xmtP, "CONN-", SDD_logP, "Connect-check" );
	if ( SDD_replyP != NULL )
	    xmtp_reply( xmtP, TRUE, 554, SDD_ecodeP, "%s",
			         SDD_replyP );
	else
	    xmtp_reply( xmtP, TRUE, 554, SDD_ecodeP,
			 "Connection to MTA not wanted.\n"
			 "Authorized clients use submission (port 587)." );
	status = SCNOTDONE;
    }

    if ( status == SCOK ) {
	/* Trap any pre-greeting input */
	time( &timenow );

	/* Limit greet delay if needed (do this late, in case of mvmcas
	   updates)
	*/
	i = cntrl_int_get( estr_tmp_str( "smtp_server_greet_delay_max" ), 30 );
	if ( SCI_t_greet > i )
	    SCI_t_greet = i;

	/* We don't do pre-greeting check if the delta has already elapsed,
	   or if the client has gotten a free pass.
	*/
	if ( ( ( timenow - time1 ) < SCI_t_greet ) &&
	     ( SCI_t_pass == 0 ) ) {
	    bio_timeout_set( &xmtP->mt_in,
			     SCI_t_greet - (timenow - time1),
			     0);

	    status = xmtp_getline( xmtP, Gbs1P );
	    if ( ( status == SCOK ) || ( status == SCNOTDONE ) ) {
		/* Got data, or too much data */
		time( &timenow );
		smtpd_log( xmtP, "PREGREET-", NULL, "time=%d (%s)",
				timenow - time1,
				Gbs1P->bs_bP );
		SDS_conn_pgtF = TRUE;
		status = SCOK;		/* In case it was overflow */
	    }
	    else if ( status == SCERR ) {
		/* Error reading; they disconnected before greeting. */
		time( &timenow );
	        smtpd_log( xmtP, "DISCONNECT-PG-", NULL, "time=%d",
				 timenow - time1 );
		SDS_conn_pgdF = TRUE;
		status = SCNOTDONE;
	    }
	    else if ( status == SCBREAK ) {
		/* Timeout, this is OK */
		status = SCOK;
	    }
	}
    }

    /* Time to say something.  How shall I greet thee? */

    /* Pre-greeting is most unwanted. */
    if ( ( status == SCOK ) && SDS_conn_pgtF ) {
	xmtp_reply( xmtP, TRUE, 421, NULL,
		     "%s Invalid pipelining (pre-greeting traffic seen.)",
		     UsehostnameP );
	status = smtpd_session_policy( SDP_DEFER, SDP_DEFER, TRUE );
    }
	
    /* Issue a go-away greeting if greet-blocked and no free pass. */
    else if ( ( status == SCOK ) &&
	      ( SCI_t_pass == 0 ) &&
              ( SCI_block == MCBLOCK_GREET ) ) {
	smtpd_log( xmtP, "GREET-", NULL, "how=%s code=%s", 
		    SCI_blockhowP,
		    SCI_blockcodeP );
	xmtp_reply( xmtP, TRUE, 421, NULL,
		    "%s Blocked : %s : code=%s",
		    UsehostnameP,
		    SCI_blockhowP,
		    SCI_blockcodeP );
	status = smtpd_session_policy( SDP_DEFER, SDP_DEFER, TRUE );
    }

    /* Deferral if greylisted and no free pass */
    else if ( ( status == SCOK ) &&
	      ( SCI_t_pass == 0 ) &&
	      ( SCI_t_gl != 0 ) && SDS_greylistF ) {
	smtpd_log( xmtP, "GREY-", NULL, "greylisted" );
	xmtp_reply( xmtP, TRUE, 421, NULL,
		    "%s Please try later.",
		    UsehostnameP );
	status = smtpd_session_policy( SDP_DEFER, SDP_DEFER, TRUE );
    }

    /* Maybe we can proceed now. */
    else if ( status == SCOK ) {
	bP = cntrl_str_get( estr_tmp_str( "smtp_server_greeting" ) );
	if ( Debug > 0 )
	    fprintf( stderr, "at greeting; %s%s%s pass=%ld g=%d\n",
			SCI_relayokF ? " relayok" : "",
			SCI_block == MCBLOCK_NONE ? "" : " block",
			SCI_t_gl == 0 ? "" : " grey",
			SCI_t_pass,
			SCI_t_greet );

	if ( bP != NULL ) {
	    xmtp_reply( xmtP, FALSE, -220, NULL,
			"%s here - MVMTR version %s;\n",
			UsehostnameP, MVMTR_VERSION );
	    xmtp_reply( xmtP, FALSE, -220, NULL, "%s", bP );
	}
	xmtp_reply( xmtP, TRUE, 220, NULL,
			"%s here - MVMTR version %s\n",
			UsehostnameP, MVMTR_VERSION );

	smtpd_log( xmtP, "GREET+", NULL, NULL );
    }

    /* If still hunky-dory, do the smtp dialog. */
    if ( status == SCOK ) {
	SDS_greetF = TRUE;		/* We issued a greeting of some kind */

	status = smtpd_session( xmtP );
    }

#ifdef	MVMCAS_IN_MVMTR
    if ( SDS_mvmcasF )
	/* Send an update message, with stop & done status so that our
	   occurances get removed.
	*/
	mvmcas_send_update( TRUE, TRUE );
#endif	/* MVMCAS_IN_MVMTR */

    return ( status );
}
/*

*//* smtpd_log( xmtP, kwdP, logP, fmtP, args... )

	Log some info about the smtpd session

Accepts :

	xmtP		Ptr to XMTP control block (optional)
	kwdP		Keyword that starts the log line
	logP		Any optional log classification string
	fmtP		printf-style format string
	args		any arguments to the format string

Returns :

	<nothing>

Notes :

	Assumes stdc with function prototypes.

	Logs smtpd session info in a standard format and in whatever logging
	  method is enabled.

	The initial keyword is separated from the format string so that
	  (a) all log lines start with a keyword that can be parsed;
	  (b) standard info will follow the keyword;
	  (c) non-standard stuff from fmtP and args follow the standard stuff

	Likewise, logP is an optional string that may add some clarity
	  to the log line.  It's useful when logging actions caused by
	  hooks, for example.  This string is passed separately in order
	  to make the logging of this text uniform in format.  This string
	  should be short.

	The 'xmtP' arg is not currently used, but is passed in case we
	  ever want to make use of it.

	Uses a static BSTR Gbs3P for buffering.  If multithreading is ever
	  needed, will probably want to fold this into the XMTP struct.

*/


void
smtpd_log ARGLIST( (xmtP, kwdP, logP, fmtP, va_decl) )
   NFARG( XMTP		*xmtP )		/* XMTP control block */
   NFARG( char		*kwdP )		/* log keyword */
   NFARG( char		*logP )		/* Extra log classification */
   NFARG( char		*fmtP )		/* message format string */
    FARG( ... )				/* stdc variable arguments */
{
	va_list		ap;		/* varargs arg pointer */
	BOOL		okF;
	UBYTE		*prognameP;

    /* Start varargs */
    va_start( ap, fmtP );

    Gbs3P->bs_bC = 0;

    /* Start with program name and keyword */
    prognameP = cntrl_str_get( estr_tmp_str( "log_program_name" ) );
    if ( prognameP == NULL )
	prognameP = "mvmtr";
    okF = bstr_printf( Gbs3P, "%s %s", prognameP, kwdP);

    /* Add the log note if one is given */
    if ( logP != NULL )
	okF = okF && bstr_printf( Gbs3P, " [%s]", logP );

    /* Add process ID */
    okF = okF && bstr_printf( Gbs3P, " pid=%d", Mypid );

    /* Add remote IP */
    okF = okF && bstr_printf( Gbs3P, " IP=%s",
			      SCI_hostipP == NULL ? "0" : SCI_hostipP );

    /* HELO name if known */
    if ( SDS_heloF )
	okF = okF && bstr_printf( Gbs3P, " HELO %s", &SDS_heloname[0] );

    /* Finish standard part with semi */
    okF = okF && bstr_printf( Gbs3P, " ; " );

    /* Add any other passed stuff */
    if ( fmtP != NULL )
	okF = okF && bstr_vprintf( Gbs3P, fmtP, ap );

    /* If overflow, substitute that. */
    if ( !okF ) {
	Gbs3P->bs_bC = 0;
	okF = bstr_printf( Gbs3P, "%s log-overflow", prognameP );
    }

    /* If still a problem, not much we can do. */
    if ( !okF )
	return;

    /* Log this buffer in whatever ways are open. */

TODO	/* Currently only output to stderr */

    fprintf( stderr, "%s\n", Gbs3P->bs_bP );

    va_end( ap );
}
/*

*//* smtpd_reset_transaction(xmtP)

	Reset the SMTP mail transaction state

Accepts :

	xmtP		Ptr to XMTP control block

Returns :

	<value>		status code

Notes :

	xmtP isn't needed as of this time but might as well
	include it in the calling sequence in case it is, ever.

	This does NOT reset the HELO state; just the transaction
	state (as per RSET/HELO/EHLO).

*/


STSCODE
smtpd_reset_transaction ARGLIST( (xmtP) )
    FARG( XMTP		*xmtP )		/* XMTP control block */
{
	STSCODE		status;
	int		i;
	LL_NODE		*nodeP;
	SMTP_RCPT	*rcptP;

    status = SCOK;

#ifdef	MVMCAS_IN_MVMTR
    /* Send an mvmcas update message */
    mvmcas_send_update( FALSE, FALSE );
#endif	/* MVMCAS_IN_MVMTR */

    /* Back to default timeout */
    bio_timeout_set( &xmtP->mt_in, SDS_to_general, 0L );

    SDS_dataF = FALSE;			/* Haven't seen DATA */
    SDS_transF = FALSE;			/* No mail transaction */

    /* Clear reverse path */
    if ( SDS_rpathP != NULL ) {
	xmtp_mbox_dealloc( SDS_rpathP );
	SDS_rpathP = NULL;
    }

    /* Clear any recipient list */
    while ( ( nodeP = SDS_rcpt.llh_firstP ) != NULL ) {
	rcptP = (SMTP_RCPT *)nodeP->lln_valP;
	dealloc( rcptP );
	ll_rem( nodeP );
    }
    SDS_rcptC = 0;

    /* Also clear the recipient counters */
    SDS_rcpt_acceptC = 0;
    SDS_rcpt_deferC = 0;
    SDS_rcpt_rejectC = 0;


    /* Clear out the destination addresses */
    avldeltree( SDS_dest_treeP );

    return ( status );
}



/***********************************************************************
 *                                                                     *
 *                     smtpd command verb handlers                     *
 *                                                                     *
 ***********************************************************************/


    /****************
     *   smtpd_cmd_xxx routines, one per SMTP command.
     *
     *   These are called as:
     *
     *      smtpd_cmd_xxx( xmtP, tailP )
     *
     *   Where
     *
     *      xmtP is the XMTP handle,
     *      tailP is the xmtp->mt_cmdtailP, passed for convenience.
     *
     *   Function shall return SCxxx status code, interpreted exactly
     *      per the comments in the smtpd_command() function, with
     *      the exception that SCCONTINUE is an alternate version of
     *      SCOK (allowing subfunctions to more easily take different
     *      return paths).
     *
     ****************/


    /* AUTH command */
static STSCODE
smtpd_cmd_auth ARGLIST( (xmtP, tailP) )
   NFARG( XMTP		*xmtP )		/* XMTP handle */
    FARG( UBYTE		*tailP )	/* Command tail */
{
	STSCODE		status;
	BOOL		okF;
	int		i;
	UBYTE		*irespP;	/* Any initial response string */
	char		method[21];	/* Method name */
	char		empty[2];	/* An empty string */

    /* Only one AUTH allowed per session */
    if ( SDS_authF ) {
	smtpd_log( xmtP, "AUTH-DUP-", NULL, "Duplicate AUTH" );
	okF = xmtp_reply( xmtP, TRUE, 503, NULL,
			  "AUTH already done.\n" );
	return ( okF ? SCCONTINUE : SCERR );
    }

    /* Must have HELO'd */
    if ( ( status = smtpd_require_helo( xmtP ) ) != SCOK )
	return ( status );

    /* Get the auth method */
    i = strtoken_ws( &tailP, &method[0], sizeof(method), FALSE );
    if ( i >= sizeof(method) ) {
	return ( smtpd_reply_syntax( xmtP) );
    }

    /* Skip to any optional initial argument */
    tailP = solbat( tailP, FALSE );
    if ( *tailP != NUL ) {
	irespP = tailP;

	/* An initial response string of "=" is taken to be an empty
	   string.  Use a dynamic buffer here so handlers can wipe it
	   out if they want to.
	*/
	if ( strcmp( irespP, "=" ) == 0 ) {
	    irespP = &empty[0];
	    *irespP = NUL;
	}
    }
    else
	irespP = NULL;

    /* Dispatch on the method */
    for ( i = 0; Smtpauths[i].au_mnameP != NULL ; ++i ) {
	if ( stricmp( &method[0], Smtpauths[i].au_mnameP ) == 0 )
	    break;
    }
    if ( Smtpauths[i].au_mnameP == NULL ) {
	/* End of table, method not known. */
	smtpd_log( xmtP, "AUTH-UNK-", NULL, "Unsupported auth method \"%s\"",
		   &method[0] );

	okF = xmtp_reply( xmtP, TRUE, 504, NULL,
			 "Unsupported auth method \"%s\"\n",
			  &method[0] );
	return ( okF ? SCCONTINUE : SCERR );
    }

    /* Dispatch and return. */
    return ( (*Smtpauths[i].au_rtc)(xmtP, irespP) );
}




    /* DATA command */
static STSCODE
smtpd_cmd_data ARGLIST( (xmtP, tailP) )
   NFARG( XMTP		*xmtP )		/* XMTP handle */
    FARG( UBYTE		*tailP )	/* Command tail */
{
	BOOL		okF;
	BOOL		doneF;
	int		mdv;		/* Mail disposition value */
	STSCODE		status;
	OMM		*ommP;
	char		mflcmd[200];	/* Command buffer for MFL command */

    if ( ( status = smtpd_require_transaction( xmtP ) ) != SCOK )
	return ( status );

    /* Don't do data if no recipients */
    if ( SDS_rcpt_acceptC == 0 ) {
	smtpd_log( xmtP, "DATA-NR-", NULL, "DATA with no valid recipients" );
	okF = xmtp_reply( xmtP, TRUE, 554, NULL, "No recipients.\n" );
	doneF = TRUE;
    }

    /* Give deferal if deferred by session policy */
    else if ( SDS_sdp_data == SDP_DEFER ) {
	smtpd_log( xmtP, "DATA-DEFER-", NULL,
		      "DATA deferred by session policy" );
	okF = xmtp_reply( xmtP, TRUE, 450, NULL, "Deferred." );
	doneF = TRUE;
    }

    /* Give reject if rejected by session policy */
    else if ( SDS_sdp_data == SDP_REJECT ) {
	smtpd_log( xmtP, "DATA-REJECT-", NULL,
			"DATA rejected by session policy" );
	okF = xmtp_reply( xmtP, TRUE, 550, NULL, "Rejected." );
	doneF = TRUE;
    }

    else
	doneF = FALSE;

    if ( doneF ) {
	/* This still resets the transaction (and results in an update) */
	SDS_dataF = TRUE;
	smtpd_reset_transaction( xmtP );

	return ( okF ? SCOK : SCNOTDONE );
    }


    /* No early returns after this point, as various things have to be
       cleaned up at the end.  Probably should fork some of this.

       'status' will keep track of state (secondary to okF):
          SCOK:  keep on going
	  SCDONE: have already replied & logged, need to disconnect
	  SCCONTINUE: have already replied & logged, keep dialog open
	  SCNOTDONE:  some operational error has been logged, no reply
	     yet given, but want to disconnect
	  SCREJECT: will reject per policy
	status may also have other termporary values that are converted
	  to one of the above.
    */


    /* Accept the data as a message */
    okF = xmtp_reply( xmtP, TRUE, 354, NULL,
"Go ahead.  Make my day.  End with <BOL>.<CRLF>\n" );

    if ( !okF ) {
	status = SCNOTDONE;
	ommP = NULL;
    }
    else {
	status = xmtp_getmsg( xmtP, &ommP,
			      cntrl_int_get(
			            estr_tmp_str( "smtp_server_msgsize_max" ),
				    0 ) );
	switch ( status ) {
	    case SCOK:			/* Message gathered fine */
		break;

	    case SCERR:			/* Failure to collect message */
		smtpd_log( xmtP, "DATA-", NULL,
			    "xmtp_getmsg returned SCERR" );
		okF = xmtp_reply( xmtP, TRUE, 451, NULL,
				  "problem getting message." );
		status = SCDONE;
		break;

	    case SCBREAK:		/* Exceeded size limit */
		smtpd_log( xmtP, "DATA-", NULL, "Message too large." );
		okF = xmtp_reply( xmtP, TRUE, 552, NULL,
				   "Message too large.\n" );
		status = SCCONTINUE;
		break;

	    case SCNOTDONE:		/* Some input failure */
		/* We'll just distinguish between timeout and other
		   error.  In any case we'll attempt to give a
		   reply message, and then disconnect.
		*/
		if ( xmtP->mt_in.bio_result == IOR_TIMEOUT ) {
		    smtpd_log( xmtP, "DATA-TO-", NULL, "Timed out" );
		    xmtp_reply( xmtP, TRUE, 421, NULL,
			        "Timed out receiving DATA" );
		    status = SCDONE;
		}
		else {
		    /* Some error, but log the numbers. */
		    smtpd_log( xmtP, "DATA-ERR-", NULL,
			        "Error in DATA input, result=%d errno=%d",
				xmtP->mt_in.bio_result,
				xmtP->mt_in.bio_errno );
		    xmtp_reply( xmtP, TRUE, 421, NULL,
			        "Input error receiving DATA" );
		    status = SCDONE;
		}
		break;

	    default:			/* No clue.. */
		smtpd_log( xmtP, "DATA-", NULL,
				"xmtp_getmsg returned %d", status );
		okF = xmtp_reply( xmtP, TRUE, 451, NULL,
				   "Problem getting message." );
		status = SCDONE;
		break;
	}
    }

    if ( status == SCOK )
	OmmP = ommP;			/* Let MFL see the message */

    mdv = MDV_DUNNO;

    /* Check for bad pipelining or bad dot-stuffing .. i.e., see if
       there is still data in the queue
    */
    if ( status == SCOK ) {
	if ( bio_getb_check( &xmtP->mt_in, 0, 0 ) ) {
TODO	    /* argh, this DATA flow needs reworking */
	    smtpd_log( xmtP, "DATA-", NULL, "input pending after DATA" );
	    mdv = MDV_REJECT;
	}
    }

    /* Apply any $hook_data() hook */
    if ( ( status == SCOK ) &&
	 ( mdv <= 0 ) &&
         ( func_lookup( "$hook_data" ) != NULL ) ) {
	if ( snprintf( &mflcmd[0], sizeof(mflcmd),
		      "%s( \"%s\" );",
		      "$hook_data",
		      mdv_to_str( mdv )
		      ) >= sizeof(mflcmd) ) {

	    smtpd_log( xmtP, "BUG-", NULL,
		    "buffer not large enough for MFL exec of $hook_data" );
	}
	else {
	    if ( smtpd_hook_dialog( &mflcmd[0], &mdv ) == SCOK ) {
		;
	    }
	}
    }

TODO    /* Apply any other checks */

    /* Firm up and validate acceptance value */
    if ( status == SCOK )
	if ( mdv < 0 )			/* If tentative */
	    mdv = -mdv;			/*  make it firm */
	switch ( mdv ) {
	    case MDV_DUNNO:		/* No decision, we'll take it */
	    case MDV_PASS:		/* Also take a "pass" */
		mdv = MDV_ACCEPT;
		break;

	    case MDV_ACCEPT:		/* OK as is */
	    case MDV_REJECT:
		break;

	    case MDV_UNKNOWN:		/* Not really relevant here */
	    case MDV_KNOWN:		/*  but translate to reject */
		mdv = MDV_REJECT;
		break;

	   default:			/* Some problem */
		status = SCERR;
		break;
	}


    /* Change non-acceptance into a status code we can use. */
    if ( ( status == SCOK ) && ( mdv != MDV_ACCEPT ) )
	status = SCREJECT;

    /* Delivery time */
    if ( ( status == SCOK ) && ( mdv == MDV_ACCEPT ) ) {
	smtpd_rcvdhdr();		/* Generate Received header */

	/* Lock the headers so we can add one or more, and do so */
	mml_mutex_lock( &OmmP->om_mmsgP->mm_topmpP->mp_hdrs.llh_lock );

	/* Add the X-Peer-Info header if wanted */
	if ( cntrl_int_get( estr_tmp_str( "add_x_peer_info" ), 0 ) != 0 )
	    if ( !msgpart_hdr_add( OmmP->om_mmsgP->mm_topmpP, 
				   estr_str( &Xphdr[0] ),
				   OmmP->om_mmsgP->mm_topmpP->mp_hdrs.llh_firstP ) ) {
		smtpd_log( xmtP, "DATA-XP-", NULL,
				"Problem adding X-Peer-Info" );
		status = SCNOTDONE;
	    }

	/* Add the Received header */
	if ( status == SCOK )
	    if ( !msgpart_hdr_add( OmmP->om_mmsgP->mm_topmpP, 
				   estr_str( &Rcvdhdr[0] ),
				   OmmP->om_mmsgP->mm_topmpP->mp_hdrs.llh_firstP ) ) {
		smtpd_log( xmtP, "DATA-XP-", NULL,
				"Problem adding Received header" );
		status = SCNOTDONE;
	    }

	mml_mutex_unlock( &OmmP->om_mmsgP->mm_topmpP->mp_hdrs.llh_lock );
    }


#ifdef	QMAIL
    /* Send the message via qmail-queue */
    if ( status == SCOK )
	status = smtpd_qmail_queue( ommP );
#endif	/* QMAIL */

#ifndef	QMAIL
    if ( status == SCOK )
	status = SCERR;
#endif	/* !QMAIL */

    switch ( status ) {
	case	SCOK:
	    smtpd_log( xmtP, "DATA+", NULL, "Message accepted." );
	    okF = xmtp_reply( xmtP, TRUE, 250, NULL,
			     "Message accepted.\n" );
	    SDS_conn_validF = TRUE;
	    break;


	case	SCREJECT:		/* Rejected by policy/filter */
	    smtpd_log( xmtP, "DATA-", SDD_logP, "Rejected by policy/filter" );

	    if ( SDD_replyP != NULL ) {
		/* Custom reply string was given */
		okF = xmtp_reply( xmtP, TRUE, 550, SDD_ecodeP, "%s",
					SDD_replyP );
	    }
	    else {
		okF = xmtp_reply( xmtP, TRUE, 550, SDD_ecodeP,
				   "Message rejected.\n" );
	    }
	    status = SCOK;
	    break;


	case	SCDONE:			/* Already processed & replied,
					   and we want to terminate. */
	    status = SCNOTDONE;
	    break;


	case	SCCONTINUE:		/* Already processed & replied */
	    status = SCOK;
	    break;


	default:
	    smtpd_log( xmtP, "DATA-", NULL, "Error, status=%d", status );
	    okF = xmtp_reply( xmtP, TRUE, 451, NULL,
			      "Sorry, I got status=%d\n", status );
	    break;
    }

    /* Done with the message now */
    if ( ommP != NULL )
	omm_free( ommP );
    ommP = NULL;
    OmmP = NULL;

    /* Say that we've seen DATA (this is very temporary) */
    SDS_dataF = TRUE;

    if ( smtpd_reset_transaction( xmtP ) != SCOK ) {
	smtpd_log( xmtP, "ERROR-", NULL,
	   "smtpd_reset_transaction failed after DATA" );
    }

    return ( status );
}



    /* EHLO command */
static STSCODE
smtpd_cmd_ehlo ARGLIST( (xmtP, tailP) )
   NFARG( XMTP		*xmtP )		/* XMTP handle */
    FARG( UBYTE		*tailP )	/* Command tail */
{
	BOOL		okF;
	STSCODE		status;

    /* Do common part of HELO/EHLO. */
    status = smtpd_cmd_helo_first( xmtP, "EHLO-", &tailP );
    if ( status != SCACCEPT )
	return ( status );

    SDS_esmtpF = TRUE;

    smtpd_log( xmtP, "EHLO+", SDD_logP, "(%s)", &SDS_heloname[0] );

    okF = xmtp_reply( xmtP, FALSE, -250, NULL,
"%s says hello back to %s and extends\n\
HELP\n",
		      UsehostnameP, &SDS_heloname[0] );
    if ( !SCI_suppress_authF ) {
	okF = okF && xmtp_reply( xmtP, FALSE, -250, NULL,
"AUTH LOGIN PLAIN\n" );

#if 1				/* AUTH= for buggy clients */
	okF = okF && xmtp_reply( xmtP, FALSE, -250, NULL,
"AUTH=LOGIN PLAIN\n" );
#endif	/* 1 */

    }

    okF = okF && xmtp_reply( xmtP, TRUE, 250, NULL,
"8BITMIME\n" );

    return ( okF ? SCOK : SCDONE );
}



    /* HELO command */
static STSCODE
smtpd_cmd_helo ARGLIST( (xmtP, tailP) )
   NFARG( XMTP		*xmtP )		/* XMTP handle */
    FARG( UBYTE		*tailP )	/* Command tail */
{
	BOOL		okF;
	STSCODE		status;

    /* Do common part of HELO/EHLO. */
    status = smtpd_cmd_helo_first( xmtP, "HELO-", &tailP );
    if ( status != SCACCEPT )
	return ( status );

    SDS_esmtpF = FALSE;

    smtpd_log( xmtP, "HELO+", SDD_logP, "(%s)", &SDS_heloname[0] );

    okF = xmtp_reply( xmtP, TRUE, 250, NULL,
"%s says hello to %s.  I am ready.\n",
		     UsehostnameP, &SDS_heloname[0] );

    return ( okF ? SCOK : SCDONE );
}



    /* HELP command */
static STSCODE
smtpd_cmd_help ARGLIST( (xmtP, tailP) )
   NFARG( XMTP		*xmtP )		/* XMTP handle */
    FARG( UBYTE		*tailP )	/* Command tail */
{
	BOOL		okF;
	int		i;
	char		topic[10];	/* Help topic wanted */

    /* Check for and collect the topic */
    if ( ( tailP != NULL ) && ( ( *solbat( (char *)tailP, FALSE ) != NUL ) ) ) {
	i = strtoken_ws( &tailP, &topic[0], sizeof(topic), FALSE );
	if ( i >= sizeof(topic) )
	    return ( smtpd_reply_syntax( xmtP ) );
    }
    else
	topic[0] = NUL;

    okF = xmtp_reply( xmtP, FALSE, -214, NULL,
		     "MVMTR here.\n" );

    okF = okF && xmtp_reply( xmtP, FALSE, -214, NULL,
		     "See http://www.mvmf.org/\n\n" );

    if ( topic[0] != NUL ) {
	/* Individual topic wanted.  We don't have individual text
	   (yet) but I suppose we could at least validate the topic.
	*/
	for ( i = 0; Smtpcmds[i].cm_nameP != NULL; ++i )
	    if ( stricmp( Smtpcmds[i].cm_nameP, &topic[0] ) == 0 )
		break;
	if ( Smtpcmds[i].cm_nameP == NULL ) {
	    smtpd_log( xmtP, "HELP-", NULL, "Invalid topic \"%s\"",
				&topic[0] );
	    okF = okF && xmtp_reply( xmtP, FALSE, -214, NULL,
				"topic \"%s\" is unknown.\n",
				&topic[0] );
	}
	else {
	    smtpd_log( xmtP, "HELP+", NULL, "Topic \"%s\"", &topic[0] );
	    okF = okF && xmtp_reply( xmtP, FALSE, -214, NULL,
"topic \"%s\" is valid, but I have no specific help text for it.\n",
				&topic[0] );
	}
    }
    else {
	/* General help wanted. */
	smtpd_log( xmtP, "HELP+", NULL, "General help" );
	okF = okF && xmtp_reply( xmtP, FALSE, -214, NULL,
				  "Supported commands:\n\n" );
	for ( i = 0; Smtpcmds[i].cm_nameP != NULL; ++i )
	    okF = okF && xmtp_reply( xmtP, FALSE, -214, NULL,
				    "    %s\n",
				      Smtpcmds[i].cm_nameP );
    }

    okF = okF && xmtp_reply( xmtP, TRUE, 214, NULL,
		     "\n(HELP needs help.)\n" );

    return ( okF ? SCOK : SCDONE );
}




    /* MAIL command */
static STSCODE
smtpd_cmd_mail ARGLIST( (xmtP, tailP) )
   NFARG( XMTP		*xmtP )		/* XMTP handle */
    FARG( UBYTE		*tailP )	/* Command tail */
{
	BOOL		okF;
	STSCODE		status;
	int		ss_helo;	/* Strict session override for helo */
	int		mdv;		/* MDV_xxx mail disposition value */
	XMTPMBX		*mbxP;		/* The mailbox */
	char		mflcmd[200];	/* Command buffer for MFL command */

    if ( ( status = smtpd_require_helo( xmtP ) ) != SCOK )
	return ( status );

    if ( strnicmp( (char *)tailP, "from:", 5 ) != 0 )
	return ( smtpd_reply_syntax( xmtP ) );

    /* Note: allowing spaces after the colon is forbidden by RFC, but
       acceptable (and sometimes needed) in practice.
    */
    tailP = (UBYTE *)solbat( (char *)tailP +5, FALSE );

    /* Error to have this command if a transaction is already started */
    if ( SDS_transF ) {
	smtpd_log( xmtP, "MAIL-", NULL, "Duplicate MAIL FROM" );
	okF = xmtp_reply( xmtP, TRUE, 503, NULL,
	  "A transaction was already started (duplicate MAIL FROM)\n" );
	return ( okF ? SCOK : SCDONE );
    }

    /* Collect the reverse path */
    if ( xmtp_parse_path( &tailP, &mbxP ) != XPR_OK ) {
	if ( mbxP != NULL )
	    xmtp_mbox_dealloc( mbxP );
	return ( smtpd_reply_syntax( xmtP ) );
    }

    /* Check for parameters that we support. */

TODO /* HACKity hack.. should fork & generalize this */
    for ( ; ; ) {
	tailP = (UBYTE *)solbat( (char *)tailP, FALSE );
	if ( *tailP == NUL )
	    break;

	if ( strnicmp( (char *)tailP, "BODY=", 5 ) == 0 ) {
	    tailP += 5;
	    if ( strnicmp( (char *)tailP, "7BIT", 4 ) == 0 )
		tailP += 4;
	    else if ( strnicmp( (char *)tailP, "8BITMIME", 8 ) == 0 )
		tailP += 8;
	    else
	        return ( smtpd_reply_syntax( xmtP ) );
	}
	else if ( strnicmp( (char *)tailP, "AUTH=", 5 ) == 0 ) {
	    tailP += 5;
	    if ( strnicmp( (char *)tailP, "<>", 2 ) == 0 )
		tailP += 2;
	    else {
		/* Just skip it */
		for ( ; *tailP != NUL; ++tailP )
		    if ( *tailP == ' ' )
			break;
	    }
	}
	else
	    break;
    }

    if ( ( status = smtpd_require_eol( xmtP, tailP ) ) != SCOK )
	return ( status );

    mdv = MDV_DUNNO;

    /* Do mail_from checks if we're not deferring until later */
    if ( ( status == SCOK ) && ( mdv <= 0 ) &&
         ( cntrl_int_get( estr_tmp_str( "smtp_server_mail_check_defer" ),
		        0 ) == 0 ) )
	mdv = mail_check( xmtP, mbxP );

    /* Firm up and validate acceptance value */
    if ( status == SCOK )
	if ( mdv < 0 )			/* If tentative */
	    mdv = -mdv;			/*  make it firm */
	switch ( mdv ) {
	    case MDV_DUNNO:		/* No decision, we'll take it */
	    case MDV_PASS:		/* Also take a "pass" */
		mdv = MDV_ACCEPT;
		break;

	    case MDV_ACCEPT:		/* OK as is */
	    case MDV_REJECT:
		break;

	    case MDV_UNKNOWN:		/* Not really relevant here */
	    case MDV_KNOWN:		/*  but translate to reject */
		mdv = MDV_REJECT;
		break;

	   default:			/* Some problem */
		status = SCERR;
		break;
	}

    /* Handle any operational problem */
    if ( status != SCOK ) {
	xmtp_reply( xmtP, TRUE, 451, NULL, "server problem.\n" );
	return ( status );
    }

    /* Handle rejection */
    if ( mdv == MDV_REJECT ) {
	smtpd_log( xmtP, "MAIL-", SDD_logP, "from <%s>", mbxP->mb_pathP );

	if ( SDD_replyP != NULL ) {
	    /* Use custom reply string */
	    xmtp_reply( xmtP, TRUE, 550, SDD_ecodeP, "%s", SDD_replyP );
	}
	else {
	    xmtp_reply( xmtP, TRUE, 550, SDD_ecodeP, "Rejected.\n" );
	}

	xmtp_mbox_dealloc( mbxP );
	return ( SCOK );
    }

    /* Remember reverse path */
    SDS_rpathP = mbxP;

    /* Transaction is now in progress */
    SDS_transF = TRUE;

    smtpd_log( xmtP, "MAIL+", NULL, "from <%s>", mbxP->mb_pathP );

    okF = xmtp_reply( xmtP, TRUE, 250, NULL, "OK\n" );

    return ( okF ? SCOK : SCDONE );
}




    /* NOOP command */
static STSCODE
smtpd_cmd_noop ARGLIST( (xmtP, tailP) )
   NFARG( XMTP		*xmtP )		/* XMTP handle */
    FARG( UBYTE		*tailP )	/* Command tail */
{
	BOOL		okF;
	STSCODE		status;

    /* Log if there is any text after the NOOP */
    if ( ( tailP != NULL ) && ( ( *solbat( (char *)tailP, FALSE ) != NUL ) ) ) {
	smtpd_log( xmtP, "NOOP+", NULL, "%s", xmtP->mt_cmbsP->bs_bP );
    }

    okF = xmtp_reply( xmtP, TRUE, 250, NULL,
		     "yuup -- noop successful.\n" );

    return ( okF ? SCOK : SCDONE );
}



    /* QUIT command */
static STSCODE
smtpd_cmd_quit ARGLIST( (xmtP, tailP) )
   NFARG( XMTP		*xmtP )		/* XMTP handle */
    FARG( UBYTE		*tailP )	/* Command tail */
{
	BOOL		okF;

    /* We don't care here if there's junk in the tail.  If they
       want to quit, let them.
    */

    okF = xmtp_reply( xmtP, TRUE, 221, NULL,
		     "have a nice day.\n" );

    return ( SCDONE );
}




    /* RCPT command */
static STSCODE
smtpd_cmd_rcpt ARGLIST( (xmtP, tailP) )
   NFARG( XMTP		*xmtP )		/* XMTP handle */
    FARG( UBYTE		*tailP )	/* Command tail */
{
	BOOL		okF;
	STSCODE		status;
	int		mdv;		/* MDV_XXX mail disposition value */
	int		i;
	XMTPMBX		*mbxP;
	SMTP_DEST	*destP;		/* Destination block */
	SMTP_RCPT	*rcptP;		/* Receipt block */
	char		mflcmd[200];	/* Command buffer for MFL command */

    if ( ( status = smtpd_require_transaction( xmtP ) ) != SCOK )
	return ( status );

    if ( strnicmp( (char *)tailP, "to:", 3 ) != 0 )
	return ( smtpd_reply_syntax( xmtP ) );

    /* Note: allowing spaces after the colon is forbidden by RFC, but
       acceptable (and sometimes needed) in practice.
    */
    tailP = (UBYTE *)solbat( (char *)tailP + 3, FALSE );

    /* Get the path */
    if ( xmtp_parse_path( &tailP, &mbxP ) != XPR_OK ) {
	if ( mbxP != NULL )
	    xmtp_mbox_dealloc( mbxP );
	return ( smtpd_reply_syntax( xmtP ) );
    }

    /* Path can't be <> */
    if ( mbxP->mb_pathP[0] == NUL ) {
	xmtp_mbox_dealloc( mbxP );
	return ( smtpd_reply_syntax( xmtP ) );
    }

    if ( ( status = smtpd_require_eol( xmtP, tailP ) ) != SCOK ) {
	xmtp_mbox_dealloc( mbxP );
	return ( status );
    }

    /* Rewrite any @[local-ip-address] forms */
    if ( ( status = smtpd_rewrite_local_ip( mbxP ) ) != SCOK ) {
	xmtp_mbox_dealloc( mbxP );
	return ( status );
    }

    /* Find and/or establish the destination struct for this */
    destP = smtpd_dest_access( mbxP );

    /* And we can be rid of this */
    xmtp_mbox_dealloc( mbxP );

    if ( destP == NULL )
	return ( SCERR );

    /* Add a receipt block for this */
    status = smtpd_rcpt_add( xmtP, destP );
    if ( status != SCOK )
	return ( status );
    rcptP = (SMTP_RCPT *)SDS_rcpt.llh_lastP->lln_valP;

    mdv = destP->sd_acceptance;

    /* If the acceptance is not already decided, decide it now.
       (Don't re-decide a previously decided destination.)
    */
    if ( mdv <= 0 ) {

	/* Try any $hook_rcpt_to() hook function.
	   Note that we give the address to the hook without reservation;
	   RFCs say that a naked "postmaster" address must be accepted,
	   but we'll let the hook violate that if that's the admin's
	   choice.
	*/
	if ( ( mdv <= 0 ) &&
	     ( func_lookup( "$hook_rcpt_to" ) != NULL ) ) {
	    if ( snprintf( &mflcmd[0], sizeof(mflcmd),
				"%s( \"%s\", \"%s\" );",
				"$hook_rcpt_to",
				mdv_to_str( mdv ),
				destP->sd_mbx.mb_pathP ) >= sizeof(mflcmd) ) {

		smtpd_log( xmtP, "BUG-", NULL,
		    "buffer not large enough for MFL exec of $hook_rcpt_to" );
	    }
	    else {
		if ( smtpd_hook_dialog( &mflcmd[0], &mdv ) == SCOK ) {
		    ;
		}
	    }
	}

	/* If not rejected for other policy reasons, Naked "postmaster"
	   must always be accepted.  Note that the surrounding system
	   must be able to accept this address as well!
	*/
	if ( ( mdv <= 0 ) &&
	     ( stricmp( destP->sd_mbx.mb_pathP, "postmaster" ) == 0 ) )
	    mdv = MDV_ACCEPT;

	/* If acceptance isn't decided, do built-in checks. */
	if ( ( status == SCOK ) && ( mdv <= 0 ) )
	    mdv = smtpd_address_check( destP->sd_mbx.mb_pathP,
						destP->sd_mbx.mb_len,
						mdv);

	/* Do a connecting client validation if we've deferred it to here.
	   This can still override a positive acceptance.
	*/
	if ( ( status == SCOK ) &&
	     ( ( mdv <= 0 ) || ( mdv == MDV_ACCEPT ) ) &&
             ( cntrl_int_get(
			estr_tmp_str( "smtp_server_client_check_defer" ),
		        0 ) != 0 ) ) {
	    if ( client_check( xmtP ) == MDV_REJECT ) {
		mdv = MDV_REJECT;
	    }
	}

	/* Validate the helo name if that check has been deferred 'til now.
	   This can also override a positive acceptance.
	*/
	if ( ( status == SCOK ) &&
	     ( ( mdv <= 0 ) || ( mdv == MDV_ACCEPT ) ) &&
             ( cntrl_int_get(
			estr_tmp_str( "smtp_server_helo_check_defer" ),
		        0 ) != 0 ) ) {
	    if ( helo_check( xmtP ) == MDV_REJECT ) {
		mdv = MDV_REJECT;
	    }
	}

	/* Validate the mail-from name if that check has been deferred.
	   This, too, can override a positive acceptance.
	*/
	if ( ( status == SCOK ) &&
	     ( ( mdv <= 0 ) || ( mdv == MDV_ACCEPT ) ) &&
             ( cntrl_int_get(
			estr_tmp_str( "smtp_server_mail_check_defer" ),
		        0 ) != 0 ) ) {
	    if ( mail_check( xmtP, SDS_rpathP ) == MDV_REJECT ) {
		mdv = MDV_REJECT;
	    }
	}


#ifdef	MVMTR_QMAIL_COMPATIBILITY
	/* Check via qmail-smtpd compatibility if the issue hasn't already
           been decided.
	*/
	if ( SDS_qmcompatF && ( status == SCOK ) && ( mdv <= 0 ) )
	    mdv = qmc_smtpd_check_rcpt_to( &destP->sd_mbx, mdv );
#endif	/* MVMTR_QMAIL_COMPATIBILITY */

	/* Firm up and validate acceptance value */
	if ( status == SCOK ) {
	    if ( mdv < 0 )
		mdv = -mdv;		/* Turn tentative into firm */

	    switch ( mdv ) {
		case MDV_UNKNOWN:	/* OK, or acceptable if relaying */
		case MDV_KNOWN:
		    if ( SCI_relayokF )	/* Accept if allowed to relay */
			mdv = MDV_ACCEPT;
		    break;

		case MDV_REJECT:
		case MDV_ACCEPT:
		    break;

		case MDV_PASS:		/* A pass turns into accept */
		    mdv = MDV_ACCEPT;
		    break;

		case MDV_DUNNO:		/* No judgement yet */
TODO		    /* If we still haven't decided, we might as well take it.
		       Maybe this should be settable?
		    */
		    mdv = MDV_ACCEPT;
		    break;

		default:		/* Some problem */
		    status = SCERR;
		    break;
	    }
	}

	/* Handle any operational problem */
	if ( status != SCOK ) {
	    xmtp_reply( xmtP, TRUE, 451, NULL, "server problem.\n" );
	    return ( status );
	}

	/* Save the new acceptance value */
	destP->sd_acceptance = mdv;
    }

    if ( mdv != MDV_ACCEPT ) {
	/* Sleep after bad RCPT, if so configured. */
	i = cntrl_int_get( estr_tmp_str( "smtp_server_badrcpt_delay" ), 0 );
	if ( i > 0 )
	    sleep( i );

	switch ( mdv ) {
	    case MDV_UNKNOWN:		/* Unknown (outside) address */
		smtpd_log( xmtP, "RCPT-RELAY-", SDD_logP, "to <%s> from <%s>",
			  destP->sd_mbx.mb_pathP,
			  SDS_rpathP->mb_pathP );

		++SDS_rcpt_rejectC;
		if ( ( SDS_rcpt_rejectC + SDS_rcpt_deferC ) >= SDS_rcpt_badmax )
		    return ( smtpd_reply_badmax( xmtP ) );

		okF = xmtp_reply( xmtP, TRUE, 553, NULL,
				 "we are not a relay for that address.\n" );
		return ( okF ? SCOK : SCDONE );

	    case MDV_KNOWN:		/* Known, but not for us */
		smtpd_log( xmtP, "RCPT+-", SDD_logP, "to <%s> from <%s>",
			  destP->sd_mbx.mb_pathP,
			  SDS_rpathP->mb_pathP );

		++SDS_rcpt_deferC;
		if ( ( SDS_rcpt_rejectC + SDS_rcpt_deferC ) >= SDS_rcpt_badmax )
		    return ( smtpd_reply_badmax( xmtP ) );

		okF = xmtp_reply( xmtP, TRUE, 450, NULL,
				  "Use primary server.\n" );
		return ( okF ? SCOK : SCDONE );

	    case MDV_REJECT:
		smtpd_log( xmtP, "RCPT-", SDD_logP, "to <%s> from <%s>",
			  destP->sd_mbx.mb_pathP,
			  SDS_rpathP->mb_pathP );

		++SDS_rcpt_rejectC;
		if ( ( SDS_rcpt_rejectC + SDS_rcpt_deferC ) >= SDS_rcpt_badmax )
		    return ( smtpd_reply_badmax( xmtP ) );

		if ( SDD_replyP != NULL )
		    okF = xmtp_reply( xmtP, TRUE, 550, SDD_ecodeP,
				 "%s", SDD_replyP );
		else
		    okF = xmtp_reply( xmtP, TRUE, 550, SDD_ecodeP,
				 "Invalid recipient.\n" );
		return ( okF ? SCOK : SCDONE );

	    case MDV_DUNNO:		/* Should not have a dunno */
	    default:
		smtpd_log( xmtP, "RCPT-ERR-", NULL, "to <%s> from <%s>",
			  destP->sd_mbx.mb_pathP,
			  SDS_rpathP->mb_pathP );

		++SDS_rcpt_deferC;
		if ( ( SDS_rcpt_rejectC + SDS_rcpt_deferC ) >= SDS_rcpt_badmax )
		    return ( smtpd_reply_badmax( xmtP ) );

		okF = xmtp_reply( xmtP, TRUE, 451, NULL, "some error.\n" );
		return ( SCERR );
	}
    }

    /* Address is OK.  We can still reject if we are rejecting by
       session policy.  This could be combined with the above but
       I'd rather have separate flow and log elements for this.
       At least as of this writing.
    */
    switch ( SDS_sdp_rcpt ) {
	case SDP_ACCEPT:		/* This is ok still */
	    break;

	case SDP_DEFER:			/* Deferred by policy */
	    smtpd_log( xmtP, "RCPT-P+-", NULL, "to <%s> from <%s>",
			  destP->sd_mbx.mb_pathP,
			  SDS_rpathP->mb_pathP );
	    ++SDS_rcpt_deferC;
	    okF = xmtp_reply( xmtP, TRUE, 450, NULL,
			      "Deferred by session policy" );
	    return ( okF ? SCOK : SCDONE );

	case SDP_REJECT:		/* Rejected by policy */
	    smtpd_log( xmtP, "RCPT-P--", NULL, "to <%s> from <%s>",
			  destP->sd_mbx.mb_pathP,
			  SDS_rpathP->mb_pathP );
	    ++SDS_rcpt_rejectC;
	    okF = xmtp_reply( xmtP, TRUE, 550, NULL,
			      "Rejected by session policy" );
	    return ( okF ? SCOK : SCDONE );

	default:			/* Yes we have to check */
	    /* Bad case.  Log it but still process the mail */
	    smtpd_log( xmtP, "BUG-", NULL,
		      "Unknown SDS_sdp_rcpt value %d in RCPT" );
	    break;
    }

    /* Finally OK.  ACK it. */
    smtpd_log( xmtP, "RCPT+", NULL, "to <%s> from <%s>",
			destP->sd_mbx.mb_pathP,
			SDS_rpathP->mb_pathP );

    okF = xmtp_reply( xmtP, TRUE, 250, NULL,
			     "Thank you, may I have another?\n" );
    ++SDS_rcpt_acceptC;

    return ( okF ? SCOK : SCDONE );
}




    /* RSET command */
static STSCODE
smtpd_cmd_rset ARGLIST( (xmtP, tailP) )
   NFARG( XMTP		*xmtP )		/* XMTP handle */
    FARG( UBYTE		*tailP )	/* Command tail */
{
	BOOL		okF;
	STSCODE		status;

    if ( ( status = smtpd_require_eol( xmtP, tailP ) ) != SCOK )
	return ( status );

    if ( smtpd_reset_transaction( xmtP ) != SCOK ) {
	(void)xmtp_reply( xmtP, TRUE, 451, NULL,
			 "can not reset state." );
	return ( SCERR );
    }

    okF = xmtp_reply( xmtP, TRUE, 250, NULL,
		     "transaction reset.\n" );

    return ( SCOK );
}

/***********************************************************************
 *                                                                     *
 *             end of smtpd_cmd_xxx command verb handlers              *
 *                                                                     *
 ***********************************************************************/



/***********************************************************************
 *                                                                     *
 *           Start of smtpd_cmd_auth_xxx SMTP AUTH handlers            *
 *                                                                     *
 ***********************************************************************/


    /****************
     *   smtpd_cmd_auth_xxx routines, one per SMTP AUTH method.
     *     each handles an SMTP AUTH method after initial processing
     *     by smtp_cmd_auth command handler.
     *
     *   These are called as:
     *
     *      smtpd_cmd_auth_xxx( xmtP, irespP )
     *
     *   Where
     *
     *      xmtP is the XMTP handle,
     *      irespP is any initial response to the AUTH,
     *        or NULL if no initial response given.  Note that
     *        no decoding has been done on this response.
     *
     *   Function shall return SCxxx status code, just like
     *      smtp_cmd_xxx() functions.
     *
     ****************/


    /* AUTH method: LOGIN
    */
static STSCODE
smtpd_cmd_auth_login ARGLIST( (xmtP, irespP) )
   NFARG( XMTP		*xmtP )		/* XMTP handle */
    FARG( UBYTE		*irespP )	/* Initial response */
{
	STSCODE		status;
	BOOL		okF;
	int		i;
	UBYTE		*encP;		/* Encoded string */
	UBYTE		*unameP;	/* Decoded username */
	UBYTE		*passwdP;	/* Decoded password */

    /* Get username if not supplied. */
    if ( irespP == NULL ) {
	/* Prompt for and get encoded username from client.
	 * % echo -n Username: | mmencode -b
	 * VXNlcm5hbWU6
	*/
	if ( !xmtp_reply( xmtP, TRUE, 334, NULL, "VXNlcm5hbWU6\n" ) )
	    return ( SCERR );

	if ( ( status = xmtp_getline( xmtP, Gbs1P ) ) != SCOK )
	    return ( status );

	encP = Gbs1P->bs_bP;

	/* "*" means abort. */
	if ( strcmp( encP, "*" ) == 0 ) {
	    okF = xmtp_reply( xmtP, TRUE, 501, NULL, "Aborted.\n" );
	    return ( okF ? SCCONTINUE : SCERR );
	}
    }
    else
	encP = irespP;

    /* Get decoded username into a tmpstr */
    i = strlen( (char *)encP );
    unameP = b64_decode_tmpstr_str_n( encP, i, NULL );
    if ( unameP == NULL ) {
	okF = xmtp_reply( xmtP, TRUE, 501, NULL,
			  "Invalid username response.\n" );
	return ( okF ? SCCONTINUE : SCERR );
    }

    /* Prompt for and get encoded password from client.
     * % echo -n Password: | mmencode -b
     * UGFzc3dvcmQ6
     */
    if ( !xmtp_reply( xmtP, TRUE, 334, NULL, "UGFzc3dvcmQ6\n" ) )
	status = SCERR;
    else {
	if ( ( status = xmtp_getline( xmtP, Gbs1P ) ) == SCOK ) {
	    encP = Gbs1P->bs_bP;
	    /* "*" is abort, again (seems dumb, probably is a reason tho) */
	    if ( strcmp( encP, "*" ) == 0 ) {
		okF = xmtp_reply( xmtP, TRUE, 501, NULL, "Aborted.\n" );
		status = SCNOTDONE;
	    }

	    /* Get decoded password */
	    i = strlen( (char *)encP );
	    passwdP = b64_decode_tmpstr_str_n( encP, i, NULL );
	    memset( encP, 0, i );	/* Don't keep encoded pw */
	    if ( passwdP == NULL ) {
		okF = xmtp_reply( xmtP, TRUE, 501, NULL,
				  "Invalid password response.\n" );
		status = okF ? SCCONTINUE : SCERR;
	    }
	}
    }

    /* Here, status is either SCOK to continue, or something else
       not to (with response already issued).  Return here if needed,
       for convenience.
    */
    if ( status != SCOK ) {
	return ( status );
    }

    /* Use checkpassword authentication */
    status = auth_checkpassword( cntrl_str_get(
				    estr_tmp_str( "prog_checkpassword" ) ),
				unameP, passwdP );

    /* Get rid of password in memory now */
    memset( passwdP, 0, strlen( (char *)passwdP ) );

    if ( status != SCOK ) {
	okF = xmtp_reply( xmtP, TRUE, 535, NULL, "Auth failed.\n" );
	return ( okF ? SCCONTINUE : SCERR );
    }

    /* Password checked... set up auth state & return. */
    return ( smtpd_authsetup( xmtP, unameP ) );
}



    /* AUTH method: PLAIN
     *
     * per RFC2595
     *
    */
static STSCODE
smtpd_cmd_auth_plain ARGLIST( (xmtP, irespP) )
   NFARG( XMTP		*xmtP )		/* XMTP handle */
    FARG( UBYTE		*irespP )	/* Initial response */
{
	STSCODE		status;
	BOOL		okF;
	int		i;
	int		strL;		/* Length of decoded string */
	UBYTE		*encP;		/* Encoded auth string */
	UBYTE		*strP;		/* The decoded string */
	UBYTE		*idP;		/* Identity string */
	UBYTE		*unameP;	/* Username */
	UBYTE		*passwdP;	/* Password */

    if ( irespP == NULL ) {
	/* Initial response not given; get it from client */
	if ( !xmtp_reply( xmtP, TRUE, 334, NULL, "\n" ) )
	    return ( SCERR );
	if ( ( status = xmtp_getline( xmtP, Gbs1P ) ) != SCOK )
	    return ( status );
	encP = Gbs1P->bs_bP;

	/* "*" is abort. */
	if ( strcmp( encP, "*" ) == 0 ) {
	    okF = xmtp_reply( xmtP, TRUE, 501, NULL, "Aborted.\n" );
	    return ( okF ? SCCONTINUE : SCERR );
	}
    }
    else
	encP = irespP;

    /* Decode the string and wipe out the encoded copy */
    i = strlen( (char *)encP );
    strP = b64_decode_tmpstr_str_n( encP, i, &strL );
    memset( encP, 0, i );
    if ( strP == NULL ) {
	okF = xmtp_reply( xmtP, TRUE, 501, NULL,
			  "Invalid response.\n" );
	return ( okF ? SCCONTINUE : SCERR );
    }

    /* Decoded string should now contain
	   id<NUL>user<NUL>passwd<NUL>
       (the final NUL added by the decoding function).
       Get pointers to each piece.
    */
    idP = strP;
    for ( i = 0; i < strL; ++i )
	if ( strP[i] == NUL )
	    break;
    unameP = &strP[ ++i ];
    for ( ; i < strL; ++i )
	if ( strP[i] == NUL )
	    break;
    passwdP = &strP[ ++i ];

    /* Make sure we have all the parts */
    if ( i > strL ) {
	okF = xmtp_reply( xmtP, TRUE, 501, NULL,
			  "Badly formatted response.\n" );
	return ( okF ? SCCONTINUE : SCERR );
    }

    /* For our purposes the ID and the username should be the same */
    if ( strcmp( idP, unameP ) != 0 ) {
#if 0
	okF = xmtp_reply( xmtP, TRUE, 501, NULL,
			  "Badly formatted response.\n" );
	return ( okF ? SCCONTINUE : SCERR );
#else
TODO	/* Should probably log this condition. */
	;
#endif
    }

    /* Use checkpassword authentication */
    status = auth_checkpassword( cntrl_str_get(
				     estr_tmp_str( "prog_checkpassword" ) ),
			       unameP, passwdP );

    /* Get rid of password in memory now */
    memset( passwdP, 0, strlen( (char *)passwdP ) );

    if ( status != SCOK ) {
	okF = xmtp_reply( xmtP, TRUE, 535, NULL, "Auth failed.\n" );
	return ( okF ? SCCONTINUE : SCERR );
    }

    /* Password checked... set up auth state & return. */
    return ( smtpd_authsetup( xmtP, unameP ) );
	
}

/***********************************************************************
 *                                                                     *
 *            End of smtpd_cmd_auth_xxx SMTP AUTH handlers             *
 *                                                                     *
 ***********************************************************************/



/***********************************************************************
 *                                                                     *
 *                     Internal support routines                       *
 *                                                                     *
 ***********************************************************************/


/*

*//* client_check( xmtP, mdv )

	Checks the client host

Accepts :

	xmtP		Ptr to XMTP control block

Returns :

	<value>		MDV_xxx acceptance value, one of:
			  MDV_ACCEPT
			  MDV_REJECT
			  MDV_ERROR

Notes :

	May also set various state variables.

	Remember that this may be deferred until somewhere into the
	  SMTP dialog.

	The individual checks are specified in a string control
	  containing a list of checks to do.  Each check is done via
	  dispatch table; each entry in the table specifies a check name
	  and a handler function.  The check name may be preceded with
	  "~" to make it a "soft" check (a result may be logged by no
	  action is taken).  The handler function is expected to return
	  one of the MDV_xxxx codes (which may include MDV_DUNNO for no
	  decision).  Any positive handler value is accepted without
	  further processing unless a soft check has been specified.

*/

static int
client_check ARGLIST( (xmtP) )
    FARG( XMTP		*xmtP )		/* XMTP control block */
{
	int		softF;		/* If "soft" component check */
	int		i;
	int		mdv, mdvret;
	char		*checkP;	/* component check name */
	UBYTE		*csP;		/* Control string ptr */
	char		buf[100];

    mdv = MDV_DUNNO;

    /* The smtp_server_client_check control, if defined, contains a list
       of connect check names that are processed in order.

       The table Clientchecks[] defines the check names and the functions
       that process each check.  These functions are expected to
       return an MDV_XXX mail disposition value:  MDV_ACCEPT means
       the helo name is unconditionally acceptable, MDV_REJECT means
       that the helo name should be rejected (but see soft check);
       other returns have no effect.

       The check name may be preceded with '~' to make the check
       a soft check.  Any MDV_REJECT from a soft check is only logged,
       but not rejected.  MDV_ACCEPT passes are still taken.
    */
    if ( mdv <= 0 ) {
	csP = cntrl_str_get( estr_tmp_str( "smtp_server_client_check" ) );
	if ( csP == NULL )
	    csP = "";
	for ( ; mdv <= 0; ) {
	    i = strtoken_ws( &csP, (UBYTE *)&buf[0], sizeof(buf), TRUE );
	    if ( i > sizeof( buf) )
		continue;		/* Ignore too-big check types */

	    if ( buf[0] == NUL )
		break;			/* No more types */

	    checkP = &buf[0];
	    if ( softF = ( *checkP == '~' ) )
		++checkP;		/* soft check */

	    /* Look up the type & dispatch if it's found.  Ignore
	       unknown type names.
	    */
	    for ( i = 0; Clientchecks[i].cc_nameP != NUL; ++i ) {
		if ( stricmp( Clientchecks[i].cc_nameP, checkP ) == 0 ) {
		    mdvret = (*Clientchecks[i].cc_rtc)(xmtP);
		    if ( mdvret == MDV_ACCEPT )
			mdv = mdvret;
		    else if ( mdvret == MDV_REJECT ) {
			if ( softF ) {	    /* Log only ? */
			    smtpd_log( xmtP, "CONNECT-+", SDD_logP,
				       "soft reject" );
			    SDD_logP = NULL;
			}
			else
			    mdv = mdvret;
		    }
		    else
			/* Just ignore this return */
			;

		    break;
		}
	    }
	}
    }

    /* Turn tentative mdv into firm one, and process it. */
    if ( mdv < 0 )
	mdv = -mdv;
    switch ( mdv ) {
	case MDV_DUNNO:			/* No decision, we'll take it */
	    mdv = MDV_ACCEPT;
	    break;

	case MDV_PASS:			/* Shouldn't see this, but... */
	    mdv = MDV_ACCEPT;
	    break;

	case MDV_ACCEPT:		/* OK as is */
	case MDV_REJECT:
	    break;

	case MDV_UNKNOWN:		/* Not really relevant here */
	case MDV_KNOWN:			/*  but translate to reject */
	    mdv = MDV_REJECT;
	    break;

	default:			/* Some problem */
	    mdv = MDV_ERROR;
	    break;
    }

    return ( mdv );
}
/*

*//* client_check_auth( xmtP )

	Client check: auth

Accepts :

	xmtP		Ptr to XMTP control block

Returns :

	<value>		MDV_xxx mail disposition value,
				per client_check interface

Notes :

	Subservient to client_check, q.v.

	This is the "auth" check, which gives a pass if the session
	  has been AUTH'd.  Useful only if client checking is
	  deferred into the SMTP dialog.

*/

static int
client_check_auth ARGLIST( (xmtP) )
    FARG( XMTP		*xmtP )		/* XMTP control block */
{
    if ( SDS_authF )			/* Give a pass if auth'd */
	return ( MDV_ACCEPT );

    return ( MDV_DUNNO );		/* no stance.. */
}
/*

*//* client_check_generic( xmtP )

	Client check: generic

Accepts :

	xmtP		Ptr to XMTP control block

Returns :

	<value>		MDV_xxx mail disposition value,
				per client_check interface

Notes :

	Subservient to client_check, q.v.

	This is the "generic" check, which looks at the connecting
	  host info to see whether it looks like a generically
	  assigned address.

*/

static int
client_check_generic ARGLIST( (xmtP) )
    FARG( XMTP		*xmtP )		/* XMTP control block */
{
	int		len;
	BOOL		okF;
	int		mdv;
	char		*revP;		/* A sample reply PTR */
	char		*replyP;	/* For reply text */
	char		qname[ NS_MAXDNAME ];
	MML_DNS		*dnsP;		/* mml-dns handle */

    mdv = MDV_DUNNO;

    dnsP = dns_handle();		/* We'll need this */

    /* Look at the reverse DNS name(s).  If *any* of the names
       aren't rejected, we'll give them all a pass.
    */
    if ( ( dnsP != NULL ) &&
         ( snprintf( &qname[0], sizeof(qname), "%d.%d.%d.%d.in-addr.arpa",
		     SCI_hostoct[3], SCI_hostoct[2],
		     SCI_hostoct[1], SCI_hostoct[0] ) < sizeof(qname) ) &&
	 ( okF = mdns_query( dnsP, &qname[0], ns_c_in, ns_t_ptr ) ) ) {
	/* Resolve one CNAME indirection */
	if ( dnsP->dn_anstype == ns_t_cname )
	    okF = mdns_query( dnsP, dnsP->dn_ansP, ns_c_in, ns_t_ptr );

	/* Look at all the returns until we find one that's NOT rejected */
	if ( okF ) {
	    revP = NULL;
	    do {
		if ( dnsP->dn_anstype == ns_t_ptr ) {
		    mdv = generic_ip_check( xmtP, SCI_hostip, dnsP->dn_ansP );
		    if ( mdv != MDV_REJECT )
			break;
		    if ( revP == NULL )
			revP = dnsP->dn_ansP;
		}
	    } while ( mdns_ans_next( dnsP ) );
	}
    }

    if ( mdv == MDV_REJECT ) {

	/* Default the log string */
	if ( SDD_logP == NULL )
	    SDD_logP = "generic";

	/* Default the reply string */
	if ( SDD_replyP == NULL ) {
	    /* Attempt to generate some kind of meaningful reply text
	       tune this to match the printf string that follows.
	    */
	    len = ( 3 * strlen( SCI_hostipP ) ) + 300;
	    replyP = (char *)tmpstr( len +1 );
	    if ( ( replyP != NULL ) &&
	         ( snprintf( replyP, len,
"Denied, see http://www.mv.com/mv/email/generic-ip.cgi?ip=%s\n"
"The connection from your IP address (%s) is being refused\n"
"as it appears to be coming from a generically assigned IP address.\n"
"Again, see http://www.mv.com/mv/email/generic-ip.cgi?ip=%s",
			SCI_hostipP, SCI_hostipP, SCI_hostipP ) < len ) )
		SDD_replyP = replyP;
	}
    }

    /* Return what we got. */
    return ( mdv );
}
/*

*//* client_check_pass( xmtP )

	Client check: pass

Accepts :

	xmtP		Ptr to XMTP control block

Returns :

	<value>		MDV_xxx mail disposition value,
				per client_check interface

Notes :

	Subservient to client_check, q.v.

	This is the "pass" check, which gives an OK if the client
	  has gotten a "pass" (e.g. from mvmcas).

*/

static int
client_check_pass ARGLIST( (xmtP) )
    FARG( XMTP		*xmtP )		/* XMTP control block */
{
    /* Return accept if there's a pass, dunno otherwise. */
    if ( SCI_t_pass != 0 )
	return ( MDV_ACCEPT );

    return ( MDV_DUNNO );
}
/*

*//* connect_check( xmtP )

	Check out the client at connect time

Accepts :

	xmtP		Ptr to XMTP control block

Returns :

	<value>		MDV_xxx mail disposition value, one of:
			   MDV_DUNNO - no opinion
			   MDV_ACCEPT - take the connection
			   MDV_REJECT - standard connection rejection
			   MDV_UNKNOWN - reject with no notice
			   

Notes :

	May also set various state variables.

	The return of MDV_UNKNOWN overloads the meaning of that value.
	  But we'll survive.

	Other MDV_xxx values might be returned, e.g. if error.

*/

static int
connect_check ARGLIST( (xmtP) )
    FARG( XMTP		*xmtP )		/* XMTP control block */
{
	int		mdv;
	STSCODE		status;
	char		mflcmd[200];	/* Command buffer for MFL command */

    mdv = MDV_DUNNO;

    /* Apply the $hook_connect() hook if it exists */
    if ( func_lookup( "$hook_connect" ) != NULL ) {
	if ( snprintf( &mflcmd[0], sizeof(mflcmd),
		      "%s( \"%s\", \"%s\" );",
		      "$hook_connect",
		      mdv_to_str( mdv ),
		      SCI_hostipP ) >= sizeof(mflcmd) ) {

	    smtpd_log( xmtP, "BUG-", NULL,
	      "buffer not large enough for MFL exec of $hook_connect" );
	}
	else {
	    if ( smtpd_hook_dialog( &mflcmd[0], &mdv ) == SCOK ) {
		;
	    }
	}
    }

    /* Firm up the acceptance value */
    if ( status == SCOK ) {
	if ( mdv < 0 )
	    mdv = -mdv;
    }

    if ( status != SCOK )
	mdv = MDV_ERROR;

    return ( mdv );
}
/*

*//* dest_avl_cmpkey( keyP, nodeP )

	Compare keys for avl tree routines, for destination address tree

Accepts :

	keyP		Ptr to the subject key (XMTPMBX struct)
	nodeP		Ptr to existing node to test key against

Returns :

	< value >	-1, 0, or 1, as dictated by the rules of
			interface between the avl tree routines and
			the key comparison routine.

Notes:


*/


static int
dest_avl_cmpkey( keyP, nodeP )
	XMTPMBX		*keyP;		/* Key to look for */
	DESTNODE	*nodeP;		/* Node to compare against */
{
	int		cmpval;
	XMTPMBX		*ekeyP;		/* Ptr to the key in the node */

    ekeyP = &nodeP->dst_dest.sd_mbx;

    /* Do case-insensitive compare on the strings.
       Something less simplistic would be needed if we treated the
       localpart differently than the domain part.
    */
    cmpval = stricmp( (char *)keyP->mb_pathP,
		       (char *)ekeyP->mb_pathP );
    return ( cmpval );
}
/*

*//* dest_avl_mknode( treeP, keyP, dataP, enodeP )

	Make an avl node on behalf of AVL tree routines,
	  for the destination address tree.

Accepts :

	treeP		Address of tree header block
	keyP		The key for the new node (XMTPMBX)
	dataP		Address of data (not used)
	enodeP		Address of any existing node

Returns :

	<value>		Address of the new node

Notes :

	It's assumed that we can steal the mailbox string from the
	key.  We'll set the key pointer to NULL after doing that.

*/

static AVLNODE *
dest_avl_mknode( treeP, keyP, dataP, enodeP )
	AVLTREE		*treeP;		/* Addr of tree header block */
	XMTPMBX		*keyP;		/* Key for the new node */
	void		*dataP;		/* Addr of new data (n/a) */
	DESTNODE	*enodeP;	/* Existing node... */
{
	DESTNODE	*nodeP;		/* New node that we make */
	SMTP_DEST	*destP;		/* Dest struct in the node */

    /* If enodeP is non-NULL, this is a call to overwrite the data.
       That should never happen, so return NULL in that case. */
    if ( enodeP != NULL )
	return( (AVLNODE *)NULL );

    /* Allocate new node */
    nodeP = (DESTNODE *)emalloc( "dest_avl_mknode", "avl node",
					 sizeof(DESTNODE) );

    /* Setup the node struct */
    destP = &nodeP->dst_dest;
    destP->sd_mbx = *keyP;
    destP->sd_acceptance = MDV_DUNNO;

    /* We've swiped the mailbox string, so clear it from the source */
    keyP->mb_pathP = NULL;

    return( (AVLNODE *)nodeP );
}
/*

*//* dest_avl_rmnode( treeP, nodeP )

	Delete an avl node on behalf of AVL tree routines,
	  for destination address tree

Accepts :

	treeP		Address of tree header block
	nodeP		Address of the node being removed

Returns :

	<value>			0 for success, -1 for failure

Notes :


*/

static int
dest_avl_rmnode( treeP, nodeP )
	AVLTREE		*treeP;
	DESTNODE	*nodeP;
{
	SMTP_DEST	*destP;

    destP = &nodeP->dst_dest;

    /* Deallocate the path buffer */
    if ( destP->sd_mbx.mb_pathP != NULL ) {
	dealloc( destP->sd_mbx.mb_pathP );
	destP->sd_mbx.mb_pathP = NULL;
    }

    /* Deallocate the node */
    dealloc( nodeP );

    /* Return OK */
    return( 0 );
}
/*

*//* generic_ip_check( xmtP, haddr, hnameP )

	Judges whether an IP address is generically allocated

Accepts :

	xmtP		Ptr to XMTP control block, for logging
	haddr		The IPv4 host address (network byte order)
	hnameP		The hostname associated with the host

Returns :

	<value>		MDV_xxx acceptance value, one of:
			  MDV_ACCEPT
			  MDV_REJECT
			  MDV_DUNNO
			  MDV_ERROR

Notes :

	Although this is an IP address check, the check mainly uses
	  the name string.

	The haddr, if given, is used to help look inside the hostname.
	  Its value may be INADDR_NONE .

	May also set various state variables.

*/

static int
generic_ip_check ARGLIST( (xmtP, haddr, hnameP) )
   NFARG( XMTP		*xmtP )		/* XMTP control block */
   NFARG( in_addr_t	haddr )		/* The IP address, if known */
    FARG( UBYTE		*hnameP )	/* The name string */
{
	int		mdv;
	ESTR		*estrP;		/* ESTR for the target string */
	ESTR		*epatP;
	UBYTE		octets[4];
	char		patbuf[500];
	char		mflcmd[200];	/* Command buffer for MFL command */

    mdv = MDV_DUNNO;

    /* Decode the IP address into its component bytes */
    mdns_ip4_iptoba( ntohl( haddr ), &octets[0] );

    estrP = estr_str( hnameP );		/* Need for strmatches */

    /* Check for particular substrings in the host name.  Some are signs of
       non-generic, some are signs of generic.
    */
    if ( estr_match_regex( NULL, estr_str( (UBYTE *)
					      "\\.dynamic\\."
					      "|pool-.*"
					      "|dhcp-.*"
					  ),
				     estrP,
				     0 ) )
	mdv = MDV_REJECT;

    else if ( estr_match_regex( NULL, estr_str( (UBYTE *)
					      "\\.static\\."
					      "|static-.*"
					  ),
				     estrP,
				     0 ) )
	mdv = MDV_DUNNO;

    if ( mdv <= 0 ) {
	/* Need for generated patterns */
	epatP = estr_str( (UBYTE *)&patbuf[0] );

	/* If we know the IP address a.b.c.d, the appearance of
	   various of the octets in various orders is a good sign of
	   generic assignment.  a*b*c or b*c*d or d*c*b or c*b*a are
	   all good indicators, either decimial or hex.
	*/
	if ( haddr != INADDR_NONE ) {
	    /* Try decimal */
	    if ( snprintf( &patbuf[0], sizeof(patbuf),
		       "%d.*%d.*%d|%d.*%d.*%d|%d.*%d.*%d|%d.*%d.*%d",
		       octets[0], octets[1], octets[2],
		       octets[1], octets[2], octets[3],
		       octets[3], octets[2], octets[1],
		       octets[2], octets[1], octets[0] ) < sizeof(patbuf) )
		if ( estr_match_regex( NULL, epatP, estrP, 0 ) )
		    mdv = MDV_REJECT;

	    /* Try hex */
	    if ( ( mdv <= 0 ) &&
	         ( snprintf( &patbuf[0], sizeof(patbuf),
		       "%x.*%x.*%x|%x.*%x.*%x|%x.*%x.*%x|%x.*%x.*%x",
		       octets[0], octets[1], octets[2],
		       octets[1], octets[2], octets[3],
		       octets[3], octets[2], octets[1],
		       octets[2], octets[1], octets[0] ) < sizeof(patbuf) ) )
		if ( estr_match_regex( NULL, epatP, estrP, MF_IGNCASE ) )
		    mdv = MDV_REJECT;
	}
    }

    /* Apply any $hook_generic_ip() if present.  This can override
       our internal decision.
    */
    if ( func_lookup( "$hook_generic_ip" ) != NULL ) {
	if ( snprintf( &mflcmd[0], sizeof(mflcmd),
		      "%s( \"%s\", \"%d.%d.%d.%d\", \"%s\" );",
		      "$hook_generic_ip",
		      mdv_to_str( mdv ),
		      octets[0], octets[1], octets[2], octets[3],
		      hnameP ) >= sizeof(mflcmd) ) {

	    smtpd_log( xmtP, "BUG-", NULL,
		    "buffer not large enough for MFL exec of $hook_generic_ip" );
	}
	else {
	    if ( smtpd_hook_dialog( &mflcmd[0], &mdv ) == SCOK ) {
		;
	    }
	}
    }

    /* We could build more in here, but that's what the hook is for, eh? */
    return ( mdv );
}
/*

*//* helo_check( xmtP )

	Checks the HELO argument.

Accepts :

	xmtP		Ptr to XMTP control block

Returns :

	<value>		MDV_xxx acceptance value, one of:
			  MDV_ACCEPT
			  MDV_REJECT
			  MDV_ERROR

Notes :

	May also set various state variables.

	May be deferred until somewhere into the SMTP dialog.

	The individual checks are specified in a string control
	  containing a list of checks to do.  Each check is done via
	  dispatch table; each entry in the table specifies a check name
	  and a handler function.  The check name may be preceded with
	  "~" to make it a "soft" check (a result may be logged by no
	  action is taken).  The handler function is expected to return
	  one of the MDV_xxxx codes (which may include MDV_DUNNO for no
	  decision).  Any positive handler value is accepted without
	  further processing unless a soft check has been specified.
*/

static int
helo_check ARGLIST( (xmtP) )
    FARG( XMTP		*xmtP )		/* XMTP control block */
{
	int		softF;		/* If "soft" component check */
	int		i;
	int		mdv;
	int		mdvret;
	char		*checkP;	/* component check name */
	UBYTE		*csP;		/* Control string ptr */
	char		buf[100];

    mdv = MDV_DUNNO;

    /* The smtp_server_helo_check control, if defined, contains a list
       of helo check names that are processed in order.

       The table Helochecks[] defines the check names and the functions
       that process each check.  These functions are expected to
       return an MDV_XXX mail disposition value:  MDV_ACCEPT means
       the helo name is unconditionally acceptable, MDV_REJECT means
       that the helo name should be rejected (but see soft check);
       other returns have no effect.

       The check name may be preceded with '~' to make the check
       a soft check.  Any MDV_REJECT from a soft check is only logged,
       but not rejected.  This might help phase in helo checking.
    */
    if ( mdv <= 0 ) {
	csP = cntrl_str_get( estr_tmp_str( "smtp_server_helo_check" ) );
	if ( csP == NULL )
	    csP = "";
	for ( ; mdv <= 0; ) {
	    i = strtoken_ws( &csP, (UBYTE *)&buf[0], sizeof(buf), TRUE );
	    if ( i > sizeof( buf) )
		continue;		/* Ignore too-big check types */

	    if ( buf[0] == NUL )
		break;			/* No more types */

	    checkP = &buf[0];
	    if ( softF = ( *checkP == '~' ) )
		++checkP;		/* soft check */

	    /* Look up the type & dispatch if it's found.  Ignore
	       unknown type names.
	    */
	    for ( i = 0; Helochecks[i].hc_nameP != NUL; ++i ) {
		if ( stricmp( Helochecks[i].hc_nameP, checkP ) == 0 ) {
		    mdvret = (*Helochecks[i].hc_rtc)(xmtP,
						  (char *)&SDS_heloname[0]);
		    if ( mdvret == MDV_ACCEPT )
			mdv = mdvret;
		    else if ( mdvret == MDV_REJECT ) {
			if ( softF ) {	    /* Log only ? */
			    smtpd_log( xmtP, "HELO-+", SDD_logP,
				       "soft reject: (%s)",
				       (char *)&SDS_heloname[0] );
			    SDD_logP = NULL;
			    SDD_replyP = NULL;
			}
			else
			    mdv = mdvret;
		    }
		    else
			/* Just ignore this return */
			;

		    break;
		}
	    }
	}
    }

    /* Turn tentative acceptance value into firm one, and process it. */
    if ( mdv < 0 )
	mdv = -mdv;
    switch ( mdv ) {
	case MDV_DUNNO:			/* No decision, we'll take it */
	    mdv = MDV_ACCEPT;
	    break;

	case MDV_PASS:			/* Shouldn't get this, but... */
	    mdv = MDV_ACCEPT;
	    break;

	case MDV_ACCEPT:		/* OK as is */
	case MDV_REJECT:
	    break;

	case MDV_UNKNOWN:		/* Not really relevant here */
	case MDV_KNOWN:			/*  but translate to reject */
	    mdv = MDV_REJECT;
	    break;

	default:			/* Some problem */
	    mdv = MDV_ERROR;
	    break;
    }

    return ( mdv );
}
/*

*//* helo_check_auth( xmtP, heloP )

	HELO check: auth

Accepts :

	xmtP		Ptr to XMTP control block
	heloP		The HELO name being checked

Returns :

	<value>		MDV_xxx mail disposition value,
				per helo_check interface

Notes :

	Subservient to helo_check, q.v.

	This is the "auth" check, which gives the HELO name a pass if
	  the session has been AUTH'd.  Useful only if HELO checking
	  is deferred past the HELO time.

*/

static int
helo_check_auth ARGLIST( (xmtP, heloP) )
   NFARG( XMTP		*xmtP )		/* XMTP control block */
    FARG( char		*heloP )	/* Helo name to check */
{
    if ( SDS_authF )			/* Give a pass if auth'd */
	return ( MDV_ACCEPT );

    return ( MDV_DUNNO );		/* no stance.. */
}
/*

*//* helo_check_dns( xmtP, heloP )

	HELO check: dns

Accepts :

	xmtP		Ptr to XMTP control block
	heloP		The HELO name being checked

Returns :

	<value>		MDV_xxx mail disposition value,
				per helo_check interface

Notes :

	Subservient to helo_check, q.v.

	This is the "dns" check, which checks that the HELO name has
	  an A record in the DNS.

	If the return is a CNAME, the one CNAME level is followed.

*/

static int
helo_check_dns ARGLIST( (xmtP, heloP) )
   NFARG( XMTP		*xmtP )		/* XMTP control block */
    FARG( char		*heloP )	/* Helo name to check */
{
	BOOL		okF;
	int		len;
	char		*replyP;
	MML_DNS		*dnsP;		/* mml-dns handle */

    /* If the HELO is a [a.b.c.d] form, we make no judgement here. */
    if ( *heloP == '[' )
	return ( MDV_DUNNO );

    if ( ( dnsP = dns_handle() ) == NULL )
	/* Can't do DNS, so we have no idea. */
	return ( MDV_DUNNO );

    /* Do a query, and resolve one CNAME indirection */
    okF = mdns_query( dnsP, heloP, ns_c_in, ns_t_a );
    if ( okF && ( dnsP->dn_anstype == ns_t_cname ) )
	okF = mdns_query( dnsP, dnsP->dn_ansP, ns_c_in, ns_t_a );

    if ( !okF ) {
	/* DNS query failed.  Return depends on the kind of failure. */

#ifdef	MML_HAVE_RES_NINIT
	switch ( dnsP->dn_statP->res_h_errno ) {
#else
	/* yeah, I know, this would work all the time */
	switch ( h_errno ) {
#endif /* MML_HAVE_RES_NINIT */

	    case HOST_NOT_FOUND:	/* Authoritatively not there. */
		SDD_logP = "helo-no-dns";
		len = strlen( heloP ) + 100;
		replyP = (char *)tmpstr( len +1 );
		if ( ( replyP != NULL ) &&
		     ( snprintf( replyP, len,
"Your HELO/EHLO name does not exist in the DNS."
				) < len ) )
		    SDD_replyP = replyP;
		return ( MDV_REJECT );

	    case NO_RECOVERY:		/* Non-recoverable error */
		SDD_logP = "helo-nrec-dns";
		len = strlen( heloP ) + 100;
		replyP = (char *)tmpstr( len +1 );
		if ( ( replyP != NULL ) &&
		     ( snprintf( replyP, len,
"Your HELO/EHLO name does not exist in the DNS."
				) < len ) )
		    SDD_replyP = replyP;
		return ( MDV_REJECT );

	    case NO_DATA:		/* Name exists, no A record. */
		SDD_logP = "helo-no-dns-A";
		len = strlen( heloP ) + 100;
		replyP = (char *)tmpstr( len +1 );
		if ( ( replyP != NULL ) &&
		     ( snprintf( replyP, len,
"Your HELO/EHLO name has no data in the DNS."
				) < len ) )
		    SDD_replyP = replyP;
		return ( MDV_REJECT );

	    default:			/* Anything else, no decision. */
		return ( MDV_DUNNO );
	}
    }

    /* If the return isn't an A, we reject */
    if ( dnsP->dn_anstype != ns_t_a ) {
	len = strlen( heloP ) + 100;
	replyP = (char *)tmpstr( len +1 );
	if ( ( replyP != NULL ) &&
	     ( snprintf( replyP, len,
"Your HELO/EHLO name does not have an A record in the DNS."
			) < len ) )
	    SDD_replyP = replyP;
	SDD_logP = "helo-dns-not-A";
	return ( MDV_REJECT );
    }

    /* Otherwise we don't reject here */
    return ( MDV_DUNNO );
}
/*

*//* helo_check_dnsclient( xmtP, heloP )

	HELO check: dnsclient

Accepts :

	xmtP		Ptr to XMTP control block
	heloP		The HELO name being checked

Returns :

	<value>		MDV_xxx mail disposition value,
				per helo_check interface

Notes :

	Subservient to helo_check, q.v.

	This is the "dnsclient" check, which checks that the HELO name has
	  an A record in the DNS, and that one of those A records matches
	  the A record that the client has connected on.

	Much of this duplicates the straight 'dns' check, maybe they
	  should be combined in some way.

*/

static int
helo_check_dnsclient ARGLIST( (xmtP, heloP) )
   NFARG( XMTP		*xmtP )		/* XMTP control block */
    FARG( char		*heloP )	/* Helo name to check */
{
	BOOL		okF;
	int		i;
	int		len;
	char		*replyP;
	MML_DNS		*dnsP;		/* mml-dns handle */
	char		buf[20];	/* Random buffer */

    /* If the HELO is a [a.b.c.d] form, we can check it directly */
    if ( *heloP == '[' ) {
	i = strlen( heloP );

	/* If there is no closing bracket (which shouldn't happen, the
	   parser shouldn't have allowed it, but what the hey), or
	   if the helo is long, we'll reject it as a format error.
	*/
	if ( ( ( i -1 ) >= sizeof(buf) ) ||
	     ( heloP[i-1] != ']' ) ) {
	    SDD_logP = "helo-dnsclient-format";
	    return ( MDV_REJECT );
	}
	memcpy( &buf[0], &heloP[1], i-2 );
	buf[i-2] = NUL;
	if ( inet_addr( &buf[0] ) != SCI_hostip ) {
	    /* No match.  reject. */
	    SDD_logP = "helo-dnsclient";
	    return ( MDV_REJECT );
	}
	else
	    /* Match.  No opinion. */
	    return ( MDV_DUNNO );
    }

    /* We'll need to do a lookup. */
    if ( ( dnsP = dns_handle() ) == NULL )
	/* Can't do DNS, so we have no idea. */
	return ( MDV_DUNNO );

    /* Do a query, and resolve one CNAME indirection */
    okF = mdns_query( dnsP, heloP, ns_c_in, ns_t_a );
    if ( okF && ( dnsP->dn_anstype == ns_t_cname ) )
	okF = mdns_query( dnsP, dnsP->dn_ansP, ns_c_in, ns_t_a );

    if ( !okF ) {
	/* DNS query failed.  Return depends on the kind of failure. */

#ifdef	MML_HAVE_RES_NINIT
	switch ( dnsP->dn_statP->res_h_errno ) {
#else
	/* dumb */
	switch ( h_errno ) {
#endif /* MML_HAVE_RES_NINIT */

	    case HOST_NOT_FOUND:	/* Authoritatively not there. */
		SDD_logP = "helo-no-dns";
		len = strlen( heloP ) + 100;
		replyP = (char *)tmpstr( len +1 );
		if ( ( replyP != NULL ) &&
		     ( snprintf( replyP, len,
"Your HELO/EHLO name does not exist in the DNS."
				) < len ) )
		    SDD_replyP = replyP;
		return ( MDV_REJECT );

	    case NO_RECOVERY:		/* Non-recoverable error */
		SDD_logP = "helo-nrec-dns";
		len = strlen( heloP ) + 100;
		replyP = (char *)tmpstr( len +1 );
		if ( ( replyP != NULL ) &&
		     ( snprintf( replyP, len,
"Your HELO/EHLO name does not exist in the DNS."
				) < len ) )
		    SDD_replyP = replyP;
		return ( MDV_REJECT );

	    case NO_DATA:		/* Name exists, no A record. */
		SDD_logP = "helo-no-dns-A";
		len = strlen( heloP ) + 100;
		replyP = (char *)tmpstr( len +1 );
		if ( ( replyP != NULL ) &&
		     ( snprintf( replyP, len,
"Your HELO/EHLO name has no data in the DNS."
				) < len ) )
		    SDD_replyP = replyP;
		return ( MDV_REJECT );

	    default:			/* Anything else, no decision. */
		return ( MDV_DUNNO );
	}
    }

    /* Check all the returns until we find an A record match,
       or fail to do so.
    */
    do {
	if ( ( dnsP->dn_anstype == ns_t_a ) &&
	     ( dnsP->dn_ipaddr == SCI_hostip ) ) {
	    /* We matched, we can return with no negative opinion.  yay. */
	    return ( MDV_DUNNO );
	}
    } while ( mdns_ans_next( dnsP ) );

    /* Failed to find a match, so we must reject. */
    SDD_logP = "helo-dnsclient";
    len = strlen( heloP ) + 100;
    replyP = (char *)tmpstr( len +1 );
    if ( ( replyP != NULL ) &&
	 ( snprintf( replyP, len,
"Your HELO/EHLO name does not resolve to your IP address."
			) < len ) )
	SDD_replyP = replyP;

    return ( MDV_REJECT );
}
/*

*//* helo_check_generic( xmtP, heloP )

	HELO check: generic

Accepts :

	xmtP		Ptr to XMTP control block
	heloP		The HELO name being checked

Returns :

	<value>		MDV_xxx mail disposition value,
				per helo_check interface

Notes :

	Subservient to helo_check, q.v.

	This is the "generic" check, which looks at the HELO name
	  (and client IP address) to see whether it looks like a
	  generically assigned address.

*/

static int
helo_check_generic ARGLIST( (xmtP, heloP) )
   NFARG( XMTP		*xmtP )		/* XMTP control block */
    FARG( char		*heloP )	/* Helo name to check */
{
	int		mdv;
	int		len;
	char		*replyP;

    /* If the HELO name is a [a.b.c.d] form, don't subject it to
       this test.
    */
    if ( *heloP == '[' )
	return ( MDV_DUNNO );

    mdv = generic_ip_check( xmtP, SCI_hostip, (UBYTE *)heloP );
    if ( mdv == MDV_REJECT ) {
	SDD_logP = "helo-generic";

	/* Provide some kind of meaningful reply text */
	/* Attempt to generate some kind of meaningful reply text
	   tune this to match the printf string that follows.
	*/
	len = strlen( SCI_hostipP ) + 100;
	replyP = (char *)tmpstr( len +1 );
	if ( ( replyP != NULL ) &&
	     ( snprintf( replyP, len,
"Denied: your HELO/EHLO name has the form of a generic assignment."
			) < len ) )
	    SDD_replyP = replyP;
    }

    /* Return what the ip check said. */
    return ( mdv );
}
/*

*//* helo_check_hook( xmtP, heloP )

	HELO check: hook

Accepts :

	xmtP		Ptr to XMTP control block
	heloP		The HELO name being checked

Returns :

	<value>		MDV_xxx mail disposition value,
				per helo_check interface

Notes :

	Subservient to helo_check, q.v.

	This is the "hook" check, which invokes an MFL hook (if present).

*/

static int
helo_check_hook ARGLIST( (xmtP, heloP) )
   NFARG( XMTP		*xmtP )		/* XMTP control block */
    FARG( char		*heloP )	/* Helo name to check */
{
	int		mdv;
	char		mflcmd[200];	/* Command buffer for MFL command */

    mdv = MDV_DUNNO;

    /* Apply any $hook_helo() if present. */
    if ( ( mdv <= 0 ) &&
         ( func_lookup( "$hook_helo" ) != NULL ) ) {
	if ( snprintf( &mflcmd[0], sizeof(mflcmd),
		      "%s( \"%s\", \"%s\" );",
		      "$hook_helo",
		      mdv_to_str( mdv ),
		      &SDS_heloname[0] ) >= sizeof(mflcmd) ) {

	    smtpd_log( xmtP, "BUG-", NULL,
		    "buffer not large enough for MFL exec of $hook_helo" );
	}
	else {
	    if ( smtpd_hook_dialog( &mflcmd[0], &mdv ) == SCOK ) {
		;
	    }
	}
    }

    return ( mdv );
}
/*

*//* helo_check_me( xmtP, heloP )

	HELO check: me

Accepts :

	xmtP		Ptr to XMTP control block
	heloP		The HELO name being checked

Returns :

	<value>		MDV_xxx mail disposition value,
				per helo_check interface

Notes :

	Subservient to helo_check, q.v.

	This is the "me" check, which gives a reject if the HELO name
	  looks like it is trying to impersonate the local server,
	  i.e., me .

*/

static int
helo_check_me ARGLIST( (xmtP, heloP) )
   NFARG( XMTP		*xmtP )		/* XMTP control block */
    FARG( char		*heloP )	/* Helo name to check */
{
	int		i;
	int		mdv;
	UBYTE		octets[4];
	char		buf[20];

    mdv = MDV_DUNNO;

    if ( mdv <= 0 ) {
	/* No client should be presenting one of our naked IP addresses
	   as their HELO name.  This is a sure malware sign.
	*/
	for ( i = 0; i < MyipC; ++i ) {
	    mdns_ip4_iptoba( ntohl( MyipP[i] ), &octets[0] );
	    snprintf( &buf[0], sizeof(buf), "%d.%d.%d.%d", 
				octets[0], octets[1], octets[2], octets[3] );
	    if ( strcmp( (char *)&SDS_heloname[0],
			 &buf[0] ) == 0 ) {
		SDD_logP = "helo-our-IP";
		SDD_replyP = "HELO/EHLO name rejected.";
		mdv = MDV_REJECT;
		break;
	    }
	}
    }

    return ( mdv );
}
/*

*//* helo_check_pass( xmtP, heloP )

	HELO check: pass

Accepts :

	xmtP		Ptr to XMTP control block
	heloP		The HELO name being checked

Returns :

	<value>		MDV_xxx mail disposition value,
				per helo_check interface

Notes :

	Subservient to helo_check, q.v.

	This is the "pass" check, which gives an OK if the client
	  has gotten a "pass" (e.g. from mvmcas).

*/

static int
helo_check_pass ARGLIST( (xmtP, heloP) )
   NFARG( XMTP		*xmtP )		/* XMTP control block */
    FARG( char		*heloP )	/* Helo name to check */
{
    /* Return accept if there's a pass, dunno otherwise. */
    if ( SCI_t_pass != 0 )
	return ( MDV_ACCEPT );

    return ( MDV_DUNNO );
}
/*

*//* helo_check_syntax( xmtP, heloP )

	HELO check: syntax

Accepts :

	xmtP		Ptr to XMTP control block
	heloP		The HELO name being checked

Returns :

	<value>		MDV_xxx mail disposition value,
				per helo_check interface

Notes :

	Subservient to helo_check, q.v.

	This is the "syntax" check, which checks that the syntax of
	  the HELO name given is conformant with RFCs (minus whatever
	  little exceptions we might allow).  Note that some checks
	  for eggregious syntax violations (like leading or trailing
	  '-' in any domain part) are already done without this.

*/

static int
helo_check_syntax ARGLIST( (xmtP, heloP) )
   NFARG( XMTP		*xmtP )		/* XMTP control block */
    FARG( char		*heloP )	/* Helo name to check */
{
    /* Make sure there is at least one '.' in the helo name */
    if ( strchr( heloP, '.' ) == NULL ) {
	SDD_logP = "syntax";
	SDD_replyP = "Your EHLO/HELO name has invalid syntax.";
	return ( MDV_REJECT );
    }

    /* OK so this isn't such an elaborate check.  */
    return ( MDV_DUNNO );
}
/*

*//* mail_check( xmtP, mbxP )

	Checks the "mail from" name

Accepts :

	xmtP		Ptr to XMTP control block
	mbxP		Ptr to the mail from mailbox

Returns :

	<value>		MDV_xxx acceptance value, one of:
			  MDV_ACCEPT
			  MDV_REJECT
			  MDV_ERROR

Notes :

	May also set various state variables.

	Remember that this may be deferred until somewhere into the
	  SMTP dialog.

	A null mailfrom is always accepted.

	The individual checks are specified in a string control
	  containing a list of checks to do.  Each check is done via
	  dispatch table; each entry in the table specifies a check name
	  and a handler function.  The check name may be preceded with
	  "~" to make it a "soft" check (a result may be logged by no
	  action is taken).  The handler function is expected to return
	  one of the MDV_xxxx codes (which may include MDV_DUNNO for no
	  decision).  Any positive handler value is accepted without
	  further processing unless a soft check has been specified.

*/

static int
mail_check ARGLIST( (xmtP, mbxP) )
   NFARG( XMTP		*xmtP )		/* XMTP control block */
    FARG( XMTPMBX	*mbxP )		/* The mailbox to check */
{
	int		softF;		/* If "soft" component check */
	int		i;
	int		mdv, mdvret;
	char		*checkP;	/* component check name */
	UBYTE		*csP;		/* Control string ptr */
	char		buf[100];

    mdv = MDV_DUNNO;

    /* A null mailfrom must always be accepted. */
    if ( mbxP->mb_len == 0 )
	mdv = MDV_ACCEPT;

    /* The smtp_server_mail_check control, if defined, contains a list
       of mail check names that are processed in order.

       The table Mailchecks[] defines the check names and the functions
       that process each check.  These functions are expected to
       return an MDV_XXX mail disposition value:  MDV_ACCEPT means
       the helo name is unconditionally acceptable, MDV_REJECT means
       that the helo name should be rejected (but see soft check);
       other returns have no effect.

       The check name may be preceded with '~' to make the check
       a soft check.  Any MDV_REJECT from a soft check is only logged,
       but not rejected.  MDV_ACCEPT passes are still taken.
    */
    if ( mdv <= 0 ) {
	csP = cntrl_str_get( estr_tmp_str( "smtp_server_mail_check" ) );
	if ( csP == NULL )
	    csP = "";
	for ( ; mdv <= 0; ) {
	    i = strtoken_ws( &csP, (UBYTE *)&buf[0], sizeof(buf), TRUE );
	    if ( i > sizeof( buf) )
		continue;		/* Ignore too-big check types */

	    if ( buf[0] == NUL )
		break;			/* No more types */

	    checkP = &buf[0];
	    if ( softF = ( *checkP == '~' ) )
		++checkP;		/* soft check */

	    /* Look up the type & dispatch if it's found.  Ignore
	       unknown type names.
	    */
	    for ( i = 0; Mailchecks[i].mc_nameP != NUL; ++i ) {
		if ( stricmp( Mailchecks[i].mc_nameP, checkP ) == 0 ) {
		    mdvret = (*Mailchecks[i].mc_rtc)(xmtP, mbxP);
		    if ( mdvret == MDV_ACCEPT )
			mdv = mdvret;
		    else if ( mdvret == MDV_REJECT ) {
			if ( softF ) {	    /* Log only ? */
			    smtpd_log( xmtP, "MAIL-+", SDD_logP,
				       "soft reject (%s)",
				       mbxP->mb_pathP );
			    SDD_logP = NULL;
			}
			else
			    mdv = mdvret;
		    }
		    else
			/* Just ignore this return */
			;

		    break;
		}
	    }
	}
    }

    /* Turn tentative mdv into firm one, and process it. */
    if ( mdv < 0 )
	mdv = -mdv;
    switch ( mdv ) {
	case MDV_DUNNO:			/* No decision, we'll take it */
	    mdv = MDV_ACCEPT;
	    break;

	case MDV_PASS:			/* Shouldn't see this, but... */
	    mdv = MDV_ACCEPT;
	    break;

	case MDV_ACCEPT:		/* OK as is */
	case MDV_REJECT:
	    break;

	case MDV_UNKNOWN:		/* Not really relevant here */
	case MDV_KNOWN:			/*  but translate to reject */
	    mdv = MDV_REJECT;
	    break;

	default:			/* Some problem */
	    mdv = MDV_ERROR;
	    break;
    }

    return ( mdv );
}
/*

*//* mail_check_auth( xmtP, mbxP )

	Mail check: auth

Accepts :

	xmtP		Ptr to XMTP control block
	mbxP		Ptr to the mail from mailbox

Returns :

	<value>		MDV_xxx acceptance value,
				per mail_check interface

Notes :

	Subservient to mail_check()

	This is the "auth" check, which gives a pass if the session
	  has been AUTH'd.

*/

static int
mail_check_auth ARGLIST( (xmtP, mbxP) )
   NFARG( XMTP		*xmtP )		/* XMTP control block */
    FARG( XMTPMBX	*mbxP )		/* The mailbox to check */
{
    if ( SDS_authF )			/* Give a pass if auth'd */
	return ( MDV_ACCEPT );

    return ( MDV_DUNNO );		/* no stance.. */
}
/*

*//* mail_check_dns( xmtP, mbxP )

	Mail check: dns

Accepts :

	xmtP		Ptr to XMTP control block
	mbxP		Ptr to the mail from mailbox

Returns :

	<value>		MDV_xxx acceptance value,
				per mail_check interface

Notes :

	Subservient to mail_check()

	This is the "dns" check, which checks to see that the domain
	  has an entry in the DNS.
*/

static int
mail_check_dns ARGLIST( (xmtP, mbxP) )
   NFARG( XMTP		*xmtP )		/* XMTP control block */
    FARG( XMTPMBX	*mbxP )		/* The mailbox to check */
{
	BOOL		okF;
	int		len;
	char		*domP;		/* Ptr to domain name */
	char		*replyP;
	MML_DNS		*dnsP;		/* mml-dns handle */

    /* Get pointer to the domain name */
    if ( ( mbxP->mb_atX < 0 ) ||
         ( *(domP = mbxP->mb_pathP + mbxP->mb_atX + 1 ) == NUL ) ) {
	SDD_logP = "[dns-nul-domain]";
	return ( MDV_REJECT );	
    }

    /* Domain must have either an MX or A record. */
TODO	/* or AAAA */

    if ( ( dnsP = dns_handle() ) == NULL )
	/* Can't do DNS, so we have no idea. */
	return ( MDV_DUNNO );

    /* Do a query, and resolve one CNAME indirection */
    okF = mdns_query( dnsP, domP, ns_c_in, ns_t_a );
    if ( okF && ( dnsP->dn_anstype == ns_t_cname ) )
	okF = mdns_query( dnsP, dnsP->dn_ansP, ns_c_in, ns_t_a );

    if ( !okF ) {
	/* DNS failed, try MX */
	okF = mdns_query( dnsP, domP, ns_c_in, ns_t_mx );
	if ( okF && ( dnsP->dn_anstype == ns_t_cname ) )
	    okF = mdns_query( dnsP, dnsP->dn_ansP, ns_c_in, ns_t_mx );
    }

    if ( !okF ) {
	/* DNS query failed.  Return depends on the kind of failure. */

#ifdef	MML_HAVE_RES_NINIT
	switch ( dnsP->dn_statP->res_h_errno ) {
#else
	/* yeah, like I said elsewhere, this would work all the time */
	switch ( h_errno ) {
#endif /* MML_HAVE_RES_NINIT */

	    case HOST_NOT_FOUND:	/* Authoritatively not there. */
		SDD_logP = "no-dns";
		len = strlen( domP ) + 100;
		replyP = (char *)tmpstr( len +1 );
		if ( ( replyP != NULL ) &&
		     ( snprintf( replyP, len,
"Your MAIL FROM domain doesn't exist in the DNS."
				) < len ) )
		    SDD_replyP = replyP;
		return ( MDV_REJECT );

	    case NO_RECOVERY:		/* Non-recoverable error */
		SDD_logP = "nrec-dns";
		len = strlen( domP ) + 100;
		replyP = (char *)tmpstr( len +1 );
		if ( ( replyP != NULL ) &&
		     ( snprintf( replyP, len,
"Your MAIL FROM name does not have an A or MX record."
				) < len ) )
		    SDD_replyP = replyP;
		return ( MDV_REJECT );

	    case NO_DATA:		/* Name exists, no record. */
		SDD_logP = "no-dns";
		len = strlen( domP ) + 100;
		replyP = (char *)tmpstr( len +1 );
		if ( ( replyP != NULL ) &&
		     ( snprintf( replyP, len,
"Your MAIL FROM domain name does not have an A or MX record."
				) < len ) )
		    SDD_replyP = replyP;
		return ( MDV_REJECT );

	    default:			/* Anything else, no decision. */
		return ( MDV_DUNNO );
	}
    }

    /* If the return isn't a A or MX, we reject */
    if ( ( dnsP->dn_anstype != ns_t_a ) &&
         ( dnsP->dn_anstype != ns_t_mx ) ) {
	len = strlen( domP ) + 100;
	replyP = (char *)tmpstr( len +1 );
	if ( ( replyP != NULL ) &&
	     ( snprintf( replyP, len,
"Your HELO/EHLO name does not have an A record in the DNS."
			) < len ) )
	    SDD_replyP = replyP;
	SDD_logP = "dns-not";
	return ( MDV_REJECT );
    }

    /* Otherwise we don't reject here */
    return ( MDV_DUNNO );
}
/*

*//* mail_check_hook( xmtP, mbxP )

	Mail check: dns

Accepts :

	xmtP		Ptr to XMTP control block
	mbxP		Ptr to the mail from mailbox

Returns :

	<value>		MDV_xxx acceptance value,
				per mail_check interface

Notes :

	Subservient to mail_check()

	This is the "hook" check, which invokes an MFL hook (if present).

*/

static int
mail_check_hook ARGLIST( (xmtP, mbxP) )
   NFARG( XMTP		*xmtP )		/* XMTP control block */
    FARG( XMTPMBX	*mbxP )		/* The mailbox to check */
{
	int		mdv;
	char		mflcmd[200];	/* Command buffer for MFL command */

    mdv = MDV_DUNNO;

    /* Apply any $hook_mail_from() if present. */
    if ( ( mdv <= 0 ) &&
         ( func_lookup( "$hook_mail_from" ) != NULL ) ) {
	if ( snprintf( &mflcmd[0], sizeof(mflcmd),
		      "%s( \"%s\", \"%s\" );",
		      "$hook_mail_from",
		      mdv_to_str( mdv ),
		      mbxP->mb_pathP ) >= sizeof(mflcmd) ) {

	    smtpd_log( xmtP, "BUG-", NULL,
		    "buffer not large enough for MFL exec of $hook_mail_from" );
	}
	else {
	    if ( smtpd_hook_dialog( &mflcmd[0], &mdv ) == SCOK ) {
		;
	    }
	}
    }

    return ( mdv );
}
/*

*//* mail_check_pass( xmtP, mbxP )

	Mail check: dns

Accepts :

	xmtP		Ptr to XMTP control block
	mbxP		Ptr to the mail from mailbox

Returns :

	<value>		MDV_xxx acceptance value,
				per mail_check interface

Notes :

	Subservient to mail_check()

	This is the "pass" check, which gives an OK if the client
	  has gotten a "pass" (e.g. from mvmcas).

*/

static int
mail_check_pass ARGLIST( (xmtP, mbxP) )
   NFARG( XMTP		*xmtP )		/* XMTP control block */
    FARG( XMTPMBX	*mbxP )		/* The mailbox to check */
{
    /* Return accept if there's a pass, dunno otherwise. */
    if ( SCI_t_pass != 0 )
	return ( MDV_ACCEPT );

    return ( MDV_DUNNO );
}
/*

*//* mail_check_qmail( xmtP, mbxP )

	Mail check: dns

Accepts :

	xmtP		Ptr to XMTP control block
	mbxP		Ptr to the mail from mailbox

Returns :

	<value>		MDV_xxx acceptance value,
				per mail_check interface

Notes :

	Subservient to mail_check()

	This is the "qmail" check, which checks via qmail-smtpd
	  compatibility.

	Note: only available if compiled with qmail compatibility.

*/

#ifdef	MVMTR_QMAIL_COMPATIBILITY

static int
mail_check_qmail ARGLIST( (xmtP, mbxP) )
   NFARG( XMTP		*xmtP )		/* XMTP control block */
    FARG( XMTPMBX	*mbxP )		/* The mailbox to check */
{
	int		mdv;

    mdv = MDV_DUNNO;

    /* Check via qmail-smtpd compatibility if the issue hasn't already
       been decided.  This check is a refusal check, it really makes no
       judgement about acceptance.
    */
    if ( SDS_qmcompatF && ( mdv <= 0 ) )
	mdv = qmc_smtpd_check_mail_from( mbxP, mdv );

    return ( mdv );
}

#endif	/* MVMTR_QMAIL_COMPATIBILITY */


/*

*//* mdv_from_str( mdvP )

	Decode a mail disposition value name (string) to its value

Accepts :

	mdvP		String containing mail disposition value

Returns :

	<value>		-1 if invalid/unknown name,
			MDV_xxx otherwise

Notes :

	name comparison is case-insensitive

*/

static int
mdv_from_str ARGLIST( (mdvP) )
    FARG( char		*mdvP )		/* Mail disposition name */
{
	int		mdv;		/* The MDV */

    if ( stricmp( mdvP, "dunno" ) == 0 )
	mdv = MDV_DUNNO;
    else if ( stricmp( mdvP, "accept" ) == 0 )
	mdv = MDV_ACCEPT;
    else if ( stricmp( mdvP, "reject" ) == 0 )
	mdv = MDV_REJECT;
    else if ( stricmp( mdvP, "pass" ) == 0 )
	mdv = MDV_PASS;
    else if ( stricmp( mdvP, "known" ) == 0 )
	mdv = MDV_KNOWN;
    else if ( stricmp( mdvP, "unknown" ) == 0 )
	mdv = MDV_UNKNOWN;
    else if ( stricmp( mdvP, "error" ) == 0 )
	mdv = MDV_ERROR;
    else
	mdv = -1;

    return ( mdv );
}
/*

*//* mdv_to_str( mdv )

	Return the name (string) of a mail disposition value

Accepts :

	mdv		The MDV_xxx value

Returns :

	<value>		The MDV name string,
			"" if unknown.

Notes :

	If the input mdv is a negative MDV_xxx value, then the
	  string returned is a "~" followed by the name.

*/

static char *
mdv_to_str ARGLIST( (mdv) )
    FARG( int		mdv )		/* Mail disposition value */
{
	char		*mdvP;		/* The string to return */

    switch ( mdv ) {
	case MDV_DUNNO:			/* This must be 0, there is no */
	    mdvP = "dunno";		/*  negative value of this. */
	    break;

	case MDV_ACCEPT:
	    mdvP = "accept";
	    break;

	case -MDV_ACCEPT:
	    mdvP = "~accept";
	    break;

	case MDV_REJECT:
	    mdvP = "reject";
	    break;

	case -MDV_REJECT:
	    mdvP = "~reject";
	    break;

	case MDV_PASS:
	    mdvP = "pass";
	    break;

	case -MDV_PASS:
	    mdvP = "~pass";
	    break;

	case MDV_KNOWN:
	    mdvP = "known";
	    break;

	case -MDV_KNOWN:
	    mdvP = "~known";
	    break;

	case MDV_UNKNOWN:
	    mdvP = "unknown";
	    break;

	case -MDV_UNKNOWN:
	    mdvP = "~unknown";
	    break;

	case MDV_ERROR:
	    mdvP = "error";
	    break;

	case -MDV_ERROR:
	    mdvP = "~error";
	    break;

	default:
	    mdvP = "";
	    break;
    }

    return ( mdvP );
}
/*

*//* mvmcas_check( tosec )

	Check for (and handle) any messages from mvmcas

Accepts :

	tosec		Number of seconds to wait before timing out

Returns :

	<value>		status code

Notes :

	The timeout applies only to the first message received.  Once
	any message is received, any other pending messages are received
	and handled.

	Although this function returns a status code, callers are not
	really expected to act on it... it's not critical that mvmcas
	communications happen, and any messages from mvmcas simply change
	various state parameters.

*/

#ifdef	MVMCAS_IN_MVMTR

static STSCODE
mvmcas_check ARGLIST( (tosec) )
    FARG( int		tosec )		/* Timeout, in seconds */
{
	STSCODE		status;
	int		i;
	int		fromL;
	int		f_port;		/* Source (from) port */
	struct in_addr	f_addr;		/* Source (from) IP address */
	struct sockaddr	from;
	fd_set		readfds;
	struct timeval	totv;
	UBYTE		readbuf[500];

    /* No-op if not enabled */
    if ( Mvmcas_fd < 0 )
	return ( SCOK );

    /* Setup timeout value for first try */
    totv.tv_sec = tosec;
    totv.tv_usec = 0;

    for ( status = SCOK; status == SCOK; totv.tv_sec = 0 ) {
	FD_ZERO( &readfds );
	FD_SET( Mvmcas_fd, &readfds );
	i = select( Mvmcas_fd +1, &readfds, NULL, NULL, &totv );
	if ( i < 0 ) {
	    smtpd_log( NULL, "MVMCAS-", NULL,
		         "mvmcas_check: error %d in select()", errno );
	    status = SCNOTDONE;
	}
	else if ( !FD_ISSET( Mvmcas_fd, &readfds ) ) {
	    /* Timeout not an error, especially if called with a zero
	       timeout.  For non-zero timeout, which implies that we are
	       actually waiting for something, I suppose we could count it
	       or note it.
	    */
	    status = SCDONE;
	}
	else {
	    fromL = sizeof( from );
	    i = recvfrom( Mvmcas_fd,
			  &readbuf[0], sizeof(readbuf),
			  MSG_WAITALL,
			  (struct sockaddr *)&from, &fromL );
	    if ( i < 0 ) {
		smtpd_log( NULL, "MVMCAS-", NULL,
			   "mvmcas_check: error %d in recvfrom()",
			   errno );
		status = SCNOTDONE;
	    }
	    else if ( i == 0 ) {
		smtpd_log( NULL, "MVMCAS-", NULL,
				"mvmcas_check: read 0 bytes." );
		status = SCDONE;
	    }
	    else {
		readbuf[i] = NUL;

		/* Check out the source.  We *should* be able to predict
		   that the message will come from the port we sent to,
		   but maybe won't go quite that far at this point in the
		   implementation.  At least trap stray DNS replies.
		   First get the address and ports in a convenient place
		   (to make the code look neater, mainly).
		*/
		f_addr = ((struct sockaddr_in *)&from)->sin_addr;
		f_port = ntohs( ((struct sockaddr_in *)&from)->sin_port);
		if ( f_port == 53 ) {		/* Yes this is hardwired */
		    /* Might as well log it */
		    smtpd_log( NULL, "MVMCAS", NULL,
			      "stray DNS reply from %s port %d",
			      inet_ntoa( f_addr ), f_port );
		    continue;
		}
		if ( Debug > 0 ) {
		    fprintf( stderr, "mvmcas_check got response, %d byte%s.\n",
			    i, i == 1 ? "" : "s" );
		    fprintf( stderr, "%s%s",
			      &readbuf[0],
			      readbuf[i-1] == '\n' ? "" : "\n" );
		}
		status = mvmcas_process( &readbuf[0] );
		if ( status != SCOK ) {
		    smtpd_log( NULL, "MVMCAS", NULL, "was from %s port %d",
					inet_ntoa( f_addr ), f_port );
		}
	    }
	}
    }

    if ( status == SCDONE )
	status = SCOK;

    return ( status );
}

#endif	/* MVMCAS_IN_MVMTR */
/*

*//* mvmcas_process( msgP )

	Process a message from mvmcas

Accepts :

	msgP		Ptr to NUL-terminated message from mvmcas

Returns :

	<value>		status code

Notes :


*/

#ifdef	MVMCAS_IN_MVMTR

static STSCODE
mvmcas_process ARGLIST( (msgP) )
    FARG( UBYTE		*msgP )		/* Ptr to message */
{
	int		i;
	UBYTE		*tailP;		/* Tail of the message, for processing */
	CAVTE		*cavteP;	/* Verb processing table ptr */
	char		verb[20];	/* verb from message */

    /* Get the verb */
    tailP = msgP;
    i = strtoken_ws( &tailP, &verb[0], sizeof(verb), TRUE );
    if ( i == 0 )
	return ( SCOK );		/* Nothing in buffer, just be OK */

    if ( i >= sizeof(verb) ) {
	smtpd_log( NULL, "MVMCAS-", NULL,
		   "mvmcas_process: verb too large in \"%s\"", msgP );
	return ( SCNOTDONE );
    }

    /* Look up the verb in the processing table */
    for( cavteP = &Mvmcas_vtbl[0]; cavteP->cav_nameP != NULL; ++cavteP ) {
	if ( stricmp( (char *)&verb[0], (char *)cavteP->cav_nameP ) == 0 )
	    break;
    }

    if ( cavteP->cav_nameP == NULL ) {
	smtpd_log( NULL, "MVMCAS-", NULL,
		   "mvmcas_process: unrecognized message \"%s\"", msgP );
	return ( SCNOTDONE );
    }

    /* Hand off to the processing function */
    return ( (*cavteP->cav_rtc)( msgP, tailP ) );
}

#endif	/* MVMCAS_IN_MVMTR */
/*

*//* mvmcas_process_hi( msgP, tailP )

	mvmcas message processing: "HI"

Accepts :

	msgP		Ptr to entire NUL-terminated message from mvmcas
	tailP		Ptr to the tail part of the message

Returns :

	<value>		status code

Notes :

	The tail pointer is the part that hasn't been processed yet.

*/

#ifdef	MVMCAS_IN_MVMTR

static STSCODE
mvmcas_process_hi ARGLIST( (msgP, tailP) )
   NFARG( UBYTE		*msgP )		/* Ptr to message */
    FARG( UBYTE		*tailP )	/* Ptr to tail that we will process */
{
    /* "HI" doesn't do anything. */
    return ( SCOK );
}

#endif	/* MVMCAS_IN_MVMTR */
/*

*//* mvmcas_process_info( msgP, tailP )

	mvmcas message processing: "INFO"

Accepts :

	msgP		Ptr to entire NUL-terminated message from mvmcas
	tailP		Ptr to the tail part of the message

Returns :

	<value>		status code

Notes :

	The tail pointer is the part that hasn't been processed yet.

*/

#ifdef	MVMCAS_IN_MVMTR

static STSCODE
mvmcas_process_info ARGLIST( (msgP, tailP) )
   NFARG( UBYTE		*msgP )		/* Ptr to message */
    FARG( UBYTE		*tailP )	/* Ptr to tail that we will process */
{
	BOOL		skipF;		/* Skipping to semicolon */
	long		lv1;
	ULONG		ulv1, ulv2;
	UBYTE		tok[ 100 ];
	UBYTE		tok1[ 100 ];
	UBYTE		tok2[ 100 ];
	UBYTE		tok3[ 100 ];

    /* Info message looks like:
	INFO mcname ; [args] END

	mcname is "IP xxx" or "AUTH xxx"
	args are argument strings, each string ending with semicolon token.
    */

    /*	Get the first two tokens of the mail client name */
    if ( ! (mvmcas_token( &tailP, &tok[0], sizeof(tok) ) &&
	    mvmcas_token( &tailP, &tok1[0], sizeof(tok1) ) ) )
	return ( SCNOTDONE );

    if ( ( stricmp( (char *)&tok[0], "ip" ) != 0 ) &&
         ( stricmp( (char *)&tok[0], "auth" ) != 0 ) ) {
	smtpd_log( NULL, "MVMCAS-", NULL,
		   "mvmcas_process_info: bad mclient type \"%s\"", &tok[0] );
	return ( SCNOTDONE );
    }

    /* Process any arguments, stop when we see END */
    skipF = TRUE;			/* Skip to first semi first */
    for ( ; ; ) {
	if ( ! mvmcas_token( &tailP, &tok[0], sizeof(tok) ) ) {
	    /* Token too large; skip to the next semi. */
	    skipF = TRUE;
	    continue;			/* too large, warning given */
	}
	if ( tok[0] == NUL )
	    break;			/* No more tokens */
	if ( skipF ) {
	    /* Skipping to semi... */
	    if ( ( tok[0] != ';' ) || ( tok[1] != NUL ) ) {
		smtpd_log( NULL, "MVMCAS-", NULL,
			   "mvmcas_process_info: expected semi, got \"%s\"",
			&tok[0] );
	    }
	    else
		skipF = FALSE;		/* Found it */
	    continue;
	}

	/* Toggle the skip flag back on; anything else handled here (valid
	   or not) will need to look for semicolon.
	*/
	skipF = TRUE;

	/* Process token.  Check end first.. */
	if ( stricmp( (char *)&tok[0], "END" ) == 0 )
	    break;

	else if ( stricmp( (char *)&tok[0], "AGG" ) == 0 )
	    /* Aggregate (ignore) */
	    ;

	else if ( stricmp( (char *)&tok[0], "AUTHREQ" ) == 0 )
	    /* AUTH is required for this source to do anything. */
	    SCI_authreqF = TRUE;

	else if ( stricmp( (char *)&tok[0], "BLOCK" ) == 0 ) {
	    /* Blocked:  how type end code */
	    if ( !mvmcas_token( &tailP, &tok1[0], sizeof(tok1) ) ) {
		smtpd_log( NULL, "MVMCAS-", NULL,
			   "mvmcas_process_info: too few arguments to BLOCK" );
		skipF = FALSE;
	    }
	    else if ( ( tok1[0] == ';' ) && ( tok1[1] == NUL ) )
		skipF = FALSE;
	    else if ( !mvmcas_token( &tailP, &tok2[0], sizeof(tok2) ) ) {
		smtpd_log( NULL, "MVMCAS-", NULL,
			   "mvmcas_process_info: too few arguments to BLOCK" );
		skipF = FALSE;
	    }
	    else if ( ( tok2[0] == ';' ) && ( tok2[1] == NUL ) )
		skipF = FALSE;
	    else if ( !mvmcas_token_uint( &tailP, &ulv1 ) )
		skipF = ( ulv1 != 1 );
	    else if ( !mvmcas_token( &tailP, &tok3[0], sizeof(tok3) ) )
		;
	    else if ( ( tok3[0] == ';' ) && ( tok3[1] == NUL ) ) {
		smtpd_log( NULL, "MVMCAS-", NULL,
			   "mvmcas_process_info: too few arguments to BLOCK" );
		skipF = FALSE;
	    }
	    else {
		if ( stricmp( (char *)&tok1[0], "conn" ) == 0 ) {
		    SCI_block = MCBLOCK_CONN;
		}
		else if ( stricmp( (char *)&tok1[0], "greet" ) == 0 ) {
		    SCI_block = MCBLOCK_GREET;
		}
		else if ( stricmp( (char *)&tok1[0], "smtp" ) == 0 ) {
		    SCI_block = MCBLOCK_SMTP;
		}
		else {
		    smtpd_log( NULL, "MVMCAS-", NULL,
			     "mvmcas_process_info: invalid block type \"%s\"",
			      &tok1[0] );
		    SCI_block = MCBLOCK_NONE;
		}

		if ( SCI_block == MCBLOCK_NONE )
		    ;
		else if ( stricmp( (char *)&tok2[0], "admin" ) == 0 )
		    SCI_blockhowP = "administrative";
		else if ( stricmp( (char *)&tok2[0], "auto" ) == 0 ) {
		    SCI_blockhowP = "automatic";

		    /* If autoblocks disabled, turn into not-blocked */
		    if ( cntrl_int_get(
		            estr_tmp_str( "smtp_server_autoblock_enable" ),
			       1 ) == 0 )
			SCI_block = MCBLOCK_NONE;
		}
		else {
		    smtpd_log( NULL, "MVMCAS-", NULL,
			       "mvmcas_process_info: invalid block how \"%s\"",
			      &tok2[0] );
		    SCI_block = MCBLOCK_NONE;
		}
		if ( SCI_block != MCBLOCK_NONE ) {
		    /* Remember the block code */
		    if ( SCI_blockcodeP != NULL )
			dealloc( SCI_blockcodeP );
		    SCI_blockcodeP = newstr( "mvmcas_process_info",
					     "block code",
					     (char *)&tok3[0] );
		}
	    }
	}

	else if ( stricmp( (char *)&tok[0], "AUTHREQ" ) == 0 )
	    /* DNSBL check should be applied */
	    SCI_dnsblF = TRUE;

	else if ( stricmp( (char *)&tok[0], "IRT" ) == 0 )
	    /* In Response To -- ignore this */
	    ;

	else if ( stricmp( (char *)&tok[0], "OCC" ) == 0 ) {
	    /* Informational -- the number of occurances */
	    if ( mvmcas_token_uint( &tailP, &ulv1 ) )
		;
	    else
		/* Problem with input value; set skip flag accordingly */
		skipF = ( ulv1 != 1 );
	}

	else if ( stricmp( (char *)&tok[0], "MAXCONN" ) == 0 )
	    /* Max connections exceeded */
	    ;

	else if ( stricmp( (char *)&tok[0], "PASS" ) == 0 ) {
	    /* Says we should give a free pass on some checks.  The
	       value is an end time/interval, which we take as a flag.
	    */
	    if ( mvmcas_token_uint( &tailP, &ulv1 ) )
		SCI_t_pass = (time_t)ulv1;
	    else
		/* Problem with input value; set skip flag accordingly */
		skipF = ( ulv1 != 1 );
	}

	else if ( stricmp( (char *)&tok[0], "PROTO" ) == 0 )
	    /* Prototype (ignore) */
	    ;

	else if ( stricmp( (char *)&tok[0], "RCPT" ) == 0 ) {
	    /* Bookkeepping counts -- sent as part of generic INFO
	       message.
	    */
	    /* Get "good" count */
	    if ( !mvmcas_token_uint( &tailP, &ulv1 ) )
		/* Problem with input value; set skip flag accordingly */
		skipF = ( ulv1 != 1 );

	    /* Get "bad" count */
	    else if ( !mvmcas_token_uint( &tailP, &ulv2 ) )
		skipF = ( ulv2 != 1 );

	    else {
		/* We don't need these values, we just ignore 'em. */
		;
	    }
	}

	else if ( stricmp( (char *)&tok[0], "RELAYOK" ) == 0 )
	    /* Relay is OK for this */
	    SCI_relayokF = TRUE;

	else if ( stricmp( (char *)&tok[0], "RGOV" ) == 0 ) {
	    /* Bookkeepping, rcpt-to governor, sent as part of
	       generic info message.
	    */
	    if ( !mvmcas_token_int( &tailP, &lv1 ) )
		/* Problem with input value; set skip flag accordingly */
		skipF = ( lv1 != 1 );
	    else {
		/* We don't need this. */
		;
	    }
	}

	else if ( stricmp( (char *)&tok[0], "RMCNAME" ) == 0 )
	    /* The originally requested mcname */
	    ;

	else if ( stricmp( (char *)&tok[0], "T.GL" ) == 0 ) {
	    /* Specifies greylisting; value is either end time or 
		interval: at any rate, it's a number.
	    */
	    if ( mvmcas_token_uint( &tailP, &ulv1 ) )
		SCI_t_gl = (time_t)ulv1;
	    else
		/* Problem with input value; set skip flag accordingly */
		skipF = ( ulv1 != 1 );
	}

	else if ( stricmp( (char *)&tok[0], "T.GREET" ) == 0 ) {
	    /* Pre-greeting delay */
	    /* OBSOLETE -- here temporarily.  Replaced by T.PGD */
	    if ( mvmcas_token_int( &tailP, &lv1 ) ) {
		if ( ( lv1 < 0 ) || ( lv1 > 255 ) )
		    smtpd_log( NULL, "MVMCAS-", NULL,
			       "mvmcas_process_info: invalid T.GREET %d",
				lv1 );
		else
		    SCI_t_greet = lv1;
	    }
	    else {
		/* Problem with input value; set skip flag accordingly */
		skipF = ( lv1 != 1 );
	    }
	}

	else if ( stricmp( (char *)&tok[0], "T.PGD" ) == 0 ) {
	    /* Pre-greeting delay */
	    if ( mvmcas_token_int( &tailP, &lv1 ) ) {
		if ( ( lv1 < 0 ) || ( lv1 > 255 ) )
		    smtpd_log( NULL, "MVMCAS-", NULL,
			       "mvmcas_process_info: invalid T.PGD %d",
				lv1 );
		else
		    SCI_t_greet = lv1;
	    }
	    else {
		/* Problem with input value; set skip flag accordingly */
		skipF = ( lv1 != 1 );
	    }
	}

	else {
	    /* Unrecognized; ignore by skipping to semi. */
	    smtpd_log( NULL, "MVMCAS-", NULL,
		      "mvmcas_process_info: unknown \"%s\"", &tok[0] );
	}
    }
	
    /* If we're here, we found the "END" or ran out of input.  At any
       rate, we're done with this message.
    */
    return ( SCOK );
}

#endif	/* MVMCAS_IN_MVMTR */
/*

*//* mvmcas_send( fmtP, args... )

	Send a message to mvmcas

Accepts :

	fmtP		printf-style format string, or NULL.
	args..		any arguments to the format string

Returns :

	<value>		status code

Notes :

	If fmtP is passed as NULL, the string to send is presumed to
	  be already prepared (in the mvmcas BSTR).  Otherwise,
	  the message is formatted according to the format string
	  and its arguments.

*/

#ifdef	MVMCAS_IN_MVMTR

static STSCODE
mvmcas_send ARGLIST( (fmtP, va_decl) )
   NFARG( char		*fmtP )		/* Format string */
    FARG( ... )				/* variable arguments */
{
	STSCODE		status;
	int		len;
	va_list		ap;		/* Varargs arg ptr */

    /* Don't send if mvmcas is not enabled */
    if ( Mvmcas_fd < 0 )
	return ( SCOK );

    /* Start varargs */
    va_start( ap, fmtP );

    if ( fmtP != NULL ) {
	/* Format the string into our bstr */
	Mvmcas_bstrP->bs_bC = 0;
	bstr_vprintf( Mvmcas_bstrP, fmtP, ap );
    }

    if ( Debug > 0 )
	fprintf( stderr, "sending \"%s\"\n", Mvmcas_bstrP->bs_bP );

    /* Send to mvmcas */
    len = sendto( Mvmcas_fd,
		  (char *)Mvmcas_bstrP->bs_bP, Mvmcas_bstrP->bs_bC,
		  0,
		  (struct sockaddr *)&Mvmcas_to, sizeof(Mvmcas_to) );

    /* No error here, but return status if sent length is not buffer length */
    if ( len == Mvmcas_bstrP->bs_bC )
	status = SCOK;
    else
	status = SCNOTDONE;

    va_end( ap );

    return ( status );
}

#endif	/* MVMCAS_IN_MVMTR */
/*

*//* mvmcas_send_update( stopF, doneF )

	Send an update message to mvmcas

Accepts :

	stopF		If the mail client is stopping
	doneF		If the session is ending

Returns :

	<value>		status code

Notes :

	Sends any mail client update message to mvmcas 

*/

#ifdef	MVMCAS_IN_MVMTR

static STSCODE
mvmcas_send_update ARGLIST( (stopF, doneF) )
   NFARG( BOOL		stopF )		/* If mail client is stopping */
    FARG( BOOL		doneF )		/* If session is ending */
{
	BOOL		okF;

    /* Return if mvmcas disabled or not connected */
    if ( ( !SDS_mvmcasF ) || ( Mvmcas_fd < 0 ) )
	return ( SCOK );

    /* Check if there is no update to be sent */
    if ( ( SDS_rcpt_acceptC == 0 ) &&
         ( SDS_rcpt_rejectC == 0 ) &&
	 !stopF ) {
	/* There is no update to send.
	   If done, send a DONE.
	*/
	if ( doneF )
	    mvmcas_send( "DONE" );
	return ( SCOK );
    }

    /* Prepare the update message */
    Mvmcas_bstrP->bs_bC = 0;

    okF = bstr_printf( Mvmcas_bstrP, "%s %s %s ;",
	stopF ? "STOP" : "UPDATE",
	SDS_authF ? "AUTH " : "IP",
	SDS_authF ? SDS_authnameP : (UBYTE *)SCI_hostipP );

    /* Add connection behaviour -- but only if this is a stop */
    if ( doneF ) {
	okF = okF && bstr_printf( Mvmcas_bstrP, " CONND" );

	if ( SDS_greetF )
	    okF = okF && bstr_printf( Mvmcas_bstrP, " GREET" );
	if ( SDS_conn_pgdF )
	    okF = okF && bstr_printf( Mvmcas_bstrP, " PGDISC" );
	if ( SDS_conn_pgtF )
	    okF = okF && bstr_printf( Mvmcas_bstrP, " PGTALK" );
	if ( SDS_conn_validF )
	    okF = okF && bstr_printf( Mvmcas_bstrP, " VALID" );

	okF = okF && bstr_printf( Mvmcas_bstrP, " ;" );
    }

    /* Add RCPT-To totals */
    if ( ( SDS_rcpt_acceptC != 0 ) ||
         ( SDS_rcpt_rejectC != 0 ) )
	okF = okF && bstr_printf( Mvmcas_bstrP, " RCPT +%d -%d ;",
				  SDS_rcpt_acceptC, SDS_rcpt_rejectC );

    /* Add DATA flag if we had data */
    if ( SDS_dataF )
	okF = okF && bstr_printf( Mvmcas_bstrP, " DATA ;" );

    /* Add DONE if we're stopping */
    if ( doneF )
	okF = okF && bstr_printf( Mvmcas_bstrP, " DONE ;" );

    /* End of message */
    okF = okF && bstr_printf( Mvmcas_bstrP, " END" );

    /* Send it */
    if ( okF )
	mvmcas_send( NULL );

    return ( SCOK );
}

#endif	/* MVMCAS_IN_MVMTR */
/*

*//* mvmcas_start( tosec )

	Initialize for communicating with mvmcas

Accepts :

	<nothing>

Returns :

	<value>		status code

Notes :


*/

#ifdef	MVMCAS_IN_MVMTR

static STSCODE
mvmcas_start NOARGLIST
{
	int		port;		/* Port to talk to */
	char		*hostnameP;	/* The server hostname */
	char		*strP;
	struct hostent	*hP;		/* Host entry.. */
	char		hostname[ PATH_MAX +1 ];

    /* Get hostname and port */
    hostnameP = (char *)cntrl_str_get( estr_tmp_str( "mvmcas_host" ) );
    if ( hostnameP == NULL )
	hostnameP = "localhost";
    port = MVMCAS_PORT;

    /* Separate hostname into hostname and port */
    if ( strlen( hostnameP ) >= sizeof(hostname) ) {
	warning( "mvmcas_start: hostname too large \"%s\"",
		  hostnameP );
	return ( SCERR );
    }
    strncpy( &hostname[0], hostnameP, sizeof(hostname) );
    strP = strrchr( &hostname[0], ':' );
    if ( strP != NULL ) {
	*strP++ = NUL;
	if ( *strP == NUL )
	    strP = NULL;
    }
    if ( strP != NULL )
	port = atod( strP );
    else
	port = MVMCAS_PORT;


    /* Set the socket to -1 to indicate no connection to mvmcas.  It
       should already be that value, but set it anyway.
    */
    Mvmcas_fd = -1;

    /* Set up the bstr we use to send to mvmcas */
    Mvmcas_bstrP = bstr_new( "mvmcas output", 100, 0 );
    if ( Mvmcas_bstrP == NULL ) {
	smtpd_log( NULL, "MVMCAS-", NULL,
		   "mvmcas_start: can't create mvmcas BSTR" );
	return ( SCERR );
    }

    hP = gethostbyname( &hostname[0] );
    if ( hP == NULL ) {
	smtpd_log( NULL, "MVMCAS-", NULL,
		  "mvmcas_start: bad mvmcas server hostname \"%s\"",
		  &hostname[0] );
	return ( SCNOTDONE );
    }

    /* Fill in the address info we use to send to mvmcas server. */
    memset( &Mvmcas_to, 0, sizeof(Mvmcas_to) );
    Mvmcas_to.sin_len = sizeof(Mvmcas_to);
    Mvmcas_to.sin_family = hP->h_addrtype;
    memcpy( (char *)&Mvmcas_to.sin_addr, hP->h_addr, hP->h_length );
    Mvmcas_to.sin_port = htons( port );

    /* Create the socket we'll use */
    Mvmcas_fd = socket( AF_INET, SOCK_DGRAM, IPPROTO_UDP );
    if ( Mvmcas_fd < 0 ) {
	smtpd_log( NULL, "MVMCAS-", NULL,
		   "mvmcas_start: error %d creating socket.", errno );
	return ( SCNOTDONE );
    }

    return ( SCOK );
}

#endif	/* MVMCAS_IN_MVMTR */
/*

*//* mvmcas_token( msgPP, tokP, tokL )

	Get next token from an mvmcas message

Accepts :

	msgPP		Ptr to the buffer pointer
	tokP		Where to put the token
	tokL		Length of token buffer

Returns :

	<value>		TRUE if successful
			FALSE otherwise.

Notes :

	More or less a wrapper around strtoken_ws() from mml;
	as with strtoken_ws,
	  - the pointer at *msgPP is updated to point after the extracted
	    token.
	  - the returned token is NUL-terminated.

	The FALSE return will happen if the buffer is too small for
	  the token, and a warning message will be given.

*/

#ifdef	MVMCAS_IN_MVMTR

static BOOL
mvmcas_token ARGLIST( (msgPP, tokP, tokL) )
   NFARG( UBYTE		**msgPP )	/* Ptr to buffer ptr */
   NFARG( UBYTE		*tokP )		/* Where to put the token */
    FARG( int		tokL )		/* Length of token buffer */
{
	int		i;
	UBYTE		*msgP;

    /* Get a copy of the message pointer, in case we need it for warning */
    msgP = *msgPP;

    /* Do token-getting */
    i = strtoken_ws( msgPP, tokP, tokL, TRUE );
    if ( i >= tokL ) {
	smtpd_log( NULL, "MVMCAS-", NULL,
		   "mvmcas_token: token too large at \"%s\"", msgP );
	return ( FALSE );
    }

    return ( TRUE );
}

#endif	/* MVMCAS_IN_MVMTR */
/*

*//* mvmcas_token_int( msgPP, longP )

	Get next token from an mvmcas message, interpret as int.

Accepts :

	msgPP		Ptr to the buffer pointer
	longP		Where to store the value

Returns :

	<value>		TRUE if successful
			FALSE otherwise.

Notes :

	At present the only validity check that is done is to see if the
	string begins with a sign or a digit.

	For a FALSE return, the return value *longP also conveys meaning:
	  0 - nothing special
	  1 - input was a semicolon token.

*/

#ifdef	MVMCAS_IN_MVMTR

static BOOL
mvmcas_token_int ARGLIST( (msgPP, longP) )
   NFARG( UBYTE		**msgPP )	/* Ptr to buffer ptr */
    FARG( long		*longP )	/* Where to store the result */
{
	UBYTE		tok[50];	/* Token for integer value */

    if ( !mvmcas_token( msgPP, &tok[0], sizeof(tok) ) ) {
	/* Bad input. */
	*longP = 0;
	return ( FALSE );
    }

    if ( ( tok[0] == '-' ) || ( tok[0] == '+' ) ||
	 ( isdigit( tok[0] ) ) ) {
	/* Input looks OK; get value and return it. */
	*longP = atod( (char *)&tok[0] );
	return ( TRUE );
    }

    /* Not valid input; check for semicolon token */
    if ( ( tok[0] == ';' ) && ( tok[1] == NULL ) )
	*longP = 1;
    else
	*longP = 0;

    return ( FALSE );
}

#endif	/* MVMCAS_IN_MVMTR */
/*

*//* mvmcas_token_uint( msgPP, ulongP )

	Get next token from an mvmcas message, interpret as unsigned int.

Accepts :

	msgPP		Ptr to the buffer pointer
	ulongP		Where to store the value

Returns :

	<value>		TRUE if successful
			FALSE otherwise.

Notes :

	At present the only validity check that is done is to see if the
	string begins with a plus or a digit.

	For a FALSE return, the return value *ulongP also conveys meaning:
	  0 - nothing special
	  1 - input was a semicolon token.

*/

#ifdef	MVMCAS_IN_MVMTR

static BOOL
mvmcas_token_uint ARGLIST( (msgPP, ulongP) )
   NFARG( UBYTE		**msgPP )	/* Ptr to buffer ptr */
    FARG( ULONG		*ulongP )	/* Where to store the result */
{
	UBYTE		tok[50];	/* Token for integer value */

    if ( !mvmcas_token( msgPP, &tok[0], sizeof(tok) ) ) {
	/* Bad input. */
	*ulongP = 0;
	return ( FALSE );
    }

    if ( isdigit( tok[0] ) || ( tok[0] == '+' ) ) {
	/* Input looks OK; get value and return it. */
	*ulongP = atod( (char *)&tok[0] );
	return ( TRUE );
    }

    /* Not valid input; check for semicolon token */
    if ( ( tok[0] == ';' ) && ( tok[1] == NULL ) )
	*ulongP = 1;
    else
	*ulongP = 0;

    return ( FALSE );
}

#endif	/* MVMCAS_IN_MVMTR */
/*

*//* smtpd_address_check( addrP, addrL, acceptance )

	Check an address to see if we'll accept mail for it.

Accepts :

	addrP		Address to check
	addrL		Length of the addrP string
	acceptance	Any acceptance value already determined

Returns :

	<value>		an MDV_XXXX acceptance code (an int).

Notes :


*/


static int
smtpd_address_check ARGLIST( (addrP, addrL, mdv) )
   NFARG( UBYTE		*addrP )	/* email address */
   NFARG( int		addrL )		/* Bytes in the address */
    FARG( int		mdv )		/* Current mail disposition */
{
	int		newmdv;

    newmdv = mdv;			/* Start with what came in */

    /* Check the address map. */
    if ( SDS_addrmapF )
	newmdv = smtpd_addrmap_check( addrP, addrL );

    /* If we're going to accept, retain any prior PASS */
    if ( newmdv == MDV_ACCEPT )
	if ( ( mdv == MDV_PASS ) || ( mdv == -MDV_PASS ) )
	    newmdv = MDV_PASS;

    /* May want to do other things here. */
    ;

    return ( newmdv );
}
/*

*//* smtpd_addrmap_check(addrP, addrL)

	Look up an address in the address map database

Accepts :

	addrP		Ptr to address string
	addrL		Length of address string

Returns :

	<value>		an MDV_XXXX acceptance code (an int):
			  MDV_ACCEPT: address is found in the address map
			      and explicitly accepted;
			  MDV_REJECT: address is found in the address map
			      and explicitly rejected;
			  MDV_KNOWN: address is found in the address map
			      without being accepted or rejected
			      (e.g. it's valid on some other system);
			  MDV_UNKNOWN: address is not found in the address map;
			  MDV_ERROR: some error occured.

Notes :

	This interface will probably need expanding when/if the
	address map database is enhanced.

*/

static int
smtpd_addrmap_check ARGLIST( (addrP, addrL) )
   NFARG( UBYTE		*addrP )	/* email address */
    FARG( int		addrL )		/* Bytes in the address */
{
	UBYTE		ch;
	int		r;
	int		atX;
	int		len;
	unsigned int	dlen;
	off_t		dpos;
	UBYTE		*bP;
	UBYTE		addr[ XMTP_PATH_MAX ];
	UBYTE		data[ 20 ];

    /* Idiot check to see if we can use our local buffer.. this should
       never fail since the parsed address should be constrained to be
       less than this length.
    */
    if ( addrL >= (sizeof(addr) -1) )
	return ( MDV_ERROR );

    /* Get lowercase copy (yes we should have a lowercase copy at
       some higher level that we pass around)
    */
    for ( bP = &addr[0], len = addrL; len-- > 0; ) {
	ch = *addrP++;
	*bP++ = isupper(ch) ? tolower(ch) : ch;
    }
    *bP = NUL;

    /* Try exact match first. */
    r = cdb_find( &Addrmapdb.cdb_cdb, (char *)&addr[0], addrL );
    if ( r == 0 ) {

	/* No exact match.  Try all substrings of the localpart looking
           for a wildcard.  thus if input is user@example.com we will try

	     use*@example.com
	     us*@example.com
	     u*@example.com

	   Although note if the appropriate conditional is enabled,
	   wildcards will only be attempted after '-' breaks.

	   Note that we do NOT try *@example.com since a "*" wildcard
	   will be represented in the file as simply the domain name, and
	   we drop through to try various blank domain name forms after
	   the address wildcards have been tried.
	*/

	atX = strbyte_n( &addr[0], addrL, '@' );
	if ( atX < addrL )
	    --atX;

	for ( len = addrL; atX > 0; ) {


#define WCBREAKS
#ifdef WCBREAKS	/* mem 20041115 enable to only try wildcards at hyphen
		   breaks.
		*/
	    if ( addr[atX -1] == '-' ) {
#endif	/* WCBREAKS */

	    addr[atX] = '*';		/* Insert wildcard before '@' */
	    r = cdb_find( &Addrmapdb.cdb_cdb, (char *)&addr[0], len );
	    if ( r != 0 )
		break;			/* Fail or err.. done */

#ifdef	WCBREAKS
	    }
#endif	/* WCBREAKS */

	    /* Shift domain part left one (including the '@') ..
	       also include the terminating NUL just to accomodate
	       any debugging/logging.
	    */
	    memmove( &addr[atX -1], &addr[atX], len +1 - atX );
	    --atX;
	    --len;
	}
    }

    /* mem 20040106 if still no match, try the exact domain
       plus any enclosing domains.  e.g. if we see x.example.com,
       we try:
         x.example.com -- the domain exactly
	 .example.com -- the next higher domain
	 .com -- the next higher domain
    */
    if ( ( r == 0 ) && ( addr[atX +1] == '@' ) ) {
	/* Try the exact domain plus any .domains found */
	for ( atX += 2; atX < len; ) {
	r = cdb_find( &Addrmapdb.cdb_cdb, (char *)&addr[atX], len - atX );
	if ( r != 0 )
	    break;

	/* Spin to the next '.' */
	for ( ++atX; ( atX < len ) && ( addr[atX] != '.' ); ++atX )
	    ;
	}
    }

    /* mem 20040106 if addr found, check associated value */
    if ( r < 0 )
	return ( MDV_ERROR );		/* Some processing error */
    if ( r == 0 )
	return ( MDV_UNKNOWN );		/* Not found */

    dlen = cdb_datalen( &Addrmapdb.cdb_cdb );
    dpos = cdb_datapos( &Addrmapdb.cdb_cdb );
    if ( dlen >= (sizeof(data) -1) )
	return ( MDV_ERROR );

    r = cdb_read( &Addrmapdb.cdb_cdb, (char *)&data[0], dlen, dpos );
    if ( r != 0 )
	return ( MDV_ERROR );

    data[dlen] = NUL;
    if ( strcmp( (char *)&data[0], "accept" ) == 0 )
	return ( MDV_ACCEPT );
    if ( strcmp( (char *)&data[0], "deny" ) == 0 )
	return ( MDV_REJECT );
    if ( strcmp( (char *)&data[0], "pass" ) == 0 )
	/* Address is OK, and also gets a pass on some sender checks */
	return ( MDV_PASS );
    if ( strcmp( (char *)&data[0], "defer" ) == 0 )
	return ( MDV_KNOWN );		/* OK address but check more */

    /* Some other tag: don't really care what it is; it's valid in
       some sense so we'll conditionally OK it without accepting it.
    */
    return ( MDV_UNKNOWN );
}
/*

*//* smtpd_addrmap_open()

	Open the addrmap database

Accepts :


Returns :

	<value>		TRUE if OK
			FALSE if not.

Notes :

*/

static BOOL
smtpd_addrmap_open NOARGLIST
{
    /* Open any cdbfiles we need */
    Addrmapdb.cdb_fd = open( MVMTRSYS_ADDRMAP, O_RDONLY );
    if ( Addrmapdb.cdb_fd < 0 ) {
	smtpd_log( NULL, "ERROR", NULL, "Can't open addrmap file \"%s\"\n",
				MVMTRSYS_ADDRMAP );
	return ( FALSE );
    }
    cdb_init( &Addrmapdb.cdb_cdb, Addrmapdb.cdb_fd );

    return ( TRUE );
}
/*

*//* smtpd_authsetup( xmtP, unameP )

	Finish SMTP AUTH setup

Accepts :

	xmtP		Ptr to XMTP control block
	unameP		The username authenticated/authorized

Returns :

	<value>		status code

Notes :

	Returns as per smtpd_cmd_xxx functions.

	This processes the final step of an SMTP AUTH command,
	  after authentication has succeeded.  It includes returning
	  the final response code to the client.

*/


static STSCODE
smtpd_authsetup ARGLIST( (xmtP, unameP) )
   NFARG( XMTP		*xmtP )		/* XMTP control block */
    FARG( UBYTE		*unameP )	/* Name of user authenticated */
{
	STSCODE		status;
	BOOL		okF;

    if ( SDS_authnameP != NULL )
	dealloc( SDS_authnameP );
    SDS_authnameP = (UBYTE *)newstr( "smtpd_authsetup", "authname", unameP );
    if ( SDS_authnameP == NULL ) {
	smtpd_log( xmtP, "ERROR", NULL,
		     "smtpd_authsetup: failure storing name." );
	okF = FALSE;
    }
    else {
	SDS_authF = TRUE;		/* We've AUTH'd */
	SCI_relayokF = TRUE;		/* We can relay */

#ifdef	MVMCAS_IN_MVMTR
	if ( SDS_mvmcasF ) {
	    /* Tell mvmcas about the change. */
	    mvmcas_send( "STOP IP %s", SCI_hostipP );
	    mvmcas_send( "START AUTH %s", SDS_authnameP );
	    Mvmcas_expectF = TRUE;
	}
#endif	/* MVMCAS_IN_MVMTR */
    
	smtpd_log( xmtP, "AUTH+", NULL, "user %s", unameP );

	okF = xmtp_reply( xmtP, TRUE, 235, NULL,
			 "Authentication successful.\n" );
    }



    return ( okF ? SCOK : SCERR );
}
/*

*//* smtpd_cmd_helo_first( xmtP, ecmdP, tailP )

	Do the first, common part of HELO/EHLO command

Accepts :

	xmtP		Ptr to XMTP control block
	ecmdP		The error command name (HELO-/EHLO-) for logging
	tailPP		Address of pointer to the command tail

Returns :

	<value>		SCxxx status code, one of:
			  SCACCEPT:  if HELO name was accepted
			  SCOK:	     HELO rejected, session should continue
			  SCDONE:    HELO rejected, session should be done
			  other:     if some error.

Notes :


*/

static STSCODE
smtpd_cmd_helo_first ARGLIST( (xmtP, ecmdP, tailPP) )
   NFARG( XMTP		*xmtP )		/* XMTP handle */
   NFARG( char		*ecmdP )	/* Error command name */
    FARG( UBYTE		**tailPP )	/* Command tail ptr */
{
	BOOL		okF;
	int		ss_helo;	/* Strict session override for helo */
	int		mdv;		/* Mail disposition value */
	STSCODE		status;
	XPRSLT		xpresult;

    /* Post-helo timeout */
    bio_timeout_set( &xmtP->mt_in, SDS_to_general, 0L );

    SDS_heloF = FALSE;			/* Not good HELO yet */
    xpresult = xmtp_parse_domain( tailPP, &SDS_heloname[0], XMTP_DOMAIN_MAX );
    if ( xpresult != XPR_OK )
	return ( smtpd_reply_syntax( xmtP ) );

    *tailPP = (UBYTE *)solbat( *(char **)tailPP, FALSE );
    if ( **tailPP != NUL )
	return ( smtpd_reply_syntax( xmtP ) );

    mdv = MDV_DUNNO;

    /* Validate helo name unless deferred until later */
    if ( cntrl_int_get( estr_tmp_str( "smtp_server_helo_check_defer" ),
		        0 ) == 0 ) {
	if ( helo_check( xmtP ) == MDV_REJECT )
	    mdv = MDV_REJECT;
    }

    /* Firm up the acceptance */
    if ( mdv < 0 )
	mdv = -mdv;

    switch ( mdv ) {
	case MDV_REJECT:		/* Outright rejection */
	    smtpd_log( xmtP, ecmdP, SDD_logP, "Rejected (%s)",
					&SDS_heloname[0] );

	    if ( SDD_replyP != NULL ) {
		/* Custom reply string was given */
		okF = xmtp_reply( xmtP, TRUE, 553, SDD_ecodeP, "%s",
					SDD_replyP );
	    }
	    else {
		okF = xmtp_reply( xmtP, TRUE, 553, SDD_ecodeP,
				    "Bad HELO/EHLO argument %s\n",
				    &SDS_heloname[0] );
	    }
	    status = SCREJECT;
	    break;


	default:			/* Anything else isn't valid here,
					   and we'll take the HELO.
					*/
	    status = SCACCEPT;
	    okF = TRUE;
	    break;
    }

    if ( status == SCREJECT ) {
	/* This helo is rejected.  Return code depends on session policy. */
	status = smtpd_session_policy( SDP_DEFER, SDP_DEFER, TRUE );

	/* May be overridden by HELO session policy. */
	ss_helo = cntrl_int_get( estr_tmp_str( "smtp_server_ss_helo" ), -1 );
	if ( ss_helo != -1 )
	    status = ( ss_helo == 0 ) ? SCDONE : SCOK;
	return ( status );
    }

    /* Successful EHLO must reset the transaction */
    if ( smtpd_reset_transaction( xmtP ) != SCOK ) {
	(void)xmtp_reply( xmtP, TRUE, 451, NULL,
			 "can not reset state." );
	return ( SCERR );
    }

    SDS_heloF = TRUE;

    return ( okF ? SCACCEPT : SCDONE );
}
/*

*//* smtpd_command(xmtP)

	Accept and process an SMTP command

Accepts :

	xmtP		Ptr to XMTP control block

Returns :

	<value>		status code

Notes :

	Codes return by this function may be interpreted as follows:

	  SCDONE means that the session should end.  It does not
	    necessarily mean that everything went OK.

	  SCBREAK means that the command input timed out.  Caller
	    might need to react further (tho probably not).

	  SCERR means some internal error.

	  SCOK means the command was OK, and another one should probably be
	    processed.

	The individual command processing routines may also return
	  those codes, except that they may also return SCCONTINUE as
	  a variation of SCOK.

*/


static STSCODE
smtpd_command ARGLIST( (xmtP) )
    FARG( XMTP		*xmtP )		/* XMTP control block */
{
	STSCODE		status;
	int		i;

#ifdef	MVMCAS_IN_MVMTR
    if ( SDS_mvmcasF ) {
	/* Note: this will probably be replaced by a select() on both
	   the smtpd input and the mvmcas input.  This is a q&d
	   first-pass to get some testing in.
	*/

	/* Check for any message from mvmcas.  If we think we're expecting
	   a message, wait a little while for one.
	*/
	if ( Mvmcas_expectF ) {
	    i = 3;
	    Mvmcas_expectF = FALSE;
	}
	else
	    i = 0;

	mvmcas_check( i );
    }
#endif		/* MVMCAS_IN_MVMTR */

    status = xmtp_getcmd( xmtP );

    /* Get any abnormal case out of the way */
    if ( status != SCOK ) {
	switch ( status ) {
	    case  SCERR:		/* Some error */
	        smtpd_log( xmtP, "DISCONNECT-", NULL, NULL );
		return ( SCDONE );

	    case SCBREAK:		/* Timeout */
	        smtpd_log( xmtP, "CMD-TO-", NULL,
				"Timeout getting command." );
		xmtp_reply( xmtP, TRUE, 321, NULL, "say what?" );
		return ( SCBREAK );

	    case SCNOTDONE:		/* Input buffer overflow */
	        smtpd_log( xmtP, "CMD-OFL-", NULL,
				"Command too large -- %s.",
				xmtP->mt_cmbsP->bs_bP );
		status = smtpd_reply_bad_command( xmtP );
		if ( status == SCOK )
		    xmtp_reply( xmtP, TRUE, 502, NULL, "command too large." );
		return ( status );

	    default:
		smtpd_log( xmtP, "ERROR", NULL,
				"smtpd_command: bad case: %d\n", status );
		return ( SCERR );
	}
    }

    /* Normal case may now proceed */
    if ( xmtP->mt_cmdL == 0 ) {
	/* Null command -- check limit and/or proceed */
	smtpd_log( xmtP, "CMD-NULL-", NULL, "Null command." );
	status = smtpd_reply_bad_command( xmtP );
	if ( status == SCOK )
	    xmtp_reply( xmtP, TRUE, 324, NULL, "null command." );
	return ( status );
    }

    /* Log command if enabled */
    if ( cntrl_int_get( estr_tmp_str( "smtp_server_log_commands" ), 0 ) != 0 )
	smtpd_log( xmtP, "CMD-LOG+", NULL, "%s", xmtP->mt_cmbsP->bs_bP );


    /* Look up the command and handle it. */
    for ( i = 0; Smtpcmds[i].cm_nameP != NULL; ++i ) {
	if ( ( xmtP->mt_cmdL == strlen( Smtpcmds[i].cm_nameP ) ) &&
	     ( strnicmp( (char *)xmtP->mt_cmdP, Smtpcmds[i].cm_nameP,
				xmtP->mt_cmdL ) == 0 ) ) {

	    /* Preclear any temporary dialog strings */
	    SDD_ecodeP = SDD_logP = SDD_replyP = NULL;

	    status = (*Smtpcmds[i].cm_rtc)( xmtP, xmtP->mt_cmdtailP );

	    /* SCCONTINUE is a synonym for SCOK at this point. */
	    if ( status == SCCONTINUE )
		status = SCOK;
	    return ( status );
	}
    }

    /* Not found, bad command */
    smtpd_log( xmtP, "CMD-UNK-", NULL, "Unrecognized command -- %s",
		xmtP->mt_cmbsP->bs_bP );

    /* Bad command- either hit limit or proceed */
    status = smtpd_reply_bad_command( xmtP );
    if ( status == SCOK )
	xmtp_reply( xmtP, TRUE, 500, NULL, "unrecognized." );

    return ( status );
}
/*

*//* smtpd_dest_access( mbxP )

	Access a destination struct

Accepts :

	mbxP		Ptr to mailbox struct 

Returns :

	<value>		Pointer to the SMTP_DEST struct corresponding
			  to the given mailbox, or
			NULL if something went awry.

Notes :


*/

static SMTP_DEST *
smtpd_dest_access ARGLIST( (mbxP) )
    FARG( XMTPMBX	*mbxP )		/* Mailbox */
{
	DESTNODE	*dnodeP;	/* AVL node ptr */
	XMTPMBX		lclmbx;		/* Local copy of mbx */

    /* Try to find the entry */
    dnodeP = (DESTNODE *)avlfind( SDS_dest_treeP, mbxP );
    if ( dnodeP == NULL ) {
	/* Can't find it, need to create it.  Take a copy first since
	   the insertion process steals the path pointer, but we need
	   it to do a lookup.
	*/
	lclmbx = *mbxP;
	avlinsert( SDS_dest_treeP, mbxP, NULL );
	dnodeP = (DESTNODE *)avlfind( SDS_dest_treeP, &lclmbx );
	if ( dnodeP == NULL ) {
	    /* Error.. */
	    return ( NULL );
	}
    }

    /* Return what we got */
    return ( &dnodeP->dst_dest );
}
/*

*//* smtpd_hook_dialog( cmdP, mdvP )

	Process an MFL hook that goes with an SMTP dialog.

Accepts :

	cmdP		The MFL command to run (call to a hook)
	mdvP		Where to update the mail disposition value

Returns :

	<value>		STSxxx stscode value

Notes :

	If the function return is other than SCOK, the other return
	  values are not relevant.

	This is for hooks that use a specific SMTP dialog interface.
	The hook is called to help with some part of the SMTP dialog,
	and it is expected to return a string that looks like this:

	    acceptance:logstring:replystring

	Leading whitespace in each section is skipped.  Trailing whitespace
	  is not.

	"acceptance" is one of the acceptance codes (aka mail disposition
	  value codes).  It may be optionally prefixed with a "~" character
	  to indicate that the opinion is tentative.

	"logstring" is any short string that may be logged by the
	  caller of the hook, as part of the general SMTP dialog logging.
	  This may be blank for no value.

	"replystring" is any SMTP dialog reply to give, overriding the
	  built-in reply string.  This may be blank for no value.  Note
	  that since this is the last part of the reply string, it may
	  contain colons.  The replystring may optionally be prefixed
	  with a "a.b.c/" string, where "a.b.c" is the extended
	  SMTP code to return.  If there is no string following the
	  '/' there is no replystring set, but the extended code is
	  remembered.  The check for the a.b.c/ string is pretty simplistic
	  at present; it can probably be fooled.

	If the acceptance code returned is blank or unrecognized,
	  the value at *mdvP is not altered.  Otherwise the
	  corresponding MDV_xxxx code is returned (negated if the
	  returned code was prefixed with ~).

	Various SMTP dialog state variables may be updated from this
	  function.  In particular the SDD_xxx strings corresponding
	  to the returned values are set.

	Not all returned values may be relevant to the caller.

*/

static STSCODE
smtpd_hook_dialog ARGLIST( (cmdP, mdvP) )
   NFARG( char		*cmdP )		/* The MFL command to execute */
    FARG( int		*mdvP )		/* Where to return the MDV */
{
	STSCODE		status;
	BOOL		negmdvF;	/* If mdv is negated */
	int		mdv;
	int		len;
	char		*rstrP;		/* NUL-terminated returned string */
	char		*accP;		/* Acceptance part of return */
	char		*ecodeP;	/* Extended code part of return */
	char		*logP;		/* Log note part of return */
	char		*replyP;	/* Reply part of string */
	char		*strP;		/* Generic string */
	ESTR		*retP;		/* Returned string value */
	VALUE		hookval;

    /* Attempt to run the MFL code */
    status = mfl_run_str( &hookval, estr_tmp_str( cmdP ),
			    estr_tmp_str( "Mc" ) );
    if ( status != SCOK )
	return ( status );

    /* Get the estr from the returned value */
    if ( ( retP = val_estr_tmp( &hookval ) ) == NULL ) {
TODO	/* Might want to log this? */
	val_release( &hookval );

	/* Note that we do not clear the various SDD_ dialog elements here. */
	return ( SCOK );
    }

    /* But we will want to work with a NUL-terminated string */
    rstrP = (char *)tmpstr_estr( retP );

    /* Get the colon-separated component parts of the string */
    rstrP = solbat( rstrP, TRUE );
    accP = tmptoken_char( (UBYTE **)&rstrP, ':' );
    rstrP = solbat( rstrP, TRUE );
    logP = tmptoken_char( (UBYTE **)&rstrP, ':' );

    /* Check for an ecode in the reply string. */
    rstrP = solbat( rstrP, TRUE );
TODO
    /* Check for an extended SMTP return code.  This is of the
       form "a.b.c/" .  We should probably validate the format,
       but we'll just do a couple of idiot checks for now.
    */
    if ( isdigit( *rstrP ) &&
	 ( ( strP = strchr( rstrP, '/' ) ) != NULL ) &&
	 ( ( len = ( strP - rstrP ) ) <= 9 ) ) {
	/* Got an ecode, make a copy. */
	ecodeP = (char *)tmpstr( len + 1 );
	memcpy( ecodeP, rstrP, len );
	ecodeP[len] = NUL;

	rstrP = solbat( rstrP + len + 1, TRUE );
    }
    else
	ecodeP = NULL;

    /* Collect any reply string */
    if ( *rstrP != NUL ) {
	replyP = (char *)tmpstr( strlen( rstrP ) +1 );
	strcpy( replyP, rstrP );
    }
    else
	replyP = NULL;

    val_release( &hookval );		/* Done with this now */

    /* Decode the acceptance value string into a mail disposition
       value code.  If prefixed with "~" the acceptance is tentative,
       which we'll indicate by negating it.
    */
    if ( *accP == '~' ) {		/* Check negated acceptance */
	negmdvF = TRUE;
	++accP;
    }
    else
	negmdvF = FALSE;
    mdv = mdv_from_str( accP );
    if ( mdv >= 0 ) {
	/* Got a valid mdv, use it. */
	if ( negmdvF )
	    mdv = -mdv;
	*mdvP = mdv;
    }
    else
TODO	/* Bad mdv return.  should we log this? */
	;

    /* Return the other values */
    SDD_ecodeP = ecodeP;
    SDD_logP = logP;
    SDD_replyP = replyP;

    return ( SCOK );
}
/*

*//* smtpd_qmail_queue( ommP )

	Send to [remaining] recipients via qmail-queue

Accepts :

	ommP		Ptr to open message thing

Returns :

	<value>		STSCODE status, one of:
			  SCOK if OK
			  SCNOTDONE if some transaction problem
			  SCERR if some internal error

Notes :


*/

#ifdef	QMAIL

STSCODE
smtpd_qmail_queue ARGLIST( (ommP) )
    FARG( OMM		*ommP )		/* Open mail message */
{
	BOOL		okF;
	int		i;
	QMQIF		*qqP;		/* qmail-queue handle */

    qqP = qmail_queue_open();
    if ( qqP == NULL )
	return ( SCNOTDONE );

    okF = okF && qmail_queue_write_message( qqP, ommP->om_mmsgP->mm_topmpP );

    /* Give reverse path / sender */
    okF = okF && qmail_queue_write_sender( qqP, SDS_rpathP->mb_pathP );

    /* Give recipient(s) via avlwalk */
    SDS_dest_treeP->t_walkP = (void *)qqP;
    avlwalk( SDS_dest_treeP, smtpd_qmail_queue_dest );

    /* Be done */
    if ( qmail_queue_close( qqP, okF ) != SCOK )
	okF = FALSE;

    return ( okF ? SCOK : SCNOTDONE );
}

#endif	/* QMAIL */
/*

*//* smtpd_qmail_queue_dest( treeP, nodeP, depth )

	Give a destination address to qmail-queue to on behalf of AVL walk 

Accepts :

	treeP		Ptr to AVL tree
	nodeP		Ptr to the destination node
	depth		depth of the node

Returns :

	<value>		0 to continue
			non-zero to stop

Notes :

	Per avlwalk() interface; see that.

*/

#ifdef	QMAIL

static int
smtpd_qmail_queue_dest ARGLIST( (treeP, nodeP, depth) )
   NFARG( AVLTREE	*treeP )	/* Ptr to tree */
   NFARG( DESTNODE	*nodeP )	/* Ptr to node */
    FARG( int		depth )		/* Depth of node */
{
	SMTP_DEST	*destP;
	QMQIF		*qqP;		/* qmail-queue handle */

    /* NULL node signals end of walk */
    if ( nodeP == NULL )
	return ( 0 );			/* no special processing */

    destP = &nodeP->dst_dest;

    /* Skip this node if the address is not accepted */
    if ( destP->sd_acceptance != MDV_ACCEPT )
	return ( 0 );

    /* Send it.  Any error is remembered in the *qqP struct. */
    qqP = (QMQIF *)SDS_dest_treeP->t_walkP;
    qmail_queue_write_recipient( qqP, destP->sd_mbx.mb_pathP );

    return ( 0 );
}

#endif	/* QMAIL */
/*

*//* smtpd_rcpt_add(xmtP, destP)

	Add one to the rcpt-to table

Accepts :

	xmtP		Ptr to XMTP control block
	destP		Ptr to the destination struct

Returns :

	<value>		status code, see below.

Notes :

	Remembers this RCPT-TO address and does any ancillary
	processing associated with remembering it.

	Returns are:

	    SCOK: rcpt-to successfully remembered.

	    SCNOTDONE: Too many rcpt blocks already seen

	    SCERR: some error, should abort.

	Each RCPT-TO is remembered so that we can report post-data
	status (e.g. for pipelining).  The remembered RCPT-TO list
	is *not* used to control delivery.

	On successful (SCOK) return, the new rcpt block is added to
	  the SDS_rcpt list; caller can fetch it from the tail.

	On non-successful return, a response is given to the xmtP
	  peer.

*/

static STSCODE
smtpd_rcpt_add ARGLIST( (xmtP, destP) )
   NFARG( XMTP		*xmtP )		/* XMTP control block */
    FARG( SMTP_DEST	*destP )	/* The destination to add */
{
	SMTP_RCPT	*rcptP;

    /* Reject softly if we already have the max rcpts */
    if ( SDS_rcptC == SMTPD_RCPTS_MAX ) {
	xmtp_reply( xmtP, TRUE, 451, NULL, "Too many RCPT TO commands." );
	return ( SCNOTDONE );
    }

    /* Allocate and initialize rcpt-to info block */
    rcptP = (SMTP_RCPT *)emalloc( "smtpd_rcpt_add", "SMTP_RCPT block",
					sizeof( SMTP_RCPT ) );
    if ( rcptP == NULL ) {
	xmtp_reply( xmtP, TRUE, 451, NULL, "Server problem." );
	return ( SCERR );
    }
    rcptP->sr_destP = destP;

    /* Add this to the end of the list */
    ll_addvt( &SDS_rcpt, rcptP );

    ++SDS_rcptC;

    return ( SCOK );
}
/*

*//* smtpd_rcvdhdr();

	Create "Received" headers in the Rcvdhdr[] buffer.

Accepts :



Returns :

	<nothing>

Notes :

	Also creates an "X-Peer-Info" line in Xphdr

*/

static void
smtpd_rcvdhdr NOARGLIST
{
	int		len;
	time_t		timenow;
	struct tm	*tmP;
	char		datestr[100];
	char		tzoff[10];

    /* Format date/time */
    time(&timenow);
    tmP = localtime( &timenow );

#if 0
    strftime( &datestr[0], sizeof(datestr),
	     "%a, %d %B %Y %H:%M:%S %z (%Z)", tmP );
#else
    /* Not all strftimes support %z so we'll do this hack.
       I suppose we could just as easily format it directly.
    */
    if ( tmP->tm_gmtoff < 0 )
	snprintf( &tzoff[0], sizeof(tzoff), "-%02d%02d",
		 (-tmP->tm_gmtoff)/3600,
		 ((-tmP->tm_gmtoff)/60)%60);
    else
	snprintf( &tzoff[0], sizeof(tzoff), "+%02d%02d",
		 (tmP->tm_gmtoff)/3600,
		 ((tmP->tm_gmtoff)/60)%60);

    len = strftime( &datestr[0], sizeof(datestr),
	     "%a, %d %B %Y %H:%M:%S ", tmP );
    len += snprintf( &datestr[len], sizeof(datestr) -len, "%s ", tzoff );
    strftime( &datestr[len], sizeof(datestr) -len,
	     "(%Z)", tmP );
#endif


    len = snprintf( (char *)&Rcvdhdr[0], sizeof(Rcvdhdr),
		   "Received: from %s (%s [%s])\n\
	by %s ([%s]) with %sSMTP%s\n\
	(MVMTR %s); %s\n",
		   &SDS_heloname[0],
		   SCI_hostnameP, SCI_hostipP,
		   SCI_myhostnameP, SCI_myhostipP,
		   SDS_esmtpF ? "E" : "",
		   SDS_authF ? "A" : "",
		   MVMTR_VERSION,
		   datestr );


    /* Create Xphdr */
    snprintf( (char *)&Xphdr[0], sizeof(Xphdr),
"X-Peer-Info: remote-ip %s local-ip %s\n  local-name %s\n",
	     SCI_hostipP, SCI_myhostipP, SCI_myhostnameP );

}
/*

*//* smtpd_reply_bad_command(xmtP)

	Preprocess a bad command response

Accepts :

	xmtP		Ptr to XMTP control block

Returns :

	<value>		SCxxx status code, including:
			  SCOK to continue;
			  SCDONE to exit immediately

Notes :

	This is a preprocessor for a bad command response.

	Mostly it counds bad commands, and checks the bad-command
	  limit.  If the limit has not yet been reached, it returns
	  SCOK.

	If the limit has been reached, it gives an SMTP reply to the
	  client, logs a termination entry, and returns SCNOTDONE.

*/

static STSCODE
smtpd_reply_bad_command ARGLIST( (xmtP) )
    FARG( XMTP		*xmtP )		/* XMTP control block */
{
	STSCODE		status;
	int		i;

    /* Check maximum bad command limit */
    i = cntrl_int_get( estr_tmp_str( "smtp_server_badcmd_max" ), 3 );
    if ( ++SDS_badcmdC >= i ) {
	smtpd_log( xmtP, "BADCMDS-", NULL, "Terminating session." );
	xmtp_reply( xmtP, TRUE, 554, NULL, "Too many bad commands." );
	status = SCNOTDONE;
    }
    else
	status = SCOK;

    return ( status );
}
/*

*//* smtpd_reply_badmax(xmtP)

	Issue a reply about too many bad recipients

Accepts :

	xmtP		Ptr to XMTP control block

Returns :

	<value>		SCxxx status code, including:
			  SCOK to continue;
			  SCDONE to exit immediately

Notes :

	This centralizes the badmax handling, including the
	adherence to strict smtp sessions.
*/

static STSCODE
smtpd_reply_badmax ARGLIST( (xmtP) )
    FARG( XMTP		*xmtP )		/* XMTP control block */
{
	BOOL		okF;
	int		rcode;
	STSCODE		status;

    status = smtpd_session_policy( SDP_DEFER, SDP_DEFER, TRUE );
    switch ( status ) {
	case SCOK:			/* Not ending, give deferral */
	    rcode = 450;
	    break;

	case SCDONE:			/* Ending, give shutdown */
	    rcode = 421;
	    break;

	default:			/* Some error, give error. */
	    rcode = 451;
	    break;
    }

    /* Log a warning */
    smtpd_log( xmtP, "RCPT--", NULL, "excessive bad rcpt-to" );

    /* Give reply */
    okF = xmtp_reply( xmtP, TRUE, rcode, NULL, "too many bad recipients.\n" );

    return ( status );
}
/*

*//* smtpd_reply_syntax(xmtP)

	Issue a "syntax error" reply

Accepts :

	xmtP		Ptr to XMTP control block

Returns :

	<value>		SCxxx status code

Notes :

	Just a place to centralize a standard syntax error reply.

	This will also do max-bad-commands processing; i.e., will
	  disconnect if too many bad commands have been accumulated.

*/

static STSCODE
smtpd_reply_syntax ARGLIST( (xmtP) )
    FARG( XMTP		*xmtP )		/* XMTP control block */
{
	BOOL		okF;
	STSCODE		status;

    /* Log a warning */
    smtpd_log( xmtP, "SYNTAX", NULL, "%s", xmtP->mt_cmbsP->bs_bP );

    /* Check the count of bad commands first */
    status = smtpd_reply_bad_command( xmtP );
    if ( status == SCOK )
	if ( !xmtp_reply( xmtP, TRUE, 501, NULL,
				"syntax error in command\n" ) )
	    status = SCDONE;

    return ( status );
}
/*

*//* smtpd_require_eol(xmtP, tailP)

	Require that a command has no [more] arguments.

Accepts :

	xmtP		Ptr to XMTP control block
	tailP		Ptr to command tail

Returns :

	<value>		SCxxx status code:
			  SCOK if it's OK to continue (end of command line)
			  SCCONTINUE if it's not OK (not end of line)
			  SCERR if some failure.

Notes :

	If HELO has not yet been seen, gives an SMTP reply
	to that effect.  

*/

static STSCODE
smtpd_require_eol ARGLIST( (xmtP, tailP) )
   NFARG( XMTP		*xmtP )		/* XMTP control block */
    FARG( UBYTE		*tailP )	/* Command tail */
{
	BOOL		okF;

    if ( ( tailP != NULL ) && ( ( *solbat( (char *)tailP, FALSE ) != NUL ) ) ) {
	/* Log a warning */
	smtpd_log( xmtP, "SYNTAX", NULL, "%s", xmtP->mt_cmbsP->bs_bP );

	okF = xmtp_reply( xmtP, TRUE, 501, NULL,
			   "syntax error (junk at end of line)" );
	if ( !okF )
	    return ( SCERR );
	return ( SCCONTINUE );
    }

    return ( SCOK );
}
/*

*//* smtpd_require_helo(xmtP)

	Require that a HELO (etc) has been given.

Accepts :

	xmtP		Ptr to XMTP control block

Returns :

	<value>		SCxxx status code:
			  SCOK if it's OK to continue (HELO has been seen);
			  SCCONTINUE if it's not OK (no HELO seen);
			  SCERR if some failure.

Notes :

	If HELO has not yet been seen, gives an SMTP reply
	to that effect.  

*/

static STSCODE
smtpd_require_helo ARGLIST( (xmtP) )
    FARG( XMTP		*xmtP )		/* XMTP control block */
{
	BOOL		okF;

    if ( !SDS_heloF ) {
	/* No HELO seen, or HELO state has been reset.  */

	/* This may be because we've done an AUTH.  RFC2554 says that
	   after AUTH, client SHOULD helo again; we'll take that as
	   allowing this to pass if client has auth'd.
	*/
	if ( SDS_authF ) {
TODO	    /* Might want to log a warning here. */
	    return ( SCOK );
	}

	okF = xmtp_reply( xmtP, TRUE, 503, NULL,
			   "say hello!  (HELO/EHLO required)\n" );
	if ( !okF )
	    return ( SCERR );
	return ( SCCONTINUE );
    }

    return ( SCOK );
}
/*

*//* smtpd_require_transaction(xmtP)

	Require that a transaction be in progress

Accepts :

	xmtP		Ptr to XMTP control block

Returns :

	<value>		SCxxx status code:
			  SCOK if it's OK to continue
			  SCCONTINUE if it's not OK (no transaction yet)
			  SCERR if some failure.

Notes :

	Implies HELO required, i.e.
	   calls smtpd_require_helo().

	XMTP reply will be given if there's an issue.

*/

static STSCODE
smtpd_require_transaction ARGLIST( (xmtP) )
    FARG( XMTP		*xmtP )		/* XMTP control block */
{
	STSCODE		status;
	BOOL		okF;

    /* HELO is also required */
    if ( ( status = smtpd_require_helo( xmtP ) ) != SCOK )
	return ( status );


    if ( !SDS_transF ) {
	okF = xmtp_reply( xmtP, TRUE, 503, NULL,
			   "No transaction!  (MAIL FROM: required)\n" );
	return ( okF ? SCCONTINUE : SCERR );
    }

    return ( SCOK );
}
/*

*//* smtpd_rewrite_local_ip( mbxP )

	Check for our IP address in user@[ip] form; rewrite if so

Accepts :

	mbxP		Ptr to mailbox info

Returns :

	<value>		status code:
			  SCOK if ok
			  other if not.

Notes :


*/


static STSCODE
smtpd_rewrite_local_ip ARGLIST( (mbxP) )
    FARG( XMTPMBX	*mbxP )		/* Mailbox to remember */
{
	int		i;
	in_addr_t	addr;
	char		ipbuf[100];	/* Buffer for IP address string */

    /* No rewrite if no local-ip-host defined */
    if ( SDS_localiphostP == NULL )
	return ( SCOK );

    /* No rewrite if not @[host] form */
    if ( ( mbxP->mb_pathP[ mbxP->mb_atX +1 ] != '[' ) ||
	 ( mbxP->mb_pathP[ mbxP->mb_len -1 ] != ']' ) )
	return ( SCOK );

    /* Copy out the IP address string and get an IP address value from it */
    i = ( mbxP->mb_len - mbxP->mb_atX ) -3;  /* @,[,] */
    if ( i >= sizeof( ipbuf ) )
	return ( SCOK );
    memcpy( &ipbuf[0], &mbxP->mb_pathP[ mbxP->mb_atX +2 ], i );
    ipbuf[i] = NUL;
    if ( !ip_decode_str( (UBYTE *)&ipbuf[0], &addr ) )
	return ( SCOK );

    /* See if that address is one that is local to this host */
    for ( i = 0; i < MyipC; ++i )
	if ( addr == MyipP[i] )
	    break;
    if ( i == MyipC )
	return ( SCOK );

    /* Given IP is local.  replace the domain name in the mbx. */
    i = strlen( SDS_localiphostP );
    if ( ( mbxP->mb_atX + i + 1 ) > mbxP->mb_len ) {
	/* Need a bigger buffer */
	mbxP->mb_pathP = (UBYTE *)erealloc( "smtpd_rewrite_local_ip",
					    "new path string",
					    (char *)mbxP->mb_pathP,
					    mbxP->mb_atX + i + 1 );
	if ( mbxP->mb_pathP == NULL )
	    return ( SCERR );
    }
    strcpy( (char *)&mbxP->mb_pathP[ mbxP->mb_atX +1 ], SDS_localiphostP );
    mbxP->mb_len = mbxP->mb_atX + i + 1;

    return ( SCOK );
}
/*

*//* smtpd_session(xmtP)

	Handle an smtpd session

Accepts :

	xmtP		Ptr to XMTP control block

Returns :

	<value>		status code

Notes :

	Caller should have already issued the SMTP greeting
	before entering here.

*/


static STSCODE
smtpd_session ARGLIST( (xmtP) )
    FARG( XMTP		*xmtP )		/* XMTP control block */
{
	STSCODE		status;
	int		i;

    /* Prep for new command stream and transaction */
    status = smtpd_reset_transaction( xmtP );

    /* But set the timeout to first-helo timeout value */
    i = cntrl_int_get( estr_tmp_str( "smtp_server_timeout_helo" ), 30 );
    bio_timeout_set( &xmtP->mt_in, i, 0L );

    SDS_heloF = FALSE;			/* No HELO seen yet */

    /* Process commands */
    for ( ; ; ) {
	status = smtpd_command( xmtP );
	if ( status != SCOK )
	    break;
    }

    /* Translate the internal "done" code to "ok" */
    if ( status == SCDONE )
	status = SCOK;

    return ( status );
}
/*

*//* smtpd_session_policy( rcpt, data, wantendF )

	Set new session acceptance policy

Accepts :

	rcpt		rcpt-to session policy
	data		data session policy
	wantendF	If we might want to end the session now.

Returns :

	<value>		status code, including:
			  SCOK to continue with session
			  SCDONE to end the session now

Notes :

	rcpt: says what to do with future RCPT-TO commands that might
	  otherwise be accepted.

	data: says what to do with future DATA commands that might otherwise
	  be accepted.

	Each of the above values acts in a ratchet function.  Once a
	  particularly high level has been set, it will not back down
	  even in the event of a future lower request.  These are sticky
	  for an entire connection/session, not just the current
	  transaction.  The ratchet order is:  SDP_DEFER, SDP_REJECT.

	The wantendF is passed as TRUE if the caller might want to
	  end the session now.  This will be used to influence the
	  return value of this function; any such request will be
	  overridden by the current "strict smtpd sessions" control.

*/


static STSCODE
smtpd_session_policy ARGLIST( (rcpt, data, wantendF) )
   NFARG( SDPOLICY	rcpt )		/* RCPT-TO policy */
   NFARG( SDPOLICY	data )		/* DATA policy */
    FARG( BOOL		wantendF )	/* If we want to end now */
{
    /* Ratchet up each value.  I suppose I could have a function for
       each of these, but it's pretty simple to do in-line at present.
       Maybe if it gets more complex..
    */
    switch ( rcpt ) {
	case SDP_DEFER:			/* Take if not already higher */
	    if ( SDS_sdp_rcpt != SDP_REJECT )
		SDS_sdp_rcpt = rcpt;
	    break;

	case SDP_REJECT:		/* Always take this policy */
	    SDS_sdp_rcpt = rcpt;
	    break;

	default:
	    warning( "smtpd_session_policy: unknown rcpt value %d", rcpt );
	    break;
    }


    switch ( data ) {
	case SDP_DEFER:			/* Take if not already higher */
	    if ( SDS_sdp_data != SDP_REJECT )
		SDS_sdp_data = data;
	    break;

	case SDP_REJECT:		/* Always take this policy */
	    SDS_sdp_data = data;
	    break;

	default:
	    warning( "smtpd_session_policy: unknown data value %d", data );
	    break;
    }

    /* Return the status depending on whether we want to end, and
       whether strict session adherence is enabled.  This control
       is checked dynamically rather than at session start, as a hook
       might change it.
    */
    if ( wantendF &&
         ( cntrl_int_get(
	       estr_tmp_str( "smtp_server_strict_sessions" ),
	       1 ) == 0 ) )
	return ( SCDONE );
    return ( SCOK );
}
/*

*//* tmptoken_char( strPP, delch )

	Get a char-delimited token out of a NUL-terminated string

Accepts :

	strPP		Ptr to the string pointer
	delch		The delimiter byte

Returns :

	<value>		Ptr to tmpstr() with token if found
			NULL if no further tokens

Notes :

	*strPP will be set to the byte in the string after the delimiter

	If the end of the string has already been reached, NULL is
	  returned.  Otherwise a tmpstr() will be allocated and filled
	  with the token, which is the part of the string up until the
	  delimiter char (or to the end of the source string if no
	  delimiter is found).

	The returned string, if any, will be NUL-terminated.

*/

static UBYTE *
tmptoken_char ARGLIST( (strPP, delch) )
   NFARG( UBYTE		**strPP )	/* Ptr to string ptr */
    FARG( UBYTE		delch )		/* delimiter char */
{
	UBYTE		ch;
	int		len;
	UBYTE		*tokP;		/* Token pointer */
	UBYTE		*strP;		/* Source string */

    strP = *strPP;			/* Source string */
    if ( *strP == NUL )
	tokP = NULL;			/* No more tokens */
    else {
	/* Find the string length */
	for ( len = 0; ( ch = strP[len] ) != NULL; ++len )
	    if ( ch == delch )
		break;

	/* Found a token, allocate a tmpstr and fill it */
	tokP = tmpstr( len +1 );
	if ( tokP != NULL ) {
	    memcpy( tokP, strP, len );
	    tokP[len] = NUL;
	}

        /* Set return string past the delimiter found */
	if ( ch != NUL )
	    *strPP += ( len +1 );
	else
	    *strPP += len;
    }

    return ( tokP );
}
